/**
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2007 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.ow2.orchestra.definition;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.xml.namespace.QName;

import org.ow2.orchestra.definition.activity.AbstractBpelActivity;
import org.ow2.orchestra.definition.activity.Flow;
import org.ow2.orchestra.definition.activity.Pick;
import org.ow2.orchestra.definition.activity.Receive;
import org.ow2.orchestra.definition.activity.Scope;
import org.ow2.orchestra.definition.activity.Sequence;
import org.ow2.orchestra.definition.element.Correlation;
import org.ow2.orchestra.definition.element.CorrelationSet;
import org.ow2.orchestra.definition.element.InboundMessageElement;
import org.ow2.orchestra.definition.element.Target;
import org.ow2.orchestra.definition.element.Correlation.Initiate;
import org.ow2.orchestra.exception.StaticAnalysisException;
import org.ow2.orchestra.facade.data.def.ProcessDefinitionData;
import org.ow2.orchestra.facade.uuid.IdFactory;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessInstanceUUID;
import org.ow2.orchestra.pvm.internal.model.ExecutionImpl;
import org.ow2.orchestra.pvm.internal.model.NodeImpl;
import org.ow2.orchestra.pvm.internal.model.ProcessDefinitionImpl;
import org.ow2.orchestra.pvm.internal.xml.Parse;
import org.ow2.orchestra.pvm.model.Node;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.services.Receiver;
import org.ow2.orchestra.services.ReceivingElement;
import org.ow2.orchestra.util.EnvTool;
import org.ow2.orchestra.util.StaticAnalysisFault;
import org.ow2.orchestra.util.StyleSheetRepository;
import org.ow2.orchestra.wsdl.WsdlsInfos;

/**
 * a bpel process definition.
 *
 * Both the initial node in the process and the activity of the aggregated scope
 * should point to the same node which is the top level activity in the process.
 *
 * @author Goulven Le Jeune
 *
 */
public class BpelProcess extends ProcessDefinitionImpl {

  /** serialVersionUID. */
  private static final long serialVersionUID = 1L;

  /**
   * specifies the query language used in the process for selection of nodes in
   * assignment. The default value for this attribute is:
   * "urn:oasis:names:tc:wsbpel:2.0:sublang:xpath1.0", which represents the
   * usage of [XPath 1.0] within WS-BPEL 2.0
   */
  protected String queryLanguage;

  /**
   * specifies the expression language used in the process element. The default
   * value for this attribute is:
   * "urn:oasis:names:tc:wsbpel:2.0:sublang:xpath1.0", which represents the
   * usage of [XPath 1.0] within WS-BPEL 2.0.
   */
  protected String expressionLanguage;

  /**
   * suppressJoinFailure attribute of the process element. determines whether
   * the joinFailure fault will be suppressed for all activities in the process.
   * The effect of the attribute at the process level can be overridden by an
   * activity using a different value for the attribute. The default for this
   * attribute is "no" at the process level. When this attribute is not
   * specified for an activity, it inherits its value from its closest enclosing
   * activity or from the process element if no enclosing activity specifies
   * this attribute.
   */
  protected boolean suppressJoinFailure = false;

  /** exitOnStandardFault attribute of the process element. */
  protected boolean exitOnStandardFault = false;

  /** targetNamespace attribute of the process element. */
  protected String targetNamespace = null;

  protected ProcessDefinitionUUID processUUID;

  /**
   * Infos collected in wsdl files used by the process like portTypes,
   * messages...
   */
  protected WsdlsInfos wsdlInfos;

  /** styleSheetRepository. */
  protected StyleSheetRepository styleSheetRepository;

  protected Map<OperationKey, InboundMessageElement> startElements;

  protected Map<OperationKey, Receiver> receivers = new HashMap<OperationKey, Receiver>();

  protected Collection<CorrelationSet> commonCorrelationSets = null;

  public BpelProcess(final String name, final String targetNamespace, final String processUUID) {
    this.name = name;
    this.targetNamespace = targetNamespace;
    this.processUUID = new ProcessDefinitionUUID(this.getQName(), processUUID);
  }

  /** For persistence * */
  protected BpelProcess() {
    super();
  }

  @Override
  protected ExecutionImpl newProcessInstance() {
    final ProcessInstanceUUID processInstanceUUID = IdFactory.getNewInstanceUUID();
    final BpelExecution instance = new BpelExecution(processInstanceUUID);
    instance.setProcessDefinition(this);
    instance.setProcessInstance(instance);
    EnvTool.getRepository().storeInstance(instance);
    return instance;
  }


  public InboundMessageElement addStartElement(final OperationKey operationKey, final InboundMessageElement receivingElement) {
    if (this.startElements == null) {
      this.startElements = new HashMap<OperationKey, InboundMessageElement>();
    }
    return this.startElements.put(operationKey, receivingElement);
  }

  public InboundMessageElement getStartElement(final OperationKey operationKey) {
    if (this.startElements != null) {
      return this.startElements.get(operationKey);
    }
    return null;
  }

  public String getExpressionLanguage() {
    return this.expressionLanguage;
  }

  public void setExpressionLanguage(final String expressionLanguage) {
    this.expressionLanguage = expressionLanguage;
  }

  public String getQueryLanguage() {
    return this.queryLanguage;
  }

  public void setQueryLanguage(final String queryLanguage) {
    this.queryLanguage = queryLanguage;
  }

  public boolean isSuppressJoinFailure() {
    return this.suppressJoinFailure;
  }

  public void setSuppressJoinFailure(final boolean suppressJoinFailure) {
    this.suppressJoinFailure = suppressJoinFailure;
  }

  public String getTargetNamespace() {
    return this.targetNamespace;
  }

  public void setTargetNamespace(final String targetNamespace) {
    this.targetNamespace = targetNamespace;
  }

  public void setWsdlInfos(final WsdlsInfos wsdlInfos) {
    this.wsdlInfos = wsdlInfos;
  }

  public WsdlsInfos getWsdlInfos() {
    return this.wsdlInfos;
  }

  public QName getQName() {
    return new QName(this.targetNamespace, this.name);
  }

  public StyleSheetRepository getStyleSheetRepository() {
    return this.styleSheetRepository;
  }

  public void setStyleSheetRepository(final StyleSheetRepository styleSheetRepository) {
    this.styleSheetRepository = styleSheetRepository;
  }

  public void staticAnalysisCheck(final Parse parse) {
    if (this.startElements == null) {
      // SA00015
      final String message = StaticAnalysisFault.getMessage("SA00015",
        "  Process should have at least one receive or pick with createInstance set to yes");
      parse.addProblem(message, new StaticAnalysisException("SA00015"));
    } else {
      // SA00056
      final Set<InboundMessageElement> startElementFound = new HashSet<InboundMessageElement>();
      // check SA00056 (check non createInstance activities are executed after a start activity)
      // store in startElementFound the start activities that can be really executed as start activities
      this.isFirstActivityCreateInstance(this.getInitial(), parse, startElementFound);

      // create a set of ReceivingElement from the start activities
      // (pick is an activiy/ onMessage is the receivingElement)
      final Set<ReceivingElement> receivingElementFound = new HashSet<ReceivingElement>();
      for (final InboundMessageElement inMsgElt : startElementFound) {
        receivingElementFound.addAll(inMsgElt.getReceivingElements());
      }

      // check all the start receiving elements are really start elements
      final int nbUnreachable = this.getStartElements().size() - receivingElementFound.size();
      if (nbUnreachable != 0) {
        final String message = StaticAnalysisFault.getMessage("SA00056",
            nbUnreachable + " activities with \"createInstance='yes'\" are not process start activities.");
        parse.addProblem(message, new StaticAnalysisException("SA00056"));
      } else {
        // SA00057
        // Check only if we have more than one start activity (!= one start element)
        if (startElementFound.size() > 1) {
          try {
            this.commonCorrelationSets = this.checkCommonCS();
          } catch (final StaticAnalysisException e) {
            parse.addProblem(e.getMessage(), e);
          }
        }
      }
    }
  }

  /**
   * Check for SA00056
   * @param node
   * @param parse
   */
  private void isFirstActivityCreateInstance(final NodeImpl node, final Parse parse, final Set<InboundMessageElement> startElementFound) {
    final AbstractBpelActivity activity = (AbstractBpelActivity) node.getBehaviour();

    // Is current activity a createInstance ?
    if (activity instanceof Receive) {
      if (((Receive) activity).isCreateInstance()) {
        startElementFound.add((InboundMessageElement) activity);
        return;
      }
    } else if (activity instanceof Pick) {
      if (((Pick) activity).isCreateInstance()) {
        startElementFound.add((InboundMessageElement) activity);
        return;
      }
    }

    // current activity is not a create instance.
    // check if this is allowed.
    if (!this.hasControlDependencyOnStartActivity(activity)) {
      if (activity instanceof Scope) {
        this.isFirstActivityCreateInstance((NodeImpl) node.getNodes().get(Scope.MAIN_ACTIVITY), parse, startElementFound);
      } else if (activity instanceof Sequence) {
        this.isFirstActivityCreateInstance((NodeImpl) node.getNodes().get(0), parse, startElementFound);
      } else if (activity instanceof Flow) {
        for (final Node child : node.getNodes()) {
          this.isFirstActivityCreateInstance((NodeImpl) child, parse, startElementFound);
        }
      } else {
        final String message = StaticAnalysisFault.getMessage("SA00056",
            "Activity " + node.getName() + " is not a start activity");
        parse.addProblem(message, new StaticAnalysisException("SA00056"));
      }
    }
  }

  private boolean hasControlDependencyOnStartActivity(final AbstractBpelActivity activity) {
    if (this.startElements.containsValue(activity)) {
      return true;
    }
    if (activity.getTargets() != null) {
      for (final Target t : activity.getTargets()) {
        if (this.hasControlDependencyOnStartActivity(t.getLink().getSourceActivity())) {
          return true;
        }
      }
    }
    return false;
  }

  /**
   * Check for SA00057
   * @param node
   * @param parse
   */
  private Set<CorrelationSet> checkCommonCS() {
    Set<CorrelationSet> commonCS = null;
    boolean initialized = false;
    for (final InboundMessageElement inboundMessageElement : this.startElements.values()) {
      for (final ReceivingElement startElement : inboundMessageElement.getReceivingElements()) {
        final List<Correlation> correlations = startElement.getCorrelations();
        if (correlations != null && !correlations.isEmpty()) {
          // If has correlations
          if (!initialized) {
            // first start element, all CS are common
            commonCS = new HashSet<CorrelationSet>();
            for (final Correlation correlation : correlations) {
              commonCS.add(correlation.getCorrelationSet());
            }
          } else {
            // check there is at least one common CS
            if (commonCS == null) {
              // previous had non CS -> error
              throw new StaticAnalysisException("SA00057");
            } else {
              // make the intersection of previous cs and element cs
              final Set<CorrelationSet> elementCs = new HashSet<CorrelationSet>();
              for (final Correlation correlation : correlations) {
                elementCs.add(correlation.getCorrelationSet());
              }
              commonCS.retainAll(elementCs);
              // if intersection is empty => error
              if (commonCS.isEmpty()) {
                throw new StaticAnalysisException("SA00057");
              }
            }
          }
        } else {
          // no CS in this start element
          // check if previous start elements had a CS
          if (commonCS != null) {
            throw new StaticAnalysisException("SA00057");
          }
        }
        initialized = true;
      }
    }

    // check if all common CS are set to initiate="join"
    // TODO: check if this method can be improved
    if (commonCS != null) {
      for (final CorrelationSet cs : commonCS) {
        for (final InboundMessageElement inboundMessageElement : this.startElements.values()) {
          for (final ReceivingElement startElement : inboundMessageElement.getReceivingElements()) {
            for (final Correlation correlation : startElement.getCorrelations()) {
              if (correlation.getCorrelationSet().equals(cs)) {
                if (!correlation.getInitiate().equals(Initiate.join)) {
                  throw new StaticAnalysisException("SA00057");
                }
              }
            }
          }
        }
      }
    }
    return commonCS;
  }

  public ProcessDefinitionUUID getUUID() {
    return this.processUUID;
  }

  public ProcessDefinitionData getProcessDefinitionData() {
    return new ProcessDefinitionData(this.getUUID(), this.name, this.targetNamespace);
  }

  /**
   * @param receiver
   */
  public void addReceiver(final Receiver receiver) {
    this.receivers.put(receiver.getOperationKey(), receiver);
  }

  /**
   * @param receiver
   */
  public void removeReceiver(final OperationKey operationKey) {
    this.receivers.remove(operationKey);
  }
  /**
   * @return the receivers
   */
  public Collection<Receiver> getReceivers() {
    return this.receivers.values();
  }

  /**
   * @param operationKey
   * @return
   */
  public Receiver getReceiver(final OperationKey operationKey) {
    return this.receivers.get(operationKey);
  }


  public Map<OperationKey, InboundMessageElement> getStartElements() {
    return this.startElements;
  }

  public Collection<CorrelationSet> getCommonCorrelationSets() {
    return this.commonCorrelationSets;
  }


  protected void setCommonCorrelationSets(final Set<CorrelationSet> commonCorrelationSets) {
    this.commonCorrelationSets = commonCorrelationSets;
  }

}
