/**
 * Copyright (C) 2009 Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/

package org.ow2.orchestra.definition.activity;

import java.util.Date;
import java.util.Map;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.facade.data.def.ActivityDefinitionData;
import org.ow2.orchestra.facade.data.runtime.ActivityInstanceData;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.facade.uuid.ActivityDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ActivityInstanceUUID;
import org.ow2.orchestra.facade.uuid.IdFactory;
import org.ow2.orchestra.pvm.activity.ActivityExecution;
import org.ow2.orchestra.pvm.activity.ExternalActivity;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.util.EnvTool;

/**
 * @author S. Ali Tokmen
 */
public abstract class AbstractActivity implements ExternalActivity {

  private static final long serialVersionUID = -2049270479220106510L;

  /** For persistence */
  protected long dbid;

  protected ActivityDefinitionUUID uuid;

  /** name attribute present for all activities. */
  protected String name = null;

  public long getDbid() {
    return this.dbid;
  }

  public ActivityDefinitionUUID getUuid() {
    return this.uuid;
  }

  public void setUuid(final ActivityDefinitionUUID uuid) {
    this.uuid = uuid;
  }

  public String getName() {
    return this.name;
  }

  public void setName(final String name) {
    this.name = name;
  }

  public abstract ActivityType getType();

  /**
   * Hook start of activity.
   *
   * @param bpelExecution
   */
  public void beforeRunning(final BpelExecution bpelExecution) {
    final ActivityInstanceUUID activityInstanceUUID = IdFactory.getNewActivityUUID();
    bpelExecution.setStartedDate(new Date());
    bpelExecution.setActivityInstanceUUID(activityInstanceUUID);

    EnvTool.getRecorder().recordActivityStarted(this.createRuntimeData(bpelExecution));

  }

  /**
   * @return a definition data
   */
  public ActivityDefinitionData createDefinitionData(final BpelProcess bpelProcess) {
    return new ActivityDefinitionData(bpelProcess.getProcessDefinitionData(), this.uuid, null,
      this.getType(), this.name);
  }

  /**
   * @param bpelExecution the execution
   * @return an instance data
   */
  public final ActivityInstanceData createRuntimeData(final BpelExecution bpelExecution) {
    return new ActivityInstanceData(this.createDefinitionData(bpelExecution.getProcessDefinition()), bpelExecution
      .getProcessInstance().getProcessInstanceData(), bpelExecution.getActivityInstanceUUID(), bpelExecution
      .getEncloserActivityInstanceUUID());
  }

  /**
   * Hook end of activity.
   *
   * @param bpelExecution
   */
  public void afterRunned(final BpelExecution bpelExecution) {
    EnvTool.getRecorder().recordActivityEnded(this.createRuntimeData(bpelExecution));
  }

  /**
   * Signal method.
   *
   * @param execution execution
   * @param signal signal
   */
  public void signal(final ActivityExecution execution, final String signal, final Map<String, Object> parameters) {
    if (signal == null || !signal.equals("target")) {
      this.signal((BpelExecution) execution, signal, parameters);
    }
  }

  /**
   * Signal method.
   *
   * @param execution execution
   * @param signal signal
   */
  public abstract void signal(final BpelExecution execution, final String signal, final Map<String, Object> parameters);

  @Override
  public int hashCode() {
    return this.uuid.hashCode();
  }

  @Override
  public boolean equals(final Object obj) {
    if (obj == null) {
      return false;
    }
    if (!(obj instanceof AbstractActivity)) {
      return false;
    }
    final AbstractActivity other = (AbstractActivity) obj;
    return other.getUuid().equals(this.uuid);
  }

}
