/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.definition.activity;

import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.orchestra.definition.element.Link;
import org.ow2.orchestra.definition.element.Source;
import org.ow2.orchestra.definition.element.Target;
import org.ow2.orchestra.facade.exception.BpelFaultException;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.facade.exception.BpelFaultException.BpelFault;
import org.ow2.orchestra.lang.JoinCondition;
import org.ow2.orchestra.pvm.Execution;
import org.ow2.orchestra.pvm.activity.ActivityExecution;
import org.ow2.orchestra.pvm.env.Environment;
import org.ow2.orchestra.pvm.session.MessageSession;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.runtime.FlowRuntime;
import org.ow2.orchestra.runtime.LinkRuntime;
import org.ow2.orchestra.services.job.ExecuteSourcesMessage;

/**
 * common parts of all the bpel activities.
 *
 * @author Tom Baeyens, Charles Souillard
 */
public abstract class AbstractBpelActivity extends AbstractActivity {

  private static final long serialVersionUID = 8757394004727477526L;

  /** log. */
  private static Logger log = Logger.getLogger(AbstractBpelActivity.class.getName());

  /** enclosing scope. */
  protected Scope enclosingScope;

  /** suppressJoinFailure attribute present for all activities. */
  protected boolean suppressJoinFailure;

  protected boolean executeSourcesAsync = false;

  /**
   * targetsJoinCondition attribute set if this activity has target(s) and the
   * joinCondition attribute is set on targets element.
   */
  protected JoinCondition targetsJoinCondition = null;

  /** list of all sources present for all activities. */
  protected List<Source> sources = null;

  /** list of all targets present for all activities. */
  protected List<Target> targets = null;

  public enum TargetBehaviour {
    /**
     * all targets are evaluated and transitionCondition is true
     */
    execute,
    /**
     * All targets are not yet evaluated
     */
    unresolved,
    /**
     * all targets are evaluated and transitionCondition is false
     */
    notexecute
  }

  /**
   * Execute method called by process virtual machine.
   *
   * @param execution current execution
   */
  public void execute(final ActivityExecution execution) {

    final BpelExecution bpelExecution = (BpelExecution) execution;

    this.beforeRunning(bpelExecution);

    final TargetBehaviour targetBehaviour = this.targetsOK(bpelExecution);
    if (targetBehaviour == TargetBehaviour.execute) {
      if (AbstractBpelActivity.log.isLoggable(Level.FINE)) {
        AbstractBpelActivity.log.fine("### Executing " + execution.getNode().getName() + " in execution " + execution);
      }
      this.executeActivity(bpelExecution);
    } else if (targetBehaviour != TargetBehaviour.notexecute) {
      bpelExecution.waitForSignal();
    }
  }

  /**
   * Signal method.
   *
   * @param execution execution
   * @param signal signal
   */
  @Override
  public void signal(final BpelExecution execution, final String signal, final Map<String, Object> parameters) {
    throw new OrchestraRuntimeException("Signal not supported.");
  }

  /**
   * Returns true if all target are evaluated and if joinCondition is true.
   *
   * @param execution execution
   * @return true if all target are evaluated and if joinCondition is true.
   */
  private TargetBehaviour targetsOK(final BpelExecution execution) {
    // verify that all links have been evaluated
    if (this.targets == null || this.targets.isEmpty()) {
      return TargetBehaviour.execute;
    }
    boolean tjcLogicalOR = false;
    for (final Target target : this.targets) {
      final String linkName = target.getLinkName();
      final FlowRuntime fr = execution.getFlowRuntimeWithLink(linkName);
      final Boolean linkStatus = fr.getLinkStatus(linkName);
      if (linkStatus == null) {
        // a link is not already evaluated
        if (AbstractBpelActivity.log.isLoggable(Level.FINE)) {
          AbstractBpelActivity.log.fine(this.getClass() + "." + this.getName() + " targetsOK, waiting for link : " + linkName);
        }
        // add this execution as a waiting execution
        fr.addWaitingExecution(linkName, execution);
        return TargetBehaviour.unresolved;
      } else {
        tjcLogicalOR = tjcLogicalOR || linkStatus;
      }
    }
    boolean tjc = tjcLogicalOR;
    if (this.targetsJoinCondition != null) {
      try {
        tjc = this.targetsJoinCondition.getEvaluator().evaluateBoolean(execution);
      } catch (final Exception e) {
        // TODO throw joinFailure
        e.printStackTrace();
        this.setAllUnsetLinks(execution);
        throw BpelFaultException.generateBpelFaultException(
            BpelFault.joinFailure,
            "bpel:joinFailure : " + this.targetsJoinCondition.getText()
          + ", exception while evaluating the targetJoinCondition",
          null);
      }
    }

    if (!tjc) {
      if (AbstractBpelActivity.log.isLoggable(Level.FINE)) {
        AbstractBpelActivity.log.fine(this.getClass() + ".targetsOK : " + false);
      }
      if (!this.suppressJoinFailure) {
        if (this.targetsJoinCondition != null) {
          throw BpelFaultException.generateBpelFaultException(
              BpelFault.joinFailure,
              "bpel:joinFailure : " + this.targetsJoinCondition.getText()
            + ", targetJoinCondition is false and suppressJoinFailure is no",
            null);
        } else {
          throw BpelFaultException.generateBpelFaultException(
              BpelFault.joinFailure,
              "bpel:joinFailure : targetJoinCondition is false and suppressJoinFailure is no",
              null);
        }
      } else {
        if (this.sources != null) {
          for (final Source source : this.sources) {
            final String linkName = source.getLinkName();
            final FlowRuntime fr = execution.getFlowRuntimeWithLink(linkName);
            final LinkRuntime lr = fr.getLink(linkName);
            if (lr.getStatus() == null) {
              lr.setStatus(false);
            }
            final BpelExecution execToWakeUp = fr.getWaitingExecution(linkName);
            if (execToWakeUp != null) {
              execToWakeUp.execute(execToWakeUp.getNode());
            }
          }
        }
        return TargetBehaviour.notexecute;
      }
      // TODO
      // set outgoingLinks (sources) status to false according to links semantic
      // (in same scope or not...)
      // see transitionCondition
      // notify waitingExecutions
      // if suppressjoinFailure : nothing (do not execute activity)
      // if not suppressJoinFailure : throw bpel:joinfFailure
    }
    if (AbstractBpelActivity.log.isLoggable(Level.FINE)) {
      AbstractBpelActivity.log.fine(this.getClass() + "." + this.getName() + " targetsOK : " + true);
    }
    return TargetBehaviour.execute;
  }

  @Override
  public void afterRunned(final BpelExecution execution) {
    super.afterRunned(execution);
    if (this.executeSourcesAsync) {
      Environment.getFromCurrent(MessageSession.class).send(new ExecuteSourcesMessage(execution));
      execution.waitForSignal();
      execution.setState(Execution.STATE_ASYNC);
    } else {
      this.executeSources(execution);
    }
  }

  /**
   * Executes sources of this activity : called when this activity is finished.
   *
   * @param execution execution
   */
  public void executeSources(final BpelExecution execution) {
    // Sources are executed once the enclosing activity is completed.
    // In the special case of an enclosing scope, this is not required to have
    // a successful completion. In that case (enclosing scope abnormally
    // completed)
    // sources must be executed only if there is a faultHandler associated to
    // the
    // enclosing scope and if this faultHandler has terminated without throwing
    // a fault.

    // if an error occurs during the TC evaluation
    // This error does not prevent enclosing activity to be set at "completed".
    // This error must be managed by the enclosing scope.
    // If the target activity is inside the source activity enclosing scope,
    // UNSET must be set on the concerned link status. (nothing to do)
    // -> done
    // If the target activity is outside the source activity enclosing scope,
    // false must be set on the concerned link status.
    // -> done
    // If source activity is a scope, the error must be propagated to its
    // father.
    // All remaining outgoing links (sources) having target activity in source
    // activity enclosing scope
    // must not have their transitionCondition evaluated (nothing to do)
    // -> done
    // All remaining outgoing links (sources) having target activity outside
    // source activity enclosing scope
    // must not have their status set to false
    // -> done
    if (AbstractBpelActivity.log.isLoggable(Level.FINE)) {
      AbstractBpelActivity.log.fine("Executing sources of activity : " + this.getClass().getName() + "." + this.name);
    }
    if (this.sources == null) {
      return;
    }
    for (final Source source : this.sources) {
      final String linkName = source.getLinkName();
      boolean tc = true;
      if (source.getTransitionCondition() != null) {
        try {
          tc = source.getTransitionCondition().getEvaluator().evaluateBoolean(execution);
        } catch (final Exception e) {
          // TODO : throw error to enclosing scope : warning if the source
          // activity is a scope, propagate to parent
          this.setAllUnsetLinks(execution);
          e.printStackTrace();
        }
      }
      execution.setLinkStatus(linkName, tc);
      if (AbstractBpelActivity.log.isLoggable(Level.FINE)) {
        AbstractBpelActivity.log.fine("executeSources, notifying executions of the link : " + linkName + "(" + tc + ")");
      }
      final FlowRuntime fr = execution.getFlowRuntimeWithLink(linkName);
      final BpelExecution execToWakeUp = fr.getWaitingExecution(linkName);
      if (execToWakeUp != null) {
        execToWakeUp.execute(execToWakeUp.getNode());
        // execToWakeUp.execute(execToWakeUp.getNode(), "target");
      }
    }
  }

  /**
   * For every LinkRuntime having status UNSET, set it to false if target
   * activity is not defined in source activity enclosing scope.
   *
   * @param execution execution
   */
  private void setAllUnsetLinks(final BpelExecution execution) {
    if (this.sources != null) {
      for (final Source source : this.sources) {
        final String linkName = source.getLinkName();
        final FlowRuntime fr = execution.getFlowRuntimeWithLink(linkName);
        final LinkRuntime lr = fr.getLink(linkName);
        if (lr.getStatus() == null) {
          final Link link = source.getLink();
          final Scope targetActivityEnclosingScope = link.getTargetActivity().getEnclosingScope();
          final Scope sourceActivityEnclosingScope = link.getSourceActivity().getEnclosingScope();
          if (!(targetActivityEnclosingScope == sourceActivityEnclosingScope)) {
            lr.setStatus(false);
          }
        }
      }
    }
  }

  /**
   * This method must be implemented by every activity. This is the business
   * behaviour of the activity.
   *
   * @param execution execution
   */
  protected abstract void executeActivity(BpelExecution execution);

  public Scope getEnclosingScope() {
    return this.enclosingScope;
  }

  public void setEnclosingScope(final Scope enclosingScope) {
    this.enclosingScope = enclosingScope;
  }

  public boolean isSuppressJoinFailure() {
    return this.suppressJoinFailure;
  }

  public void setSuppressJoinFailure(final boolean suppressJoinFailure) {
    this.suppressJoinFailure = suppressJoinFailure;
  }

  public List<Source> getSources() {
    return this.sources;
  }

  public void setSources(final List<Source> sources) {
    this.sources = sources;
  }

  public List<Target> getTargets() {
    return this.targets;
  }

  public void setTargets(final List<Target> targets) {
    this.targets = targets;
  }

  public JoinCondition getTargetsJoinCondition() {
    return this.targetsJoinCondition;
  }

  public void setTargetsJoinCondition(final JoinCondition targetsJoinCondition) {
    this.targetsJoinCondition = targetsJoinCondition;
  }
}
