/**
 * Copyright (C) 2009 Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/

package org.ow2.orchestra.definition.activity;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

import org.ow2.orchestra.exception.FaultWithMessageVariable;
import org.ow2.orchestra.facade.exception.BpelFaultException;
import org.ow2.orchestra.facade.exception.FaultWithElement;
import org.ow2.orchestra.facade.runtime.ScopeState;
import org.ow2.orchestra.facade.uuid.ActivityDefinitionUUID;
import org.ow2.orchestra.pvm.internal.model.NodeImpl;
import org.ow2.orchestra.pvm.model.Node;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.runtime.ScopeRuntime;
import org.ow2.orchestra.util.Misc;
import org.w3c.dom.Element;

/**
 * @author S. Ali Tokmen
 */
public abstract class BaseCatchActivity extends ActivityWithSingleChild {

  private static final long serialVersionUID = 660890927633975015L;

  private static Logger log = Logger.getLogger(BaseCatchActivity.class.getName());

  protected BaseCatchActivity(final NodeImpl child, final ActivityDefinitionUUID uuid, final String name) {
    super(child, uuid, name);
  }

  @Override
  public void signal(final BpelExecution execution) {
    final ScopeRuntime scopeRuntime = execution.getCurrentScopeRuntime();
    scopeRuntime.setState(ScopeState.COMPLETED_UNSUCCESSFUL);
    scopeRuntime.setLastException(null);
  }

  public static Node findFaultHandler(final ScopeRuntime scopeRuntime, final List<Node> nodes,
    final BpelFaultException bpelFaultException) {
    if (nodes.size() > Scope.BIGGEST_ACTIVITY_INDEX) {
      final Map<Node, CatchActivity> catches = new HashMap<Node, CatchActivity>();
      for (int i = Scope.BIGGEST_ACTIVITY_INDEX; i < nodes.size(); i++) {
        final NodeImpl node = (NodeImpl) nodes.get(i);
        final CatchActivity activity = (CatchActivity) node.getBehaviour();
        if (activity == null) {
          Misc.unreachableStatement("The behaviour of a catch node cannot be null");
        }
        catches.put(node, activity);
      }

      if (bpelFaultException instanceof FaultWithMessageVariable) {
        final org.ow2.orchestra.var.MessageVariable bfm = ((FaultWithMessageVariable) bpelFaultException).getFaultMessage();
        if (bfm != null && bfm.getParts() != null && bfm.getParts().size() != 0) {
          // Fault thrown with associated data
          // (wsbpel-v2.0-OS, page 129, part 2)

          if (bpelFaultException.getFaultName() != null) {
            for (final Map.Entry<String, Element> part : bfm.getParts().entrySet()) {
              if (part.getValue() != null) {
                for (final Map.Entry<Node, CatchActivity> catch_ : catches.entrySet()) {
                  final CatchActivity cd = catch_.getValue();
                  final QName type = cd.getFaultMessageType();
                  if (bpelFaultException.getFaultName().equals(cd.getFaultName()) && type != null
                    && type.getNamespaceURI().equals(part.getValue().getNamespaceURI())
                    && type.getLocalPart().equals(part.getValue().getLocalName())) {
                    // wsbpel-v2.0-OS, page 129, part 2, case 2
                    return catch_.getKey();
                  }
                }
              }
            }

            for (final Map.Entry<Node, CatchActivity> catch_ : catches.entrySet()) {
              final CatchActivity cd = catch_.getValue();
              if (bpelFaultException.getFaultName().equals(cd.getFaultName()) && cd.getFaultVariable() == null) {
                // wsbpel-v2.0-OS, page 129, part 2, case 3
                return catch_.getKey();
              }
            }
          }

          // Starting from here, fault.getFaultName() can be null

          for (final Map.Entry<String, Element> part : bfm.getParts().entrySet()) {
            if (part.getValue() != null) {
              for (final Map.Entry<Node, CatchActivity> catch_ : catches.entrySet()) {
                final CatchActivity cd = catch_.getValue();
                final QName type = cd.getFaultMessageType();
                if (cd.getFaultName() == null && type != null
                  && type.getNamespaceURI().equals(part.getValue().getNamespaceURI())
                  && type.getLocalPart().equals(part.getValue().getLocalName())) {
                  // wsbpel-v2.0-OS, page 129, part 2, case 5
                  // this is the same as case 2 but without the faultName
                  return catch_.getKey();
                }
              }
            }
          }
        }
      } else if (bpelFaultException instanceof FaultWithElement) {
        final Element bfe = ((FaultWithElement) bpelFaultException).getFaultElement();
        if (bfe != null) {
          // Fault thrown with associated data
          // (wsbpel-v2.0-OS, page 129, part 2)

          if (bpelFaultException.getFaultName() != null) {
            for (final Map.Entry<Node, CatchActivity> catch_ : catches.entrySet()) {
              final CatchActivity cd = catch_.getValue();
              final QName type = cd.getFaultElement();
              if (bpelFaultException.getFaultName().equals(cd.getFaultName()) && type != null
                && type.getNamespaceURI().equals(bfe.getNamespaceURI())
                && type.getLocalPart().equals(bfe.getLocalName())) {
                // wsbpel-v2.0-OS, page 129, part 2, case 1
                return catch_.getKey();
              }
            }

            for (final Map.Entry<Node, CatchActivity> catch_ : catches.entrySet()) {
              final CatchActivity cd = catch_.getValue();
              if (bpelFaultException.getFaultName().equals(cd.getFaultName()) && cd.getFaultVariable() == null) {
                // wsbpel-v2.0-OS, page 129, part 2, case 3
                return catch_.getKey();
              }
            }
          }

          // Starting from here, fault.getFaultName() can be null

          for (final Map.Entry<Node, CatchActivity> catch_ : catches.entrySet()) {
            final CatchActivity cd = catch_.getValue();
            final QName type = cd.getFaultElement();
            if (cd.getFaultName() == null && type != null
              && type.getNamespaceURI().equals(bfe.getNamespaceURI())
              && type.getLocalPart().equals(bfe.getLocalName())) {
              // wsbpel-v2.0-OS, page 129, part 2, case 4
              // this is the same as case 1 but without the faultName
              return catch_.getKey();
            }
          }
        }
      }

      // Fault thrown without associated data
      // (wsbpel-v2.0-OS, page 129, part 1)
      if (bpelFaultException.getFaultName() != null) {
        for (final Map.Entry<Node, CatchActivity> catch_ : catches.entrySet()) {
          final CatchActivity cd = catch_.getValue();
          if (bpelFaultException.getFaultName().equals(cd.getFaultName())) {
            // wsbpel-v2.0-OS, page 129, part 1, case 1
            return catch_.getKey();
          }
        }
      }
    }

    return null;
  }
}
