/**
 * Copyright (C) 2009  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.definition.activity;

import java.util.ListIterator;
import java.util.Map;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.facade.data.def.CompensateActivityDefinitionData;
import org.ow2.orchestra.facade.exception.BpelFaultException;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.facade.runtime.ScopeState;
import org.ow2.orchestra.pvm.model.Node;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.runtime.ScopeRuntime;
import org.ow2.orchestra.util.Misc;

/**
 * the runtime behaviour of the compensate activity.
 *
 * @author S. Ali Tokmen
 */
public class Compensate extends AbstractBpelActivity {

  private static final long serialVersionUID = 1587165958245350103L;

  public static final String EXCEPTION_PARAMETER = "exception";

  public Compensate() {
    super();
  }

  /**
   * Handles compensation.
   *
   * @param execution Current execution.
   */
  @Override
  public void executeActivity(final BpelExecution execution) {
    // This activity compensates the immediately
    // enclosed scopes of the scope associated with the FCT-Handler
    // Get execution/scope associated with the FCT handler
    final BpelExecution scopeExecution = this.findCompensateParentScopeExecution(execution);

    // Find one immediately enclosed scope to compensate
    final ScopeRuntime sr = scopeExecution.getScopeRuntime();
    if (!sr.getEndedChildren().isEmpty()) {
      // iterate over the list of scope runtime in the reverse order
      final ListIterator<ScopeRuntime> srIterator = sr.getEndedChildren().listIterator(sr.getEndedChildren().size());
      while (srIterator.hasPrevious()) {
        final ScopeRuntime enclosedScopeRuntime = srIterator.previous();
        if (enclosedScopeRuntime.getState().equals(ScopeState.COMPLETED_SUCCESSFUL)) {
          // we found a child to compensate -> compensate it
          final BpelExecution childExecution = (BpelExecution) scopeExecution.createExecution();
          childExecution.pushScopeRuntime(enclosedScopeRuntime);
          childExecution.setActivityInstanceUUID(enclosedScopeRuntime.getActivityInstanceUUID());

          enclosedScopeRuntime.setCompensateExecution(execution);

          final Node enclosedScopeNode = enclosedScopeRuntime.getScopeDefinition().getScopeNode();
          final Node chNode = enclosedScopeNode.getNodes().get(Scope.COMPENSATION_ACTIVITY);

          childExecution.createExecution().execute(chNode);
          execution.waitForSignal();
          return;
        }
      }
    }
  }

  @Override
  public void signal(final BpelExecution execution, final String signal, final Map<String, Object> parameters) {
    if (parameters != null) {
      final BpelFaultException exception = (BpelFaultException) parameters.get(Compensate.EXCEPTION_PARAMETER);
      if (exception != null) {
        throw exception;
      }
    }
    this.executeActivity(execution);
  }

  /**
   * Compensate can only be used in catch, catchAll, compensationHandler,
   * terminationHandler<br/>
   * <br/>
   * The scope associated must be in state: terminating / handling fault /
   * compensating<br/>
   * <br/>
   * This method returns the execution of the scope associated.
   */
  private BpelExecution findCompensateParentScopeExecution(final BpelExecution execution) {
    if (execution == null) {
      Misc.unreachableStatement("No scope is currently handling fault, terminating or compensating");
    }
    final ScopeRuntime sr = execution.getScopeRuntime();
    if (sr != null) {
      if (sr.getState().equals(ScopeState.HANDLING_FAULT) || sr.getState().equals(ScopeState.COMPENSATING)
        || sr.getState().equals(ScopeState.TERMINATING)) {
        return execution;
      }
    }
    return this.findCompensateParentScopeExecution(execution.getParent());
  }

  @Override
  public ActivityType getType() {
    return ActivityType.COMPENSATE;
  }

  @Override
  public CompensateActivityDefinitionData createDefinitionData(final BpelProcess bpelProcess) {
    return new CompensateActivityDefinitionData(bpelProcess.getProcessDefinitionData(), this.uuid, null, this.name);
  }

}
