/**
 * Copyright (C) 2009  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.definition.activity;

import java.util.ListIterator;
import java.util.Map;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.facade.data.def.CompensateScopeActivityDefinitionData;
import org.ow2.orchestra.facade.exception.BpelFaultException;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.facade.runtime.ScopeState;
import org.ow2.orchestra.pvm.internal.model.NodeImpl;
import org.ow2.orchestra.pvm.model.Node;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.runtime.ScopeRuntime;
import org.ow2.orchestra.util.Misc;

/**
 * the runtime behaviour of the compensateScope activity.
 *
 * @author S. Ali Tokmen
 */
public class CompensateScope extends AbstractBpelActivity {

  private static final long serialVersionUID = 5427405633738152258L;

  private String target;

  public CompensateScope() {
    super();
  }

  /**
   * Handles compensation of a scope.
   *
   * @param execution Current execution.
   */
  @Override
  public void executeActivity(final BpelExecution execution) {
    // Find the scope which contains the immediately enclosed scope named
    // "target" that we need to compensate
    final BpelExecution callerScopeExecution = this.findCompensateScopeParentScopeExecution(execution);

    // This activity compensates the immediately
    // enclosed scopes of the scope associated with the FCT-Handler
    // Get execution/scope associated with the FCT handler
    final BpelExecution scopeExecution = this.findCompensateParentScopeExecution(callerScopeExecution);

    // Find one immediately enclosed scope to compensate
    final ScopeRuntime sr = scopeExecution.getScopeRuntime();
    if (!sr.getEndedChildren().isEmpty()) {
      // iterate over the list of scope runtime in the reverse order
      final ListIterator<ScopeRuntime> srIterator = sr.getEndedChildren().listIterator(sr.getEndedChildren().size());
      while (srIterator.hasPrevious()) {
        final ScopeRuntime enclosedScopeRuntime = srIterator.previous();
        final Node enclosedScopeNode = enclosedScopeRuntime.getScopeDefinition().getScopeNode();
        final String enclosedScopeName = enclosedScopeRuntime.getScopeDefinition().getName();

        if (target.equals(enclosedScopeName)) {
          if (enclosedScopeRuntime.getState().equals(ScopeState.COMPLETED_SUCCESSFUL)) {
            // we found a child to compensate -> compensate it
            final BpelExecution childExecution = (BpelExecution) scopeExecution.createExecution();
            childExecution.pushScopeRuntime(enclosedScopeRuntime);
            childExecution.setActivityInstanceUUID(enclosedScopeRuntime.getActivityInstanceUUID());

            enclosedScopeRuntime.setCompensateExecution(execution);

            final Node chNode = enclosedScopeNode.getNodes().get(Scope.COMPENSATION_ACTIVITY);

            childExecution.createExecution().execute(chNode);
            execution.waitForSignal();
          }
          return;
        }
      }
    }

    throw new OrchestraRuntimeException("The scope \"" + sr.getScopeDefinition().getName()
      + "\" doesn't have any immediate terminated children called \"" + target + "\"");
  }

  @Override
  public void signal(final BpelExecution execution, final String signal, final Map<String, Object> parameters) {
    if (parameters != null) {
      final BpelFaultException exception = (BpelFaultException) parameters.get(Compensate.EXCEPTION_PARAMETER);
      if (exception != null) {
        throw exception;
      }
    }
    this.executeActivity(execution);
  }

  /**
   * The compensateScope's parent scope execution is the activity with type
   * that's a catch, catchAll, terminationHandler or compensationHandler
   */
  private BpelExecution findCompensateScopeParentScopeExecution(final BpelExecution execution) {
    if (execution == null) {
      Misc.unreachableStatement("No parent scope is a fault, termination or compensation handler");
    }
    final NodeImpl executionNode = execution.getNode();
    final AbstractActivity executionActivity = (AbstractActivity) executionNode.getBehaviour();
    final ActivityType executionActivityType = executionActivity.getType();
    if (executionActivityType == ActivityType.CATCH_HANDLER || executionActivityType == ActivityType.CATCHALL_HANDLER
      || executionActivityType == ActivityType.TERMINATION_HANDLER
      || executionActivityType == ActivityType.COMPENSATION_HANDLER) {
      return execution;
    }
    return this.findCompensateScopeParentScopeExecution(execution.getParent());
  }

  /**
   * Compensate can only be used in catch, catchAll, compensationHandler,
   * terminationHandler<br/>
   * <br/>
   * The scope associated must be in state: terminating / handling fault /
   * compensating<br/>
   * <br/>
   * This method returns the execution of the scope associated.
   */
  private BpelExecution findCompensateParentScopeExecution(final BpelExecution execution) {
    if (execution == null) {
      Misc.unreachableStatement("No scope is currently handling fault, terminating or compensating");
    }
    final ScopeRuntime sr = execution.getScopeRuntime();
    if (sr != null) {
      if (sr.getState().equals(ScopeState.HANDLING_FAULT) || sr.getState().equals(ScopeState.COMPENSATING)
        || sr.getState().equals(ScopeState.TERMINATING)) {
        return execution;
      }
    }
    return this.findCompensateParentScopeExecution(execution.getParent());
  }

  @Override
  public ActivityType getType() {
    return ActivityType.COMPENSATE_SCOPE;
  }

  @Override
  public CompensateScopeActivityDefinitionData createDefinitionData(final BpelProcess bpelProcess) {
    return new CompensateScopeActivityDefinitionData(bpelProcess.getProcessDefinitionData(), this.uuid, null, this.name,
      target);
  }

  public void setTarget(final String target) {
    this.target = target;
  }

  public String getTarget() {
    return target;
  }

}
