/**
 * Copyright (C) 2009  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.definition.activity;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.datatype.Duration;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.element.InboundMessageElement;
import org.ow2.orchestra.definition.element.OnAlarm;
import org.ow2.orchestra.definition.element.OnEvent;
import org.ow2.orchestra.facade.data.runtime.TimerData;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.facade.runtime.ScopeState;
import org.ow2.orchestra.pvm.Execution;
import org.ow2.orchestra.pvm.activity.ActivityExecution;
import org.ow2.orchestra.pvm.internal.model.ExecutionImpl;
import org.ow2.orchestra.pvm.internal.model.NodeImpl;
import org.ow2.orchestra.pvm.model.OpenExecution;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.services.MessageCarrier;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.services.Receiver;
import org.ow2.orchestra.services.ReceivingElement;
import org.ow2.orchestra.services.SignalExecUtil;
import org.ow2.orchestra.util.EnvTool;
import org.ow2.orchestra.var.MessageVariable;


/**
 * @author Guillaume Porcher
 *
 */
public class EventHandlerActivity extends AbstractActivity implements InboundMessageElement {

  public static final String ON_ALARM_PARAM = "onAlarm";

  protected List<OnEvent> onEvents = new ArrayList<OnEvent>();
  protected List<OnAlarm> onAlarms = new ArrayList<OnAlarm>();

  @Override
  public ActivityType getType() {
    return ActivityType.SCOPE_EVENT_HANDLER;
  }

  public List<ReceivingElement> getReceivingElements() {
    return new ArrayList<ReceivingElement>(this.onEvents);
  }

  @Override
  public void signal(final BpelExecution execution, final String signal, final Map<String, Object> parameters) {
    if (parameters != null) {
      final OnEvent onEvent = (OnEvent) parameters.get(SignalExecUtil.RECEIVING_ELEMENT_PARAM);
      final NodeImpl onAlarmActivity = (NodeImpl) parameters.get(EventHandlerActivity.ON_ALARM_PARAM);
      if (execution.getCurrentScopeRuntime().getState().equals(ScopeState.RUNNING)) {
        // Enable other waiting executions
        execution.getProcessInstance().getWaitingExecutions().add(execution);
      }
      if (onEvent != null) {
        final MessageVariable message = (MessageVariable) parameters.get(SignalExecUtil.MESSAGE_PARAM);
        final MessageCarrier messageCarrier = (MessageCarrier) parameters.get(SignalExecUtil.MESSAGE_CARRIER_PARAM);
        final NodeImpl node = onEvent.getActivity();
        // execute onMessage activity in a child execution
        final ExecutionImpl child = execution.createExecution();
        // set this execution as active to allow concurrent events
        execution.unlock();
        child.createVariable(EventHandlerScope.MESSAGE, message);
        child.createVariable(EventHandlerScope.MESSAGE_CARRIER, messageCarrier);
        child.createVariable(EventHandlerScope.ON_EVENT, onEvent);
        child.execute(node);
      } else if (onAlarmActivity != null) {
        // execute onAlarm activity in a child execution
        final ExecutionImpl child = execution.createExecution();
        // set this execution as active to allow concurrent events
        execution.unlock();
        child.execute(onAlarmActivity);
      } else {
        throw new OrchestraRuntimeException("Unsupported signal parameters");
      }
    } else if (signal == null) {
      // this is the child execution returning
      // end child execution
      execution.end(Execution.STATE_ENDED);

      final BpelExecution parentExecution = execution.getParent();

      // remove finished execution from parent
      parentExecution.removeExecution(execution);

      // parentExecution is active only when all children are finished
      if (parentExecution.getExecutions().isEmpty()) {
        // signal must be non-null to make the distinction with child executions
        // in the first line of this method
        parentExecution.signal("all-finished");
      }
    } else if (signal.equals("all-finished")) {
      final Collection<OpenExecution> scopeExecutions = execution.getParent().getExecutions();
      if (scopeExecutions.size() > 1) {
        // scope main activity is still running
        execution.waitForSignal();
      } else {
        // scope is ended, only event handler execution
        this.cancelEvents(execution);
      }
    } else if (signal.equals(Scope.SCOPE_FINISHED_SIGNAL)) {
      this.cancelEvents(execution);
    } else {
      boolean found = false;
      for (final OnAlarm onAlarm : this.onAlarms) {
        if (signal.equals(onAlarm.getUuid())) {
          found = true;
          final Map<String, Object> params = new HashMap<String, Object>();
          params.put(EventHandlerActivity.ON_ALARM_PARAM, onAlarm.getActivity());
          this.signal(execution, null, params);
        }
      }
      if (!found) {
        throw new OrchestraRuntimeException("Unsupported signal " + signal);
      }
    }
  }

  public void execute(final ActivityExecution execution) throws Exception {
    if (this.onEvents.isEmpty() && this.onAlarms.isEmpty()) {
      // nothing to do
      execution.end(Execution.STATE_ENDED);

      final BpelExecution parentExecution = (BpelExecution) execution.getParent();

      // remove finished execution from parent
      parentExecution.removeExecution(execution);
      return;
    }
    this.registerEvents((BpelExecution) execution);
  }


  public void registerEvents(final BpelExecution execution) {
    // register on events
    // add to list of child executions waiting for a signal
    execution.getProcessInstance().addWaitingExecution(execution);
    // register on alarms
    if (this.onAlarms != null) {
      for (final OnAlarm onAlarm : this.onAlarms) {
        final Date date = OnAlarm.computeTimer(execution,
            onAlarm.getDeadlineExpression(),
            onAlarm.getDurationExpression(),
            onAlarm.getRepeatEveryExpression());
        String repeat = null;
        if (onAlarm.getRepeatEveryExpression() != null) {
          final Duration repeatDuration = onAlarm.getRepeatEveryExpression().getEvaluator().evaluateDuration(execution);
          final GregorianCalendar d = new GregorianCalendar();
          d.setTimeInMillis(0);
          repeatDuration.addTo(d);
          final long repeatMillis = d.getTimeInMillis();
          repeat = Long.toString(repeatMillis) + " milliseconds";
        }
        execution.createTimer(null, onAlarm.getUuid(), null, date, repeat, null, null);
        EnvTool.getRecorder().recordTimerStarted(new TimerData(execution.getActivityInstanceUUID(), null, date));
      }
    }
    // wait for signal
    execution.waitForSignal();
  }

  public void cancelEvents(final BpelExecution execution) {
    // end of event handler:
    // cancel alarms
    execution.getJobs().clear();

    // remove onEvents
    final BpelProcess process = execution.getProcessDefinition();
    for (final OnEvent onEvt : this.onEvents) {
      final OperationKey opKey = onEvt.getOperationKey();
      final Receiver receiver = process.getReceiver(opKey);
      receiver.removeWaitingExecution(execution);
    }
    // Remove from list of child executions waiting for a signal
    execution.getProcessInstance().getWaitingExecutions().remove(execution);
  }

  /**
   * @param onAlarms the onAlarms to set
   */
  public void setOnAlarms(final List<OnAlarm> onAlarms) {
    this.onAlarms = onAlarms;
  }


  /**
   * @param onEvents the onEvents to set
   */
  public void setOnEvents(final List<OnEvent> onEvents) {
    this.onEvents = onEvents;
  }
}
