/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.definition.activity;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.ow2.orchestra.definition.element.Link;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.pvm.Execution;
import org.ow2.orchestra.pvm.model.Node;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.runtime.FlowRuntime;

/**
 * the runtime behaviour of the flow activity.
 *
 * @author Charles Souillard, Tom Baeyens, Goulven Le Jeune, Guillaume Porcher
 * @version $Id: Flow.java 2901 2009-09-01 10:31:20Z porcherg $
 */
public class Flow extends AbstractBpelActivity {

  /** serialVersionUID. */
  private static final long serialVersionUID = 1L;

  /** list of links of the flow. */
  protected Map<String, Link> links = null;

  public Flow() {
    super();
  }


  /**
   * business logic.
   *
   * @param execution
   *            execution
   */
  @Override
  public void executeActivity(final BpelExecution execution) {
    execution.pushFlowRuntime(new FlowRuntime(this.links, execution.getActivityInstanceUUID()));
    final List<Node> nesteNodes = execution.getNode().getNodes();

    if (nesteNodes != null && !nesteNodes.isEmpty()) {
      final Map<Node, BpelExecution> childs = new HashMap<Node, BpelExecution>();
      // create all child executions (because we check the child executions for flow completion)
      for (final Node child : nesteNodes) {
        final BpelExecution childExecution = (BpelExecution) execution.createExecution();
        childs.put(child, childExecution);
      }
      // execute child nodes
      for (final Entry<Node,BpelExecution> child : childs.entrySet()) {
        child.getValue().execute(child.getKey());
      }
    }
  }

  /**
   * Signal method.
   *
   * @param execution
   *            execution
   * @param signal
   *            signal
   */
  @Override
  public void signal(final BpelExecution execution, final String signal, final Map<String, Object> parameters) {
    if (signal == null) {
      // end the execution
      execution.end(Execution.STATE_ENDED);

      // this is the child execution returning
      final BpelExecution flowExecution = execution.getParent();
      // remove finished execution
      flowExecution.removeExecution(execution);

      // flowExecution is active only when all children are finished
      if (flowExecution.isActive()) {
        // signal must be non-null to make the distinction with child executions
        // in the first line of this method
        flowExecution.signal("all-finished");
      }
    } else {
      this.afterRunned(execution);
    }
  }

  public Map<String, Link> getLinks() {
    return this.links;
  }

  /**
   * @param linkName
   *            linkName
   * @return the link having the given name
   */
  public Link getLink(final String linkName) {
    return this.links.get(linkName);
  }

  public void setLinks(final Map<String, Link> links) {
    this.links = links;
  }

  @Override
  public ActivityType getType() {
    return ActivityType.FLOW;
  }
}
