/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.definition.activity;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.ow2.orchestra.facade.data.runtime.ForeachInitializedData;
import org.ow2.orchestra.facade.exception.BpelFaultException;
import org.ow2.orchestra.facade.exception.BpelFaultException.BpelFault;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.facade.runtime.ScopeState;
import org.ow2.orchestra.lang.Expression;
import org.ow2.orchestra.pvm.Execution;
import org.ow2.orchestra.pvm.model.Node;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.runtime.ForEachRuntime;
import org.ow2.orchestra.util.EnvTool;
import org.ow2.orchestra.util.Misc;

/**
 * the runtime behaviour of the forEach activity.
 *
 * @author Tom Baeyens, Charles Souillard, Guillaume Porcher
 */
public class ForEach extends AbstractBpelActivity {

  /** serialVersionUID. */
  private static final long serialVersionUID = 1L;

  private String counterName;

  private boolean parallel;

  private Expression startCounterValue;

  private Expression finalCounterValue;

  private Expression completionCondition;

  private boolean successfulBranchesOnly = false;

  public ForEach() {
    super();
  }

  @Override
  public void signal(final BpelExecution execution, final String signal,
      final Map<String, Object> parameters) {
    if (signal == null) {
      // this is the child execution returning

      // check if scope was successfully completed
      final boolean isSuccessful = execution.getVariable(BpelExecution.SCOPE_STATE_VARIABLE).equals(ScopeState.COMPLETED_SUCCESSFUL);
      execution.removeVariable(BpelExecution.SCOPE_STATE_VARIABLE);
      // end child execution
      execution.end(Execution.STATE_ENDED);

      final BpelExecution foreachExecution = execution.getParent();
      // remove finished execution from parent
      foreachExecution.removeExecution(execution);

      final ForEachRuntime forEachRuntime = foreachExecution.getForEachRuntime();
      // count one more branch as successful
      forEachRuntime.addTerminatedBranch(isSuccessful);
      // check if completion condition is true
      if (!forEachRuntime.isCompleted()) {
        // completion condition is not reached
        if (!this.parallel) {
          // Sequential ForEach
          // TODO: check if remaining executions can set completion condition to true
          // throw completionConditionFailure otherwise
          if (!this.sequentialFor(foreachExecution, forEachRuntime)) {
            // if false, no child has been created
            // it means completion condition can never be true
            throw BpelFaultException.generateBpelFaultException(BpelFault.completionConditionFailure);
          }
        } else {
          // Parallel ForEach
          // foreachExecution is active only when all children are finished
          if (foreachExecution.isActive()) {
            // if this code is executed, it means completion condition can never be true
            throw BpelFaultException.generateBpelFaultException(BpelFault.completionConditionFailure);
          }
        }
      } else {
        // Completion condition is true
        // stop for each execution
        foreachExecution.terminate();
      }
    } else if (signal.equals("all-finished")) {
      this.afterRunned(execution);
    }
  }

  /**
   * business logic.
   *
   * @param execution
   *          execution
   */
  @Override
  public void executeActivity(final BpelExecution execution) {
    // Evaluate start counter value
    int startCounterRuntimeValue;
    try {
      startCounterRuntimeValue = this.startCounterValue.getEvaluator()
          .evaluateUnsignedInt(execution);
    } catch (final Exception e) {
      throw BpelFaultException.generateBpelFaultException(
          BpelFault.invalidExpressionValue,
          "startCounterValue : " + this.startCounterValue,
          e);
    }
    // Evaluate final counter value
    int finalCounterRuntimeValue;
    try {
      finalCounterRuntimeValue = this.finalCounterValue.getEvaluator()
          .evaluateUnsignedInt(execution);
    } catch (final Exception e) {
      throw BpelFaultException.generateBpelFaultException(
          BpelFault.invalidExpressionValue,
          "finalCounterValue : " + this.finalCounterValue,
          e);
    }

    // For each scope should execute n = N + 1 times
    final int n = Math.max(0, finalCounterRuntimeValue - startCounterRuntimeValue + 1);

    Integer completionConditionValue = null;
    // Evaluate completion condition B
    int completionConditionRuntime;
    if (this.completionCondition != null) {
      try {
        completionConditionRuntime = this.completionCondition.getEvaluator()
            .evaluateUnsignedInt(execution);
      } catch (final Exception e) {
        throw BpelFaultException.generateBpelFaultException(
            BpelFault.invalidExpressionValue,
            "completionCondition : " + this.completionCondition,
            e);
      }
      // Check B > N + 1
      // reject with bpel:invalidBranchCondition otherwise
      if (completionConditionRuntime > n) {
        throw BpelFaultException.generateBpelFaultException(BpelFault.invalidBranchCondition);
      }
      completionConditionValue = completionConditionRuntime;
    } else {
      // when no completion condition is set,
      // wait for all branches (N + 1)
      completionConditionRuntime = n;
    }

    // Create ForEach runtime
    final ForEachRuntime forEachRuntime = new ForEachRuntime(startCounterRuntimeValue,
        finalCounterRuntimeValue, completionConditionRuntime, this.successfulBranchesOnly);
    EnvTool.getRecorder().recordForeachInitialized(
      new ForeachInitializedData(execution.getActivityInstanceUUID(), startCounterRuntimeValue, finalCounterRuntimeValue,
        completionConditionValue));

    execution.pushForEachRuntime(forEachRuntime);

    // set execution to wait state
    execution.waitForSignal();

    // Execute for each scopes if completion condition is false
    if (completionConditionRuntime != 0) {
      if (this.parallel) {
        // launch all executions in parallel
        this.parallelFor(execution, forEachRuntime);
      } else {
        // launch one execution
        if (!this.sequentialFor(execution, forEachRuntime)) {
          // if returns false, no execution has been created
          // this is a bug
          Misc.unreachableStatement();
        }
      }
    } else {
      // if completion condition is true (or n = 0)
      // signal execution (no scope execution to create)
      execution.signal("all-finished");
    }
  }

  /**
   *
   * @param execution parent execution
   * @return a child execution
   */
  private BpelExecution createChildExecution(final BpelExecution execution, final int counterValue) {
    // create ADDITIONNAL_SCOPE_VARIABLES. The enclosed scope will use variables defined in
    // this map to add variables in the scope.
    // The variables are previously defined in binding.
    final Map<String, Object> additionnalVariables = new HashMap<String, Object>();
    additionnalVariables.put(this.counterName, counterValue);

    final BpelExecution childExecution = (BpelExecution) execution.createExecution();
    childExecution.createVariable(BpelExecution.ADDITIONNAL_SCOPE_VARIABLES, additionnalVariables);
    return childExecution;
  }


  /**
   * @param execution the execution
   * @param forEachRuntime the foreachRuntime
   * @return true if a child execution has been created
   */
  private boolean sequentialFor(final BpelExecution execution,
      final ForEachRuntime forEachRuntime) {
    final Node node = execution.getNode();
    final int counterValue = forEachRuntime.getNextCurrentCounterValue();
    if (counterValue != ForEachRuntime.NO_MORE_COUNTER_VALUE) {
      final Node scope = node.getNodes().get(0);
      // go to enclosed scope
      this.createChildExecution(execution, counterValue).execute(scope);
      return true;
    }
    return false;
  }

  private void parallelFor(final BpelExecution execution, final ForEachRuntime forEachRuntime) {
    final Node node = execution.getNode();
    final Node scope = node.getNodes().get(0);

    final List<BpelExecution> childBpelExecutions = new ArrayList<BpelExecution>();

    // create each child executions.
    int counterValue = forEachRuntime.getNextCurrentCounterValue();
    while (counterValue != ForEachRuntime.NO_MORE_COUNTER_VALUE) {
      childBpelExecutions.add(this.createChildExecution(execution, counterValue));
      counterValue = forEachRuntime.getNextCurrentCounterValue();
    }

    // launch child executions.
    for (final BpelExecution childBpelExecution : childBpelExecutions) {
      if (execution.getExecutions().contains(childBpelExecution)) {
        childBpelExecution.execute(scope);
      }
    }

  }

  public String getCounterName() {
    return this.counterName;
  }

  public void setCounterName(final String counterName) {
    this.counterName = counterName;
  }

  public boolean isParallel() {
    return this.parallel;
  }

  public void setParallel(final boolean parallel) {
    this.parallel = parallel;
  }

  public Expression getStartCounterValue() {
    return this.startCounterValue;
  }

  public void setStartCounterValue(final Expression startCounterValue) {
    this.startCounterValue = startCounterValue;
  }

  public Expression getFinalCounterValue() {
    return this.finalCounterValue;
  }

  public void setFinalCounterValue(final Expression finalCounterValue) {
    this.finalCounterValue = finalCounterValue;
  }

  public Expression getCompletionCondition() {
    return this.completionCondition;
  }

  public void setCompletionCondition(final Expression completionCondition) {
    this.completionCondition = completionCondition;
  }

  public boolean isSuccessfulBranchesOnly() {
    return this.successfulBranchesOnly;
  }

  public void setSuccessfulBranchesOnly(final boolean successfulBranchesOnly) {
    this.successfulBranchesOnly = successfulBranchesOnly;
  }

  @Override
  public ActivityType getType() {
    return ActivityType.FOR_EACH;
  }

}
