/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.definition.activity;

import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.orchestra.definition.element.Else;
import org.ow2.orchestra.definition.element.Elseif;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.lang.Expression;
import org.ow2.orchestra.pvm.Execution;
import org.ow2.orchestra.pvm.internal.model.NodeImpl;
import org.ow2.orchestra.runtime.BpelExecution;

/**
 * the runtime behaviour of the if activity.
 *
 * @author Tom Baeyens, Charles Souillard
 */
public class If extends AbstractBpelActivity {

  /** serialVersionUID. */
  private static final long serialVersionUID = 1L;
  private static Logger log = Logger.getLogger(If.class.getName());

  protected Expression condition;
  protected NodeImpl activity;
  protected List<Elseif> elseifs;
  protected Else elseField;

  public If() {
    super();
  }
  /**
   * business logic.
   *
   * @param execution execution
   */
  @Override
  public void executeActivity(final BpelExecution execution) {
    boolean ifResult = condition.getEvaluator().evaluateBoolean(execution);
    if (ifResult) {
      if (log.isLoggable(Level.FINE)) {
        log.fine("if condition : " + condition + " evaluated to true");
      }
      execution.createExecution().execute(activity);
    } else {
      // manage elseifs
      for (Elseif elseif : elseifs) {
        boolean elseifResult = elseif.getCondition().getEvaluator().evaluateBoolean(execution);
        if (elseifResult) {
          if (log.isLoggable(Level.FINE)) {
            log.fine("elseif condition : " + elseif.getCondition() + " evaluated to true");
          }
          execution.createExecution().execute(elseif.getActivity());
          return;
        }
      }
      // execute else
      if (elseField != null) {
        if (log.isLoggable(Level.FINE)) {
          log.fine("Executing else...");
        }
        execution.createExecution().execute(elseField.getActivity());
      }
    }
  }

  /**
   * Signal method of the NodeBehaviour.
   *
   * @param execution execution
   * @param signal signal
   */
  @Override
  public void signal(final BpelExecution execution, final String signal, final Map<String, Object> parameters) {
    if (signal == null) {
      // this is the child execution returning
      // end child execution
      execution.end(Execution.STATE_ENDED);

      final BpelExecution ifExecution = execution.getParent();
      // remove finished execution from parent
      ifExecution.removeExecution(execution);

      // signal must be non-null to make the distinction with child
      // executions
      // in the first line of this method
      ifExecution.signal("finished");
    } else if (signal.equals("finished")) {
      afterRunned(execution);
      // execution.proceed();
    }
  }
  public Expression getCondition() {
    return condition;
  }
  public void setCondition(final Expression condition) {
    this.condition = condition;
  }
  public List<Elseif> getElseifs() {
    return elseifs;
  }
  public void setElseifs(final List<Elseif> elseifs) {
    this.elseifs = elseifs;
  }
  public Else getElse() {
    return elseField;
  }
  public void setElse(final Else else1) {
    this.elseField = else1;
  }
  public NodeImpl getActivity() {
    return activity;
  }
  public void setActivity(final NodeImpl activity) {
    this.activity = activity;
  }

  @Override
  public ActivityType getType() {
    return ActivityType.IF;
  }

}
