/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.definition.activity;

import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.element.Copy;
import org.ow2.orchestra.definition.element.Correlation;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.facade.data.def.InvokeActivityDefinitionData;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.pvm.env.Environment;
import org.ow2.orchestra.pvm.session.MessageSession;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.runtime.CorrelationSetRuntime;
import org.ow2.orchestra.runtime.PartnerLinkRuntime;
import org.ow2.orchestra.runtime.ScopeRuntime;
import org.ow2.orchestra.runtime.VariableRuntime;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.services.job.ExecuteInvokeJob;
import org.ow2.orchestra.var.MessageVariable;

/** the runtime behaviour of the invoke activity.
 * @author Charles Souillard
 * @author Tom Baeyens
 * @author Guillaume Porcher
 */
public class Invoke extends AbstractBpelActivity {
  /** serialVersionUID. */
  private static final long serialVersionUID = 1L;
  private static Logger log = Logger.getLogger(Invoke.class.getName());


  /**
   * <xsd:element ref="catch" minOccurs="0" maxOccurs="unbounded"/>
   * <xsd:element ref="catchAll" minOccurs="0"/>
   * <xsd:element ref="compensationHandler" minOccurs="0"/>
   */
  /** partnerLink attribute of this invoke element. */
  protected String partnerLink;
  /** portType attribute of this invoke element. */
  protected QName portTypeQName;
  /** operation attribute of this invoke element. */
  protected String operationName;
  /** inputVariable attribute of this invoke element. */
  protected String inputVariable;
  /** outputVariable attribute of this invoke element. */
  protected String outputVariable;

  /**
   * list of correlations having pattern "request" or "request-response"
   * of this invoke element.
   */
  protected List<Correlation> requestCorrelations = null;
  /**
   * list of correlations having pattern "response" or "request-response"
   * of this invoke element.
   */
  protected List<Correlation> responseCorrelations = null;

  protected boolean isOneWay;
  protected List<Copy> outCopies;
  protected List<Copy> inCopies;
  protected Variable anonymousOutboundMessageVariable;
  protected Variable anonymousInboundMessageVariable;
  public static final String RECEIVED_MESSAGE = "SIGNAL_PARAMETER_RECEIVED_MSG";
  public static final String RECEIVED_EXCEPTION = "SIGNAL_PARAMETER_RECEIVED_EXCEPTION";

  public Invoke() {
    super();
  }

  /**
   * business logic.
   * @param execution execution
   */
  @Override
  public void executeActivity(final BpelExecution execution) {
    final PartnerLinkRuntime partnerLinkRuntime = execution.getPartnerLinkRuntime(this.partnerLink);
    final OperationKey operationKey = new OperationKey(
        execution.getProcessDefinition().getQName(),
        this.portTypeQName,
        this.operationName);

    final ScopeRuntime scopeRuntime = execution.getCurrentScopeRuntime();
    final VariableRuntime anonymousOutboundMessageVariableRuntime =
      scopeRuntime.createVariable(execution, this.anonymousOutboundMessageVariable);

    // Perform copy operations
    for (final Copy copy : this.outCopies) {
      copy.execute(execution);
    }

    if (this.requestCorrelations != null) {
      //we verify correlation
      for (final Correlation correlation : this.requestCorrelations) {
        // Check correlation on outgoing message
        final CorrelationSetRuntime correlationSetRuntime =
          execution.getCorrelationSetByName(correlation.getCorrelationSet().getName());
        correlationSetRuntime.checkCSValues(
            execution,
            (MessageVariable) anonymousOutboundMessageVariableRuntime.getValue(),
            this.anonymousOutboundMessageVariable.getMessageType(),
            correlation.getInitiate());
      }
    }
    scopeRuntime.removeVariable(execution, this.anonymousOutboundMessageVariable);

    if (Invoke.log.isLoggable(Level.FINE)) {
      Invoke.log.fine("Invoke, inputMessage " + anonymousOutboundMessageVariableRuntime.getValue());
    }

    final ExecuteInvokeJob invokeJob = new ExecuteInvokeJob(
        (MessageVariable) anonymousOutboundMessageVariableRuntime.getValue(),
        partnerLinkRuntime.getPartnerRoleEndPointReference(),
        operationKey);
    invokeJob.setExecution(execution);

    Environment.getFromCurrent(MessageSession.class).send(invokeJob);

    execution.waitForSignal();

  }

  @Override
  public void signal(final BpelExecution execution, final String signal, final Map<String, Object> parameters) {

    if (parameters.containsKey(Invoke.RECEIVED_EXCEPTION)) {
      final RuntimeException exception = (RuntimeException) parameters.get(Invoke.RECEIVED_EXCEPTION);
      throw exception;
    }
    if (!this.isOneWay) {
      final MessageVariable returnedMessage = (MessageVariable) parameters.get(Invoke.RECEIVED_MESSAGE);

      if (Invoke.log.isLoggable(Level.FINE)) {
        Invoke.log.fine("Invoke, returnedMessage " + returnedMessage);
      }

      final ScopeRuntime scopeRuntime = execution.getCurrentScopeRuntime();

      // Receive response message before correlation validation
      final VariableRuntime anonymousInboundMessageVariableRuntime =
        scopeRuntime.createVariable(execution, this.anonymousInboundMessageVariable);
      anonymousInboundMessageVariableRuntime.setValue(returnedMessage);

      for (final Copy copy : this.inCopies) {
        copy.execute(execution);
      }

      scopeRuntime.removeVariable(execution, this.anonymousInboundMessageVariable);

      // validate correlations
      if (this.responseCorrelations != null) {
        for (final Correlation correlation : this.responseCorrelations) {
          final CorrelationSetRuntime correlationSetRuntime =
            execution.getCorrelationSetByName(correlation.getCorrelationSet().getName());
          correlationSetRuntime.checkCSValues(execution, returnedMessage, this.anonymousInboundMessageVariable.getMessageType(),
              correlation.getInitiate());
        }
      }
    } else {
      if (Invoke.log.isLoggable(Level.FINE)) {
        Invoke.log.fine("Invoke one way signalled");
      }
    }

    this.afterRunned(execution);
  }

  @Override
  public InvokeActivityDefinitionData createDefinitionData(final BpelProcess bpelProcess) {
    return new InvokeActivityDefinitionData(bpelProcess.getProcessDefinitionData(),
      this.uuid, null, this.name);
  }

  public String getInputVariable() {
    return this.inputVariable;
  }
  public void setInputVariable(final String inputVariable) {
    this.inputVariable = inputVariable;
  }
  public String getOperation() {
    return this.operationName;
  }
  public void setOperation(final String operation) {
    this.operationName = operation;
  }
  public String getOutputVariable() {
    return this.outputVariable;
  }
  public void setOutputVariable(final String outputVariable) {
    this.outputVariable = outputVariable;
  }
  public String getPartnerLink() {
    return this.partnerLink;
  }
  public void setPartnerLink(final String partnerLink) {
    this.partnerLink = partnerLink;
  }
  public QName getPortTypeQName() {
    return this.portTypeQName;
  }
  public void setPortTypeQName(final QName portType) {
    this.portTypeQName = portType;
  }
  public List<Correlation> getRequestCorrelations() {
    return this.requestCorrelations;
  }
  public void setRequestCorrelations(final List<Correlation> correlations) {
    this.requestCorrelations = correlations;
  }
  public List<Correlation> getResponseCorrelations() {
    return this.responseCorrelations;
  }
  public void setResponseCorrelations(final List<Correlation> correlations) {
    this.responseCorrelations = correlations;
  }
  public boolean isOneWay() {
    return this.isOneWay;
  }
  public void setOneWay(final boolean oneWay) {
    this.isOneWay = oneWay;
  }
  public List<Copy> getOutCopies() {
    return this.outCopies;
  }
  public void setOutCopies(final List<Copy> outCopies) {
    this.outCopies = outCopies;
  }
  public List<Copy> getInCopies() {
    return this.inCopies;
  }
  public void setInCopies(final List<Copy> inCopies) {
    this.inCopies = inCopies;
  }
  public Variable getAnonymousOutboundMessageVariable() {
    return this.anonymousOutboundMessageVariable;
  }
  public void setAnonymousOutboundMessageVariable(
      final Variable anonymousOutboundMessageVariable) {
    this.anonymousOutboundMessageVariable = anonymousOutboundMessageVariable;
  }
  public Variable getAnonymousInboundMessageVariable() {
    return this.anonymousInboundMessageVariable;
  }
  public void setAnonymousInboundMessageVariable(
      final Variable anonymousInboundMessageVariable) {
    this.anonymousInboundMessageVariable = anonymousInboundMessageVariable;
  }
  @Override
  public ActivityType getType() {
    return ActivityType.INVOKE;
  }
}
