/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.definition.activity;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.element.InboundMessageElement;
import org.ow2.orchestra.definition.element.OnAlarm;
import org.ow2.orchestra.definition.element.OnMessage;
import org.ow2.orchestra.facade.data.runtime.TimerData;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.pvm.Execution;
import org.ow2.orchestra.pvm.internal.model.NodeImpl;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.services.MessageCarrier;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.services.Receiver;
import org.ow2.orchestra.services.ReceivingElement;
import org.ow2.orchestra.services.SignalExecUtil;
import org.ow2.orchestra.util.EnvTool;
import org.ow2.orchestra.util.ReceivingElementUtil;
import org.ow2.orchestra.var.MessageVariable;

/**
 * the runtime behaviour of the pick activity.
 *
 * @author Tom Baeyens, Charles Souillard, Guillaume Porcher
 */
public class Pick extends AbstractBpelActivity implements InboundMessageElement {

  /** serialVersionUID. */
  private static final long serialVersionUID = 1L;

  public static final String ON_ALARM_PARAM = "onAlarm";

  protected List<OnMessage> onMessages;
  protected List<OnAlarm> onAlarms;
  /** createInstance attribute of this pick element. */
  protected boolean createInstance = false;

  public Pick() {
    super();
    this.executeSourcesAsync = true;
  }


  /**
   * business logic.
   *
   * @param execution
   *            execution
   */
  @Override
  public void executeActivity(final BpelExecution execution) {
    // add to list of child executions waiting for a signal
    execution.getProcessInstance().addWaitingExecution(execution);
    if (this.onAlarms != null) {
      for (final OnAlarm onAlarm : this.onAlarms) {
        final Date date = OnAlarm.computeTimer(execution,
            onAlarm.getDeadlineExpression(),
            onAlarm.getDurationExpression(),
            onAlarm.getRepeatEveryExpression());
        execution.createTimer(null, onAlarm.getUuid(), null, date, null, null, null);
        EnvTool.getRecorder().recordTimerStarted(new TimerData(execution.getActivityInstanceUUID(), null, date));
      }
    }
    execution.waitForSignal();
  }

  @Override
  public void signal(final BpelExecution execution, final String signal,
      final Map<String, Object> parameters) {
    if (parameters != null) {
      final OnMessage onMessage = (OnMessage) parameters.get(SignalExecUtil.RECEIVING_ELEMENT_PARAM);
      final NodeImpl onAlarmActivity = (NodeImpl) parameters.get(Pick.ON_ALARM_PARAM);
      if (onMessage != null || onAlarmActivity != null) {
        // Remove other waiting executions
        final BpelProcess process = execution.getProcessDefinition();
        for (final OnMessage onMsg : this.onMessages) {
          final OperationKey opKey = onMsg.getOperationKey();
          final Receiver receiver = process.getReceiver(opKey);
          receiver.removeWaitingExecution(execution);
        }
        // Remove from list of child executions waiting for a signal
        execution.getProcessInstance().getWaitingExecutions().remove(execution);
      }
      if (onMessage != null) {
        final MessageVariable message = (MessageVariable) parameters.get(SignalExecUtil.MESSAGE_PARAM);
        final MessageCarrier messageCarrier = (MessageCarrier) parameters.get(SignalExecUtil.MESSAGE_CARRIER_PARAM);
        ReceivingElementUtil.receive(onMessage, message, messageCarrier, execution);
        final NodeImpl node = onMessage.getActivity();
        // execute onMessage activity in a child execution
        execution.createExecution().execute(node);
      } else if (onAlarmActivity != null) {
        // execute onAlarm activity in a child execution
        execution.createExecution().execute(onAlarmActivity);
      } else {
        throw new OrchestraRuntimeException("Unsupported signal parameters");
      }
    } else if (signal != null) {
      if (signal.equals("finished")) {
        this.afterRunned(execution);
        return;
      }
      boolean found = false;
      for (final OnAlarm onAlarm : this.onAlarms) {
        if (signal.equals(onAlarm.getUuid())) {
          found = true;
          final Map<String, Object> params = new HashMap<String, Object>();
          params.put(Pick.ON_ALARM_PARAM, onAlarm.getActivity());
          this.signal(execution, null, params);
        }
      }
      if (!found) {
        throw new OrchestraRuntimeException("Unsupported signal " + signal);
      }
    } else {
      // this is the child execution returning
      // end child execution
      execution.end(Execution.STATE_ENDED);

      final BpelExecution pickExecution = execution.getParent();
      // remove finished execution from parent
      pickExecution.removeExecution(execution);

      // signal must be non-null to make the distinction with child
      // executions
      // in the first line of this method
      pickExecution.signal("finished");
    }
  }
  public boolean isCreateInstance() {
    return this.createInstance;
  }
  public void setCreateInstance(final boolean createInstance) {
    this.createInstance = createInstance;
  }
  public List<OnMessage> getOnMessages() {
    return this.onMessages;
  }
  public void setOnMessages(final List<OnMessage> onMessages) {
    this.onMessages = onMessages;
  }
  public List<OnAlarm> getOnAlarms() {
    return this.onAlarms;
  }
  public void setOnAlarms(final List<OnAlarm> onAlarms) {
    this.onAlarms = onAlarms;
  }
  public List<ReceivingElement> getReceivingElements() {
    return new ArrayList<ReceivingElement>(this.onMessages);
  }

  @Override
  public ActivityType getType() {
    return ActivityType.PICK;
  }
}
