/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.definition.activity;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.element.Copy;
import org.ow2.orchestra.definition.element.Correlation;
import org.ow2.orchestra.definition.element.InboundMessageElement;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.facade.data.def.ReceiveActivityDefinitionData;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.services.MessageCarrier;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.services.ReceivingElement;
import org.ow2.orchestra.services.SignalExecUtil;
import org.ow2.orchestra.util.ReceivingElementUtil;
import org.ow2.orchestra.util.XmlConstants;
import org.ow2.orchestra.var.MessageVariable;

/**
 * the runtime behaviour of the receive activity.
 *
 * @author Charles Souillard, Tom Baeyens
 * @version $Id: Receive.java 3047 2009-09-18 09:51:51Z porcherg $
 */
public class Receive extends AbstractBpelActivity implements ReceivingElement,
    InboundMessageElement {

  /** serialVersionUID. */
  private static final long serialVersionUID = 1L;

  /** log. */
  private static final Logger LOG = Logger.getLogger(Receive.class.getName());

  /** partnerLink attribute of this receive element. */
  protected String partnerLinkName = null;

  /** operation attribute of this receive element. */
  protected OperationKey operationKey = null;

  /** variable attribute of this receive element. */
  protected String variable = null;

  /** createInstance attribute of this receive element. */
  protected boolean createInstance = false;

  /** messageExchange attribute of this receive element. */
  protected String messageExchange = XmlConstants.BPEL_DEFAULT_MESSAGE_EXCHANGE;

  /** list of correlations of this receive. */
  protected List<Correlation> correlations = null;

  /** fromParts management. */
  protected List<Copy> inCopies;

  protected Variable anonymousInboundMessageVariable;

  public Receive() {
    super();
    this.executeSourcesAsync = true;
  }


  /**
   * business logic.
   *
   * @param execution
   *            execution
   */
  @Override
  public void executeActivity(final BpelExecution execution) {
    execution.getProcessInstance().addWaitingExecution(execution);

    if (Receive.LOG.isLoggable(Level.FINE)) {
      Receive.LOG.fine("Receive " + this.operationKey + " : waiting...");
    }
    execution.waitForSignal();
  }

  @Override
  public void signal(final BpelExecution execution, final String signal, final Map<String, Object> parameters) {
    final MessageVariable message = (MessageVariable) parameters.get(SignalExecUtil.MESSAGE_PARAM);
    final MessageCarrier messageCarrier = (MessageCarrier) parameters.get(SignalExecUtil.MESSAGE_CARRIER_PARAM);
    ReceivingElementUtil.receive(this, message, messageCarrier, execution);
    execution.getProcessInstance().getWaitingExecutions().remove(execution);
    this.afterRunned(execution);
  }

  public String getPartnerLinkName() {
    return this.partnerLinkName;
  }

  public String getOperation() {
    return this.getOperationKey().getOperationName();
  }

  public QName getPortTypeQName() {
    return this.getOperationKey().getPortTypeQName();
  }

  public String getVariable() {
    return this.variable;
  }

  public void setVariable(final String variable) {
    this.variable = variable;
  }

  public void setPartnerLinkName(final String partnerLinkName) {
    this.partnerLinkName = partnerLinkName;
  }

  public List<Correlation> getCorrelations() {
    return this.correlations;
  }

  public void setCorrelations(final List<Correlation> correlations) {
    this.correlations = correlations;
  }

  public boolean isCreateInstance() {
    return this.createInstance;
  }

  public void setCreateInstance(final boolean createInstance) {
    this.createInstance = createInstance;
  }

  public String getMessageExchange() {
    return this.messageExchange;
  }

  public void setMessageExchange(final String messageExchange) {
    this.messageExchange = messageExchange;
  }

  public Variable getAnonymousInboundMessageVariable() {
    return this.anonymousInboundMessageVariable;
  }

  public void setAnonymousInboundMessageVariable(
      final Variable anonymousInboundMessageVariable) {
    this.anonymousInboundMessageVariable = anonymousInboundMessageVariable;
  }

  public List<Copy> getInCopies() {
    return this.inCopies;
  }

  public void setInCopies(final List<Copy> inCopies) {
    this.inCopies = inCopies;
  }

  public OperationKey getOperationKey() {
    return this.operationKey;
  }
  /**
   * @param operationKey the operationKey to set
   */
  public void setOperationKey(final OperationKey operationKey) {
    this.operationKey = operationKey;
  }

  public List<ReceivingElement> getReceivingElements() {
    final List<ReceivingElement> list = new ArrayList<ReceivingElement>();
    list.add(this);
    return list;
  }

  @Override
  public ActivityType getType() {
    return ActivityType.RECEIVE;
  }


  @Override
  public ReceiveActivityDefinitionData createDefinitionData(final BpelProcess bpelProcess) {
    return new ReceiveActivityDefinitionData(bpelProcess.getProcessDefinitionData(),
      this.uuid, null, this.name, null, null, null);
  }

}
