/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.definition.activity;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.Map.Entry;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.element.CorrelationSet;
import org.ow2.orchestra.definition.element.PartnerLink;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.facade.data.def.ScopeActivityDefinitionData;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.facade.runtime.ScopeState;
import org.ow2.orchestra.parsing.binding.ActivityBinding;
import org.ow2.orchestra.parsing.binding.ScopeBinding;
import org.ow2.orchestra.parsing.binding.exception.CorrelationSetNotFoundException;
import org.ow2.orchestra.pvm.Execution;
import org.ow2.orchestra.pvm.internal.model.NodeImpl;
import org.ow2.orchestra.pvm.model.Node;
import org.ow2.orchestra.pvm.model.OpenExecution;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.runtime.ScopeRuntime;
import org.ow2.orchestra.runtime.VariableRuntime;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.var.MessageVariable;

/**
 * the runtime behaviour of the scope activity.
 *
 * @author Charles Souillard, Tom Baeyens
 * @version $Id: Scope.java 3062 2009-09-23 08:47:39Z porcherg $
 */
public class Scope extends AbstractBpelActivity {

  /** serialVersionUID. */
  private static final long serialVersionUID = -1892215735882878127L;

  /**
   * Index of the main activity in the list of nodes in the Scope execution
   * node.<br/>
   * <br/>
   * <b>IMPORTANT</b>: To be compatible with the
   * {@link ActivityBinding#createActivityNodeWithChildren} method, this
   * constant needs to be set to zero.
   */
  public static final int MAIN_ACTIVITY = 0;

  /**
   * Index of the event handler activity in the list of nodes in the Scope
   * execution node.<br/>
   * <br/>
   * <b>IMPORTANT</b>: To be compatible with the {@link ScopeBinding#parse}
   * method, this constant needs to be set to one.
   */
  public static final int EVENT_ACTIVITY = 1;

  /**
   * Index of the termination handler activity in the list of nodes in the Scope
   * execution node.<br/>
   * <br/>
   * <b>IMPORTANT</b>: To be compatible with the {@link ScopeBinding#parse}
   * method, this constant needs to be set to 2.
   */
  public static final int TERMINATION_ACTIVITY = 2;

  /**
   * Index of the compensation handler activity in the list of nodes in the
   * Scope execution node.<br/>
   * <br/>
   * <b>IMPORTANT</b>: To be compatible with the {@link ScopeBinding#parse}
   * method, this constant needs to be set to 3.
   */
  public static final int COMPENSATION_ACTIVITY = 3;

  /**
   * Index of the catchAll fault handling activity in the list of nodes in the
   * Scope execution node.<br/>
   * <br/>
   * <b>NOTE</b>: All nodes with an index superior to CATCHALL_ACTIVITY are
   * catch activities.<br/>
   * <br/>
   * <b>IMPORTANT</b>: To be compatible with the
   * {@link ActivityBinding#createActivityNodeWithChildren} method, this
   * constant needs to be set to 4.
   */
  public static final int CATCHALL_ACTIVITY = 4;

  /**
   * The biggest activity index in the list of nodes in the Scope execution
   * node.
   */
  public static final int BIGGEST_ACTIVITY_INDEX = Scope.CATCHALL_ACTIVITY + 1;

  /**
   * Signal sent to EnventHandler execution when main activity is finished
   */
  public static final String SCOPE_FINISHED_SIGNAL = "scope-finished";

  /**
   * List of variable name that must be initialized at the beginning of the
   * scope. (variable with an inline from-spec).
   */
  protected List<String> listOfVariableToBeInitialized;

  /** isolated attribute of this scope. */
  protected boolean isolated = false;

  /** exitOnStandardFault attribute of this scope. */
  protected boolean exitOnStandardFault;

  /** list of variables of this scope. */
  protected Map<String, Variable> variables;

  /** list of partnerLinks of this scope. */
  protected Map<String, PartnerLink> partnerLinks;

  /** list of messageExchanges of this scope. */
  protected Set<String> messageExchanges;

  /** list of correlationSets of this scope. */
  protected Map<String, CorrelationSet> correlationSets;

  protected NodeImpl scopeNode;

  public Scope() {
    super();
  }

  /**
   * business logic.
   *
   * @param execution execution
   */
  @Override
  @SuppressWarnings("unchecked")
  public void executeActivity(final BpelExecution execution) {
    final ScopeRuntime sr = new ScopeRuntime(this, execution, UUID.randomUUID().toString());
    execution.pushScopeRuntime(sr);
    if (this.listOfVariableToBeInitialized != null) {
      for (final String variableName : this.listOfVariableToBeInitialized) {
        execution.getVariableByName(variableName).initializeXmlType(execution);
      }
    }

    // Add additional variables (used in ForEach)
    final Map<String, Object> additionalVariables = (Map<String, Object>) execution
      .getVariable(BpelExecution.ADDITIONNAL_SCOPE_VARIABLES);
    if (additionalVariables != null) {
      for (final Entry<String, Object> var : additionalVariables.entrySet()) {
        final VariableRuntime varRun = execution.getVariableByName(var.getKey());
        if (var.getValue() instanceof MessageVariable) {
          varRun.setValue((MessageVariable) var.getValue());
        } else {
          varRun.setValue(var.getValue(), false);
        }
        // record variable updated
        varRun.updated(execution.getActivityInstanceUUID());
      }
      execution.removeVariable(BpelExecution.ADDITIONNAL_SCOPE_VARIABLES);
    }

    this.executeEventHandlers(execution);

    // set scope running
    sr.setState(ScopeState.RUNNING);

    // execute scope in a child execution
    this.executeMainActivity(execution);
  }

  protected void executeEventHandlers(final BpelExecution execution) {
    final List<Node> nodes = execution.getNode().getNodes();
    // launch event handlers
    execution.createExecution().execute(nodes.get(Scope.EVENT_ACTIVITY));
  }

  public void executeMainActivity(final BpelExecution execution) {
    // execute scope in a child execution
    // get the current sequence node
    final List<Node> nodes = execution.getNode().getNodes();
    final Node mainActivityNode = nodes.get(Scope.MAIN_ACTIVITY);
    execution.createExecution().execute(mainActivityNode);
  }

  /**
   * Signal method.
   *
   * @param execution execution
   * @param signal signal
   */
  @Override
  public void signal(final BpelExecution execution, final String signal, final Map<String, Object> parameters) {
    if (signal == null) {
      // this is the child execution returning
      // end child execution
      execution.end(Execution.STATE_ENDED);

      final BpelExecution scopeExecution = execution.getParent();

      // remove finished execution from parent
      scopeExecution.removeExecution(execution);

      if (scopeExecution.isActive()) {
        // signal must be non-null to make the distinction with child
        // executions in the first line of this method
        scopeExecution.signal("finished");
      } else {
        if (scopeExecution.getExecutions() != null) {
          for (final OpenExecution scopeChildExecution : new ArrayList<OpenExecution>(scopeExecution.getExecutions())) {
            if (scopeChildExecution.getNode().equals(this.scopeNode.getNodes().get(Scope.EVENT_ACTIVITY))) {
              if (scopeChildExecution.getExecutions().isEmpty()) {
                ((BpelExecution)scopeChildExecution).signal(Scope.SCOPE_FINISHED_SIGNAL);
              } else {
                final EventHandlerActivity eventHandlerActivity =
                  (EventHandlerActivity) ((NodeImpl)scopeChildExecution.getNode()).getBehaviour();
                eventHandlerActivity.cancelEvents((BpelExecution)scopeChildExecution);
              }
            }
          }
        }
      }
    } else if (signal.equals("finished")) {
      // signal is not null, this is therefore the parent execution.
      // Dereference the ScopeRuntime in order to avoid FK_CONSTRAINT exceptions.
      final ScopeRuntime scopeRuntime = execution.popScopeRuntime(true);

      switch (scopeRuntime.getState()) {
        case RUNNING:
          scopeRuntime.setState(ScopeState.COMPLETED_SUCCESSFUL);
          break;

        case TERMINATING:
        case HANDLING_FAULT:
          scopeRuntime.setState(ScopeState.COMPLETED_UNSUCCESSFUL);
          break;

        case COMPLETED_SUCCESSFUL:
        case COMPLETED_UNSUCCESSFUL:
          break;

        default:
          Misc.unreachableStatement("Unexpected scope state: " + scopeRuntime.getState());

      };

      // add scope state in a temporary variable. Used in ForEach
      final AbstractActivity parentActivity = (AbstractActivity)execution.getNode().getParentNode().getBehaviour();
      if (parentActivity.getType().equals(ActivityType.FOR_EACH)) {
        execution.createVariable(BpelExecution.SCOPE_STATE_VARIABLE, scopeRuntime.getState());
      }
      this.afterRunned(execution);
    }
  }

  /**
   * Searched for a correlation having the given name. Search first in this
   * scope and if nothing is found, search in father scopes...
   *
   * @param correlationSetName correlationSetName
   * @return the found correlation
   * @throws CorrelationSetNotFoundException when correlations et has not been found
   */
  public CorrelationSet findCorrelationSet(final String correlationSetName) throws CorrelationSetNotFoundException {
    if (this.correlationSets != null && this.correlationSets.containsKey(correlationSetName)) {
      return this.correlationSets.get(correlationSetName);
    } else if (this.enclosingScope != null) {
      return this.enclosingScope.findCorrelationSet(correlationSetName);
    }
    throw new CorrelationSetNotFoundException("correlation " + correlationSetName + " declaration not found");
  }
  /**
   * Searched for a partnerLink having the given name. Search first in this
   * scope and if nothing is found, search in father scopes...
   *
   * @param partnerLinkName partnerLinkName
   * @return the found partnerLink
   */
  public PartnerLink findPartnerLink(final String partnerLinkName) {
    if (this.partnerLinks != null && this.partnerLinks.containsKey(partnerLinkName)) {
      return this.partnerLinks.get(partnerLinkName);
    } else if (this.enclosingScope != null) {
      return this.enclosingScope.findPartnerLink(partnerLinkName);
    }
    // TODO : throw an exception
    return null;
  }
  /**
   * Searched for a variable having the given name. Search first in this scope
   * and if nothing is found, search in father scopes...
   *
   * @param variableName variableName
   * @return the found variable
   */
  public Variable findVariable(final String variableName) {
    if (this.variables != null && this.variables.containsKey(variableName)) {
      return this.variables.get(variableName);
    } else if (this.enclosingScope != null) {
      return this.enclosingScope.findVariable(variableName);
    }
    // TODO : throw an exception
    return null;
  }
  /**
   * Add a variable in the list of variables that the scope needs to initialize.
   *
   * @param variableName variableName
   */
  public void addVariableToBeInitialized(final String variableName) {
    if (this.listOfVariableToBeInitialized == null) {
      this.listOfVariableToBeInitialized = new ArrayList<String>();
    }
    this.listOfVariableToBeInitialized.add(variableName);
  }

  public Variable getVariable(final String variableName) {
    return this.variables.get(variableName);
  }
  public Map<String, PartnerLink> getPartnerLinks() {
    return this.partnerLinks;
  }
  public Map<String, Variable> getVariables() {
    return this.variables;
  }
  public boolean isExitOnStandardFault() {
    return this.exitOnStandardFault;
  }
  public void setExitOnStandardFault(final boolean exitOnStandardFault) {
    this.exitOnStandardFault = exitOnStandardFault;
  }
  public boolean isIsolated() {
    return this.isolated;
  }
  public void setIsolated(final boolean isolated) {
    this.isolated = isolated;
  }
  public void setPartnerLinks(final Map<String, PartnerLink> partnerLinks) {
    this.partnerLinks = partnerLinks;
  }
  public void setVariables(final Map<String, Variable> variables) {
    this.variables = variables;
  }
  public Set<String> getMessageExchanges() {
    return this.messageExchanges;
  }
  public void setMessageExchanges(final Set<String> messageExchanges) {
    this.messageExchanges = messageExchanges;
  }
  public Map<String, CorrelationSet> getCorrelationSets() {
    return this.correlationSets;
  }
  public void setCorrelationSets(final Map<String, CorrelationSet> correlationSets) {
    this.correlationSets = correlationSets;
  }


  public NodeImpl getScopeNode() {
    return this.scopeNode;
  }


  public void setScopeNode(final NodeImpl scopeNode) {
    this.scopeNode = scopeNode;
  }

  @Override
  public String toString() {
    return Scope.class.getName()
        + "[listOfVariableToBeInitialized: " + this.listOfVariableToBeInitialized
        + ", isolated: " + this.isolated
        + ", exitOnStandardFault: " + this.exitOnStandardFault
        + ", variables: " + this.variables
        + ", partnerLinks: " + this.partnerLinks
        + ", messageExchanges: " + this.messageExchanges
        + ", correlationSets: " + this.correlationSets
        + "]";
  }

  @Override
  public ActivityType getType() {
    return ActivityType.SCOPE;
  }

  @Override
  public ScopeActivityDefinitionData createDefinitionData(final BpelProcess bpelProcess) {
    return new ScopeActivityDefinitionData(bpelProcess.getProcessDefinitionData(), this.uuid, null, this.name, null);
  }

}
