/**
 * Copyright (C) 2007-2009 Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.definition.activity;

import java.util.logging.Logger;

import javax.xml.namespace.QName;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.exception.FaultWithMessageVariable;
import org.ow2.orchestra.facade.data.def.ThrowActivityDefinitionData;
import org.ow2.orchestra.facade.exception.BpelFaultException;
import org.ow2.orchestra.facade.exception.FaultWithElement;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.pvm.internal.model.ExecutionImpl.Propagation;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.runtime.ScopeRuntime;
import org.ow2.orchestra.runtime.VariableRuntime;
import org.ow2.orchestra.util.XmlConstants;
import org.ow2.orchestra.var.MessageVariable;
import org.w3c.dom.Element;

/**
 * the runtime behaviour of the throw activity.
 *
 * @author Tom Baeyens, Charles Souillard, S. Ali Tokmen
 */
public class Throw extends AbstractBpelActivity {

  private static final long serialVersionUID = -4706956432645598341L;

  private static Logger log = Logger.getLogger(Throw.class.getName());

  /** faultQName attribute of this throw element. */
  private QName faultQName;

  /** faultVariable attribute of this throw element. */
  private String faultVariable;

  public Throw() {
    super();
  }

  /**
   * throws a Bpel exception
   *
   * @param execution execution
   */
  @Override
  public void executeActivity(final BpelExecution execution) {
    execution.setPropagation(Propagation.EXPLICIT);

    if (this.isBpelFault()) {
      final BpelFaultException res = new BpelFaultException(this.faultQName);
      if (res != null) {
        Throw.log.fine("Bpel Fault thrown: " + res.getClass().getName());
        throw res;
      } else {
        throw new OrchestraRuntimeException("No Bpel Fault corresponds to this fault: " + this.getFaultName());
      }
    } else {
      final BpelFaultException userFault = this.getUserFaultVariable(execution);
      Throw.log.fine("User Fault thrown: " + userFault);
      throw userFault;
    }
  }

  private BpelFaultException getUserFaultVariable(final BpelExecution execution) {
    BpelFaultException userFault = null;
    if (this.getFaultVariable() != null) {
      // get the fault variable
      final ScopeRuntime scopeRuntime = execution.getScopeRuntimeWithVariable(this.getFaultVariable());
      final VariableRuntime faultV = scopeRuntime.getVariable(this.getFaultVariable());

      if (faultV != null) {
        final Object val = faultV.getValue();
        if (val != null) {
          if (val instanceof Element) {
            final Element element = (Element) val;
            userFault = new FaultWithElement(this.getFaultName(), element);
          } else if (val instanceof MessageVariable) {
            final MessageVariable message = ((MessageVariable) val).duplicate();
            userFault = new FaultWithMessageVariable(this.getFaultName(), message);
          } else {
            throw new OrchestraRuntimeException("The type of the value of the variable is unknown: " + val.getClass());
          }
        }
      } else {
        throw new OrchestraRuntimeException("The variable is unknown: " + this.getFaultVariable());
      }
    } else {
      userFault = new BpelFaultException(this.getFaultName());
    }

    if (userFault == null) {
      throw new OrchestraRuntimeException("No user fault corresponds to this fault: [name=" + this.getFaultName()
        + ", variable=" + this.getFaultVariable() + "]");
    }
    return userFault;
  }

  private boolean isBpelFault() {
    boolean res = false;
    if (this.getFaultName().getNamespaceURI().equals(XmlConstants.XMLNS_BPEL_2_0_EXECUTABLE)) {
      res = true;
    }
    return res;
  }

  public QName getFaultName() {
    return this.faultQName;
  }

  public void setFaultName(final QName faultName) {
    this.faultQName = faultName;
  }

  public String getFaultVariable() {
    return this.faultVariable;
  }

  public void setFaultVariable(final String faultVariable) {
    this.faultVariable = faultVariable;
  }

  @Override
  public ActivityType getType() {
    return ActivityType.THROW;
  }

  @Override
  public ThrowActivityDefinitionData createDefinitionData(final BpelProcess bpelProcess) {
    return new ThrowActivityDefinitionData(bpelProcess.getProcessDefinitionData(), this.uuid, null, this.name,
      this.faultQName, this.faultVariable);
  }

}
