/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.jmx;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.logging.Logger;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import javax.xml.namespace.QName;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.activity.Exit;
import org.ow2.orchestra.deployment.Deployer;
import org.ow2.orchestra.env.EnvFactoryRepository;
import org.ow2.orchestra.facade.Deployment;
import org.ow2.orchestra.facade.def.ActivityDefinition;
import org.ow2.orchestra.facade.def.ProcessDefinition;
import org.ow2.orchestra.facade.def.ProcessState;
import org.ow2.orchestra.facade.def.full.ActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.ProcessFullDefinition;
import org.ow2.orchestra.facade.def.impl.ProcessDefinitionImpl;
import org.ow2.orchestra.facade.exception.ActivityNotFoundException;
import org.ow2.orchestra.facade.exception.InstanceNotFoundException;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.facade.exception.ProcessNotFoundException;
import org.ow2.orchestra.facade.jmx.RemoteDeployerMBean;
import org.ow2.orchestra.facade.runtime.ActivityInstance;
import org.ow2.orchestra.facade.runtime.ActivityState;
import org.ow2.orchestra.facade.runtime.ProcessInstance;
import org.ow2.orchestra.facade.runtime.full.ActivityFullInstance;
import org.ow2.orchestra.facade.runtime.full.ProcessFullInstance;
import org.ow2.orchestra.facade.runtime.impl.ProcessInstanceImpl;
import org.ow2.orchestra.facade.uuid.ActivityDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessInstanceUUID;
import org.ow2.orchestra.pvm.env.Environment;
import org.ow2.orchestra.pvm.env.EnvironmentFactory;
import org.ow2.orchestra.pvm.internal.cmd.Command;
import org.ow2.orchestra.pvm.internal.cmd.CommandService;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.services.Querier;
import org.ow2.orchestra.services.itf.Repository;
import org.ow2.orchestra.util.DeploymentUtil;
import org.ow2.orchestra.util.EnvTool;
import org.ow2.orchestra.util.Misc;

public class RemoteDeployerImpl implements RemoteDeployerMBean {

  private static final Logger LOG = Logger.getLogger(RemoteDeployerImpl.class.getName());

  public RemoteDeployerImpl() {
  }

  protected CommandService getCommandService() {
    final EnvironmentFactory environmentFactory = EnvFactoryRepository.get();
    Misc.badStateIfNull(environmentFactory, "environmentFactory is null.");

    return environmentFactory.get(CommandService.class);
  }

  public ProcessDefinition deploy(final Deployment bpelDeployment) {

    return this.getCommandService().execute(new Command<ProcessDefinition>() {

      public ProcessDefinition execute(final Environment environment) throws Exception {
        final Deployer deployer = new Deployer();

        final ProcessDefinition bpelProcess = new ProcessDefinitionImpl(deployer.deploy(bpelDeployment, environment));
        if (bpelProcess == null) {
          throw new OrchestraRuntimeException("Process is null");
        }
        return bpelProcess;
      }
    });

  }

  public boolean undeploy(final QName processQName) {
    return this.getCommandService().execute(new Command<Boolean>() {

      public Boolean execute(final Environment environment) throws Exception {
        final Deployer deployer = new Deployer();
        deployer.undeploy(processQName, environment);
        return true;
      }
    });
  }

  public boolean undeploy(final ProcessDefinitionUUID pduuid) {
    return this.getCommandService().execute(new Command<Boolean>() {

      public Boolean execute(final Environment environment) throws Exception {
        final Deployer deployer = new Deployer();
        deployer.undeploy(pduuid, environment);
        return true;
      }
    });
  }

  public List<ProcessDefinition> findProcessDefinitions() {
    return this.getCommandService().execute(new Command<List<ProcessDefinition>>() {

      public List<ProcessDefinition> execute(final Environment environment) throws Exception {
        final List<ProcessDefinition> result = new ArrayList<ProcessDefinition>();
        final Querier querier = EnvTool.getQuerier();
        for (final ProcessFullDefinition processFullDefinition : querier.findProcessDefinitions()) {
          result.add(new ProcessDefinitionImpl(processFullDefinition));
        }
        return result;
      }
    });
  }

  public List<ProcessDefinition> findProcessDefinitions(final ProcessState state) {
    return this.getCommandService().execute(new Command<List<ProcessDefinition>>() {

      public List<ProcessDefinition> execute(final Environment environment) throws Exception {
        final List<ProcessDefinition> result = new ArrayList<ProcessDefinition>();
        final Querier querier = EnvTool.getQuerier();
        for (final ProcessFullDefinition processFullDefinition : querier.findProcessDefinitions(state)) {
          result.add(new ProcessDefinitionImpl(processFullDefinition));
        }
        return result;
      }
    });
  }

  public List<ProcessDefinition> findProcessDefinitions(final QName processName, final ProcessState state) {
    return this.getCommandService().execute(new Command<List<ProcessDefinition>>() {

      public List<ProcessDefinition> execute(final Environment environment) throws Exception {
        final List<ProcessDefinition> result = new ArrayList<ProcessDefinition>();
        final Querier querier = EnvTool.getQuerier();
        for (final ProcessFullDefinition processFullDefinition : querier.findProcessDefinitions(processName, state)) {
          result.add(new ProcessDefinitionImpl(processFullDefinition));
        }
        return result;
      }
    });
  }

  public List<ProcessDefinition> findProcessDefinitions(final QName processName) {
    return this.getCommandService().execute(new Command<List<ProcessDefinition>>() {

      public List<ProcessDefinition> execute(final Environment environment) throws Exception {
        final List<ProcessDefinition> result = new ArrayList<ProcessDefinition>();
        final Querier querier = EnvTool.getQuerier();
        for (final ProcessFullDefinition processFullDefinition : querier.findProcessDefinitions(processName)) {
          result.add(new ProcessDefinitionImpl(processFullDefinition));
        }
        return result;
      }
    });
  }

  public ProcessDefinition getProcessDefinition(final ProcessDefinitionUUID processUUID) throws ProcessNotFoundException {
    return this.getCommandService().execute(new Command<ProcessDefinition>() {

      public ProcessDefinition execute(final Environment environment) throws Exception {
        final ProcessFullDefinition process = EnvTool.getQuerier().getProcessDefinition(processUUID);
        if (process == null) {
          throw new ProcessNotFoundException(processUUID);
        }
        return new ProcessDefinitionImpl(process);
      }
    });
  }

  public ActivityDefinition getActivityDefinition(final ActivityDefinitionUUID activityDefUUID) throws ActivityNotFoundException {
    return this.getCommandService().execute(new Command<ActivityDefinition>() {

      public ActivityDefinition execute(final Environment environment) throws Exception {
        final ActivityFullDefinition activityDefinition = EnvTool.getQuerier().getActivityDefinition(activityDefUUID);
        if (activityDefinition == null) {
          throw new ActivityNotFoundException(activityDefUUID);
        }
        return activityDefinition.copy();
      }
    });
  }

  public ProcessDefinition deployBar(final byte[] barFile) {
    try {
      // Create temp dir
      new File(System.getProperty("java.io.tmpdir")).mkdirs();
      final File tmpDir = File.createTempFile("orch", null, null);
      tmpDir.delete();
      if (!tmpDir.mkdirs()) {
        throw new IOException("Cannot create the temporary directory '" + tmpDir + "'.");
      }

      final List<URL> wsdlUrls = new ArrayList<URL>();
      URL bpelFile = null;

      // Unzip file
      final InputStream in = new ByteArrayInputStream(barFile);
      final ZipInputStream zipInput = new ZipInputStream(in);
      ZipEntry zipFile;
      while ((zipFile = zipInput.getNextEntry()) != null) {
        // write the file to the disk
        // Create file
        final File outFile = new File(tmpDir.getAbsolutePath() + File.separator + zipFile.getName());
        final FileOutputStream outputFile = new FileOutputStream(outFile);
        try {
          // Copy file content
          int c;
          while ((c = zipInput.read()) != -1) {
            outputFile.write(c);
          }
        } finally {
          outputFile.close();
        }

        if (zipFile.getName().endsWith(".wsdl")) {
          wsdlUrls.add(outFile.toURL());
        } else if (zipFile.getName().endsWith(".bpel")) {
          if (bpelFile != null) {
            throw new OrchestraRuntimeException("Bar contains more than one bpel file.");
          }
          bpelFile = outFile.toURL();
        }
      }
      zipInput.close();
      in.close();

      final Deployment deployment = DeploymentUtil.generateDeployment(bpelFile, wsdlUrls);

      // Deploy the process
      final ProcessDefinition process = this.deploy(deployment);

      // Delete temp dir
      Misc.deleteDir(tmpDir);

      return process;
    } catch (final Exception e) {
      // TODO: Handle exceptions.
      throw new OrchestraRuntimeException("Unable to deploy bar", e);
    }
  }

  public void initialize() {
    this.getCommandService().execute(new Command<Object>() {

      public Object execute(final Environment environment) throws Exception {
        final Repository repository = environment.get(Repository.class);
        final Deployer deployer = new Deployer();
        RemoteDeployerImpl.LOG.info("Orchestra initialization.");
        for (final BpelProcess process : repository.getProcesses()) {
          deployer.redeploy(process, environment);
        }
        RemoteDeployerImpl.LOG.info("Orchestra initialized.");
        return null;
      }
    });
  }

  private void verifyInstance(final ProcessInstanceUUID processUUID) throws InstanceNotFoundException {
    final InstanceNotFoundException instanceNotFoundException = this.getCommandService().execute(
      new Command<InstanceNotFoundException>() {

        public InstanceNotFoundException execute(final Environment environment) {
          final Querier querier = EnvTool.getQuerier();
          if (querier.getProcessInstance(processUUID) == null) {
            return new InstanceNotFoundException(processUUID);
          }
          return null;
        }

      });

    if (instanceNotFoundException != null) {
      throw instanceNotFoundException;
    }
  }

  public Set<ActivityInstance> findActivityInstances(final ProcessInstanceUUID processUUID) throws InstanceNotFoundException {
    this.verifyInstance(processUUID);
    return this.getCommandService().execute(new Command<Set<ActivityInstance>>() {

      public Set<ActivityInstance> execute(final Environment environment) throws Exception {
        final Set<ActivityInstance> result = new HashSet<ActivityInstance>();
        final Querier querier = EnvTool.getQuerier();
        querier.getProcessInstance(processUUID);
        for (final ActivityFullInstance activityFullInstance : querier.findActivityInstances(processUUID)) {
          result.add(activityFullInstance.copy());
        }
        return result;
      }
    });
  }

  public Set<ActivityInstance> findActivityInstances(final ProcessInstanceUUID processUUID, final ActivityState activityState)
  throws InstanceNotFoundException {
    this.verifyInstance(processUUID);
    return this.getCommandService().execute(new Command<Set<ActivityInstance>>() {

      public Set<ActivityInstance> execute(final Environment environment) throws Exception {
        final Set<ActivityInstance> result = new HashSet<ActivityInstance>();
        final Querier querier = EnvTool.getQuerier();
        for (final ActivityFullInstance activityFullInstance : querier.findActivityInstances(processUUID, activityState)) {
          result.add(activityFullInstance.copy());
        }
        return result;
      }
    });
  }

  public Set<ProcessInstance> findProcessInstances() {
    return this.getCommandService().execute(new Command<Set<ProcessInstance>>() {

      public Set<ProcessInstance> execute(final Environment environment) throws Exception {
        final Set<ProcessInstance> result = new HashSet<ProcessInstance>();
        final Querier querier = EnvTool.getQuerier();
        for (final ProcessFullInstance processFullInstance : querier.findProcessInstances()) {
          result.add(new ProcessInstanceImpl(processFullInstance));
        }
        return result;
      }
    });
  }

  public Set<ProcessInstance> findProcessInstances(final ProcessDefinitionUUID processDefinitionUUID)
  throws ProcessNotFoundException {
    return this.getCommandService().execute(new Command<Set<ProcessInstance>>() {

      public Set<ProcessInstance> execute(final Environment environment) throws Exception {
        final Set<ProcessInstance> result = new HashSet<ProcessInstance>();
        final Querier querier = EnvTool.getQuerier();
        for (final ProcessFullInstance processFullInstance : querier.findProcessInstances(processDefinitionUUID)) {
          result.add(new ProcessInstanceImpl(processFullInstance));
        }
        return result;
      }
    });
  }

  public Set<ProcessInstance> findProcessInstances(final ProcessDefinitionUUID processDefinitionUUID,
    final ActivityState processInstanceState) throws ProcessNotFoundException {
    return this.getCommandService().execute(new Command<Set<ProcessInstance>>() {

      public Set<ProcessInstance> execute(final Environment environment) throws Exception {
        final Set<ProcessInstance> result = new HashSet<ProcessInstance>();
        final Querier querier = EnvTool.getQuerier();
        for (final ProcessFullInstance processFullInstance : querier.findProcessInstances(processDefinitionUUID,
          processInstanceState)) {
          result.add(new ProcessInstanceImpl(processFullInstance));
        }
        return result;
      }
    });
  }

  public Set<ProcessInstance> findProcessInstances(final ActivityState processInstanceState) {
    return this.getCommandService().execute(new Command<Set<ProcessInstance>>() {

      public Set<ProcessInstance> execute(final Environment environment) throws Exception {
        final Set<ProcessInstance> result = new HashSet<ProcessInstance>();
        final Querier querier = EnvTool.getQuerier();
        for (final ProcessFullInstance processFullInstance : querier.findProcessInstances(processInstanceState)) {
          result.add(new ProcessInstanceImpl(processFullInstance));
        }
        return result;
      }
    });
  }


  /**
   * Exit the given instance.
   *
   * @param processInstanceUUID
   *          the Id of the instance to exit.
   * @throws InstanceNotFoundException
   *           thrown if the given instance to exit is not found.
   */
  public void exit(final ProcessInstanceUUID processInstanceUUID) throws InstanceNotFoundException {
    final InstanceNotFoundException infe = this.getCommandService().execute(new Command<InstanceNotFoundException>() {

      public InstanceNotFoundException execute(final Environment environment) throws Exception {
        final List<BpelExecution> instances = EnvTool.getRepository().getInstances();
        final Iterator<BpelExecution> iter = instances.iterator();
        BpelExecution instance = null;
        BpelExecution tmp = null;
        while (instance == null && iter.hasNext()) {
          tmp = iter.next();
          if (tmp.getProcessInstanceUUID().equals(processInstanceUUID)) {
            instance = tmp.getProcessInstance();
          }
        }
        if (instance == null) {
          return new InstanceNotFoundException(processInstanceUUID);
        }
        if (instance.getProcessInstanceState().equals(ActivityState.RUNNING)
            || instance.getProcessInstanceState().equals(ActivityState.SUSPENDED)) {
          new Exit().executeActivity(instance);
        }

        return null;
      }
    });

    if (infe != null) {
      throw infe;
    }
  }

  /**
   * Resume a previously suspend instance of a process.
   *
   * @param processInstanceUUID
   *          the process instance UUID that identify the instance to resume.
   * @throws InstanceNotFoundException
   *           thrown if the given instance to resume is not found.
   */
  public void resume(final ProcessInstanceUUID processInstanceUUID) throws InstanceNotFoundException {
    final InstanceNotFoundException infe = this.getCommandService().execute(new Command<InstanceNotFoundException>() {

      public InstanceNotFoundException execute(final Environment environment) throws Exception {
        final BpelExecution instance = EnvTool.getRepository().getInstance(processInstanceUUID);
        if (instance == null) {
          return new InstanceNotFoundException(processInstanceUUID);
        }

        instance.resumeInstance();

        return null;
      }
    });

    if (infe != null) {
      throw infe;
    }


  }

  /**
   * Suspend the given instance of a process.
   *
   * @param processInstanceUUID
   *          the process instance UUID that identify the instance to suspend.
   * @throws InstanceNotFoundException
   *           thrown if the given instance to suspend is not found.
   */
  public void suspend(final ProcessInstanceUUID processInstanceUUID) throws InstanceNotFoundException {

    final InstanceNotFoundException infe = this.getCommandService().execute(new Command<InstanceNotFoundException>() {

      public InstanceNotFoundException execute(final Environment environment) throws Exception {
        final BpelExecution instance = EnvTool.getRepository().getInstance(processInstanceUUID);
        if (instance == null) {
          return new InstanceNotFoundException(processInstanceUUID);
        }

        instance.suspendInstance();

        return null;
      }
    });

    if (infe != null) {
      throw infe;
    }
  }

}
