/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the JBPM BPEL PUBLIC LICENSE AGREEMENT as
 * published by JBoss Inc.; either version 1.0 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
package org.ow2.orchestra.lang.function;

import java.util.List;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

import org.jaxen.Context;
import org.jaxen.ContextSupport;
import org.jaxen.Function;
import org.jaxen.FunctionCallException;
import org.jaxen.NamespaceContext;
import org.jaxen.Navigator;
import org.jaxen.function.StringFunction;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.lang.jaxen.BpelVariableContext;

/**
 * The <strong>getVariableProperty</strong> function extracts global property values from
 * variables.
 * <p>
 * <code><i>any</i> bpws:getVariableProperty(<i>string</i> variableName,
 * <i>string</i> propertyName)</code>
 * </p>
 * @author Alejandro Guizar
 * @version $Revision: 1.5 $ $Date: 2007/07/26 00:16:08 $
 */
public class GetVariablePropertyFunction implements Function {

  private static final Logger LOG = Logger.getLogger(GetVariablePropertyFunction.class.getName());

  /**
   * Extracts the value of a global property from a variable.
   * @param context the context at the point in the expression when the function is called
   * @param args a list containing the variable name and the property qualified name
   * @return the property vlaue
   * @throws FunctionCallException if <code>args</code> does not have exactly two items
   */
  public Object call(final Context context, final List args) throws FunctionCallException {
    if (args.size() != 2) {
      throw new FunctionCallException("getVariableProperty() requires two arguments");
    }

    return GetVariablePropertyFunction.evaluate(args.get(0), args.get(1), context);
  }

  /**
   * Extracts the value of a global property from a variable.
   * @param variableArg the variable name
   * @param propertyArg the property qualified name
   * @param context the context at the point in the expression when the function is called
   * @return the property value
   * @throws FunctionCallException if <code>variableArg</code> does not match any variable
   */
  public static Object evaluate(final Object variableArg, final Object propertyArg, final Context context)
    throws FunctionCallException {
    GetVariablePropertyFunction.LOG.finest("parameters: " + variableArg + ", " + propertyArg);

    // convert arguments to strings
    final Navigator nav = context.getNavigator();
    final ContextSupport sup = context.getContextSupport();

    // find variable definition
    final String variableName = StringFunction.evaluate(variableArg, nav);
    final BpelVariableContext variableContext = (BpelVariableContext) sup.getVariableContext();
    final Variable variableDefinition = variableContext.findVariableDefinition(variableName);
    if (variableDefinition == null) {
      throw new FunctionCallException("variable not found: " + variableName);
    }

    // resolve property name
    final String propertyName = StringFunction.evaluate(propertyArg, nav);
    final NamespaceContext nsContext = sup.getNamespaceContext();
    final QName propertyQName = GetVariablePropertyFunction.toQName(propertyName, nsContext);

    // extract property value
    final Object propertyValue = variableContext.getPropertyValue(variableName, propertyQName);
    GetVariablePropertyFunction.LOG.finest("return value: " + propertyValue);
    return propertyValue;
  }

  private static QName toQName(final String prefixedName, final NamespaceContext namespaceContext) {
    final int colonIndex = prefixedName.indexOf(':');
    if (colonIndex == -1) {
      return new QName(prefixedName);
    }
    final String localName = prefixedName.substring(colonIndex + 1);
    final String prefix = prefixedName.substring(0, colonIndex);
    final String namespaceURI = namespaceContext.translateNamespacePrefixToUri(prefix);
    return new QName(namespaceURI, localName, prefix);
  }
}
