/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.lang.jaxen;

import java.io.Serializable;
import java.util.Map;

import javax.xml.namespace.QName;

import org.jaxen.FunctionCallException;
import org.jaxen.UnresolvableException;
import org.jaxen.VariableContext;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.lang.PropertyQuery;
import org.ow2.orchestra.pvm.activity.Activity;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.runtime.VariableRuntime;
import org.ow2.orchestra.util.BpelXmlUtil;
import org.ow2.orchestra.util.XmlConstants;
import org.ow2.orchestra.var.MessageVariable;
import org.ow2.orchestra.wsdl.PropertyAlias;
import org.w3c.dom.Element;

/**
 * An implementation of {@link VariableContext} that resolves variable bindings
 * from the values in the scope of the enclosed
 *
 * @author Charles Souillard
 * @version $Revision: 1.1 $ $Date: 2007/05/29 01:09:34 $
 */
public class BpelVariableContext implements VariableContext, Serializable {

  private final BpelExecution bpelExecution;

  private static final long serialVersionUID = 1L;

  /**
   * Creates a variable context that wraps the given token.
   *
   * @param token
   *            the enclosed token
   * @throws NullPointerException
   *             if <code>token</code> is <code>null</code>
   */
  public BpelVariableContext(final BpelExecution bpelExecution) {
    if (bpelExecution == null) {
      throw new NullPointerException("bpelExecution must not be null");
    }

    this.bpelExecution = bpelExecution;
  }

  /**
   * Gets the enclosed token.
   *
   * @return the enclosed token
   */
  public BpelExecution getBpelExecution() {
    return this.bpelExecution;
  }

  public Object getValueForAssign(final String variableName) {
    return this.bpelExecution.getVariableValue(variableName);
  }

  public void setValue(final String variableName, final Object value) {
    this.bpelExecution.getVariableByName(variableName).setValue(value, false);
  }

  public Object getPropertyValue(final String variableName, final QName propertyQName)
    throws FunctionCallException {

    VariableRuntime variable = null;

    variable = this.bpelExecution.getVariableByName(variableName);
    if (variable == null) {
      throw new FunctionCallException(
          "Unable to locate a bound variable for variableName : "
              + variableName);
    }
    PropertyAlias propertyAlias = null;

    Element value = null;
    final Variable variableDefinition = variable.getVariableDefinition();
    if (variable.getVariableDefinition().getMessageType() != null) {
      propertyAlias = this.bpelExecution.getPropertyAliasWithMessageType(
          propertyQName, variableDefinition.getMessageType());
    } else if (variable.getVariableDefinition().getElement() != null) {
      propertyAlias = this.bpelExecution.getPropertyAliasWithElement(propertyQName,
          variableDefinition.getElement());
    } else if (variable.getVariableDefinition().getXmlType() != null) {
      propertyAlias = this.bpelExecution.getPropertyAliasWithType(propertyQName,
          variableDefinition.getXmlType());
    } else {
      throw new IllegalStateException("Invalid variable type");
    }

    if (variable.getVariableDefinition().getMessageType() != null) {
      final MessageVariable varValue = (MessageVariable) variable.getValue();
      value = varValue.getPartValue(propertyAlias.getPart());
    } else {
      value = (Element) variable.getValue();
    }
    final PropertyQuery query = propertyAlias.getQuery();
    if (query != null) {
      return query.getEvaluator().evaluate(value, this.bpelExecution);
    } else {
      return value;
    }
  }

  /**
   * Finds the definition of the specified variable in the scope of the
   * {@linkplain Activity activity} where the enclosed token is.
   *
   * @param variableName
   *            name of the desired variable
   * @return the variable definition or <code>null</code> if there is no match
   */
  public Variable findVariableDefinition(final String variableName) {
    final VariableRuntime variableRuntime = this.bpelExecution
        .getVariableByName(variableName);
    if (variableRuntime == null) {
      return null;
    }
    return variableRuntime.getVariableDefinition();
  }

  /**
   * Returns the value of a process variable based on the given local name.
   *
   * @return the value of the resolved variable
   * @throws UnresolvableException
   *             in any of the following situations:
   *             <ul>
   *             <li><code>namespaceURI</code> is not <code>null</code>;
   *             process variables are always unqualified</li>
   *             <li><code>localName</code> does not match any process
   *             variable</li>
   *             </ul>
   */
  public Object getVariableValue(final String namespaceURI, final String prefix,
      final String localName) throws UnresolvableException {
    if (namespaceURI != null) {
      throw new UnresolvableException("variable not found: " + prefix + ':'
          + localName);
    }

    Object value = null;
    // look for a dot in the variable name, indicating a message part access
    final int dotIndex = localName.indexOf('.');
    if (dotIndex == -1) {
      // dotless name => variable must be defined as schema type or element
      // find variable definition
      final Variable definition = this.findVariableDefinition(localName);
      if (definition == null) {
        throw new UnresolvableException("variable not found: " + localName);
      }

      // extract variable value
      value = this.bpelExecution.getVariableValue(localName);

      // prevent direct access to message variable
      if (definition.getMessageType() != null) {
        throw new UnresolvableException(
            "illegal access to message variable: " + localName);
      }

      if (definition.getXmlType() != null && BpelXmlUtil.isDefaultElement((Element) value)) {
        final Element valueElement = (Element) value;
        if (XmlConstants.XMLNS_XMLSCHEMA.equals(definition.getXmlType().getNamespaceURI())) {
          if (definition.getXmlType().getLocalPart().equals("boolean")) {
            value = Boolean.valueOf(valueElement.getTextContent());
          } else if (definition.getXmlType().getLocalPart().equals("float")
              || definition.getXmlType().getLocalPart().equals("int")
              || definition.getXmlType().getLocalPart().equals("unsignedInt")) {
            value = Double.valueOf(valueElement.getTextContent());
          } else {
            value = valueElement.getTextContent();
          }
        }
      }
    } else {
      // dotted name => variable must be defined as messsage
      final String messageName = localName.substring(0, dotIndex);
      // find variable definition
      final Variable definition = this.findVariableDefinition(messageName);
      if (definition == null) {
        throw new UnresolvableException("variable not found: " + messageName);
      }

      // extract variable value
      value = this.bpelExecution.getVariableValue(messageName);

      // prevent access to part of non-message variable
      if (!(definition.getMessageType() != null)) {
        throw new UnresolvableException(
            "illegal access to part of non-message variable: " + localName);
      }

      // retrieve part value
      final MessageVariable messageValue = (MessageVariable) value;
      final String partName = localName.substring(dotIndex + 1);
      value = messageValue.getPartValue(partName);

      final javax.wsdl.Message wsdlMessage = this.bpelExecution.getWsdlInfos().getMessage(definition.getMessageType());
      if (wsdlMessage.getPart(partName).getTypeName() != null && BpelXmlUtil.isDefaultElement((Element) value)) {
        final Element valueElement = (Element) value;
        final QName typeName = wsdlMessage.getPart(partName).getTypeName();
        if (XmlConstants.XMLNS_XMLSCHEMA.equals(typeName.getNamespaceURI())) {
          if (typeName.getLocalPart().equals("boolean")) {
            value = Boolean.valueOf(valueElement.getTextContent());
          } else if (typeName.getLocalPart().equals("float")
              || typeName.getLocalPart().equals("int")
              || typeName.getLocalPart().equals("unsignedInt")) {
            value = Double.valueOf(valueElement.getTextContent());
          } else {
            value = valueElement.getTextContent();
          }
        }
      }
    }
    return value;
  }

  public Map<String, byte[]> getStyleSheets() {
    return this.bpelExecution.getStyleSheets();
  }
}
