/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the JBPM BPEL PUBLIC LICENSE AGREEMENT as
 * published by JBoss Inc.; either version 1.0 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
package org.ow2.orchestra.lang.jaxen;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jaxen.BaseXPath;
import org.jaxen.Context;
import org.jaxen.ContextSupport;
import org.jaxen.FunctionContext;
import org.jaxen.JaxenException;
import org.jaxen.SimpleFunctionContext;
import org.jaxen.XPathFunctionContext;
import org.jaxen.dom.DocumentNavigator;
import org.jaxen.expr.CommentNodeStep;
import org.jaxen.expr.LocationPath;
import org.jaxen.expr.NameStep;
import org.jaxen.expr.Predicate;
import org.jaxen.expr.ProcessingInstructionNodeStep;
import org.jaxen.expr.Step;
import org.jaxen.expr.TextNodeStep;
import org.jaxen.saxpath.Axis;
import org.ow2.orchestra.facade.exception.BpelFaultException;
import org.ow2.orchestra.facade.exception.BpelFaultException.BpelFault;
import org.ow2.orchestra.lang.function.DoXslTranformationFunction;
import org.ow2.orchestra.lang.function.GetVariablePropertyFunction;
import org.ow2.orchestra.util.BpelXmlUtil;
import org.ow2.orchestra.util.XmlConstants;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * @author Alejandro Guizar
 * @version $Revision: 1.9 $ $Date: 2007/09/04 06:42:26 $
 */
abstract class XPathEvaluator extends BaseXPath {

  private static final FunctionContext EMPTY_FUNCTION_CONTEXT = new SimpleFunctionContext();

  private static final Logger LOG = Logger.getLogger(XPathEvaluator.class.getName());

  private static final long serialVersionUID = 1L;

  XPathEvaluator(final String text) throws JaxenException {
    super(text, DocumentNavigator.getInstance());
  }

  protected List selectOrCreateNodes(final LocationPath location,
      final Context context) throws JaxenException {
    List contextNodes = context.getNodeSet();
    // empty context nodeset -> empty selection
    if (contextNodes.isEmpty()) {
      return Collections.EMPTY_LIST;
    }
    final ContextSupport support = context.getContextSupport();

    // absolute path?
    if (location.isAbsolute()) {
      // start from the root node
      final Object rootNode = support.getNavigator().getDocumentNode(
          contextNodes.get(0));
      contextNodes = Collections.singletonList(rootNode);
    }

    final Iterator stepIter = location.getSteps().iterator();
    while (stepIter.hasNext()) {
      // prepare the context for the current step
      final Context stepContext = new Context(support);
      stepContext.setNodeSet(contextNodes);
      // evaluate the step, capture the selected nodes
      final Step step = (Step) stepIter.next();
      contextNodes = step.evaluate(stepContext);
      // no node was selected?
      if (contextNodes.isEmpty()) {
        // try to create the missing node
        final Node newNode = this.createNode(step, stepContext);
        // create a new list, since the existing empty list is immutable
        contextNodes = Collections.singletonList(newNode);
      }
    }
    return contextNodes;
  }

  private Node createNode(final Step step, final Context context) {
    final List nodeset = context.getNodeSet();
    if (nodeset.size() != 1) {
      XPathEvaluator.LOG
          .severe("cannot create node for context node set of size other than one: "
              + nodeset);
      throw BpelFaultException.generateBpelFaultException(BpelFault.selectionFailure);
    }

    final Object contextNode = nodeset.get(0);
    if (!(contextNode instanceof Element)) {
      XPathEvaluator.LOG.severe("cannot create node for non-element context node: "
          + contextNode);
      throw BpelFaultException.generateBpelFaultException(BpelFault.selectionFailure);
    }

    final Element contextElem = (Element) contextNode;
    Node newNode;

    switch (step.getAxis()) {
      case Axis.ATTRIBUTE:
        newNode = this.createAttribute(step, context, contextElem);
        break;
      case Axis.CHILD:
        if (step instanceof NameStep) {
          newNode = this.createElement((NameStep) step, context, contextElem);
        } else {
          newNode = this.createNonElementChild(step, contextElem);
        }
        contextElem.appendChild(newNode);
        break;
      default:
        XPathEvaluator.LOG.severe("cannot create node on the specified axis: " + step);
        throw BpelFaultException.generateBpelFaultException(BpelFault.selectionFailure);
    }
    return newNode;
  }

  private Attr createAttribute(final Step step, final Context context,
      final Element contextElem) {
    if (!step.getPredicates().isEmpty()) {
      XPathEvaluator.LOG.severe("cannot create attribute for step with predicates: " + step);
      throw BpelFaultException.generateBpelFaultException(BpelFault.selectionFailure);
    }

    if (!(step instanceof NameStep)) {
      XPathEvaluator.LOG.severe("cannot create attribute for non-name node test: " + step);
      throw BpelFaultException.generateBpelFaultException(BpelFault.selectionFailure);
    }

    final NameStep nameStep = (NameStep) step;
    final String localName = nameStep.getLocalName();

    if ("*".equals(localName)) {
      XPathEvaluator.LOG.severe("cannot create attribute for any-name node test: " + nameStep);
      throw BpelFaultException.generateBpelFaultException(BpelFault.selectionFailure);
    }

    // BPEL-191: preserve source prefix
    final String prefix = nameStep.getPrefix();
    final String namespaceURI = context.translateNamespacePrefixToUri(prefix);

    final Attr attribute = contextElem.getOwnerDocument().createAttributeNS(
        namespaceURI,
        namespaceURI != null ? prefix + ':' + localName : localName);
    contextElem.setAttributeNodeNS(attribute);

    return attribute;
  }

  private Element createElement(final NameStep step, final Context context,
      final Element contextElem) {
    final String localName = step.getLocalName();
    if ("*".equals(localName)) {
      XPathEvaluator.LOG.severe("cannot create node for any-name node tests: " + step);
      throw BpelFaultException.generateBpelFaultException(BpelFault.selectionFailure);
    }

    // BPEL-191: preserve source prefix
    final String prefix = step.getPrefix();
    final String namespaceURI = context.translateNamespacePrefixToUri(prefix);

    final List predicates = step.getPredicates();
    switch (predicates.size()) {
      case 0:
        break;
      case 1: {
        // prepare the context for the predicate
        final List contextNodes = this.getElements(contextElem, namespaceURI, localName);
        final Context predicateContext = new Context(context.getContextSupport());
        predicateContext.setNodeSet(contextNodes);

        try {
          // evaluate the predicate, capture the result
          final Predicate predicate = (Predicate) predicates.get(0);
          final Object result = predicate.evaluate(predicateContext);

          if (!(result instanceof Number)) {
            XPathEvaluator.LOG
                .severe("cannot create element for step with non-numeric predicate: "
                    + step);
            throw BpelFaultException.generateBpelFaultException(BpelFault.selectionFailure);
          }

          if (((Number) result).intValue() != contextNodes.size() + 1) {
            XPathEvaluator.LOG.severe("cannot create element for step with numeric predicate "
                + "beyond the next position: " + step);
            throw BpelFaultException.generateBpelFaultException(BpelFault.selectionFailure);
          }
        } catch (final JaxenException e) {
          XPathEvaluator.LOG.log(Level.SEVERE, "predicate evaluation failed", e);
          throw BpelFaultException.generateBpelFaultException(BpelFault.subLanguageExecutionFault);

        }
        break;
      }
      default:
        XPathEvaluator.LOG.severe("cannot create element for step with multiple predicates: "
            + step);
      throw BpelFaultException.generateBpelFaultException(BpelFault.selectionFailure);
    }

    final String qualifiedName = namespaceURI != null ? prefix + ':' + localName
        : localName;
    final Element newElt = contextElem.getOwnerDocument().createElementNS(namespaceURI,
        qualifiedName);
    contextElem.appendChild(newElt);
    BpelXmlUtil.ensureOwnNamespaceDeclared(newElt);
    return newElt;
  }

  private Node createNonElementChild(final Step step, final Element contextElem) {
    if (!step.getPredicates().isEmpty()) {
      throw BpelFaultException.generateBpelFaultException(
          BpelFault.selectionFailure,
          "cannot create node for step with predicates: " + step,
          null);
    }

    final Document contextDoc = contextElem.getOwnerDocument();

    if (step instanceof TextNodeStep) {
      return contextDoc.createTextNode("");
    }

    if (step instanceof ProcessingInstructionNodeStep) {
      final ProcessingInstructionNodeStep processingStep = (ProcessingInstructionNodeStep) step;
      return contextDoc.createProcessingInstruction(processingStep.getName(),
          "");
    }

    if (step instanceof CommentNodeStep) {
      return contextDoc.createComment("");
    }

    throw BpelFaultException.generateBpelFaultException(
        BpelFault.selectionFailure,
        "cannot create node for any-node tests on the child axis: " + step,
        null);
  }

  private List getElements(final Element parent, final String namespaceURI,
      final String localName) {
    final ArrayList elements = new ArrayList();
    for (Node child = parent.getFirstChild(); child != null; child = child
        .getNextSibling()) {
      if (BpelXmlUtil.nodeNameEquals(child, namespaceURI, localName)) {
        elements.add(child);
      }
    }
    return elements;
  }

  protected Node narrowToSingleNode(final List nodeset) {
    XPathEvaluator.LOG.finest("narrowing to single node: " + nodeset);
    if (nodeset == null || nodeset.size() != 1) {
      throw BpelFaultException.generateBpelFaultException(BpelFault.selectionFailure,
          "selection of size other than one: " + nodeset,
          null);
    }
    final Object singleResult = nodeset.get(0);
    if (!(singleResult instanceof Node)) {
      throw BpelFaultException.generateBpelFaultException(
          BpelFault.selectionFailure,
          "selection is not a node: " + singleResult,
          null);
    }
    return (Node) singleResult;
  }

  protected static FunctionContext readFunctionLibrary() {
    final SimpleFunctionContext fc = new XPathFunctionContext();
    fc.registerFunction(XmlConstants.XMLNS_BPEL_2_0_EXECUTABLE,
        "getVariableProperty", new GetVariablePropertyFunction());
    fc.registerFunction(XmlConstants.XMLNS_BPEL_2_0_EXECUTABLE,
        "doXslTransform", new DoXslTranformationFunction());
    return fc;
  }
}
