/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the JBPM BPEL PUBLIC LICENSE AGREEMENT as
 * published by JBoss Inc.; either version 1.0 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
package org.ow2.orchestra.lang.jaxen;

import java.util.Collections;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.Duration;
import javax.xml.datatype.XMLGregorianCalendar;

import org.jaxen.Context;
import org.jaxen.ContextSupport;
import org.jaxen.FunctionContext;
import org.jaxen.JaxenException;
import org.jaxen.expr.Expr;
import org.jaxen.expr.FilterExpr;
import org.jaxen.expr.FunctionCallExpr;
import org.jaxen.expr.LiteralExpr;
import org.jaxen.expr.LocationPath;
import org.jaxen.expr.NumberExpr;
import org.jaxen.expr.Predicate;
import org.jaxen.expr.PredicateSet;
import org.jaxen.expr.VariableReferenceExpr;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.facade.exception.BpelFaultException;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.facade.exception.BpelFaultException.BpelFault;
import org.ow2.orchestra.lang.evaluator.ExpressionEvaluator;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.util.BpelXmlUtil;
import org.ow2.orchestra.var.MessageVariable;
import org.w3c.dom.Node;

/**
 * Evaluator for general expressions. When XPath 1.0 is used as an expression
 * language, there is no context node available.
 *
 * @author Alejandro Guizar
 * @version $Revision: 1.10 $ $Date: 2007/09/12 23:20:16 $
 */
class XPathExpressionEvaluator extends XPathEvaluator implements
    ExpressionEvaluator {

  // Create only one instance.
  // TODO: Is it thread safe? Not sure! Check!
  private static DatatypeFactory datatypeFactory;

  static {
    try {
      XPathExpressionEvaluator.datatypeFactory = DatatypeFactory.newInstance();
    } catch (final DatatypeConfigurationException e) {
      throw new ExceptionInInitializerError(e);
    }
  }

  /**
   * The jBPM configuration object name for the resource describing the function
   * library available to expressions.
   */

  private static final long serialVersionUID = 1L;

  private static final Logger LOG = Logger.getLogger(XPathExpressionEvaluator.class.getName());

  private static final FunctionContext FUNTION_LIBRARY = XPathEvaluator.readFunctionLibrary();

  XPathExpressionEvaluator(final String text) throws JaxenException {
    super(text);
  }

  /**
   * Evaluates this expression. The context of XPath evaluation is:
   * <ul>
   * <li>the context node, position and size are unspecified</li>
   * <li>the function library, namespace declaration and variable bindings are
   * obtained by a call to {@link #getContext(Object)} with the given
   * <code>contextInfo</code> as argument</li>
   * <ul>
   */
  public Object evaluate(final BpelExecution bpelExecution) {
    try {
      final List nodeSet = this.selectNodes(bpelExecution);
      if (nodeSet != null) {
        switch (nodeSet.size()) {
          case 0:
            break;
          case 1:
            return nodeSet.get(0);
          default:
            return nodeSet;
        }
      }
      return null;
    } catch (final JaxenException e) {
      XPathExpressionEvaluator.LOG.log(Level.INFO, "expression evaluation failed", e);
      throw BpelFaultException.generateBpelFaultException(BpelFault.subLanguageExecutionFault);
    }
  }

  public String assign(final BpelExecution bpelExecution, final Object value, final boolean keepSrcEltName) {
    final Expr rootExpr = this.getRootExpr();
    final Context context = this.getContext(bpelExecution);
    // direct variable assignment?
    if (rootExpr instanceof VariableReferenceExpr) {
      return XPathExpressionEvaluator.assignVariable((VariableReferenceExpr) rootExpr, context, value, keepSrcEltName);
    } else {
      try {
        return new ExprAssigner(this).assign(rootExpr, context, value, keepSrcEltName);
      } catch (final JaxenException e) {
        XPathExpressionEvaluator.LOG.log(Level.INFO, "expression assignment failed", e);
        throw BpelFaultException.generateBpelFaultException(BpelFault.subLanguageExecutionFault);
      }
    }
  }

  @Override
  protected FunctionContext createFunctionContext() {
    return XPathExpressionEvaluator.FUNTION_LIBRARY;
  }

  /**
   * Gets a context for XPath evaluation. The context for an expression
   * expression is:
   * <ul>
   * <li>the variable bindings are the variables in the scope of the given
   * execution</li>
   * <li>the function library is {@link #FUNTION_LIBRARY}</li>
   * <li>the namespace declarations are taken from the snippet</li>
   * </ul>
   *
   * @param execution
   */
  @Override
  protected Context getContext(final Object o) {
    if (!(o instanceof BpelExecution)) {
      return super.getContext(o);
    }
    final BpelExecution bpelExecution = (BpelExecution) o;
    final ContextSupport support = this.getContextSupport();
    support.setVariableContext(new BpelVariableContext(bpelExecution));

    /*
     * ContextSupport support = new ContextSupport(getNamespaceContext(),
     * getFunctionContext(), new BpelVariableContext(bpelExecution),
     * getNavigator());
     */

    return new Context(support);
  }

  private static String assignVariable(final VariableReferenceExpr rootExpr,
      final Context context, final Object value, final boolean keepSrcEltName) {
    final String variableName = rootExpr.getVariableName();
    final BpelVariableContext variableContext = (BpelVariableContext) context
        .getContextSupport().getVariableContext();
    // look for a dot in the variable name, indicating a message part access
    final int dotIndex = variableName.indexOf('.');
    if (dotIndex == -1) {
      // dotless name => variable must be defined as schema xmlType or element
      // find variable definition
      final Variable variableDefinition = variableContext
          .findVariableDefinition(variableName);
      if (variableDefinition == null) {
        throw new OrchestraRuntimeException("variable not found: " + variableName);
      }

      // prevent direct access to a message variable
      if (variableDefinition.getMessageType() != null) {
        throw new OrchestraRuntimeException("illegal access to message variable: "
            + variableName);
      }

      // assign variable value
      variableContext.setValue(variableName, value);
      return variableName;
    } else {
      // dotted name => variable must be defined as messsage
      final String messageName = variableName.substring(0, dotIndex);

      // find variable definition
      final Variable variableDefinition = variableContext
          .findVariableDefinition(messageName);
      if (variableDefinition == null) {
        throw new OrchestraRuntimeException("variable not found: " + messageName);
      }

      // extract variable value
      final Object valueForAssign = variableContext.getValueForAssign(messageName);

      // prevent access to part of non-message variable
      if (!(variableDefinition.getMessageType() != null)) {
        throw new OrchestraRuntimeException(
            "illegal access to part of non-message variable: " + variableName);
      }

      // assign part value
      final MessageVariable messageValue = (MessageVariable) valueForAssign;
      final String partName = variableName.substring(dotIndex + 1);
      messageValue.setPart(partName, value, keepSrcEltName);
      return messageName;
    }
  }

  private static class ExprAssigner extends ExprVisitor {

    private Context myContext;

    private Object result;

    private String variableName;

    private JaxenException jaxenException;

    private final XPathEvaluator xpathEvaluator;


    public ExprAssigner(final XPathEvaluator xpathEvaluator) {
      this.xpathEvaluator = xpathEvaluator;
    }

    public String assign(final Expr pathExpr, final Context context, final Object value, final boolean keepSrcEltName)
      throws JaxenException {
      this.myContext = context;
      this.visit(pathExpr);

      if (this.jaxenException != null) {
        throw this.jaxenException;
      }
      Node node = null;
      if (this.result instanceof List) {
        node = this.xpathEvaluator.narrowToSingleNode((List) this.result);
      } else {
        node = (Node) this.result;
      }
      BpelXmlUtil.setObjectValue(node, value, keepSrcEltName);
      return this.variableName;
    }

    @Override
    public void visit(final FilterExpr filterExpr) {
      this.visit(filterExpr.getExpr());
      // if result is not a list, then there is nothing to filter
      if (this.result instanceof List) {
        try {
          this.result = this.evaluatePredicates(filterExpr.getPredicateSet(),
              (List< ? >) this.result, this.myContext.getContextSupport());
        } catch (final JaxenException e) {
          this.jaxenException = e;
        }
      }
    }

    @Override
    public void visit(final LocationPath locationPath) {
      if (this.result instanceof List) {
        this.myContext.setNodeSet((List< ? >) this.result);
      } else {
        this.myContext.setNodeSet(Collections.singletonList(this.result));
      }
      try {
        this.result = this.xpathEvaluator.selectOrCreateNodes(locationPath, this.myContext);
      } catch (final JaxenException e) {
        this.jaxenException = e;
      }
    }

    @Override
    public void visit(final VariableReferenceExpr varExpr) {
      final String variableName = varExpr.getVariableName();
      final BpelVariableContext variableContext = (BpelVariableContext) this.myContext
          .getContextSupport().getVariableContext();
      // look for a dot in the variable name, indicating a message part access
      final int dotIndex = variableName.indexOf('.');
      if (dotIndex == -1) {
        // dotless name => variable must be defined as schema xmlType or element
        // find variable definition
        final Variable variableDefinition = variableContext
            .findVariableDefinition(variableName);
        if (variableDefinition == null) {
          throw new OrchestraRuntimeException("variable not found: " + variableName);
        }
        // variable value for assign
        this.result = variableContext.getValueForAssign(variableName);
        this.variableName = variableName;
      } else {
        // dotted name => variable must be defined as messsage
        final String messageName = variableName.substring(0, dotIndex);

        // find variable definition
        final Variable definition = variableContext
            .findVariableDefinition(messageName);
        if (definition == null) {
          throw new OrchestraRuntimeException("variable not found: " + messageName);
        }
        // prevent access to a non-existent part
        if (definition.getMessageType() == null) {
          throw new OrchestraRuntimeException(
              "illegal access to part of non-message variable: " + variableName);
        }
        // part value for assign
        final MessageVariable messageValue = (MessageVariable) variableContext
            .getValueForAssign(messageName);
        final String partName = variableName.substring(dotIndex + 1);
        if (!messageValue.hasPart(partName)) {
          throw new OrchestraRuntimeException("illegal access to message part " + partName);
        }
        this.result = messageValue.getPartValue(partName);
        this.variableName = messageName;
      }
    }

    @Override
    public void visit(final LiteralExpr literalExpr) {
      try {
        this.result = literalExpr.evaluate(this.myContext);
      } catch (final JaxenException e) {
        this.jaxenException = e;
      }
    }

    @Override
    public void visit(final NumberExpr numberExpr) {
      try {
        this.result = numberExpr.evaluate(this.myContext);
      } catch (final JaxenException e) {
        this.jaxenException = e;
      }
    }

    @Override
    public void visit(final FunctionCallExpr callExpr) {
      try {
        this.result = callExpr.evaluate(this.myContext);
      } catch (final JaxenException e) {
        this.jaxenException = e;
      }
    }

    @SuppressWarnings("unchecked")
    private List< ? > evaluatePredicates(final PredicateSet predicateSet,
        List< ? > nodes, final ContextSupport support) throws JaxenException {
      final List<Predicate> predicates = predicateSet
          .getPredicates();
      for (final Predicate predicate : predicates) {
        nodes = predicateSet.applyPredicate(predicate, nodes, support);
      }
      return nodes;
    }
  }

  public boolean evaluateBoolean(final BpelExecution bpelExecution) throws BpelFaultException {
    try {
      //return booleanValueOf(evaluate(bpelExecution));
      // TODO: why can't we use the booleanValueOf() method here? Unclear to me!
      return (Boolean) this.evaluate(bpelExecution);
    //} catch (JaxenException e) {
    } catch (final ClassCastException e) {
      throw BpelFaultException.generateBpelFaultException(
          BpelFault.invalidExpressionValue,
          "expression is not a boolean",
          e);
    }
  }

  public XMLGregorianCalendar evaluateDeadline(final BpelExecution bpelExecution)
    throws BpelFaultException {
    try {
      final String s = this.stringValueOf(this.evaluate(bpelExecution));
      return XPathExpressionEvaluator.datatypeFactory.newXMLGregorianCalendar(s);
    } catch (final JaxenException e) {
      throw BpelFaultException.generateBpelFaultException(
          BpelFault.invalidExpressionValue,
          e.getMessage(),
          e);
    } catch (final IllegalArgumentException iae) {
      throw BpelFaultException.generateBpelFaultException(
          BpelFault.invalidExpressionValue,
          iae.getMessage(),
          iae);
    }
  }

  public Duration evaluateDuration(final BpelExecution bpelExecution)
    throws BpelFaultException {
    try {
      final String s = this.stringValueOf(this.evaluate(bpelExecution));
      return XPathExpressionEvaluator.datatypeFactory.newDuration(s);
    } catch (final JaxenException e) {
      throw BpelFaultException.generateBpelFaultException(
          BpelFault.invalidExpressionValue,
          e.getMessage(),
          e);
    } catch (final IllegalArgumentException iae) {
      throw BpelFaultException.generateBpelFaultException(
          BpelFault.invalidExpressionValue,
          iae.getMessage(),
          iae);
    }
  }


  public int evaluateUnsignedInt(final BpelExecution bpelExecution)
    throws BpelFaultException {
    try {
      final Number nb = this.numberValueOf(this.evaluate(bpelExecution));
      if (nb.equals(Double.NaN)) {
        throw BpelFaultException.generateBpelFaultException(
            BpelFault.invalidExpressionValue,
            ExpressionEvaluator.NaNExceptionMessage,
            null);
      }
      final int n = nb.intValue();
      if (n < 0) {
        throw BpelFaultException.generateBpelFaultException(
            BpelFault.invalidExpressionValue,
            ExpressionEvaluator.NegativeIntExceptionMessage,
            null);
      }
      return n;
    } catch (final JaxenException e) {
      throw BpelFaultException.generateBpelFaultException(
          BpelFault.invalidExpressionValue,
          e.getMessage(),
          e);
    }
  }
}

