/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 *
 * ConfigurationFileParser.java
 *
 * @author Goulven Le Jeune, Charles Souillard, Candy Chlad, Stanislas Giraudet De Boudemange
 *
 * Created on : Feb 21st 07
 *
 * $Id: BasicFileParser.java 2506 2009-04-29 12:52:28Z porcherg $
 **/
package org.ow2.orchestra.parsing;

import java.io.InputStream;
import java.net.URL;

import javax.xml.XMLConstants;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;

import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;

/**
 * BasicFileParser. Parses the engine and processes configuration files to
 * configure an object.
 * 
 * @author Goulven Le Jeune, Charles Souillard, Candy Chlad, Stanislas Giraudet
 *         De Boudemange, Guillaume Porcher
 * 
 * Created on : Feb 21st 07
 */
public abstract class BasicFileParser {

  /**
   * Returns the schema to use to parse the conf file.
   * 
   * @return the schema to use to parse the conf file.
   */
  protected abstract String getSchema();

  /**
   * Set up a SAX parser and return the object configured according to the
   * configuration file.
   * 
   * @param file
   *            The configuration file to be parsed
   * @return The process Configuration
   */
  public Object parseAndSetObject(Object target, final InputStream inputStream) {
    BasicHandler handler;
    try {
      // Create the Parser factory
      handler = new BasicHandler(target);

      SAXParserFactory parserFactory = SAXParserFactory.newInstance();

      // Create a Schema with a Schema factory
      SchemaFactory schemaFactory = SchemaFactory
          .newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);

      // look for the schema in the Orchestra archive.
      URL schemaUrl = getClass().getClassLoader().getResource(getSchema());

      Schema schema = schemaFactory.newSchema(schemaUrl);

      parserFactory.setSchema(schema);
      parserFactory.setNamespaceAware(true);

      // Parse the configuration file with the validating parser
      SAXParser saxParser = parserFactory.newSAXParser();

      saxParser.parse(inputStream, handler);

      // Return the Configuration filled by the parser
      return handler.getResult();

    } catch (Exception e) {
      throw new OrchestraRuntimeException(e);
    }
  }
}
