package org.ow2.orchestra.parsing;


import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;

import javax.wsdl.Definition;
import javax.wsdl.Operation;
import javax.wsdl.PortType;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.exception.StaticAnalysisException;
import org.ow2.orchestra.facade.def.full.ProcessFullDefinition;
import org.ow2.orchestra.facade.def.full.impl.ProcessFullDefinitionImpl;
import org.ow2.orchestra.facade.exception.BpelFaultException;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.facade.wsdl.WsdlDefinitions;
import org.ow2.orchestra.parsing.binding.AssignBinding;
import org.ow2.orchestra.parsing.binding.CompensateBinding;
import org.ow2.orchestra.parsing.binding.CompensateScopeBinding;
import org.ow2.orchestra.parsing.binding.DefinitionKeeper;
import org.ow2.orchestra.parsing.binding.EmptyBinding;
import org.ow2.orchestra.parsing.binding.ExitBinding;
import org.ow2.orchestra.parsing.binding.ExtensionActivityBinding;
import org.ow2.orchestra.parsing.binding.FlowBinding;
import org.ow2.orchestra.parsing.binding.ForEachBinding;
import org.ow2.orchestra.parsing.binding.IfBinding;
import org.ow2.orchestra.parsing.binding.InvokeBinding;
import org.ow2.orchestra.parsing.binding.PickBinding;
import org.ow2.orchestra.parsing.binding.ProcessBinding;
import org.ow2.orchestra.parsing.binding.ReceiveBinding;
import org.ow2.orchestra.parsing.binding.RepeatUntilBinding;
import org.ow2.orchestra.parsing.binding.ReplyBinding;
import org.ow2.orchestra.parsing.binding.RethrowBinding;
import org.ow2.orchestra.parsing.binding.ScopeBinding;
import org.ow2.orchestra.parsing.binding.SequenceBinding;
import org.ow2.orchestra.parsing.binding.ThrowBinding;
import org.ow2.orchestra.parsing.binding.ValidateBinding;
import org.ow2.orchestra.parsing.binding.WaitBinding;
import org.ow2.orchestra.parsing.binding.WhileBinding;
import org.ow2.orchestra.pvm.PvmException;
import org.ow2.orchestra.pvm.internal.model.NodeImpl;
import org.ow2.orchestra.pvm.internal.util.UrlEntity;
import org.ow2.orchestra.pvm.internal.xml.Bindings;
import org.ow2.orchestra.pvm.internal.xml.Entity;
import org.ow2.orchestra.pvm.internal.xml.Parse;
import org.ow2.orchestra.pvm.internal.xml.Parser;
import org.ow2.orchestra.runtime.FaultHandlerThatExits;
import org.ow2.orchestra.util.StaticAnalysisFault;
import org.ow2.orchestra.util.StyleSheetRepository;
import org.ow2.orchestra.util.XmlConstants;
import org.ow2.orchestra.util.XmlUtil;
import org.ow2.orchestra.wsdl.BpelWsdlParser;
import org.ow2.orchestra.wsdl.WsdlsInfos;
import org.w3c.dom.Element;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/** parses the dom tree of the bpel document and stores the
 * builds the BpelProcess.   The Deployer will make the BpelProcess
 * available in the object stack of the parse.
 *
 * @author Tom Baeyens, Charles Souillard
 */
public class BpelParser extends Parser implements ErrorHandler {

  /** path to the directory containing all xsd files. */
  private static final String RESSOURCES_DIR = "";
  /** all schema resources used to parse bpel files. */
  private static final String[] SCHEMA_RESOURCES = {
    BpelParser.RESSOURCES_DIR + "wsdl.xsd", BpelParser.RESSOURCES_DIR + "xml-schema.xsd",
    BpelParser.RESSOURCES_DIR + "ws-bpel_abstract_common_base.xsd", BpelParser.RESSOURCES_DIR + "ws-bpel_executable.xsd"};

  public static final String CATEGORY_ACTIVITY = "activities";

  /** all bindings used for parsing. */
  private static Bindings defaultBindings = BpelParser.getDefaultBindings();
  /** schema document URIs as stringsF. */
  private static String[] schemaSources = BpelParser.getSchemaSources();

  /** all allowed tag names for process top level activities. */
  //public static final Set<String> processActivityTagNames = getProcessActivityTagNames();

  // the default entities are initialized at the bottom of this file.
  private static Map<String, Entity> defaultEntities = BpelParser.getDefaultEntities();

  private BpelWsdlParser wsdlParser = null;
  /**
   * Default constructor.
   */
  public BpelParser() {
    super(BpelParser.defaultBindings, BpelParser.defaultEntities);
    this.getDocumentBuilderFactory().setNamespaceAware(true);
    this.wsdlParser = new BpelWsdlParser();
  }

  /**
   * Parses a the dom document representing the bpel file.
   * @param processElement root element of the document : the process element
   * @param parse parse
   * @return the resulting bpelProcess
   */
  @Override
  public Object parseDocumentElement(final Element processElement, final Parse parse) {
    final BpelProcess bpelProcess = this.parseProcessAttributes(processElement, parse);
    parse.pushObject(bpelProcess);
    try {
    bpelProcess.setStyleSheetRepository(parse.findObject(StyleSheetRepository.class));
    bpelProcess.setWsdlInfos(parse.findObject(WsdlsInfos.class));
    final FaultHandlerThatExits faultHandlerThatExits = new FaultHandlerThatExits();
    faultHandlerThatExits.setExceptionClassName(BpelFaultException.class.getName());
    bpelProcess.addExceptionHandler(faultHandlerThatExits);

    final List<Element> elements = XmlUtil.elements(processElement);
    if (elements == null || elements.isEmpty()) {
      parse.addProblem("no elements in process");
      return bpelProcess;
    }

    this.parseProcessElements(processElement, bpelProcess, parse);

    // creates a scope associated to the process : in the bindings : process is mapped to ScopeBinding
    // let's create the top level node for the bpel process.

    final NodeImpl processScope = (NodeImpl) this.parseElement(processElement, parse, BpelParser.CATEGORY_ACTIVITY);
    bpelProcess.setInitial(processScope);
    } finally {
      parse.popObject();
    }
    return bpelProcess;
  }

  /**
   * Parses all process specific attributes (those not common with scope).
   * @param processElement process element (dom)
   * @param bpelProcess bpelProcess
   * @param parse parse
   */
  private BpelProcess parseProcessAttributes(final Element processElement, final Parse parse) {
    final String name = XmlUtil.attribute(processElement, "name");

    final String targetNamespace = XmlUtil.attribute(processElement, "targetNamespace");

    final BpelProcess bpelProcess = new BpelProcess(name, targetNamespace, UUID.randomUUID().toString());

    final String queryLanguage = XmlUtil.attribute(processElement, "queryLanguage");
    bpelProcess.setQueryLanguage(queryLanguage);

    final String expressionLanguage = XmlUtil.attribute(processElement, "expressionLanguage");
    bpelProcess.setExpressionLanguage(expressionLanguage);

    final ProcessDefinitionUUID processDefinitionUUID = bpelProcess.getUUID();


    final ProcessFullDefinition processFullDefinition =
      new ProcessFullDefinitionImpl(
        processDefinitionUUID,
        targetNamespace,
        name,
        queryLanguage,
        expressionLanguage,
        new Date());

    final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
    definitionKeeper.setProcessFullDefinition(processFullDefinition);

    return bpelProcess;
  }

  /**
   * Parses all process specific children elements (those not common with scope).
   * @param elements list of all process children elements
   * @param bpelProcess bpelProcess
   * @param parse parse
   */
  private void parseProcessElements(final Element processElement, final BpelProcess bpelProcess, final Parse parse) {
    final Element extensions = XmlUtil.element(processElement, "extensions");
    if (extensions != null) {
      parse.addProblem("extensions element not yet supported on process");
    }

    final List<Element> importElements = XmlUtil.elements(processElement, "import");
    if (importElements != null) {
      for (final Element importElement : importElements) {
        this.parseImportElement(importElement, bpelProcess, parse);
      }
    }
  }

  /**
   * Parse Bpel process import element.
   */
  private void parseImportElement(final Element importElement, final BpelProcess bpelProcess, final Parse parse) {
    String namespace = null;
    if (importElement.hasAttribute("namespace")) {
      namespace = importElement.getAttribute("namespace");
    }
    /*String location = null;
    if (importElement.hasAttribute("location")) {
      location = importElement.getAttribute("location");
    }*/

    String importType = null;
    if (importElement.hasAttribute("importType")) {
      importType = importElement.getAttribute("importType");
    }

    if (importType == null) {
      parse.addProblem("missing importType attribute in import element " + XmlUtil.toString(importElement));
    } else if (importType.equals(XmlConstants.XMLNS_WSDL)) {
      // Import a WSDL from Deployment
      final WsdlDefinitions wsdlDefinitions = parse.findObject(WsdlDefinitions.class);
      Set<Definition> definitionsForNS = null;
      if (wsdlDefinitions != null) {
        definitionsForNS = wsdlDefinitions.getWsdlDefinitions(namespace);
      }
      if (definitionsForNS == null) {
        parse.addProblem("No wsdl file to import in import element " + XmlUtil.toString(importElement));
      } else {
        for (final Definition wsdlDef : definitionsForNS) {
          // perform static analysis
          if (namespace == null && wsdlDef.getTargetNamespace() != null) {
            parse.addProblem("invalid namespace in imported wsdl (SA00012) :"
                + " no target namespace expected, got \""
                + wsdlDef.getTargetNamespace() + "\"");
          } else if (namespace != null
              && !namespace.equals(wsdlDef.getTargetNamespace())) {
            parse.addProblem("invalid namespace in imported wsdl (SA00011) :"
                + " expected \"" + namespace + "\", got \""
                + wsdlDef.getTargetNamespace() + "\"");
          }

          // Static Analysis
          final Map<QName, PortType> portTypes = wsdlDef.getAllPortTypes();
          for (final PortType pt : portTypes.values()) {
            final List<Operation> operations = pt.getOperations();
            for (int index = 0; index < operations.size(); index++) {
              final Operation op = operations.get(index);
              final String name = op.getName();
              int cmpt = 1;
              for (int i = index + 1; i < operations.size(); i++) {
                if (operations.get(i).getName().equals(name)) {
                  cmpt++;
                }
              }
              if (cmpt > 1) {
                // SA00002
                final String message = StaticAnalysisFault.getMessage("SA00002",
                    "  PartnerLink " + pt.getQName().toString() + " has " + cmpt + " operations with name " + name);
                parse.addProblem(message, new StaticAnalysisException("SA00002"));
              }
            }
          }


          this.wsdlParser.parse(
              wsdlDef,
              parse,
              bpelProcess.getQueryLanguage(),
              bpelProcess,
              bpelProcess.getWsdlInfos());
        }
      }
    } else {
      parse.addProblem("unsupported importType attribute in import element " + XmlUtil.toString(importElement));
    }
  }

  /**
   * returns a configured documentBuilderFactory using schemaLanguage, xsd resources...
   * @return the build DocumentBuilderFactory
   */
  @Override
  public synchronized DocumentBuilderFactory getDocumentBuilderFactory() {
    this.documentBuilderFactory = this.newDocumentBuilderFactory();
    this.documentBuilderFactory.setNamespaceAware(true);
    this.documentBuilderFactory.setValidating(true);
    // ignore white space can only be set if parser is validating
    this.documentBuilderFactory.setIgnoringElementContentWhitespace(true);
    // select xml schema as the schema language (a.o.t. DTD)
    this.documentBuilderFactory.setAttribute(
      "http://java.sun.com/xml/jaxp/properties/schemaLanguage",
      "http://www.w3.org/2001/XMLSchema");
    // set schema sources
    this.documentBuilderFactory.setAttribute(
        "http://java.sun.com/xml/jaxp/properties/schemaSource", BpelParser.schemaSources);
    return this.documentBuilderFactory;
  }

  /**
   * Returns all necessary bindings that the parser should use.
   * @return the build Bindings object
   */
  public static Bindings getDefaultBindings() {
    final Bindings bindings = new Bindings();
    bindings.addBinding(new ProcessBinding());
    bindings.addBinding(new SequenceBinding());
    bindings.addBinding(new ScopeBinding());
    bindings.addBinding(new ReceiveBinding());
    bindings.addBinding(new ReplyBinding());
    bindings.addBinding(new FlowBinding());
    bindings.addBinding(new EmptyBinding());
    bindings.addBinding(new AssignBinding());
    bindings.addBinding(new InvokeBinding());
    bindings.addBinding(new IfBinding());
    bindings.addBinding(new CompensateBinding());
    bindings.addBinding(new CompensateScopeBinding());
    bindings.addBinding(new ExitBinding());
    bindings.addBinding(new ForEachBinding());
    bindings.addBinding(new PickBinding());
    bindings.addBinding(new RepeatUntilBinding());
    bindings.addBinding(new ThrowBinding());
    bindings.addBinding(new RethrowBinding());
    bindings.addBinding(new ValidateBinding());
    bindings.addBinding(new WaitBinding());
    bindings.addBinding(new WhileBinding());

    bindings.addBinding(new ExtensionActivityBinding());
    return bindings;
  }

  protected static QName getBPELQName(final String localName) {
    return new QName(XmlConstants.XMLNS_BPEL_2_0_EXECUTABLE, localName);
  }

  public static Set<QName> getAllActivityQNames() {
    final Set<String> allActivityTagNames = BpelParser.getDefaultBindings().getTagNames(BpelParser.CATEGORY_ACTIVITY);
    final Set<QName> allActivityQNames = new HashSet<QName>();
    for (final String tagName : allActivityTagNames) {
      allActivityQNames.add(BpelParser.getBPELQName(tagName));
    }
    return allActivityQNames;
  }

  /**
   * Returns all necessary entities that the parser should use.
   * @return a map containing all entities : key : url on the internet, value : local resource.
   */
  private static Map<String, Entity> getDefaultEntities() {
    final Map<String, Entity> defaultSchemaCatalog = new HashMap<String, Entity>();
    final ClassLoader resourceLoader = BpelParser.class.getClassLoader();
    defaultSchemaCatalog.put("http://schemas.xmlsoap.org/wsdl/",
        new UrlEntity(BpelParser.RESSOURCES_DIR + "wsdl.xsd", resourceLoader));
    defaultSchemaCatalog.put("http://www.w3.org/2001/XMLSchema",
        new UrlEntity(BpelParser.RESSOURCES_DIR + "xml-schema.xsd", resourceLoader));
    defaultSchemaCatalog.put("http://www.w3.org/2001/xml.xsd",
        new UrlEntity(BpelParser.RESSOURCES_DIR + "xml-schema.xsd", resourceLoader));
    defaultSchemaCatalog.put("http://docs.oasis-open.org/wsbpel/2.0/process/abstract",
        new UrlEntity(BpelParser.RESSOURCES_DIR + "ws-bpel_abstract_common_base.xsd", resourceLoader));
    defaultSchemaCatalog.put("http://docs.oasis-open.org/wsbpel/2.0/process/executable",
        new UrlEntity(BpelParser.RESSOURCES_DIR + "ws-bpel_executable.xsd", resourceLoader));
    return defaultSchemaCatalog;
  }

  /**
   * Build a string array containing all schema resources to use when parsing a bpel file.
   * @return the build string array
   */
  private static String[] getSchemaSources() {
    final String[] schemas = new String[BpelParser.SCHEMA_RESOURCES.length];
    final ClassLoader resourceLoader = BpelParser.class.getClassLoader();
    for (int i = 0; i < BpelParser.SCHEMA_RESOURCES.length; i++) {
      schemas[i] = resourceLoader.getResource(BpelParser.SCHEMA_RESOURCES[i]).toExternalForm();
    }
    return schemas;
  }

  protected void setErrorHandler(final DocumentBuilder documentBuilder) {
    documentBuilder.setErrorHandler(this);
  }

  public void error(final SAXParseException exception) throws SAXException {
    throw exception;
  }

  public void fatalError(final SAXParseException exception) throws SAXException {
    throw exception;
  }

  public void warning(final SAXParseException exception) throws SAXException {
    throw exception;
  }

  @Override
  public void checkProblems(final String description, final Parse parse) {
    try {
      super.checkProblems(description, parse);
    } catch (final PvmException e) {
      throw new OrchestraRuntimeException(e.getMessage(), e);
    }
  }
}
