package org.ow2.orchestra.parsing;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.io.Serializable;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.helpers.DefaultHandler;

/**
 * extracts relevant information to determine automatically wether a byte array
 * represents a bpel document or a wsdl document. The information is collected
 * by a minimal SAX parse.
 *
 * @author Tom Baeyens
 */
public final class XmlDocumentInfo implements Serializable {

  /** serialVersionUID. */
  private static final long serialVersionUID = 1L;

  /** log. */
  private static Logger log = Logger.getLogger(XmlDocumentInfo.class.getName());

  /** saxParserFactory. */
  private static SAXParserFactory saxParserFactory = SAXParserFactory
      .newInstance();

  /** uri. */
  private String uri;

  /** localName. */
  private String localName;

  /** targetNamespace. */
  private String targetNamespace;

  /** DefaultHandler. */
  private static class XmlDocumentInfoHandler extends DefaultHandler {

    /** isRoot. */
    private boolean isRoot = true;

    /** xmlDocumentInfo. */
    private XmlDocumentInfo xmlDocumentInfo;

    /**
     * Default constructor.
     *
     * @param xmlDocumentInfo
     *            xmlDocumentInfo to use
     */
    public XmlDocumentInfoHandler(XmlDocumentInfo xmlDocumentInfo) {
      this.xmlDocumentInfo = xmlDocumentInfo;
    }

    /**
     * Method called by xml parser.
     *
     * @param uri
     *            uri
     * @param localName
     *            localName
     * @param qName
     *            qName
     * @param attributes
     *            attributes
     */
    public void startElement(String uri, String localName, String qName,
        Attributes attributes) {
      if (isRoot) {
        isRoot = false;
        xmlDocumentInfo.uri = uri;
        xmlDocumentInfo.localName = localName;
        String targetNamespaceAttribute = attributes
            .getValue("targetNamespace");
        if (targetNamespaceAttribute != null
            && !"".equals(targetNamespaceAttribute)) {
          xmlDocumentInfo.targetNamespace = targetNamespaceAttribute;
        }
      }
    }
  }

  /**
   * Default constructor.
   */
  private XmlDocumentInfo() {
  }

  /**
   * Creates xml document info by scanning the xml document given as a byte
   * array.
   *
   * @param bytes
   *            byte array source
   * @return the build XmlDocumentInfo
   */
  public static XmlDocumentInfo createXmlDocumentInfo(byte[] bytes) {
    XmlDocumentInfo xmlDocumentInfo = new XmlDocumentInfo();
    try {
      SAXParser saxParser = saxParserFactory.newSAXParser();
      InputStream inputStream = new ByteArrayInputStream(bytes);
      XmlDocumentInfoHandler xmlDocumentInfoHandler = new XmlDocumentInfoHandler(
          xmlDocumentInfo);
      saxParser.parse(inputStream, xmlDocumentInfoHandler);
    } catch (Exception e) {
      if (log.isLoggable(Level.INFO)) {
        log.info("couldn't determine document xmlType: " + e.getMessage());
      }
    }
    return xmlDocumentInfo;
  }

  /**
   * Getter.
   *
   * @return field value
   */
  public static long getSerialVersionUID() {
    return serialVersionUID;
  }

  /**
   * Getter.
   *
   * @return field value
   */
  public String getLocalName() {
    return localName;
  }

  /**
   * Getter.
   *
   * @return field value
   */
  public String getTargetNamespace() {
    return targetNamespace;
  }

  /**
   * Getter.
   *
   * @return field value
   */
  public String getURI() {
    return uri;
  }

  /**
   * Getter.
   *
   * @return field value
   */
  public boolean isBpelProcess() {
    return false;
  }

  /**
   * Getter.
   *
   * @return field value
   */
  public boolean isWsdl() {
    return false;
  }
}
