/**
 * Copyright (C) 2009  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.parsing.binding;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;

import org.ow2.orchestra.definition.activity.AbstractBpelActivity;
import org.ow2.orchestra.definition.activity.CatchActivity;
import org.ow2.orchestra.definition.activity.CatchAllActivity;
import org.ow2.orchestra.definition.activity.Compensate;
import org.ow2.orchestra.definition.activity.CompensationHandlerActivity;
import org.ow2.orchestra.definition.activity.Rethrow;
import org.ow2.orchestra.definition.activity.Sequence;
import org.ow2.orchestra.definition.activity.TerminationHandlerActivity;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.exception.StaticAnalysisException;
import org.ow2.orchestra.facade.def.full.BaseCatchFullDefinition;
import org.ow2.orchestra.facade.def.full.CatchAllFullDefinition;
import org.ow2.orchestra.facade.def.full.CatchFullDefinition;
import org.ow2.orchestra.facade.def.full.CompensateActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.CompensationHandlerFullDefinition;
import org.ow2.orchestra.facade.def.full.RethrowActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.ScopeActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.SequenceActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.TerminationHandlerFullDefinition;
import org.ow2.orchestra.facade.def.full.impl.CatchAllFullDefinitionImpl;
import org.ow2.orchestra.facade.def.full.impl.CatchFullDefinitionImpl;
import org.ow2.orchestra.facade.def.full.impl.CompensateActivityFullDefinitionImpl;
import org.ow2.orchestra.facade.def.full.impl.CompensationHandlerFullDefinitionImpl;
import org.ow2.orchestra.facade.def.full.impl.RethrowActivityFullDefinitionImpl;
import org.ow2.orchestra.facade.def.full.impl.SequenceActivityFullDefinitionImpl;
import org.ow2.orchestra.facade.def.full.impl.TerminationHandlerFullDefinitionImpl;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.facade.uuid.ActivityDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.pvm.internal.model.NodeImpl;
import org.ow2.orchestra.pvm.internal.xml.Parse;
import org.ow2.orchestra.pvm.internal.xml.Parser;
import org.ow2.orchestra.runtime.FaultInCHHandler;
import org.ow2.orchestra.runtime.OrchestraFaultHandler;
import org.ow2.orchestra.util.StaticAnalysisFault;
import org.ow2.orchestra.util.XmlUtil;
import org.w3c.dom.Element;

/**
 * Activity binding for an activity with Fault, Compensation and Termination
 * handling.
 *
 * @author S. Ali Tokmen
 */
public abstract class ActivityWithFCTHBinding extends ActivityBinding {

  private static Logger log = Logger.getLogger(ActivityWithFCTHBinding.class.getName());

  protected ActivityWithFCTHBinding(final String tagName, final ActivityType activityType) {
    super(tagName, activityType);
  }

  private AbstractBpelActivity fillActivity(final Parse parse, final AbstractBpelActivity activity) {
    activity.setName("autoGenerated_activity_" + activity.getClass().getSimpleName());

    final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
    final ProcessDefinitionUUID processDefinitionUUID = definitionKeeper.getProcessFullDefinition().getUUID();
    final ActivityDefinitionUUID activityDefinitionUUID = new ActivityDefinitionUUID(processDefinitionUUID, definitionKeeper
      .getNextActivityCount(), activity.getType(), activity.getName());
    activity.setUuid(activityDefinitionUUID);

    return activity;
  }

  /**
   * Parses the terminationHandler element in an invoke or scope.
   *
   * @param element The element in which to look for the terminationHandler
   *          Handler element.
   * @param parse Parse object.
   * @param parser Parser object.
   * @param scopeNode The node in which to add the activities.
   * @param activityFullDefinition Scope Activity Definition object in which to
   *          add all the termination handler.
   */
  protected void parseTermination(final Element element, final Parse parse, final Parser parser, final NodeImpl scopeNode,
    final ScopeActivityFullDefinition scopeActivityFullDefinition) {
    TerminationHandlerFullDefinition terminationHandlerDefinition;

    final Element terminationHandlerElement = XmlUtil.element(element, "terminationHandler");
    if (terminationHandlerElement != null) {
      final String name = "terminationHandler_" + scopeNode.getName();
      final NodeImpl activity = scopeNode.createNode(name);

      parse.pushObject(activity);
      try {
        final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
        final ProcessDefinitionUUID processDefinitionUUID = definitionKeeper.getProcessFullDefinition().getUUID();
        final ActivityDefinitionUUID activityDefinitionUUID = new ActivityDefinitionUUID(processDefinitionUUID, definitionKeeper
          .getNextActivityCount(), ActivityType.CATCHALL_HANDLER, name);

        terminationHandlerDefinition = new TerminationHandlerFullDefinitionImpl(activityDefinitionUUID, processDefinitionUUID,
            name);
        definitionKeeper.setActivityFullDefinition(terminationHandlerDefinition);
        final NodeImpl terminationHandlerActivity = this.parseChildActivity(terminationHandlerElement, parse, parser);
        final TerminationHandlerActivity terminationHandlerBehavior = new TerminationHandlerActivity(terminationHandlerActivity,
          activityDefinitionUUID, name, scopeNode);
        activity.setBehaviour(terminationHandlerBehavior);

      } finally {
        parse.popObject();
      }
    } else {
      // Add the default terminationHandler node in order to respect indices
      terminationHandlerDefinition = this.addDefaultTH(parse, scopeNode);
    }

    scopeActivityFullDefinition.setTerminationHandler(terminationHandlerDefinition);
  }

  /**
   * Adds a node with the default termination handler activity to the scopeNode
   * given as argument.
   *
   * @param parse Parse object.
   * @param scopeNode Node in which to add the default termination handler
   *          activity node.
   */
  protected TerminationHandlerFullDefinition addDefaultTH(final Parse parse, final NodeImpl scopeNode) {
    final String name = "autoGenerated_DefaultTH_" + scopeNode.getName();
    final NodeImpl activity = scopeNode.createNode(name);

    final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
    final ProcessDefinitionUUID processDefinitionUUID = definitionKeeper.getProcessFullDefinition().getUUID();
    final ActivityDefinitionUUID activityDefinitionUUID = new ActivityDefinitionUUID(processDefinitionUUID, definitionKeeper
      .getNextActivityCount(), ActivityType.TERMINATION_HANDLER, name);

    final NodeImpl compensate = activity.createNode(name + "_compensate");
    compensate.setBehaviour(this.fillActivity(parse, new Compensate()));
    final CompensateActivityFullDefinition compensateActivityFullDefinition =
      new CompensateActivityFullDefinitionImpl(
          processDefinitionUUID,
          ((AbstractBpelActivity)compensate.getBehaviour()).getUuid(),
          compensate.getName());

    final TerminationHandlerActivity terminationHandlerBehavior = new TerminationHandlerActivity(compensate, activityDefinitionUUID,
      name, scopeNode);
    activity.setBehaviour(terminationHandlerBehavior);

    final TerminationHandlerFullDefinition terminationHandlerFullDefinition =
      new TerminationHandlerFullDefinitionImpl(activityDefinitionUUID, processDefinitionUUID, name);
    terminationHandlerFullDefinition.setEnclosedActivity(compensateActivityFullDefinition);

    return terminationHandlerFullDefinition;
  }

  /**
   * Parses the compensationHandler element in an invoke or scope.
   *
   * @param element The element in which to look for the compensationHandler
   *          Handler element.
   * @param parse Parse object.
   * @param parser Parser object.
   * @param scopeNode The node in which to add the activities.
   * @param activityFullDefinition Scope Activity Definition object in which to
   *          add all the compensation handler.
   */
  protected void parseCompensation(final Element element, final Parse parse, final Parser parser, final NodeImpl scopeNode,
    final ScopeActivityFullDefinition scopeActivityFullDefinition) {
    CompensationHandlerFullDefinition compensationHandlerDefinition;

    final Element compensationHandlerElement = XmlUtil.element(element, "compensationHandler");
    if (compensationHandlerElement != null) {
      final String name = "compensationHandler_" + scopeNode.getName();
      final NodeImpl activity = scopeNode.createNode(name);

      parse.pushObject(activity);
      try {
        final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
        final ProcessDefinitionUUID processDefinitionUUID = definitionKeeper.getProcessFullDefinition().getUUID();
        final ActivityDefinitionUUID activityDefinitionUUID = new ActivityDefinitionUUID(processDefinitionUUID, definitionKeeper
          .getNextActivityCount(), ActivityType.CATCHALL_HANDLER, name);

        compensationHandlerDefinition = new CompensationHandlerFullDefinitionImpl(activityDefinitionUUID,
            processDefinitionUUID, name);
        definitionKeeper.setActivityFullDefinition(compensationHandlerDefinition);

        final NodeImpl compensationHandlerActivity = this.parseChildActivity(compensationHandlerElement, parse, parser);
        final CompensationHandlerActivity compensationHandlerBehavior = new CompensationHandlerActivity(compensationHandlerActivity,
          activityDefinitionUUID, name);
        activity.setBehaviour(compensationHandlerBehavior);
        activity.addExceptionHandler(new FaultInCHHandler());

      } finally {
        parse.popObject();
      }
    } else {
      // Add the default compensationHandler node in order to respect indices
      compensationHandlerDefinition = this.addDefaultCH(parse, scopeNode);
    }

    scopeActivityFullDefinition.setCompensationHandler(compensationHandlerDefinition);
  }

  /**
   * Adds a node with the default compensation handler activity to the scopeNode
   * given as argument.
   *
   * @param parse Parse object.
   * @param scopeNode Node in which to add the default compensation handler
   *          activity node.
   */
  protected CompensationHandlerFullDefinition addDefaultCH(final Parse parse, final NodeImpl scopeNode) {
    final String name = "autoGenerated_DefaultCH_" + scopeNode.getName();
    final NodeImpl activity = scopeNode.createNode(name);

    final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
    final ProcessDefinitionUUID processDefinitionUUID = definitionKeeper.getProcessFullDefinition().getUUID();
    final ActivityDefinitionUUID activityDefinitionUUID = new ActivityDefinitionUUID(processDefinitionUUID, definitionKeeper
      .getNextActivityCount(), ActivityType.COMPENSATION_HANDLER, name);

    final NodeImpl compensate = activity.createNode(name + "_compensate");
    compensate.setBehaviour(this.fillActivity(parse, new Compensate()));
    final CompensateActivityFullDefinition emptyActivityFullDefinition =
      new CompensateActivityFullDefinitionImpl(
          processDefinitionUUID,
          ((AbstractBpelActivity)compensate.getBehaviour()).getUuid(),
          compensate.getName());

    final CompensationHandlerActivity compensationHandlerBehavior = new CompensationHandlerActivity(compensate, activityDefinitionUUID,
      name);
    activity.setBehaviour(compensationHandlerBehavior);
    activity.addExceptionHandler(new FaultInCHHandler());


    final CompensationHandlerFullDefinition compensationHandlerFullDefinition =
      new CompensationHandlerFullDefinitionImpl(activityDefinitionUUID, processDefinitionUUID, name);
    compensationHandlerFullDefinition.setEnclosedActivity(emptyActivityFullDefinition);

    return compensationHandlerFullDefinition;
  }

  /**
   * Parses all catch and catchAll elements in the faultHandlersElement. That
   * element can be a real faultHandlers (for example, if we're parsing a scope)
   * or directly the current element (for example, if we're parsing an invoke).
   *
   * @param faultHandlersElement The element in which to look for the catch and
   *          catchAll elements
   * @param parse Parse object.
   * @param parser Parser object.
   * @param scopeNode The node in which to add the activities.
   * @param activityFullDefinition Scope Activity Definition object in which to
   *          add all the handlers.
   *
   * @return Number of catch and catchAll activities parsed.
   */
  protected int parseCatchAndCatchAll(final Element faultHandlersElement, final Parse parse, final Parser parser,
    final NodeImpl scopeNode, final ScopeActivityFullDefinition scopeActivityFullDefinition) {
    int totalDefined = 0;
    final List<BaseCatchFullDefinition< ? >> faultHandlers = new ArrayList<BaseCatchFullDefinition< ? >>();

    final Element catchAllElement = XmlUtil.element(faultHandlersElement, "catchAll");
    if (catchAllElement != null) {
      final CatchAllFullDefinition catchAllDefinition = this.parseCatchAll(catchAllElement, parse, parser, scopeNode);
      faultHandlers.add(catchAllDefinition);
      totalDefined++;
    } else {
      // Add the default faultHandler node in order to respect indices
      final BaseCatchFullDefinition< ? > defaultFHDefinition = this.addDefaultFH(parse, scopeNode);
      faultHandlers.add(defaultFHDefinition);
    }

    final List<Element> catchElements = XmlUtil.elements(faultHandlersElement, "catch");
    if (catchElements != null) {
      final List<CatchFullDefinition> catchDefinitions = this.parseCatchs(catchElements, parse, parser, scopeNode);
      faultHandlers.addAll(catchDefinitions);
      totalDefined += catchDefinitions.size();
    }

    scopeActivityFullDefinition.setFaultHandlers(faultHandlers);
    scopeNode.addExceptionHandler(new OrchestraFaultHandler());

    return totalDefined;
  }

  /**
   * Adds a node with the default fault handler activity to the scopeNode given
   * as argument.
   *
   * @param parse Parse object.
   * @param scopeNode Node in which to add the default fault handler activity
   *          node.
   */
  protected BaseCatchFullDefinition< ? > addDefaultFH(final Parse parse, final NodeImpl scopeNode) {
    final String name = "autoGenerated_DefaultFH_" + scopeNode.getName();
    final NodeImpl activity = scopeNode.createNode(name);

    final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
    final ProcessDefinitionUUID processDefinitionUUID = definitionKeeper.getProcessFullDefinition().getUUID();
    final ActivityDefinitionUUID activityDefinitionUUID = new ActivityDefinitionUUID(processDefinitionUUID, definitionKeeper
      .getNextActivityCount(), ActivityType.CATCHALL_HANDLER, name);

    final NodeImpl sequence = activity.createNode(name + "_sequence");
    sequence.setBehaviour(this.fillActivity(parse, new Sequence()));
    sequence.setPreviousNeeded(true);
    final SequenceActivityFullDefinition sequenceActivityFullDefinition =
      new SequenceActivityFullDefinitionImpl(
          processDefinitionUUID,
          ((AbstractBpelActivity)sequence.getBehaviour()).getUuid(),
          sequence.getName());

    final NodeImpl compensate = sequence.createNode(name + "_compensate");
    compensate.setBehaviour(this.fillActivity(parse, new Compensate()));
    final CompensateActivityFullDefinition compensateActivityFullDefinition =
      new CompensateActivityFullDefinitionImpl(
          processDefinitionUUID,
          ((AbstractBpelActivity)compensate.getBehaviour()).getUuid(),
          compensate.getName());
    sequenceActivityFullDefinition.addEnclosedActivity(compensateActivityFullDefinition);

    final NodeImpl rethrow = sequence.createNode(name + "_rethrow");
    rethrow.setBehaviour(this.fillActivity(parse, new Rethrow()));
    final RethrowActivityFullDefinition rethrowActivityFullDefinition =
      new RethrowActivityFullDefinitionImpl(
          processDefinitionUUID,
          ((AbstractBpelActivity)rethrow.getBehaviour()).getUuid(),
          rethrow.getName());
    sequenceActivityFullDefinition.addEnclosedActivity(rethrowActivityFullDefinition);

    final CatchAllActivity catchAllBehavior = new CatchAllActivity(sequence, activityDefinitionUUID, name);
    activity.setBehaviour(catchAllBehavior);

    final CatchAllFullDefinition catchAllFullDefinition =
      new CatchAllFullDefinitionImpl(activityDefinitionUUID, processDefinitionUUID, name);
    catchAllFullDefinition.setEnclosedActivity(sequenceActivityFullDefinition);

    return catchAllFullDefinition;
  }

  private CatchAllFullDefinition parseCatchAll(final Element catchAllElement, final Parse parse, final Parser parser,
    final NodeImpl scopeNode) {
    final String name = "catchAll_" + scopeNode.getName();
    final NodeImpl activity = scopeNode.createNode(name);

    parse.pushObject(activity);
    try {
      final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
      final ProcessDefinitionUUID processDefinitionUUID = definitionKeeper.getProcessFullDefinition().getUUID();
      final ActivityDefinitionUUID activityDefinitionUUID = new ActivityDefinitionUUID(processDefinitionUUID, definitionKeeper
        .getNextActivityCount(), ActivityType.CATCHALL_HANDLER, name);

      final CatchAllFullDefinition catchAllFullDefinition =
        new CatchAllFullDefinitionImpl(activityDefinitionUUID, processDefinitionUUID, name);
      definitionKeeper.setActivityFullDefinition(catchAllFullDefinition);

      final NodeImpl catchAllActivity = this.parseChildActivity(catchAllElement, parse, parser);
      final CatchAllActivity catchAllBehavior = new CatchAllActivity(catchAllActivity, activityDefinitionUUID, name);
      activity.setBehaviour(catchAllBehavior);

      return catchAllFullDefinition;
    } finally {
      parse.popObject();
    }
  }

  private List<CatchFullDefinition> parseCatchs(final List<Element> catchsElements, final Parse parse, final Parser parser,
    final NodeImpl scopeNode) {
    final List<CatchFullDefinition> result = new ArrayList<CatchFullDefinition>(catchsElements.size());

    for (final Element catchElmt : catchsElements) {
      result.add(this.parseCatch(catchElmt, parse, parser, scopeNode));
    }

    for (int i = 0; i < result.size(); i++) {
      for (int j = i + 1; j < result.size(); j++) {
        if (ActivityWithFCTHBinding.areEqual(result.get(i).getFaultName(), result.get(j).getFaultName())
          && ActivityWithFCTHBinding.areEqual(result.get(i).getFaultMessageType(), result.get(j).getFaultMessageType())
          && ActivityWithFCTHBinding.areEqual(result.get(i).getFaultElement(), result.get(j).getFaultElement())) {
          final String message = StaticAnalysisFault.getMessage("SA00093", "  Each <catch> element must be unique.");
          parse.addProblem(message, new StaticAnalysisException("SA00093"));
          return result;
        }
      }
    }

    return result;
  }

  private CatchFullDefinition parseCatch(final Element catchElement, final Parse parse, final Parser parser,
    final NodeImpl scopeNode) {
    final String name = "catch_" + scopeNode.getName();
    final NodeImpl activity = scopeNode.createNode(name);

    parse.pushObject(activity);
    try {
      final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
      final ProcessDefinitionUUID processDefinitionUUID = definitionKeeper.getProcessFullDefinition().getUUID();
      final ActivityDefinitionUUID activityDefinitionUUID = new ActivityDefinitionUUID(processDefinitionUUID, definitionKeeper
        .getNextActivityCount(), ActivityType.CATCH_HANDLER, name);

      final CatchFullDefinitionImpl catch1 = new CatchFullDefinitionImpl(activityDefinitionUUID, processDefinitionUUID, name);

      catch1.setFaultName(XmlUtil.attributeQName(catchElement, "faultName"));
      catch1.setFaultVariable(this.createCatchVariable(catchElement, parse, parser));
      catch1.setFaultMessageType(XmlUtil.attributeQName(catchElement, "faultMessageType"));
      catch1.setFaultElement(XmlUtil.attributeQName(catchElement, "faultElement"));

      if (catch1.getFaultMessageType() != null && catch1.getFaultElement() != null) {
        final String message = StaticAnalysisFault.getMessage("SA00081",
          "  The faultMessageType and faultElement attributes are mutually exclusive.");
        parse.addProblem(message, new StaticAnalysisException("SA00081"));
        return null;
      }
      if (catch1.getFaultVariable() != null) {
        if (catch1.getFaultMessageType() == null && catch1.getFaultElement() == null) {
          final String message = StaticAnalysisFault.getMessage("SA00081",
            "  The faultVariable attribute must be accompanied by a faultMessageType or faultElement attribute.");
          parse.addProblem(message, new StaticAnalysisException("SA00081"));
          return null;
        }
      } else if (catch1.getFaultMessageType() != null || catch1.getFaultElement() != null) {
        final String message = StaticAnalysisFault.getMessage("SA00081",
          "  The faultMessageType and faultElement attributes must be accompanied by a faultVariable attribute.");
        parse.addProblem(message, new StaticAnalysisException("SA00081"));
        return null;
      }

      definitionKeeper.setActivityFullDefinition(catch1);

      final NodeImpl catchActivity = this.parseChildActivity(catchElement, parse, parser);
      final CatchActivity catchBehavior = new CatchActivity(catchActivity, activityDefinitionUUID, name, catch1);
      activity.setBehaviour(catchBehavior);

      return catch1;
    } finally {
      parse.popObject();
    }
  }

  private String createCatchVariable(final Element catchElement, final Parse parse, final Parser parser) {
    final String name = XmlUtil.attribute(catchElement, "faultVariable");

    if (name != null) {
      final Variable variable = new Variable();
      variable.setMessageType(XmlUtil.attributeQName(catchElement, "faultMessageType"));
      variable.setElement(XmlUtil.attributeQName(catchElement, "faultElement"));

      if (!(((variable.getElement() != null) && (variable.getMessageType() == null) && (variable.getXmlType() == null))
        || ((variable.getElement() == null) && (variable.getMessageType() != null) && (variable.getXmlType() == null)) || ((variable
        .getElement() == null)
        && (variable.getMessageType() == null) && (variable.getXmlType() != null)))) {
        // Static Analysis SA00025
        final String message = StaticAnalysisFault.getMessage("SA00025", "  Type of variable " + name + " not defined properly.");
        parse.addProblem(message, new StaticAnalysisException("SA00025"));
      }
    }

    return name;
  }

  private static boolean areEqual(final Object e1, final Object e2) {
    if (e1 == null) {
      return (e2 == null);
    } else {
      return e1.equals(e2);
    }
  }
}
