/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.parsing.binding;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.activity.Assign;
import org.ow2.orchestra.definition.activity.Scope;
import org.ow2.orchestra.definition.element.Copy;
import org.ow2.orchestra.definition.element.From;
import org.ow2.orchestra.definition.element.FromTo;
import org.ow2.orchestra.definition.element.Literal;
import org.ow2.orchestra.definition.element.To;
import org.ow2.orchestra.facade.def.CopyDefinition;
import org.ow2.orchestra.facade.def.LiteralDefinition;
import org.ow2.orchestra.facade.def.Variant;
import org.ow2.orchestra.facade.def.full.ActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.AssignActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.BpelActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.CopyFullDefinition;
import org.ow2.orchestra.facade.def.full.FromToFullDefinition;
import org.ow2.orchestra.facade.def.full.ToFullDefinition;
import org.ow2.orchestra.facade.def.full.impl.AssignActivityFullDefinitionImpl;
import org.ow2.orchestra.facade.def.full.impl.CopyFullDefinitionImpl;
import org.ow2.orchestra.facade.def.full.impl.FromFullDefinitionImpl;
import org.ow2.orchestra.facade.def.full.impl.ToFullDefinitionImpl;
import org.ow2.orchestra.facade.def.impl.LiteralDefinitionImpl;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.facade.uuid.ActivityDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.lang.Expression;
import org.ow2.orchestra.lang.PropertyQuery;
import org.ow2.orchestra.lang.VariableQuery;
import org.ow2.orchestra.pvm.internal.xml.Parse;
import org.ow2.orchestra.pvm.internal.xml.Parser;
import org.ow2.orchestra.util.BpelUtil;
import org.ow2.orchestra.util.BpelXmlUtil;
import org.ow2.orchestra.util.XmlConstants;
import org.ow2.orchestra.util.XmlUtil;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

/**
 * @author Charles Souillard
 */
public class AssignBinding extends ActivityBinding {

  /** log. */
  private static Logger log = Logger.getLogger(AssignBinding.class.getName());

  public AssignBinding() {
    super("assign", ActivityType.ASSIGN);
  }
  /**
   * Parses the given dom element to build a Node and a behaviour.
   *
   * @param element dom element to parse
   * @param parse parse
   * @param parser parser
   * @return the build behaviour!
   */
  public Object parse(final Element element, final Parse parse, final Parser parser) {
    if (AssignBinding.log.isLoggable(Level.FINE)) {
      AssignBinding.log.fine("parsing element = " + element);
    }
    final Assign assign = new Assign();
    this.setEnclosingScope(assign, parse);
    this.parseStandardAttributes(element, assign, parse);

    final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
    final ActivityFullDefinition activityFullDefinition = definitionKeeper.getActivityFullDefinition();

    this.parseStandardElements(element, assign, parse);

    // attributes
    final boolean validate = BpelUtil.booleanValue(XmlUtil.attribute(element, "validate"));
    if (validate) {
      parse.addProblem("validate attribute not yet supported on assign activity");
      assign.setValidate(validate);
    }
    ((AssignActivityFullDefinition) activityFullDefinition).setValidate(assign.isValidate());

    // copy elements
    final List<Element> copyElements = XmlUtil.elements(element, "copy");
    this.parseCopies(copyElements, assign, parse);

    // extensionAssignOperation elements
    final List<Element> extensionAssignOperationElements = XmlUtil.elements(element, "extensionAssignOperation");
    if (extensionAssignOperationElements != null) {
      parse.addProblem("extensionAssignOperation not yet supported on assign activity");
    }

    final org.ow2.orchestra.pvm.model.Node activityNode = this.createActivityNode(element, parse, parser, assign);
    return activityNode;
  }

  /**
   * Parses copy elements of the given assign.
   *
   * @param copyElements a the list of all copies of this assign
   * @param parse parse
   * @param assign assign to analyse
   */
  private void parseCopies(final List<Element> copyElements, final Assign assign, final Parse parse) {
    final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
    final ActivityFullDefinition activityFullDefinition = definitionKeeper.getActivityFullDefinition();
    final List<CopyDefinition> copyDefinitionList = ((AssignActivityFullDefinition) activityFullDefinition).getCopyDefinitions();

    final List<Copy> copies = new ArrayList<Copy>();

    if (copyElements != null) {
      for (final Element copyElement : copyElements) {
        copies.add(this.parseCopy(copyElement, parse, assign.getEnclosingScope(), copyDefinitionList));
      }
    }
    assign.setCopies(copies);

    ((AssignActivityFullDefinition) activityFullDefinition).setCopyDefinitions(copyDefinitionList);
  }

  /**
   * Parses a copy element.
   *
   * @param copyElement copy dom element to parse
   * @param parse parse
   * @param enclosingScope enclosingScope
   * @return the build Copy
   */
  private Copy parseCopy(final Element copyElement, final Parse parse, final Scope enclosingScope,
    final List<CopyDefinition> copyDefinitionList) {
    final CopyFullDefinition copyFullDefinition = new CopyFullDefinitionImpl();

    final Copy copy = new Copy();
    // attributes
    final boolean keepSrcElementName = BpelUtil.booleanValue(XmlUtil.attribute(copyElement, "keepSrcElementName"));
    copy.setKeepSrcElementName(keepSrcElementName);

    final boolean ignoreMissingFromData = BpelUtil.booleanValue(XmlUtil.attribute(copyElement, "ignoreMissingFromData"));
    if (ignoreMissingFromData) {
      parse.addProblem("ignoreMissingFromData attribute not yet supported on copy element");
    }
    copy.setIgnoreMissingFromData(ignoreMissingFromData);

    AssignBinding.parseFrom(XmlUtil.element(copyElement, "from"), copy, parse, enclosingScope, copyFullDefinition);
    this.parseTo(XmlUtil.element(copyElement, "to"), copy, parse, enclosingScope, copyFullDefinition);

    copyDefinitionList.add(copyFullDefinition);

    return copy;
  }
  /**
   * Parses a from element.
   *
   * @param fromElement from dom element to parse
   * @param parse parse
   * @param enclosingScope enclosingScope
   * @param copy enclosing copy
   */
  protected static void parseFrom(final Element fromElement, final Copy copy, final Parse parse, final Scope enclosingScope,
    final CopyFullDefinition copyFullDefinition) {

    final FromFullDefinitionImpl fromFullDefinition = new FromFullDefinitionImpl();

    final From from = new From();
    // attributes
    from.setEndpointReference(XmlUtil.attribute(fromElement, "endpointReference"));

    // parse from and to common elements/attributes
    AssignBinding.parseFromTo(fromElement, from, parse, enclosingScope, fromFullDefinition);

    // parse literal
    final Element literalElement = XmlUtil.element(fromElement, "literal");
    if (literalElement != null) {
      from.setLiteral(AssignBinding.parseLiteral(literalElement));
      from.setVariant(Variant.LITERAL);
      fromFullDefinition.setVariant(Variant.LITERAL);
    }
    copy.setFrom(from);

    // Definition
    final LiteralDefinition literalDefinition = new LiteralDefinitionImpl(literalElement);
    fromFullDefinition.setLiteral(literalDefinition);
    fromFullDefinition.setEndpointReference(from.getEndpointReference());
    copyFullDefinition.addFromDefinition(fromFullDefinition);
  }

  /**
   * Parses a to element.
   *
   * @param toElement to dom element to parse
   * @param parse parse
   * @param enclosingScope enclosingScope
   * @param copy enclosing copy
   */
  private void parseTo(final Element toElement, final Copy copy, final Parse parse, final Scope enclosingScope,
    final CopyFullDefinition copyFullDefinition) {

    final ToFullDefinition toFullDefinition = new ToFullDefinitionImpl();

    final To to = new To();
    // parse from and to common elements/attributes
    AssignBinding.parseFromTo(toElement, to, parse, enclosingScope, toFullDefinition);
    copy.setTo(to);

    // Definition
    copyFullDefinition.addToDefinition(toFullDefinition);
  }

  /**
   * Parses common attributes/children elements of from and to elements.
   *
   * @param fromToElement dom element to parse (from or to)
   * @param enclosingScope enclosingScope
   * @param parse parse
   * @param fromTo build fromTo
   */
  protected static void parseFromTo(final Element fromToElement, final FromTo fromTo, final Parse parse,
    final Scope enclosingScope, final FromToFullDefinition fromToFullDefinition) {

    // attributes
    fromTo.setExpressionLanguage(XmlUtil.attribute(fromToElement, "expressionLanguage"));
    fromTo.setVariable(XmlUtil.attribute(fromToElement, "variable"));
    fromTo.setPart(XmlUtil.attribute(fromToElement, "part"));
    fromTo.setProperty(XmlUtil.attributeQName(fromToElement, "property"));
    fromTo.setPartnerLink(XmlUtil.attribute(fromToElement, "partnerLink"));

    // Definition
    fromToFullDefinition.setExpressionLanguage(fromTo.getExpressionLanguage());
    fromToFullDefinition.setPart(fromTo.getPart());
    fromToFullDefinition.setVariable(fromTo.getVariable());
    fromToFullDefinition.setProperty(fromTo.getProperty());
    fromToFullDefinition.setPartnerLink(fromTo.getPartnerLink());

    final BpelProcess bpelProcess = parse.findObject(BpelProcess.class);
    if (fromTo.getVariable() != null && fromTo.getProperty() == null) {
      // we are in the variable variant
      fromTo.setVariant(Variant.VAR);

      final Element queryElement = XmlUtil.element(fromToElement, "query");
      if (queryElement != null) {
        String queryLanguage = XmlUtil.attribute(queryElement, "queryLanguage");
        if (queryLanguage == null) {
          queryLanguage = bpelProcess.getQueryLanguage();
        }
        final VariableQuery query = new VariableQuery();
        query.setLanguage(queryLanguage);
        query.setNamespaces(BpelUtil.getAllNameSpaces(queryElement));
        query.setText(queryElement.getTextContent());
        fromTo.setVariableQuery(query);

        // Definition
        fromToFullDefinition.setQuery(fromTo.getVariableQuery().getText());
        fromToFullDefinition.setQueryLanguage(fromTo.getVariableQuery().getLanguage());
      }
      fromToFullDefinition.setVariant(Variant.VAR);
    } else if (fromTo.getVariable() != null) {
      // we are in the property variant
      fromTo.setVariant(Variant.PROP);
      final String queryString = XmlConstants.BPEL_FUNC_PREFIX + ":getVariableProperty(\"" + fromTo.getVariable() + "\",\""
        + fromTo.getProperty().getPrefix() + ":" + fromTo.getProperty().getLocalPart() + "\")";
      final PropertyQuery query = new PropertyQuery();
      query.setLanguage(bpelProcess.getQueryLanguage());
      query.setNamespaces(BpelUtil.getAllNameSpaces(fromToElement));
      query.setText(queryString);

      fromTo.setPropertyQuery(query);

      // Definition
      fromToFullDefinition.setVariant(Variant.PROP);

    } else if (fromTo.getPartnerLink() != null) {
      // we are in the partnerLink variant
      fromToFullDefinition.setVariant(Variant.PLNK);
      fromTo.setVariant(Variant.PLNK);
    } else if (fromTo instanceof To || (fromTo instanceof From && ((From) fromTo).getLiteral() == null)) {
      fromTo.setVariant(Variant.EXPR);
      final Expression expression = new Expression();
      expression.setLanguage(bpelProcess.getExpressionLanguage());
      expression.setNamespaces(BpelUtil.getAllNameSpaces(fromToElement));
      expression.setText(fromToElement.getTextContent());
      fromTo.setExpression(expression);

      // Definition
      fromToFullDefinition.setVariant(Variant.EXPR);
      fromToFullDefinition.setExpression(fromTo.getExpression().getText());
      if (fromTo.getExpressionLanguage() == null) {
        fromTo.setExpressionLanguage(bpelProcess.getExpressionLanguage());

        // Definition
        fromToFullDefinition.setExpressionLanguage(fromTo.getExpressionLanguage());
      }
    }
  }

  /**
   * Parses literal dom element.
   *
   * @param literalElement dom element to parse
   * @return build Literal object
   */
  protected static Literal parseLiteral(final Element literalElement) {
    final Literal literal = new Literal();
    Element value = null;

    final List<Element> elements = XmlUtil.elements(literalElement);
    if (elements.size() > 1) {
      // TODO : maybe it is a list ?
      throw new OrchestraRuntimeException("Invalid literal element : it has more than one element child");
    } else if (elements.size() == 1) {
      value = elements.get(0);
    }
    if (value == null) {
      // we need to look for Texts and CDATA
      final NodeList nodeChilds = literalElement.getChildNodes();
      String data = null;
      for (int i = 0; i < nodeChilds.getLength(); i++) {
        final Node node = nodeChilds.item(i);
        if (node.getNodeType() == Node.TEXT_NODE || node.getNodeType() == Node.CDATA_SECTION_NODE) {
          data = ((Text) node).getData();
          value = BpelXmlUtil.createElementWithContent(data);
          // we only take the first
          break;
        }
      }
    }
    if (value == null) {
      // we create an empty text node
      value = BpelXmlUtil.createElementWithContent("");
    }
    literal.setValue(value);
    return literal;
  }

  @Override
  public BpelActivityFullDefinition createActivityFullDefinition(final ProcessDefinitionUUID processDefinitionUUID,
    final ActivityDefinitionUUID activityDefinitionUUID, final String name, final Parse parse) {
    return new AssignActivityFullDefinitionImpl(processDefinitionUUID, activityDefinitionUUID, name);
  }

}
