/**
 * Copyright (C) 2009  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.parsing.binding;

import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.orchestra.definition.activity.CompensateScope;
import org.ow2.orchestra.exception.StaticAnalysisException;
import org.ow2.orchestra.facade.def.ActivityDefinition;
import org.ow2.orchestra.facade.def.full.BpelActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.impl.CompensateScopeActivityFullDefinitionImpl;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.facade.uuid.ActivityDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.pvm.internal.model.NodeImpl;
import org.ow2.orchestra.pvm.internal.xml.Parse;
import org.ow2.orchestra.pvm.internal.xml.Parser;
import org.ow2.orchestra.util.StaticAnalysisFault;
import org.ow2.orchestra.util.XmlUtil;
import org.w3c.dom.Element;

/**
 * @author S. Ali Tokmen
 */
public class CompensateScopeBinding extends ActivityBinding {

  /**
   * logger.
   */
  private static Logger log = Logger.getLogger(CompensateScopeBinding.class.getName());

  public CompensateScopeBinding() {
    super("compensateScope", ActivityType.COMPENSATE_SCOPE);
  }

  /**
   * Parses the given dom element to build a Node and the CompensateScope
   * behavior
   *
   * @param element dom element to parse
   * @param parse parse
   * @param parser parser
   * @return the CompensateScope behavior
   */
  public Object parse(final Element element, final Parse parse, final Parser parser) {
    if (log.isLoggable(Level.FINE)) {
      log.fine("parsing element = " + element);
    }

    CompensateScope compensateScope = new CompensateScope();
    setEnclosingScope(compensateScope, parse);
    parseStandardAttributes(element, compensateScope, parse);
    parseStandardElements(element, compensateScope, parse);

    // attributes
    compensateScope.setTarget(XmlUtil.attribute(element, "target"));

    NodeImpl activityNode = createActivityNode(element, parse, parser, compensateScope);

    // Static Analysis SA00008
    boolean foundFHorCHorTH = false;
    final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
    for (Object parent : definitionKeeper.getParentsStack()) {
      ActivityDefinition parentActivity = (ActivityDefinition) parent;
      ActivityType parentActivityType = parentActivity.getType();
      if (parentActivityType == ActivityType.CATCH_HANDLER || parentActivityType == ActivityType.CATCHALL_HANDLER
        || parentActivityType == ActivityType.COMPENSATION_HANDLER || parentActivityType == ActivityType.TERMINATION_HANDLER) {
        foundFHorCHorTH = true;
        break;
      }
    }
    if (!foundFHorCHorTH) {
      final String message = StaticAnalysisFault.getMessage("SA00008", "  The <compensateScope> element \""
        + compensateScope.getName() + "\" is not in a <catch>, <catchAll>, <compensationHandler> or "
        + "<terminationHandler> element.");
      parse.addProblem(message, new StaticAnalysisException("SA00008"));
      return null;
    }

    return activityNode;
  }

  @Override
  public BpelActivityFullDefinition createActivityFullDefinition(final ProcessDefinitionUUID processDefinitionUUID,
    final ActivityDefinitionUUID activityDefinitionUUID, final String name, final Parse parse) {
    return new CompensateScopeActivityFullDefinitionImpl(processDefinitionUUID, activityDefinitionUUID, name);
  }

}
