/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.parsing.binding;

import java.util.Collections;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.activity.ForEach;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.exception.StaticAnalysisException;
import org.ow2.orchestra.facade.def.CompletionConditionDefinition;
import org.ow2.orchestra.facade.def.full.BpelActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.ForeachActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.impl.ForeachActivityFullDefinitionImpl;
import org.ow2.orchestra.facade.def.impl.CompletionConditionDefinitionImpl;
import org.ow2.orchestra.facade.exception.BpelFaultException;
import org.ow2.orchestra.facade.exception.BpelFaultException.BpelFault;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.facade.uuid.ActivityDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.lang.Expression;
import org.ow2.orchestra.lang.evaluator.ExpressionEvaluator;
import org.ow2.orchestra.parsing.BpelParser;
import org.ow2.orchestra.parsing.binding.exception.DuplicateVariableException;
import org.ow2.orchestra.pvm.internal.xml.Parse;
import org.ow2.orchestra.pvm.internal.xml.Parser;
import org.ow2.orchestra.util.BpelUtil;
import org.ow2.orchestra.util.StaticAnalysisFault;
import org.ow2.orchestra.util.XmlConstants;
import org.ow2.orchestra.util.XmlUtil;
import org.w3c.dom.Element;

public class ForEachBinding extends ActivityBinding {

  /** log. */
  private static Logger log = Logger.getLogger(IfBinding.class.getName());

  public ForEachBinding() {
    super("forEach", ActivityType.FOR_EACH);
  }
  public Object parse(final Element element, final Parse parse, final Parser parser) {
    if (ForEachBinding.log.isLoggable(Level.FINE)) {
      ForEachBinding.log.fine("parsing element = " + element);
    }
    final ForEach forEach = new ForEach();
    final org.ow2.orchestra.pvm.model.Node forEachNode = this.createActivityNode(element, parse, parser, forEach);

    this.setEnclosingScope(forEach, parse);
    this.parseStandardAttributes(element, forEach, parse);

    final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
    final ForeachActivityFullDefinition foreachActivityFullDefinition = (ForeachActivityFullDefinition) definitionKeeper
      .getActivityFullDefinition();

    this.parseStandardElements(element, forEach, parse);

    parse.pushObject(forEachNode);
    parse.pushObject(forEach);
    try {
      final BpelProcess bpelProcess = parse.findObject(BpelProcess.class);

      // Parse parallel attribute
      final boolean parallel = BpelUtil.booleanValue(XmlUtil.attribute(element, "parallel"));
      forEach.setParallel(parallel);
      // Parse counter name
      forEach.setCounterName(XmlUtil.attribute(element, "counterName"));

      boolean startConstant = false;
      boolean finalConstant = false;
      int constantStartValue = -1;
      int constantFinalValue = -1;

      // Parse start counter value expression
      final Element startCounterValueElement = XmlUtil.element(element, "startCounterValue");
      // text expression
      final String startCounterValueSt = startCounterValueElement.getTextContent();
      // expression language
      String startCounterValueLanguage = XmlUtil.attribute(startCounterValueElement, "expressionLanguage");
      if (startCounterValueLanguage == null) {
        startCounterValueLanguage = bpelProcess.getExpressionLanguage();
      }

      // Definition
      foreachActivityFullDefinition.setStartCounterExpressionLanguage(startCounterValueLanguage);
      foreachActivityFullDefinition.setStartCounterExpression(startCounterValueSt);

      // create Start counter value expression
      final Expression startCounterValueExpression = new Expression();
      startCounterValueExpression.setLanguage(startCounterValueLanguage);
      startCounterValueExpression.setNamespaces(BpelUtil.getAllNameSpaces(startCounterValueElement));
      startCounterValueExpression.setText(startCounterValueSt);
      try {
        if (startCounterValueSt.length() != 0) {
          constantStartValue = startCounterValueExpression.getEvaluator().evaluateUnsignedInt(null);
        }
        startConstant = true;
      } catch (final BpelFaultException e) {
        if (
            e.getFaultName().equals(BpelFault.invalidExpressionValue.getQName())
            &&
            e.getMessage() != null
            &&
            (e.getMessage().equals(ExpressionEvaluator.NegativeIntExceptionMessage)
              ||
             e.getMessage().equals(ExpressionEvaluator.NaNExceptionMessage)
            )
           ) {
         startConstant = true;
        } else {
          // probably not constant expression (using variables for example)
          ForEachBinding.log.finest("startCounterValueSt is not constant");
        }
      } catch (final Exception e) {
        // probably not constant expression (using variables for example)
        ForEachBinding.log.finest("startCounterValueSt is not constant");
      }
      if (startConstant && constantStartValue < 0) {
        // SA00074
        final String message = StaticAnalysisFault.getMessage("SA00074", "Invalid startCounterValue");
        parse.addProblem(message, new StaticAnalysisException("SA00074"));
      }

      forEach.setStartCounterValue(startCounterValueExpression);

      // Parse final counter value expression
      final Element finalCounterValueElement = XmlUtil.element(element, "finalCounterValue");
      // Final counter value text expression
      final String finalCounterValueSt = finalCounterValueElement.getTextContent();
      // Final counter value expression language
      String finalCounterValueLanguage = XmlUtil.attribute(finalCounterValueElement, "expressionLanguage");
      if (finalCounterValueLanguage == null) {
        finalCounterValueLanguage = bpelProcess.getExpressionLanguage();
      }

      // Definition
      foreachActivityFullDefinition.setFinalCounterExpressionLanguage(finalCounterValueLanguage);
      foreachActivityFullDefinition.setFinalCounterExpression(finalCounterValueSt);

      // Create Final counter value expression
      final Expression finalCounterValueExpression = new Expression();
      finalCounterValueExpression.setLanguage(finalCounterValueLanguage);
      finalCounterValueExpression.setNamespaces(BpelUtil.getAllNameSpaces(finalCounterValueElement));
      finalCounterValueExpression.setText(finalCounterValueSt);
      try {
        if (finalCounterValueSt.length() != 0) {
          constantFinalValue = finalCounterValueExpression.getEvaluator().evaluateUnsignedInt(null);
        }
        finalConstant = true;
      } catch (final Exception e) {
        // probably not constant expression (using variables for example)
        ForEachBinding.log.finest("finalCounterValueSt is not constant");
      }

      if (finalConstant && constantFinalValue < 0) {
        // SA00074
        final String message = StaticAnalysisFault.getMessage("SA00074", "Invalid finalCounterValue");
        parse.addProblem(message, new StaticAnalysisException("SA00074"));
      }

      forEach.setFinalCounterValue(finalCounterValueExpression);

      // Parse completion condition element
      final Element completionConditionElement = XmlUtil.element(element, "completionCondition");
      if (completionConditionElement != null) {
        final Element branchesElement = XmlUtil.element(completionConditionElement, "branches");
        if (branchesElement == null) {
          // BPEL spec 11.7: if there are no branches elements in completion
          // condition,
          // we must do as if there was no completion condition
          // So there is only a warning there, and no error in parsing.
          parse.addWarning("Completion condition does not contain a \"branches\" node. It will be ignored.");
        } else {
          // completion condition successfulBranchesOnly attribute
          final Boolean successfulBranchesOnly = BpelUtil.booleanValue(XmlUtil.attribute(branchesElement, "successfulBranchesOnly"));
          // set false by default
          if (successfulBranchesOnly != null) {
            forEach.setSuccessfulBranchesOnly(successfulBranchesOnly);
          }
          // condition expression text
          final String completionConditionSt = branchesElement.getTextContent();
          // condition expression language
          String completionConditionLanguage = XmlUtil.attribute(branchesElement, "expressionLanguage");
          if (completionConditionLanguage == null) {
            completionConditionLanguage = bpelProcess.getExpressionLanguage();
          }
          // create completion condition
          final Expression completionCondition = new Expression();
          completionCondition.setLanguage(completionConditionLanguage);
          completionCondition.setNamespaces(BpelUtil.getAllNameSpaces(branchesElement));
          completionCondition.setText(completionConditionSt);

          if (startConstant && finalConstant) {
            boolean branchesConstant = false;
            int constantBranchesValue = -1;
            try {
              constantBranchesValue = completionCondition.getEvaluator().evaluateUnsignedInt(null);
              branchesConstant = true;
            } catch (final Exception e) {
              // probably not constant expression (using variables for example)
              ForEachBinding.log.finest("completionConditionSt is not constant");
            }
            if (branchesConstant && Math.max(constantFinalValue - constantStartValue + 1, 0) < constantBranchesValue) {
              // SA00075
              final String message = StaticAnalysisFault.getMessage("SA00075",
                "branches value is larger than number of directly enclosed activities");
              parse.addProblem(message, new StaticAnalysisException("SA00075"));
            }
          }

          forEach.setCompletionCondition(completionCondition);

          // Definition
          final CompletionConditionDefinition completionConditionDefinition = new CompletionConditionDefinitionImpl(
            completionCondition.getText(), completionCondition.getLanguage(), forEach.isSuccessfulBranchesOnly());
          foreachActivityFullDefinition.setCompletionConditionDefintion(completionConditionDefinition);
        }
      }

      // Add counter variable to forEach enclosed scope
      // This is done by giving additional variables as input parameter
      // during scope parsing.

      // Fetch enclosed scope ...
      final List<Element> activityElements = XmlUtil.elementsQName(element, Collections.singleton(new QName(
        XmlConstants.XMLNS_BPEL_2_0_EXECUTABLE, "scope")));
      if (activityElements.size() == 0) {
        parse.addProblem("No scope was specified under this foreach element.");
        return null;
      }
      // ... in childrenElement
      final Element childrenElement = activityElements.get(0);

      // Create an unsigned variable (the counter)
      final Variable variable = new Variable();
      variable.setName(forEach.getCounterName());
      variable.setXmlType(new QName(XmlConstants.XMLNS_XMLSCHEMA, "unsignedInt"));

      // Get binding of enclosed scope
      final ScopeBinding binding = (ScopeBinding) parser.getBinding(childrenElement, BpelParser.CATEGORY_ACTIVITY);
      // and parse the scope with method which allow to define some additional
      // variables

      try {
        definitionKeeper.pushParent();
        binding.parse(childrenElement, Collections.singleton(variable), parse, parser);
      } catch (final DuplicateVariableException e) {
        // SA00076
        final String message = StaticAnalysisFault.getMessage("SA00076",
          "enclosed scope declare a variable with counterName of foreach activity");
        parse.addProblem(message, new StaticAnalysisException("SA00076"));
      } finally {
        definitionKeeper.popParent();
      }

    } finally {
      parse.popObject();
      parse.popObject();
    }

    // Definition
    foreachActivityFullDefinition.setCounterName(forEach.getCounterName());
    foreachActivityFullDefinition.setParallelValue(forEach.isParallel());
    return forEachNode;
  }

  @Override
  public BpelActivityFullDefinition createActivityFullDefinition(final ProcessDefinitionUUID processDefinitionUUID,
    final ActivityDefinitionUUID activityDefinitionUUID, final String name, final Parse parse) {
    return new ForeachActivityFullDefinitionImpl(processDefinitionUUID, activityDefinitionUUID, name);
  }

}
