/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.parsing.binding;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.activity.If;
import org.ow2.orchestra.definition.element.Else;
import org.ow2.orchestra.definition.element.Elseif;
import org.ow2.orchestra.facade.def.full.BpelActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.ElseFullDefinition;
import org.ow2.orchestra.facade.def.full.ElseIfFullDefinition;
import org.ow2.orchestra.facade.def.full.IfActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.impl.ElseFullDefinitionImpl;
import org.ow2.orchestra.facade.def.full.impl.ElseIfFullDefinitionImpl;
import org.ow2.orchestra.facade.def.full.impl.IfActivityFullDefinitionImpl;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.facade.uuid.ActivityDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.lang.Expression;
import org.ow2.orchestra.pvm.internal.model.NodeImpl;
import org.ow2.orchestra.pvm.internal.xml.Parse;
import org.ow2.orchestra.pvm.internal.xml.Parser;
import org.ow2.orchestra.util.BpelUtil;
import org.ow2.orchestra.util.XmlUtil;
import org.w3c.dom.Element;

/**
 * @author Charles Souillard
 */
public class IfBinding extends ActivityBinding {

  /** log. */
  private static Logger log = Logger.getLogger(IfBinding.class.getName());

  public IfBinding() {
    super("if", ActivityType.IF);
  }

  /**
   * Parses the given dom element to build a Node and a behaviour.
   *
   * @param element dom element to parse
   * @param parse parse
   * @param parser parser
   * @return the build behaviour!
   */
  public Object parse(final Element element, final Parse parse, final Parser parser) {
    if (log.isLoggable(Level.FINE)) {
      log.fine("parsing element = " + element);
    }
    If ifObject = new If();
    org.ow2.orchestra.pvm.model.Node ifNode = createActivityNode(element, parse, parser, ifObject);

    setEnclosingScope(ifObject, parse);
    parseStandardAttributes(element, ifObject, parse);

    DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
    IfActivityFullDefinition ifActivityFullDefinition = (IfActivityFullDefinition) definitionKeeper
      .getActivityFullDefinition();

    parseStandardElements(element, ifObject, parse);

    parse.pushObject(ifNode);
    parse.pushObject(ifObject);
    try {
      Expression expression = parseCondition(element, parse);
      ifObject.setCondition(expression);
      NodeImpl nodeImpl = parseChildActivity(element, parse, parser);
      ifObject.setActivity(nodeImpl);

      // Definition
      ifActivityFullDefinition.setExpression(expression.getText());
      ifActivityFullDefinition.setExpressionLanguage(expression.getLanguage());

      // elseif elements
      List<Element> elseifElements = XmlUtil.elements(element, "elseif");
      parseElseifs(elseifElements, ifObject, parse, parser, ifActivityFullDefinition);

      // else element
      Element elseElement = XmlUtil.element(element, "else");
      parseElse(elseElement, ifObject, parse, parser, ifActivityFullDefinition);
    } finally {
      parse.popObject();
      parse.popObject();
    }
    return ifNode;
  }

  private void parseElseifs(final List<Element> elseifElements, final If ifObject, final Parse parse, final Parser parser,
    final IfActivityFullDefinition ifActivityFullDefinition) {
    List<Elseif> elseifs = new ArrayList<Elseif>();
    if (elseifElements != null) {
      for (Element elseifElement : elseifElements) {
        // Definition
        ElseIfFullDefinition elseIfFullDefinition = new ElseIfFullDefinitionImpl();

        DefinitionKeeper defKeeper = parse.findObject(DefinitionKeeper.class);
        defKeeper.pushParent(elseIfFullDefinition);

        Elseif elseif = parseElseif(elseifElement, parse, parser);
        elseifs.add(elseif);

        elseIfFullDefinition.setExpression(elseif.getCondition().getText());
        elseIfFullDefinition.setExpressionLanguage(elseif.getCondition().getLanguage());
        ifActivityFullDefinition.addElseIfFullDefinition(elseIfFullDefinition);
        defKeeper.popParent();
      }
    }
    ifObject.setElseifs(elseifs);
  }

  private Elseif parseElseif(final Element elseifElement, final Parse parse, final Parser parser) {
    Elseif elseif = new Elseif();
    // condition
    elseif.setCondition(parseCondition(elseifElement, parse));
    elseif.setActivity(parseChildActivity(elseifElement, parse, parser));
    return elseif;
  }

  private void parseElse(final Element elseElement, final If ifObject, final Parse parse, final Parser parser,
    final IfActivityFullDefinition ifActivityFullDefinition) {
    if (elseElement != null) {

      ElseFullDefinition elseFullDefinition = new ElseFullDefinitionImpl();
      DefinitionKeeper defKeeper = parse.findObject(DefinitionKeeper.class);
      defKeeper.pushParent(elseFullDefinition);

      Else else1 = new Else();
      else1.setActivity(parseChildActivity(elseElement, parse, parser));
      ifObject.setElse(else1);

      ifActivityFullDefinition.setElseFullDefinition(elseFullDefinition);
      defKeeper.popParent();
    }
  }

  private Expression parseCondition(final Element parentElement, final Parse parse) {
    Element conditionElement = XmlUtil.element(parentElement, "condition");
    if (conditionElement != null) {
      BpelProcess bpelProcess = parse.findObject(BpelProcess.class);
      String expressionLanguage = XmlUtil.attribute(conditionElement, "expressionLanguage");
      if (expressionLanguage == null) {
        expressionLanguage = bpelProcess.getExpressionLanguage();
      }
      Expression expression = new Expression();
      expression.setLanguage(expressionLanguage);
      expression.setNamespaces(BpelUtil.getAllNameSpaces(conditionElement));
      expression.setText(conditionElement.getTextContent());
      return expression;
    }
    return null;
  }

  @Override
  public BpelActivityFullDefinition createActivityFullDefinition(final ProcessDefinitionUUID processDefinitionUUID,
    final ActivityDefinitionUUID activityDefinitionUUID, final String name, final Parse parse) {
    return new IfActivityFullDefinitionImpl(processDefinitionUUID, activityDefinitionUUID, name);
  }

}
