/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.parsing.binding;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.wsdl.Operation;
import javax.xml.namespace.QName;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.activity.EventHandlerActivity;
import org.ow2.orchestra.definition.activity.Invoke;
import org.ow2.orchestra.definition.activity.Scope;
import org.ow2.orchestra.definition.element.Copy;
import org.ow2.orchestra.definition.element.Correlation;
import org.ow2.orchestra.definition.element.From;
import org.ow2.orchestra.definition.element.FromPart;
import org.ow2.orchestra.definition.element.PartnerLink;
import org.ow2.orchestra.definition.element.To;
import org.ow2.orchestra.definition.element.ToPart;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.exception.StaticAnalysisException;
import org.ow2.orchestra.facade.def.Variant;
import org.ow2.orchestra.facade.def.full.InvokeActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.ScopeActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.TerminationHandlerFullDefinition;
import org.ow2.orchestra.facade.def.full.impl.InvokeActivityFullDefinitionImpl;
import org.ow2.orchestra.facade.def.impl.FromPartDefinitionImpl;
import org.ow2.orchestra.facade.def.impl.ToPartDefinitionImpl;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.facade.uuid.ActivityDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.parsing.binding.exception.CorrelationSetNotFoundException;
import org.ow2.orchestra.pvm.internal.model.NodeImpl;
import org.ow2.orchestra.pvm.internal.xml.Parse;
import org.ow2.orchestra.pvm.internal.xml.Parser;
import org.ow2.orchestra.util.StaticAnalysisFault;
import org.ow2.orchestra.util.XmlUtil;
import org.w3c.dom.Element;

/**
 * @author Charles Souillard, Goulven Le Jeune
 */
public class InvokeBinding extends ActivityWithFCTHBinding {

  /** log. */
  private static Logger log = Logger.getLogger(InvokeBinding.class.getName());

  private final ImplicitScopeBinding surroundingScopeInvokeBinding;

  public InvokeBinding() {
    super("invoke", ActivityType.INVOKE);
    this.surroundingScopeInvokeBinding = new ImplicitScopeBinding();
  }
  /**
   * Parses the given dom element to build a Node and a behaviour.
   *
   * @param element dom element to parse
   * @param parse parse
   * @param parser parser
   * @return the build behaviour!
   */
  public Object parse(final Element element, final Parse parse, final Parser parser) {
    if (InvokeBinding.log.isLoggable(Level.FINE)) {
      InvokeBinding.log.fine("parsing element = " + element);
    }
    final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);

    final Scope scope = new Scope();
    this.setEnclosingScope(scope, parse);
    this.surroundingScopeInvokeBinding.parseStandardAttributes(element, scope, parse);

    final ScopeActivityFullDefinition scopeActivityFullDefinition = (ScopeActivityFullDefinition) definitionKeeper.getActivityFullDefinition();

    final NodeImpl scopeNode = this.createActivityNode("null", parse, parser, scope);
    scope.setScopeNode(scopeNode);

    parse.pushObject(scope);
    parse.pushObject(scopeNode);

    try {
      // surroundingInvoke_ becomes parent
      definitionKeeper.pushParent();

      final Invoke invoke = new Invoke();

      // WARNING: We need to create all activities BEFORE parsing any attribute
      // of the Invoke. That's because the parseStandardAttributes method will
      // change the setActivityFullDefinition to the Invoke call, whereas the
      // activities' activityFullDefinition should remain the surrounding scope

      // Create the main activity
      this.createActivityNode(element, parse, parser, invoke);

      // Add a default (empty) EventHandler node
      final EventHandlerActivity activity = new EventHandlerActivity();
      scopeNode.createNode("eventHandler_" + scopeNode.getName()).setBehaviour(activity);

      final ProcessDefinitionUUID processDefinitionUUID = definitionKeeper.getProcessFullDefinition().getUUID();
      final ActivityDefinitionUUID activityDefinitionUUID = new ActivityDefinitionUUID(processDefinitionUUID, definitionKeeper
        .getNextActivityCount(), activity.getType(), activity.getName());
      activity.setUuid(activityDefinitionUUID);

      // Add the default terminationHandler node in order to respect indices
      final TerminationHandlerFullDefinition terminationHandlerDefinition = this.addDefaultTH(parse, scopeNode);
      scopeActivityFullDefinition.setTerminationHandler(terminationHandlerDefinition);

      // Analyze the compensationHandler
      this.parseCompensation(element, parse, parser, scopeNode, scopeActivityFullDefinition);

      // Analyze the catch and catchAll elements
      this.parseCatchAndCatchAll(element, parse, parser, scopeNode, scopeActivityFullDefinition);

      this.setEnclosingScope(invoke, parse);
      this.parseStandardAttributes(element, invoke, parse);
      // link sources / targets should be set on implicit scope
      this.parseStandardElements(element, scope, parse);

      final InvokeActivityFullDefinition invokeActivityFullDefinition = (InvokeActivityFullDefinition) definitionKeeper.getActivityFullDefinition();

      scope.setName("surroundingInvoke_" + invoke.getUuid());
      scopeNode.setName(this.getNodeName(Scope.class.getSimpleName(), "surroundingInvoke_" + invoke.getUuid()));

      final String partnerLinkName = XmlUtil.attribute(element, "partnerLink");
      invoke.setPartnerLink(partnerLinkName);
      final String operationSt = XmlUtil.attribute(element, "operation");
      invoke.setOperation(operationSt);
      final String inputVariable = XmlUtil.attribute(element, "inputVariable");
      invoke.setInputVariable(inputVariable);
      final String outputVariable = XmlUtil.attribute(element, "outputVariable");
      invoke.setOutputVariable(outputVariable);

      invokeActivityFullDefinition.setPartnerLink(partnerLinkName);
      invokeActivityFullDefinition.setInputVariable(inputVariable);
      invokeActivityFullDefinition.setOutputVariable(outputVariable);
      invokeActivityFullDefinition.setOperation(operationSt);

      // get partnerLinkDef and get plnkType attribute and myRole attribute
      final PartnerLink partnerLink = this.findPartnerLinkDefinition(parse, partnerLinkName);
      if (partnerLink == null) {
        parse.addProblem("No partnerLink " + partnerLinkName + " defined.");
        return null;
      }
      final QName plportType = partnerLink.getPartnerRolePortTypeQName();
      if (plportType == null) {
        parse.addProblem("No partnerRole defined for partnerLink: " + partnerLinkName);
        return null;
      }
      invoke.setPortTypeQName(plportType);
      invokeActivityFullDefinition.setPortType(plportType);

      if (XmlUtil.attribute(element, "portType") != null) {
        final QName portType = XmlUtil.attributeQName(element, "portType");
        if (!plportType.equals(portType)) {
          // SA00005
          final String message = StaticAnalysisFault.getMessage("SA00005", "  invoke " + invoke.getName()
              + " has specified a PortType that is different from the one specified by the partnerLink");
          parse.addProblem(message, new StaticAnalysisException("SA00005"));
        }
      }

      // fromParts
      final Element fromPartsElement = XmlUtil.element(element, "fromParts");
      List<Element> fromPartElements = null;
      if (fromPartsElement != null) {
        fromPartElements = XmlUtil.elements(fromPartsElement, "fromPart");
        if (invoke.getOutputVariable() != null) {
          // SA00052
          final String message = StaticAnalysisFault.getMessage("SA00052", "  invoke " + invoke.getName()
              + " has specified both an outputVariable and a fromPart");
          parse.addProblem(message, new StaticAnalysisException("SA00052"));
        }
      }

      // toParts
      final Element toPartsElement = XmlUtil.element(element, "toParts");
      List<Element> toPartElements = null;
      if (toPartsElement != null) {
        toPartElements = XmlUtil.elements(toPartsElement, "toPart");
        if (invoke.getInputVariable() != null) {
          // SA00051
          final String message = StaticAnalysisFault
              .getMessage("SA00051", "  invoke " + invoke.getName() + " has specified both an inputVariable and a toPart");
          parse.addProblem(message, new StaticAnalysisException("SA00051"));
        }
      }

      // one way
      final BpelProcess bpelProcess = parse.findObject(BpelProcess.class);
      final Operation operation = bpelProcess.getWsdlInfos().getPortType(invoke.getPortTypeQName()).getOperation(invoke.getOperation(), null, null);
      QName inputMessageQName = null;
      if (operation.getInput() != null) {
        inputMessageQName = operation.getInput().getMessage().getQName();
      }
      QName outputMessageQName = null;
      if (operation.getOutput() != null) {
        outputMessageQName = operation.getOutput().getMessage().getQName();
        if (fromPartElements == null && invoke.getOutputVariable() == null) {
          parse.addProblem("Cannot define Invoke one way " + invoke.getName() + " on a two way operation " + operation.getName());
        }
        invoke.setOneWay(false);
      } else {
        invoke.setOneWay(true);
      }

      // correlations (needs one way)
      final Element correlationsElement = XmlUtil.element(element, "correlations");
      if (correlationsElement != null) {
        final List<Element> correlationElements = XmlUtil.elements(correlationsElement, "correlation");
        this.parseCorrelations(correlationElements, invoke, parse);
      }

      // Deploy copy elements for the outbound message
      this.setOutCopies(toPartElements, invoke, inputMessageQName, invokeActivityFullDefinition);
      this.setInCopies(fromPartElements, invoke, outputMessageQName, invokeActivityFullDefinition);
    } finally {
      parse.popObject();
      parse.popObject();
      definitionKeeper.popParent();
    }

    return scopeNode;
  }

  private void setOutCopies(final List<Element> toPartElements, final Invoke invoke, final QName inputMessageQName,
      final InvokeActivityFullDefinition invokeActivityFullDefinition) {
    final Variable anonymousOutboundMessageVariable = new Variable();
    anonymousOutboundMessageVariable.setName(Variable.OUTBOUND_ANONYMOUS_WSDL_MESSAGE_NAME);
    anonymousOutboundMessageVariable.setMessageType(inputMessageQName);
    invoke.setAnonymousOutboundMessageVariable(anonymousOutboundMessageVariable);
    final String anonymousOutboundMessageVariableName = anonymousOutboundMessageVariable.getName();

    final String inputVariable = invoke.getInputVariable();
    final List<Copy> outCopies = new ArrayList<Copy>();
    if (inputVariable != null) {
      final From from = new From();
      from.setVariable(inputVariable);
      from.setVariant(Variant.VAR);
      final To to = new To();
      to.setVariable(anonymousOutboundMessageVariableName);
      to.setVariant(Variant.VAR);

      final Copy copy = new Copy();
      copy.setFrom(from);
      copy.setTo(to);
      outCopies.add(copy);
    } else {
      // Use the toParts elements
      final List<ToPart> toParts = new ArrayList<ToPart>();
      if (toPartElements != null) {
        for (final Element toPartElement : toPartElements) {
          final ToPart toPart = new ToPart();
          final String part = XmlUtil.attribute(toPartElement, "part");
          toPart.setPart(part);
          final String fromVariable = XmlUtil.attribute(toPartElement, "fromVariable");
          toPart.setFromVariable(fromVariable);
          invokeActivityFullDefinition.addToPartDefinition(new ToPartDefinitionImpl(fromVariable, part));
          toParts.add(toPart);
        }
      }

      for (final ToPart toPart : toParts) {
        final From from = new From();
        from.setVariable(toPart.getFromVariable());
        from.setVariant(Variant.VAR);
        final To to = new To();
        to.setVariable(anonymousOutboundMessageVariableName);
        to.setPart(toPart.getPart());
        to.setVariant(Variant.VAR);
        final Copy copy = new Copy();
        copy.setFrom(from);
        copy.setTo(to);
        outCopies.add(copy);
      }
    }
    if (outCopies.size() != 0) {
      invoke.setOutCopies(outCopies);
    }
  }

  private void setInCopies(final List<Element> fromPartElements, final Invoke invoke, final QName outputMessageQName,
      final InvokeActivityFullDefinition invokeActivityFullDefinition) {
    final String outputVariable = invoke.getOutputVariable();
    final List<Copy> inCopies = new ArrayList<Copy>();
    // If the <invoke> is a request/response invocation, also deploy copy
    // elements for the inbound message
    if (!invoke.isOneWay()) {
      final Variable anonymousInboundMessageVariable = new Variable();
      anonymousInboundMessageVariable.setName(Variable.INBOUND_ANONYMOUS_WSDL_MESSAGE_NAME);
      anonymousInboundMessageVariable.setMessageType(outputMessageQName);
      invoke.setAnonymousInboundMessageVariable(anonymousInboundMessageVariable);
      final String anonymousInboundMessageVariableName = anonymousInboundMessageVariable.getName();
      if (outputVariable != null) {
        final From from = new From();
        from.setVariable(anonymousInboundMessageVariableName);
        from.setVariant(Variant.VAR);
        final To to = new To();
        to.setVariable(outputVariable);
        to.setVariant(Variant.VAR);
        final Copy copy = new Copy();
        copy.setFrom(from);
        copy.setTo(to);
        inCopies.add(copy);
      } else {
        final List<FromPart> fromParts = new ArrayList<FromPart>();
        if (fromPartElements != null) {
          for (final Element fromPartElement : fromPartElements) {
            final FromPart fromPart = new FromPart();
            final String part = XmlUtil.attribute(fromPartElement, "part");
            fromPart.setPart(part);
            final String toVariable = XmlUtil.attribute(fromPartElement, "toVariable");
            fromPart.setToVariable(toVariable);
            invokeActivityFullDefinition.addFromPartDefinition(new FromPartDefinitionImpl(part, toVariable));
            fromParts.add(fromPart);
          }
        }
        // Use the fromParts elements
        for (final FromPart fromPart : fromParts) {
          final From from = new From();
          from.setVariable(anonymousInboundMessageVariableName);
          from.setPart(fromPart.getPart());
          from.setVariant(Variant.VAR);
          final To to = new To();
          to.setVariable(fromPart.getToVariable());
          to.setVariant(Variant.VAR);
          final Variable toVariable = invoke.getEnclosingScope().findVariable(fromPart.getToVariable());
          if (toVariable.getMessageType() != null) {
            to.setPart(fromPart.getPart());
          }
          final Copy copy = new Copy();
          copy.setFrom(from);
          copy.setTo(to);
          inCopies.add(copy);
        }
      }
    }
    if (inCopies.size() != 0) {
      invoke.setInCopies(inCopies);
    }
  }

  /**
   * Parses correlation elements of the given receive.
   *
   * @param correlationElements the list of all correlation elements of the
   *          given receive
   * @param invoke invoke
   */
  private void parseCorrelations(final List<Element> correlationElements, final Invoke invoke, final Parse parse) {
    final List<Correlation> requestCorrelations = new ArrayList<Correlation>();
    final List<Correlation> responseCorrelations = new ArrayList<Correlation>();
    if (correlationElements != null) {
      for (final Element correlationElement : correlationElements) {
        final Correlation correlation = new Correlation();
        final String set = XmlUtil.attribute(correlationElement, "set");
        final Scope enclosingScope = invoke.getEnclosingScope();
        try {
          correlation.setCorrelationSet(enclosingScope.findCorrelationSet(set));
        } catch (final CorrelationSetNotFoundException e) {
          parse.addProblem("A correlation has not been defined in enclosing scopes" , e);
        }
        final String initiate = XmlUtil.attribute(correlationElement, "initiate");
        correlation.setInitiate(initiate);
        if (invoke.isOneWay()) {
          // TODO: if pattern != null, error during static analysis
          requestCorrelations.add(correlation);
        } else {
          final String pattern = XmlUtil.attribute(correlationElement, "pattern");
          // TODO: if pattern == null, error during static analysis
          if ("request".equals(pattern) || "request-response".equals(pattern)) {
            requestCorrelations.add(correlation);
          }
          if ("response".equals(pattern) || "request-response".equals(pattern)) {
            responseCorrelations.add(correlation);
          }
        }
      }
    }
    invoke.setRequestCorrelations(requestCorrelations);
    invoke.setResponseCorrelations(responseCorrelations);
  }

  @Override
  public InvokeActivityFullDefinition createActivityFullDefinition(final ProcessDefinitionUUID processDefinitionUUID,
      final ActivityDefinitionUUID activityDefinitionUUID, final String name, final Parse parse) {
    return new InvokeActivityFullDefinitionImpl(processDefinitionUUID, activityDefinitionUUID, name);
  }
}
