/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.parsing.binding;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.activity.Pick;
import org.ow2.orchestra.definition.element.Copy;
import org.ow2.orchestra.definition.element.OnAlarm;
import org.ow2.orchestra.definition.element.OnMessage;
import org.ow2.orchestra.definition.element.PartnerLink;
import org.ow2.orchestra.exception.StaticAnalysisException;
import org.ow2.orchestra.facade.def.ForDefinition;
import org.ow2.orchestra.facade.def.UntilDefinition;
import org.ow2.orchestra.facade.def.full.BpelActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.OnAlarmFullDefinition;
import org.ow2.orchestra.facade.def.full.OnMessageFullDefinition;
import org.ow2.orchestra.facade.def.full.PickActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.impl.OnAlarmFullDefinitionImpl;
import org.ow2.orchestra.facade.def.full.impl.OnMessageFullDefinitionImpl;
import org.ow2.orchestra.facade.def.full.impl.PickActivityFullDefinitionImpl;
import org.ow2.orchestra.facade.def.impl.ForDefinitionImpl;
import org.ow2.orchestra.facade.def.impl.UntilDefinitionImpl;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.facade.uuid.ActivityDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.lang.Expression;
import org.ow2.orchestra.parsing.binding.exception.CorrelationSetNotFoundException;
import org.ow2.orchestra.pvm.internal.model.NodeImpl;
import org.ow2.orchestra.pvm.internal.xml.Parse;
import org.ow2.orchestra.pvm.internal.xml.Parser;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.util.BpelUtil;
import org.ow2.orchestra.util.StaticAnalysisFault;
import org.ow2.orchestra.util.XmlUtil;
import org.w3c.dom.Element;

public class PickBinding extends ActivityBinding {

  /** log. */
  private static Logger log = Logger.getLogger(PickBinding.class.getName());

  public PickBinding() {
    super("pick", ActivityType.PICK);
  }

  public Object parse(final Element element, final Parse parse, final Parser parser) {
    final Pick pick = new Pick();
    this.setEnclosingScope(pick, parse);
    this.parseStandardAttributes(element, pick, parse);

    // Definition
    final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
    final PickActivityFullDefinition pickActivityFullDefinition = (PickActivityFullDefinition) definitionKeeper
      .getActivityFullDefinition();
    final List<OnMessageFullDefinition> onMessageFullDefinitionList = new ArrayList<OnMessageFullDefinition>();
    final List<OnAlarmFullDefinition> onAlarmFullDefinitionList = new ArrayList<OnAlarmFullDefinition>();

    this.parseStandardElements(element, pick, parse);

    pick.setCreateInstance(BpelUtil.isCreateInstance(element));

    final NodeImpl pickNode = this.createActivityNode(element, parse, parser, pick);

    final List<Element> onMessageElements = XmlUtil.elements(element, "onMessage");
    if (onMessageElements == null || onMessageElements.isEmpty()) {
      parse.addProblem("pick element " + element + "should contain at least one onMessage element.");
      return null;
    }
    final List<Element> onAlarmElements = XmlUtil.elements(element, "onAlarm");
    if (onAlarmElements != null && !onAlarmElements.isEmpty() && pick.isCreateInstance()) {
      final String message = StaticAnalysisFault.getMessage("SA00062", "  pick " + pick
        + " has a createInstance attribute with a value of yes. It must contain only onMessage events.");
      parse.addProblem(message, new StaticAnalysisException("SA00062"));
    }
    final List<OnMessage> onMessages = new ArrayList<OnMessage>();
    List<OnAlarm> onAlarms = null;
    parse.pushObject(pickNode);
    try {
      for (final Element onMessageElement : onMessageElements) {
        final OnMessageFullDefinition onMessageFullDefinition = new OnMessageFullDefinitionImpl();

        final OnMessage onMessage = this.parseOnMessage(onMessageElement, pickNode, pick, parse, parser, onMessageFullDefinition);
        onMessages.add(onMessage);

        onMessageFullDefinitionList.add(onMessageFullDefinition);
      }
      if (onAlarmElements != null) {
        onAlarms = new ArrayList<OnAlarm>();
        for (final Element onAlarmElement : onAlarmElements) {
          final OnAlarmFullDefinition onAlarmFullDefinition = new OnAlarmFullDefinitionImpl();

          final OnAlarm onAlarm = this.parseOnAlarm(onAlarmElement, pickNode, pick, parse, parser, onAlarmFullDefinition);
          onAlarms.add(onAlarm);

          onAlarmFullDefinitionList.add(onAlarmFullDefinition);
        }
      }
    } finally {
      parse.popObject();
    }
    pick.setOnMessages(onMessages);
    pick.setOnAlarms(onAlarms);

    pickActivityFullDefinition.setOnMessageDefinitionList(onMessageFullDefinitionList);
    pickActivityFullDefinition.setOnAlarmDefinitionList(onAlarmFullDefinitionList);
    pickActivityFullDefinition.setCreateInstance(pick.isCreateInstance());

    return pickNode;
  }

  public OnAlarm parseOnAlarm(final Element element, final NodeImpl pickNode, final Pick pick, final Parse parse,
    final Parser parser, final OnAlarmFullDefinition onAlarmFullDefinition) {
    if (PickBinding.log.isLoggable(Level.FINE)) {
      PickBinding.log.fine("parsing element = " + element);
    }
    final OnAlarm onAlarm = new OnAlarm();

     // Definition
    final DefinitionKeeper defKeeper = parse.findObject(DefinitionKeeper.class);
    defKeeper.pushParent(onAlarmFullDefinition);

    onAlarm.setActivity(this.parseChildActivity(element, parse, parser));

    defKeeper.popParent();

    final BpelProcess bpelProcess = parse.findObject(BpelProcess.class);
    final Element deadlineElement = XmlUtil.element(element, "until");
    final Element durationElement = XmlUtil.element(element, "for");
    if (durationElement != null && deadlineElement != null) {
      parse.addProblem("Error in element: " + element + " only one element 'for' or 'until' should be used.");
    }
    if (durationElement != null) {
      final Expression expression = PickBinding.parseExpression(durationElement, bpelProcess);
      onAlarm.setDurationExpression(expression);

      final ForDefinition forDefinition = new ForDefinitionImpl(expression.getLanguage(), durationElement.getTextContent());

      onAlarmFullDefinition.setForDefinition(forDefinition);
    } else if (deadlineElement != null) {
      final Expression expression = PickBinding.parseExpression(deadlineElement, bpelProcess);
      onAlarm.setDeadlineExpression(expression);

      final UntilDefinition untilDefinition = new UntilDefinitionImpl(expression.getLanguage(), deadlineElement.getTextContent());
      onAlarmFullDefinition.setUntilDefinition(untilDefinition);
    } else {
      parse.addProblem("Error in element: " + element + " one element 'for' or 'until' should be used.");
    }

    return onAlarm;
  }

  public OnMessage parseOnMessage(final Element element, final NodeImpl pickNode, final Pick pick, final Parse parse,
    final Parser parser, final OnMessageFullDefinition onMessageFullDefinition) {
    if (PickBinding.log.isLoggable(Level.FINE)) {
      PickBinding.log.fine("parsing element = " + element);
    }
    final OnMessage onMessage = new OnMessage();
    final String variable = XmlUtil.attribute(element, "variable");
    final String partnerLinkName = XmlUtil.attribute(element, "partnerLink");
    final String operation = XmlUtil.attribute(element, "operation");

    onMessage.setVariable(variable);
    onMessage.setPartnerLinkName(partnerLinkName);

    // Definition
    onMessageFullDefinition.setOperation(operation);
    onMessageFullDefinition.setPartnerLink(partnerLinkName);
    onMessageFullDefinition.setVariable(variable);

    final String messageExchange = XmlUtil.attribute(element, "messageExchange");
    if (messageExchange != null) {
      onMessage.setMessageExchange(messageExchange);
      onMessageFullDefinition.setMessageExchange(messageExchange);
    }

    // get partnerLinkDef and get plnkType attribute and myRole attribute
    final PartnerLink partnerLink = this.findPartnerLinkDefinition(parse, partnerLinkName);
    if (partnerLink == null) {
      parse.addProblem("No partnerLink " + partnerLinkName + " defined.");
      return null;
    }
    final QName plportType = partnerLink.getMyRolePortTypeQName();
    if (plportType == null) {
      parse.addProblem("No myRole defined for partnerLink: " + partnerLinkName);
      return null;
    }

    final BpelProcess bpelProcess = parse.findObject(BpelProcess.class);
    onMessage.setOperationKey(new OperationKey(bpelProcess.getQName(), plportType, operation));
    if (XmlUtil.attribute(element, "portType") != null) {
      if (!plportType.equals(XmlUtil.attributeQName(element, "portType"))) {
        // SA00005
        final String message = StaticAnalysisFault.getMessage("SA00005", "  onMessage " + onMessage
          + " has specified a PortType that is different from the one specified by the partnerLink");
        parse.addProblem(message, new StaticAnalysisException("SA00005"));
      }
    }

    // Definition
    onMessageFullDefinition.setPortType(onMessage.getPortTypeQName());

    // fromParts
    final Element fromPartsElement = XmlUtil.element(element, "fromParts");
    if (fromPartsElement != null) {
      if (onMessage.getVariable() != null) {
        // SA00055
        final String message = StaticAnalysisFault.getMessage("SA00063", "  onMessage " + onMessage
          + " has specified both a Variable and a fromPart");
        parse.addProblem(message, new StaticAnalysisException("SA00063"));
      }
      final List<Element> fromPartElements = XmlUtil.elements(fromPartsElement, "fromPart");
      onMessage.setAnonymousInboundMessageVariable(ReceiveBinding.getAnonymousVariable(onMessage, parse));
      final List<Copy> inCopies = ReceiveBinding.getInCopies(fromPartElements, pick.getEnclosingScope(), parse,
        onMessageFullDefinition);
      if (inCopies.size() != 0) {
        onMessage.setInCopies(inCopies);
      }
    }

    // Definition
    final DefinitionKeeper defKeeper = parse.findObject(DefinitionKeeper.class);
    defKeeper.pushParent(onMessageFullDefinition);

    // correlations
    final Element correlationsElement = XmlUtil.element(element, "correlations");
    if (correlationsElement != null) {
      final List<Element> correlationElements = XmlUtil.elements(correlationsElement, "correlation");
      try {
        onMessage.setCorrelations(ReceiveBinding.parseCorrelations(correlationElements, pick.getEnclosingScope(), parse,
          onMessageFullDefinition));
      } catch (final CorrelationSetNotFoundException e) {
        parse.addProblem("A correlation has not been defined in enclosing scopes" , e);
      }
    }

    if (pick.isCreateInstance()) {
      final OperationKey operationKey = onMessage.getOperationKey();
      if (bpelProcess.addStartElement(operationKey, pick) != null) {
        parse.addProblem(
            "More than one start activity found for operationKey: " + operationKey + ". This is not allowed.",
            new StaticAnalysisException("SA10001"));
      }
    }

    onMessage.setActivity(this.parseChildActivity(element, parse, parser));

    defKeeper.popParent();

    return onMessage;
  }
  protected static Expression parseExpression(final Element element, final BpelProcess bpelProcess) {
    String expressionLanguage = XmlUtil.attribute(element, "expressionLanguage");
    if (expressionLanguage == null) {
      expressionLanguage = bpelProcess.getExpressionLanguage();
    }
    final Expression expression = new Expression();
    expression.setLanguage(expressionLanguage);
    expression.setNamespaces(BpelUtil.getAllNameSpaces(element));
    expression.setText(element.getTextContent());
    return expression;
  }

  @Override
  public BpelActivityFullDefinition createActivityFullDefinition(final ProcessDefinitionUUID processDefinitionUUID,
    final ActivityDefinitionUUID activityDefinitionUUID, final String name, final Parse parse) {
    return new PickActivityFullDefinitionImpl(processDefinitionUUID, activityDefinitionUUID, name);
  }
}
