/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.parsing.binding;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.wsdl.Operation;
import javax.xml.namespace.QName;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.activity.Receive;
import org.ow2.orchestra.definition.activity.Scope;
import org.ow2.orchestra.definition.element.Copy;
import org.ow2.orchestra.definition.element.Correlation;
import org.ow2.orchestra.definition.element.From;
import org.ow2.orchestra.definition.element.FromPart;
import org.ow2.orchestra.definition.element.PartnerLink;
import org.ow2.orchestra.definition.element.To;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.exception.StaticAnalysisException;
import org.ow2.orchestra.facade.def.CorrelationDefinition;
import org.ow2.orchestra.facade.def.FromPartDefinition;
import org.ow2.orchestra.facade.def.Variant;
import org.ow2.orchestra.facade.def.full.ActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.BpelActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.OnEventFullDefinition;
import org.ow2.orchestra.facade.def.full.OnMessageFullDefinition;
import org.ow2.orchestra.facade.def.full.ReceiveActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.impl.ReceiveActivityFullDefinitionImpl;
import org.ow2.orchestra.facade.def.impl.CorrelationDefinitionImpl;
import org.ow2.orchestra.facade.def.impl.FromPartDefinitionImpl;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.facade.uuid.ActivityDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.parsing.binding.exception.CorrelationSetNotFoundException;
import org.ow2.orchestra.pvm.internal.model.NodeImpl;
import org.ow2.orchestra.pvm.internal.xml.Parse;
import org.ow2.orchestra.pvm.internal.xml.Parser;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.services.ReceivingElement;
import org.ow2.orchestra.util.BpelUtil;
import org.ow2.orchestra.util.StaticAnalysisFault;
import org.ow2.orchestra.util.XmlUtil;
import org.w3c.dom.Element;

/**
 * @author Charles Souillard, Goulven Le Jeune
 */
public class ReceiveBinding extends ActivityBinding {

  /** log. */
  private static Logger log = Logger.getLogger(ReceiveBinding.class.getName());

  public ReceiveBinding() {
    super("receive", ActivityType.RECEIVE);
  }
  /**
   * Parses the given dom element to build a Node and a behaviour.
   *
   * @param element dom element to parse
   * @param parse parse
   * @param parser parser
   * @return the build behaviour!
   */
  public Object parse(final Element element, final Parse parse, final Parser parser) {
    if (ReceiveBinding.log.isLoggable(Level.FINE)) {
      ReceiveBinding.log.fine("parsing element = " + element);
    }
    final Receive receive = new Receive();
    this.setEnclosingScope(receive, parse);
    this.parseStandardAttributes(element, receive, parse);

    final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
    final ActivityFullDefinition activityFullDefinition = definitionKeeper.getActivityFullDefinition();

    this.parseStandardElements(element, receive, parse);
    final BpelProcess bpelProcess = parse.findObject(BpelProcess.class);

    receive.setVariable(XmlUtil.attribute(element, "variable"));
    final String partnerLinkName = XmlUtil.attribute(element, "partnerLink");
    receive.setPartnerLinkName(partnerLinkName);
    final String operation = XmlUtil.attribute(element, "operation");
    receive.setCreateInstance(BpelUtil.booleanValue(XmlUtil.attribute(element, "createInstance")));
    final String messageExchange = XmlUtil.attribute(element, "messageExchange");
    if (messageExchange != null) {
      receive.setMessageExchange(messageExchange);
    }
    // get partnerLinkDef and get plnkType attribute and myRole attribute
    final PartnerLink partnerLink = this.findPartnerLinkDefinition(parse, partnerLinkName);
    if (partnerLink == null) {
      parse.addProblem("No partnerLink " + partnerLinkName + " defined.");
      return null;
    }
    final QName plportType = partnerLink.getMyRolePortTypeQName();
    if (plportType == null) {
      // SA00005
      final String message = StaticAnalysisFault.getMessage("SA00005", "  receive " + receive.getName()
        + " has specified a partnerLink that not corresponds to myRole");
      parse.addProblem(message, new StaticAnalysisException("SA00005"));
      return null;
    }

    receive.setOperationKey(new OperationKey(bpelProcess.getQName(), plportType, operation));
    // Add attributes to ReceiveActivityFullDefinition
    ((ReceiveActivityFullDefinition) activityFullDefinition).setVariable(receive.getVariable());
    ((ReceiveActivityFullDefinition) activityFullDefinition).setPartnerLink(receive.getPartnerLinkName());
    ((ReceiveActivityFullDefinition) activityFullDefinition).setOperation(receive.getOperation());
    ((ReceiveActivityFullDefinition) activityFullDefinition).setCreateInstance(receive.isCreateInstance());
    ((ReceiveActivityFullDefinition) activityFullDefinition).setMessageExchange(receive.getMessageExchange());
    ((ReceiveActivityFullDefinition) activityFullDefinition).setPortType(receive.getPortTypeQName());

    if (XmlUtil.attribute(element, "portType") != null) {
      if (!plportType.equals(XmlUtil.attributeQName(element, "portType"))) {
        // SA00005
        final String message = StaticAnalysisFault.getMessage("SA00005", "  receive " + receive.getName()
          + " has specified a PortType that is different from the one specified by the partnerLink");
        parse.addProblem(message, new StaticAnalysisException("SA00005"));
      }
    }
    // fromParts
    final Element fromPartsElement = XmlUtil.element(element, "fromParts");
    if (fromPartsElement != null) {
      if (receive.getVariable() != null) {
        // SA00055
        final String message = StaticAnalysisFault.getMessage("SA00055", "  receive " + receive.getName()
          + " has specified both a Variable and a fromPart");
        parse.addProblem(message, new StaticAnalysisException("SA00055"));
        return null;
      }
      final List<Element> fromPartElements = XmlUtil.elements(fromPartsElement, "fromPart");
      receive.setAnonymousInboundMessageVariable(ReceiveBinding.getAnonymousVariable(receive, parse));
      final List<Copy> inCopies = ReceiveBinding.getInCopies(fromPartElements, receive.getEnclosingScope(), parse, null);
      if (inCopies.size() != 0) {
        receive.setInCopies(inCopies);
      }
    }
    // correlations
    final Element correlationsElement = XmlUtil.element(element, "correlations");
    if (correlationsElement != null) {
      final List<Element> correlationElements = XmlUtil.elements(correlationsElement, "correlation");
      try {
        receive.setCorrelations(ReceiveBinding.parseCorrelations(correlationElements, receive.getEnclosingScope(), parse, null));
      } catch (final CorrelationSetNotFoundException e) {
        parse.addProblem("A correlation has not been defined in enclosing scopes" , e);
      }
    }
    final NodeImpl receiveNode = this.createActivityNode(element, parse, parser, receive);
    if (BpelUtil.isCreateInstance(element)) {
      final OperationKey operationKey = receive.getOperationKey();
      if (bpelProcess.addStartElement(operationKey, receive) != null) {
        parse.addProblem(
            "More than one start activity found for operationKey: " + operationKey + ". This is not allowed.",
            new StaticAnalysisException("SA10001"));
      }
    }
    return receiveNode;
  }

  /**
   * Parses correlation elements.
   *
   * @param correlationElements the list of all correlation elements of the
   *          given receivingElement
   * @param Scope enclosing scope of the current activity
   */
  protected static List<Correlation> parseCorrelations(final List<Element> correlationElements, final Scope enclosingScope,
    final Parse parse, final Object onMessageFullDefinition) throws CorrelationSetNotFoundException {
    final List<Correlation> correlations = new ArrayList<Correlation>();

    final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
    final ActivityFullDefinition activityFullDefinition = definitionKeeper.getActivityFullDefinition();
    final List<CorrelationDefinition> correlationDefinitionList = new ArrayList<CorrelationDefinition>();

    if (correlationElements != null) {
      for (final Element correlationElement : correlationElements) {
        final Correlation correlation = new Correlation();
        final String set = XmlUtil.attribute(correlationElement, "set");
        correlation.setCorrelationSet(enclosingScope.findCorrelationSet(set));
        final String initiate = XmlUtil.attribute(correlationElement, "initiate");
        correlation.setInitiate(initiate);

        correlations.add(correlation);

        final CorrelationDefinitionImpl correlationDefinition = new CorrelationDefinitionImpl(set, initiate);
        correlationDefinitionList.add(correlationDefinition);
      }
      if (ActivityType.RECEIVE.equals(activityFullDefinition.getType())) {
        ((ReceiveActivityFullDefinition) activityFullDefinition).setCorrelationDefinition(correlationDefinitionList);
      } else if (onMessageFullDefinition instanceof OnMessageFullDefinition) {
        ((OnMessageFullDefinition) onMessageFullDefinition).setCorrelationDefinition(correlationDefinitionList);
      } else if (onMessageFullDefinition instanceof OnEventFullDefinition) {
        ((OnEventFullDefinition) onMessageFullDefinition).setCorrelationDefinition(correlationDefinitionList);
      }

    }
    return correlations;
  }

  /**
   * Creates anonymous inbound variable of the receiving element.
   */
  protected static Variable getAnonymousVariable(final ReceivingElement receivingElement, final Parse parse) {
    final BpelProcess bpelProcess = parse.findObject(BpelProcess.class);
    final Operation operation = bpelProcess.getWsdlInfos().getPortType(receivingElement.getPortTypeQName()).getOperation(
      receivingElement.getOperation(), null, null);
    final QName messageQName = operation.getInput().getMessage().getQName();

    final Variable anonymousInboundMessageVariable = new Variable();
    anonymousInboundMessageVariable.setName(Variable.INBOUND_ANONYMOUS_WSDL_MESSAGE_NAME);
    anonymousInboundMessageVariable.setMessageType(messageQName);
    return anonymousInboundMessageVariable;
  }

  /**
   * Parses inCopies elements
   *
   * @param fromPartElements the list of all fromParts elements of the given
   *          receivingElement
   * @param enclosingScope enclosing scope of the current activity
   */
  protected static List<Copy> getInCopies(final List<Element> fromPartElements, final Scope enclosingScope, final Parse parse,
    final Object onMessageFullDefinition) {
    final List<Copy> inCopies = new ArrayList<Copy>();

    final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
    final ActivityFullDefinition activityFullDefinition = definitionKeeper.getActivityFullDefinition();
    final List<FromPartDefinition> fromPartDefinitionList = new ArrayList<FromPartDefinition>();

    if (fromPartElements != null) {
      final List<FromPart> fromParts = new ArrayList<FromPart>();
      for (final Element fromPartElement : fromPartElements) {
        final FromPart fromPart = new FromPart();
        final String part = XmlUtil.attribute(fromPartElement, "part");
        fromPart.setPart(part);
        final String toVariable = XmlUtil.attribute(fromPartElement, "toVariable");
        fromPart.setToVariable(toVariable);
        fromParts.add(fromPart);

        // Definition
        final FromPartDefinitionImpl fromPartDefinition = new FromPartDefinitionImpl(fromPart.getPart(), fromPart.getToVariable());
        fromPartDefinitionList.add(fromPartDefinition);
      }

      if (ActivityType.RECEIVE.equals(activityFullDefinition.getType())) {
        ((ReceiveActivityFullDefinition) activityFullDefinition).setFromPartDefinition(fromPartDefinitionList);
      } else if (onMessageFullDefinition instanceof OnMessageFullDefinition) {
        ((OnMessageFullDefinition) onMessageFullDefinition).setFromPartDefinition(fromPartDefinitionList);
      } else if (onMessageFullDefinition instanceof OnEventFullDefinition) {
        ((OnEventFullDefinition) onMessageFullDefinition).setFromPartDefinition(fromPartDefinitionList);
      }

      // Use the fromParts elements
      for (final FromPart fromPart : fromParts) {
        final From from = new From();
        from.setVariable(Variable.INBOUND_ANONYMOUS_WSDL_MESSAGE_NAME);
        from.setPart(fromPart.getPart());
        from.setVariant(Variant.VAR);
        final To to = new To();
        to.setVariable(fromPart.getToVariable());
        to.setVariant(Variant.VAR);
        // TODO: check variable has valid type
        final Copy copy = new Copy();
        copy.setKeepSrcElementName(true);
        copy.setFrom(from);
        copy.setTo(to);
        inCopies.add(copy);
      }
    }

    return inCopies;
  }
  @Override
  public BpelActivityFullDefinition createActivityFullDefinition(final ProcessDefinitionUUID processDefinitionUUID,
    final ActivityDefinitionUUID activityDefinitionUUID, final String name, final Parse parse) {
    return new ReceiveActivityFullDefinitionImpl(processDefinitionUUID, activityDefinitionUUID, name);
  }
}
