/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.parsing.binding;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.wsdl.Operation;
import javax.xml.namespace.QName;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.activity.Reply;
import org.ow2.orchestra.definition.activity.Scope;
import org.ow2.orchestra.definition.element.Copy;
import org.ow2.orchestra.definition.element.Correlation;
import org.ow2.orchestra.definition.element.From;
import org.ow2.orchestra.definition.element.PartnerLink;
import org.ow2.orchestra.definition.element.To;
import org.ow2.orchestra.definition.element.ToPart;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.exception.StaticAnalysisException;
import org.ow2.orchestra.facade.def.CorrelationDefinition;
import org.ow2.orchestra.facade.def.ToPartDefinition;
import org.ow2.orchestra.facade.def.Variant;
import org.ow2.orchestra.facade.def.full.ActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.BpelActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.impl.ReplyActivityFullDefinitionImpl;
import org.ow2.orchestra.facade.def.impl.CorrelationDefinitionImpl;
import org.ow2.orchestra.facade.def.impl.ToPartDefinitionImpl;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.facade.uuid.ActivityDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.parsing.binding.exception.CorrelationSetNotFoundException;
import org.ow2.orchestra.pvm.internal.model.NodeImpl;
import org.ow2.orchestra.pvm.internal.xml.Parse;
import org.ow2.orchestra.pvm.internal.xml.Parser;
import org.ow2.orchestra.util.StaticAnalysisFault;
import org.ow2.orchestra.util.XmlUtil;
import org.w3c.dom.Element;

public class ReplyBinding extends ActivityBinding {

  /** log. */
  private static Logger log = Logger.getLogger(ReplyBinding.class.getName());


  public ReplyBinding() {
    super("reply", ActivityType.REPLY);
  }

  /**
   * Parses the given dom element to build a Node and a behaviour.
   *
   * @param element
   *            dom element to parse
   * @param parse
   *            parse
   * @param parser
   *            parser
   * @return the build behaviour!
   */
  public Object parse(final Element element, final Parse parse, final Parser parser) {
    if (ReplyBinding.log.isLoggable(Level.FINE)) {
      ReplyBinding.log.fine("parsing element = " + element);
    }
    final Reply reply = new Reply();
    this.setEnclosingScope(reply, parse);
    this.parseStandardAttributes(element, reply, parse);

    final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
    final ActivityFullDefinition activityFullDefinition = definitionKeeper.getActivityFullDefinition();

    this.parseStandardElements(element, reply, parse);

    reply.setVariable(XmlUtil.attribute(element, "variable"));
    final String partnerLinkName = XmlUtil.attribute(element, "partnerLink");
    reply.setPartnerLinkName(partnerLinkName);
    reply.setOperation(XmlUtil.attribute(element, "operation"));
    final QName faultQName = XmlUtil.attributeQName(element, "faultName");
    if (faultQName != null) {
      reply.setFaultQName(faultQName);
    }
    final String messageExchange = XmlUtil.attribute(element, "messageExchange");
    if (messageExchange != null) {
      reply.setMessageExchange(messageExchange);
    }

    ((ReplyActivityFullDefinitionImpl) activityFullDefinition).setVariable(reply.getVariable());
    ((ReplyActivityFullDefinitionImpl) activityFullDefinition).setPartnerLink(reply.getPartnerLinkName());
    ((ReplyActivityFullDefinitionImpl) activityFullDefinition).setOperation(reply.getOperation());
    ((ReplyActivityFullDefinitionImpl) activityFullDefinition).setFaultName(reply.getFaultQName());
    ((ReplyActivityFullDefinitionImpl) activityFullDefinition).setMessageExchange(reply.getMessageExchange());

    // get partnerLinkDef and get plnkType attribute and myRole attribute
    final PartnerLink partnerLink = this.findPartnerLinkDefinition(parse,
        partnerLinkName);
    if (partnerLink == null) {
      parse.addProblem("No partnerLink " + partnerLinkName + " defined.");
      return null;
    }
    final QName plportType = partnerLink.getMyRolePortTypeQName();
    if (plportType == null) {
      parse.addProblem("No myRole defined for partnerLink: " + partnerLinkName);
      return null;
    }
    reply.setPortTypeQName(plportType);

    if (XmlUtil.attribute(element, "portType") != null) {
      if (!plportType.equals(XmlUtil.attributeQName(element, "portType"))) {
        // SA00005
        final String message = StaticAnalysisFault.getMessage("SA00005",
            "  reply " + reply.getName()
            + " has specified a PortType that is different from the one specified by the partnerLink");
        parse.addProblem(message, new StaticAnalysisException("SA00005"));
      }
    }

    ((ReplyActivityFullDefinitionImpl) activityFullDefinition).setPortType(reply.getPortTypeQName());

    // toParts
    final Element toPartsElement = XmlUtil.element(element, "toParts");
    if (toPartsElement != null) {
      if (reply.getVariable() != null) {
        // SA00059
        final String message = StaticAnalysisFault.getMessage("SA00059",
            "  reply " + reply.getName() + " has specified both a Variable and a toPart");
        parse.addProblem(message, new StaticAnalysisException("SA00059"));
      }
      final List<Element> toPartElements = XmlUtil.elements(toPartsElement, "toPart");
      this.setOutCopies(toPartElements, reply, parse);
    }

    // correlations
    final Element correlationsElement = XmlUtil.element(element, "correlations");
    if (correlationsElement != null) {
      final List<Element> correlationElements = XmlUtil.elements(correlationsElement,
          "correlation");
      this.parseCorrelations(correlationElements, reply, parse);
    }

    final NodeImpl replyNode = this.createActivityNode(element, parse, parser, reply);
    return replyNode;
  }

  private void setOutCopies(final List<Element> toPartElements, final Reply reply,
      final Parse parse) {

    final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
    final ActivityFullDefinition activityFullDefinition = definitionKeeper.getActivityFullDefinition();
    final List<ToPartDefinition> toPartDefinitionList = new ArrayList<ToPartDefinition>();

    if (toPartElements != null) {
      final List<ToPart> toParts = new ArrayList<ToPart>();
      for (final Element toPartElement : toPartElements) {
        final ToPart toPart = new ToPart();
        final String part = XmlUtil.attribute(toPartElement, "part");
        toPart.setPart(part);
        final String fromVariable = XmlUtil.attribute(toPartElement, "fromVariable");
        toPart.setFromVariable(fromVariable);
        toParts.add(toPart);

        final ToPartDefinitionImpl toPartDefinition = new ToPartDefinitionImpl(toPart.getFromVariable(), toPart.getPart());
        toPartDefinitionList.add(toPartDefinition);
      }
      final BpelProcess bpelProcess = parse.findObject(BpelProcess.class);
      final Operation operation = bpelProcess.getWsdlInfos().getPortType(
          reply.getPortTypeQName()).getOperation(reply.getOperation(), null,
          null);
      final QName messageQName = operation.getOutput().getMessage().getQName();

      final Variable anonymousOutboundMessageVariable = new Variable();
      anonymousOutboundMessageVariable
          .setName(Variable.OUTBOUND_ANONYMOUS_WSDL_MESSAGE_NAME);
      anonymousOutboundMessageVariable.setMessageType(messageQName);
      reply
          .setAnonymousOutboundMessageVariable(anonymousOutboundMessageVariable);
      final String anonymousOutboundMessageVariableName = anonymousOutboundMessageVariable
          .getName();

      ((ReplyActivityFullDefinitionImpl) activityFullDefinition).setToPartDefinition(toPartDefinitionList);

      List<Copy> outCopies = null;
      for (final ToPart toPart : toParts) {
        if (outCopies == null) {
          outCopies = new ArrayList<Copy>();
        }
        final From from = new From();
        from.setVariable(toPart.getFromVariable());
        from.setVariant(Variant.VAR);
        final To to = new To();
        to.setVariable(anonymousOutboundMessageVariableName);
        to.setPart(toPart.getPart());
        to.setVariant(Variant.VAR);
        final Copy copy = new Copy();
        copy.setFrom(from);
        copy.setTo(to);
        outCopies.add(copy);
      }
      reply.setOutCopies(outCopies);
    }
  }

  private void parseCorrelations(final List<Element> correlationElements, final Reply reply, final Parse parse) {
    final List<Correlation> correlations = new ArrayList<Correlation>();

    final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
    final ActivityFullDefinition activityFullDefinition = definitionKeeper.getActivityFullDefinition();
    final List<CorrelationDefinition> correlationDefinitionList = new ArrayList<CorrelationDefinition>();

    if (correlationElements != null) {
      for (final Element correlationElement : correlationElements) {
        final Correlation correlation = new Correlation();
        final String set = XmlUtil.attribute(correlationElement, "set");
        final Scope enclosingScope = reply.getEnclosingScope();
        try {
          correlation.setCorrelationSet(enclosingScope.findCorrelationSet(set));
        } catch (final CorrelationSetNotFoundException e) {
          parse.addProblem("A correlation has not been defined in enclosing scopes" , e);
        }
        final String initiate = XmlUtil.attribute(correlationElement, "initiate");
        correlation.setInitiate(initiate);

        correlations.add(correlation);

        final CorrelationDefinitionImpl correlationDefinition = new CorrelationDefinitionImpl(set, initiate);
        correlationDefinitionList.add(correlationDefinition);
      }
    }
    reply.setCorrelations(correlations);

    ((ReplyActivityFullDefinitionImpl) activityFullDefinition).setCorrelationDefinition(correlationDefinitionList);
  }

  @Override
  public BpelActivityFullDefinition createActivityFullDefinition(final ProcessDefinitionUUID processDefinitionUUID,
    final ActivityDefinitionUUID activityDefinitionUUID, final String name, final Parse parse) {
    return new ReplyActivityFullDefinitionImpl(processDefinitionUUID, activityDefinitionUUID, name);
  }
}
