/**
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2007 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.parsing.binding;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.wsdl.Message;
import javax.wsdl.Operation;
import javax.wsdl.Part;
import javax.xml.namespace.QName;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.activity.AbstractActivity;
import org.ow2.orchestra.definition.activity.EventHandlerActivity;
import org.ow2.orchestra.definition.activity.Scope;
import org.ow2.orchestra.definition.element.Copy;
import org.ow2.orchestra.definition.element.CorrelationSet;
import org.ow2.orchestra.definition.element.From;
import org.ow2.orchestra.definition.element.OnAlarm;
import org.ow2.orchestra.definition.element.OnEvent;
import org.ow2.orchestra.definition.element.PartnerLink;
import org.ow2.orchestra.definition.element.To;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.exception.StaticAnalysisException;
import org.ow2.orchestra.facade.def.CorrelationSetDefinition;
import org.ow2.orchestra.facade.def.ForDefinition;
import org.ow2.orchestra.facade.def.PartnerLinkDefinition;
import org.ow2.orchestra.facade.def.RepeatEveryDefinition;
import org.ow2.orchestra.facade.def.UntilDefinition;
import org.ow2.orchestra.facade.def.VariableDefinition;
import org.ow2.orchestra.facade.def.Variant;
import org.ow2.orchestra.facade.def.full.ActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.BpelActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.CopyFullDefinition;
import org.ow2.orchestra.facade.def.full.OnAlarmFullDefinition;
import org.ow2.orchestra.facade.def.full.OnEventFullDefinition;
import org.ow2.orchestra.facade.def.full.ScopeActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.impl.CopyFullDefinitionImpl;
import org.ow2.orchestra.facade.def.full.impl.OnAlarmFullDefinitionImpl;
import org.ow2.orchestra.facade.def.full.impl.OnEventFullDefinitionImpl;
import org.ow2.orchestra.facade.def.full.impl.ScopeActivityFullDefinitionImpl;
import org.ow2.orchestra.facade.def.impl.CorrelationSetDefinitionImpl;
import org.ow2.orchestra.facade.def.impl.ForDefinitionImpl;
import org.ow2.orchestra.facade.def.impl.PartnerLinkDefinitionImpl;
import org.ow2.orchestra.facade.def.impl.RepeatEveryDefinitionImpl;
import org.ow2.orchestra.facade.def.impl.UntilDefinitionImpl;
import org.ow2.orchestra.facade.def.impl.VariableDefinitionImpl;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.facade.uuid.ActivityDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.lang.Expression;
import org.ow2.orchestra.parsing.binding.exception.CorrelationSetNotFoundException;
import org.ow2.orchestra.parsing.binding.exception.DuplicateVariableException;
import org.ow2.orchestra.pvm.internal.model.NodeImpl;
import org.ow2.orchestra.pvm.internal.xml.Parse;
import org.ow2.orchestra.pvm.internal.xml.Parser;
import org.ow2.orchestra.pvm.model.Node;
import org.ow2.orchestra.runtime.OrchestraFaultHandler;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.util.BpelUtil;
import org.ow2.orchestra.util.StaticAnalysisFault;
import org.ow2.orchestra.util.XmlConstants;
import org.ow2.orchestra.util.XmlUtil;
import org.ow2.orchestra.wsdl.PartnerLinkType;
import org.ow2.orchestra.wsdl.Property;
import org.ow2.orchestra.wsdl.WsdlsInfos;
import org.w3c.dom.Element;

/**
 * @author Charles Souillard, Goulven Le Jeune
 */

public class ScopeBinding extends ActivityWithFCTHBinding {

  /** log. */
  private static Logger log = Logger.getLogger(ScopeBinding.class.getName());

  public ScopeBinding() {
    super("scope", ActivityType.SCOPE);
  }

  public ScopeBinding(final String tagName, final ActivityType activityType) {
    super(tagName, activityType);
  }

  /**
   * Parses the given dom element to build a Node and a behaviour.
   *
   * @param element
   *          dom element to parse
   * @param parse
   *          parse
   * @param parser
   *          parser
   * @return the build behaviour!
   */
  public Object parse(final Element element, final Parse parse, final Parser parser) {
    try {
      return this.parse(element, null, parse, parser);
    } catch (final DuplicateVariableException e) {
      parse.addProblem("Duplicate variable ", e);
      return null;
    }
  }

  /**
   * Factory method (reused by {@link ProcessBinding})
   *
   * @return {@link org.ow2.orchestra.definition.activity.AbstractBpelActivity}
   *         associated with this node (must extend {@link Scope})
   */
  protected Scope createScopeActivity() {
    return new Scope();
  }

  /**
   * Parses the given dom element to build a Node and a behaviour.
   *
   * @param element
   *          dom element to parse
   * @param additionnal
   *          variables to add
   * @param parse
   *          parse
   * @param parser
   *          parser
   * @return the build behaviour!
   * @throws DuplicateVariableException
   *           if duplicated variabled
   */
  public Object parse(final Element element, final Set<Variable> moreVariables, final Parse parse, final Parser parser)
    throws DuplicateVariableException {
    if (ScopeBinding.log.isLoggable(Level.FINE)) {
      ScopeBinding.log.fine("parsing element = " + element);
    }

    final Scope scope = this.createScopeActivity();
    this.setEnclosingScope(scope, parse);
    this.parseStandardAttributes(element, scope, parse);

    final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
    final ScopeActivityFullDefinition activityFullDefinition = (ScopeActivityFullDefinition) definitionKeeper
      .getActivityFullDefinition();

    this.parseStandardElements(element, scope, parse);

    final WsdlsInfos wsdlInfos = parse.findObject(BpelProcess.class).getWsdlInfos();

    // parse attributes
    final String isolatedAttr = XmlUtil.attribute(element, "isolated");
    if (isolatedAttr != null) {
      final boolean isolated = BpelUtil.booleanValue(isolatedAttr);
      scope.setIsolated(isolated);
      if (isolated) {
        parse.addProblem("isolated attribute not yet supported on scope activity");
      }
    }

    boolean defaultExitOnStandardFault = false;
    if (scope.getEnclosingScope() != null) {
      defaultExitOnStandardFault = scope.getEnclosingScope().isExitOnStandardFault();
    }
    final boolean exitOnStandardFault = BpelUtil.booleanValue(XmlUtil.attribute(element, "exitOnStandardFault"),
      defaultExitOnStandardFault);
    scope.setExitOnStandardFault(exitOnStandardFault);
    if (exitOnStandardFault) {
      ScopeBinding.log.warning("exitOnStandardFault attribute not yet supported on scope activity");
    }
    activityFullDefinition.setExitOnStandardFault(exitOnStandardFault);

    // partnerLinks
    final Element partnerLinksElement = XmlUtil.element(element, "partnerLinks");
    if (partnerLinksElement != null) {
      final List<Element> partnerLinkElements = XmlUtil.elements(partnerLinksElement, "partnerLink");
      this.parsePartnerLinks(partnerLinkElements, wsdlInfos, scope, parse);
    }

    // messageExchanges
    final Element messageExchangesElement = XmlUtil.element(element, "messageExchanges");
    if (messageExchangesElement != null) {
      final List<Element> messageExchangeElements = XmlUtil.elements(messageExchangesElement, "messageExchange");
      this.parseMessageExchanges(messageExchangeElements, scope, parse);
    }

    // variables
    final Element variablesElement = XmlUtil.element(element, "variables");
    if (variablesElement != null) {
      final List<Element> variableElements = XmlUtil.elements(variablesElement, "variable");
      this.parseVariables(variableElements, scope, parse);
    }

    // add more variables
    if (moreVariables != null) {
      Map<String, Variable> variables = scope.getVariables();
      final List<VariableDefinition> variableDefinitionList = new ArrayList<VariableDefinition>();

      if (variables == null) {
        variables = new HashMap<String, Variable>();
      }
      for (final Variable var : moreVariables) {
        if (variables.put(var.getName(), var) != null) {
          throw new DuplicateVariableException();
        } else {
          // Add variable definition to the scope definition list
          final VariableDefinitionImpl variableDefinition = new VariableDefinitionImpl(var.getElement(), var.getMessageType(),
            var.getName(), var.getXmlType());
          variableDefinitionList.add(variableDefinition);
        }
      }
      scope.setVariables(variables);
      activityFullDefinition.setVariableDefinition(variableDefinitionList);
    }

    // correlationSets
    final Element correlationSetsElement = XmlUtil.element(element, "correlationSets");
    if (correlationSetsElement != null) {
      final List<Element> correlationSetElements = XmlUtil.elements(correlationSetsElement, "correlationSet");
      this.parseCorrelationSets(correlationSetElements, scope, wsdlInfos, parse);
    }

    // Create Scope node and child activity node (child 0)
    final NodeImpl scopeNode = this.createActivityNodeWithChildren(element, parse, parser, scope);

    scope.setScopeNode(scopeNode);

    try {
      parse.pushObject(scope);
      parse.pushObject(scopeNode);

      // eventHandlers (child 1)
      final String name = "eventHandler_" + scopeNode.getName();
      final NodeImpl activity = scopeNode.createNode(name);
      final EventHandlerActivity eventHandlerActivity = new EventHandlerActivity();
      final ProcessDefinitionUUID processDefinitionUUID = definitionKeeper.getProcessFullDefinition().getUUID();
      final ActivityDefinitionUUID activityDefinitionUUID = new ActivityDefinitionUUID(processDefinitionUUID, definitionKeeper
        .getNextActivityCount(), eventHandlerActivity.getType(), eventHandlerActivity.getName());
      eventHandlerActivity.setUuid(activityDefinitionUUID);

      activity.setBehaviour(eventHandlerActivity);
      final Element eventHandlersElement = XmlUtil.element(element, "eventHandlers");
      if (eventHandlersElement != null) {
        parse.pushObject(eventHandlerActivity);
        parse.pushObject(activity);
        try {
          final int elements = this.parseEventHandlers(eventHandlersElement,scope, parse, parser, eventHandlerActivity);
          if (elements < 1) {
            final String message = StaticAnalysisFault.getMessage("SA00083",
              "The <eventHandlers> element does not have any <onEvent> or <onAlarm> element");
            parse.addProblem(message, new StaticAnalysisException("SA00083"));
          }
        } finally {
          parse.popObject();
          parse.popObject();
        }

      }

      // Analyze the terminationHandler create TH node (child 2)
      this.parseTermination(element, parse, parser, scopeNode, activityFullDefinition);

      // Analyze the compensationHandler (child 3)
      this.parseCompensation(element, parse, parser, scopeNode, activityFullDefinition);

      // Analyze the faultHandlers (catch all child 4, other catch child 5..n)
      final Element faultHandlersElement = XmlUtil.element(element, "faultHandlers");
      if (faultHandlersElement != null) {
        parse.pushObject(faultHandlersElement);

        try {
          final int elements = this.parseCatchAndCatchAll(faultHandlersElement, parse, parser, scopeNode, activityFullDefinition);

          // Static Analysis SA00080
          if (elements < 1) {
            final String message = StaticAnalysisFault.getMessage("SA00080", "  The <faultHandlers> element for scope \""
              + scope.getName() + "\" does not have any <catch> or <catchAll> elements.");
            parse.addProblem(message, new StaticAnalysisException("SA00080"));
            return null;
          }
        } finally {
          parse.popObject();
        }
      } else {
        // Add the default faultHandler node in order to respect indices
        this.addDefaultFH(parse, scopeNode);

        // Since we did not call parseCatchAndCatchAll, we need to add the
        // exception handler to the scopeNode manually
        scopeNode.addExceptionHandler(new OrchestraFaultHandler());
      }

    } finally {
      parse.popObject();
      parse.popObject();
    }

    final List<Scope> immediatelyEnclosingScopes = this.getImmediatelyEnclosingScopes(scopeNode);
    for (final Scope scope1 : immediatelyEnclosingScopes) {
      if (scope1.getName() != null && !XmlConstants.BPEL_DEFAULT_ACTIVITY_NAME.equals(scope1.getName())) {
        for (final Scope scope2 : immediatelyEnclosingScopes) {
          if (!scope1.getUuid().equals(scope2.getUuid())) {
            if (scope1.getName().equals(scope2.getName())) {
              final String message = StaticAnalysisFault.getMessage("SA00092", "  The scope name \"" + scope1.getName()
                + "\" is used twice in the immediately enclosing scopes of scope \"" + scope.getName() + "\"");
              parse.addProblem(message, new StaticAnalysisException("SA00092"));
            }
          }
        }
      }
    }

    return scopeNode;
  }

  /**
   * A scope A is considered to immediately enclose another scope B, if B is
   * enclosed in A and B is not enclosed in any other scope or FCT-handler that
   * is itself enclosed in the outer scope A. Other structured activities (e.g.
   * &lt;sequence&gt; or &lt;forEach&gt;) and event handlers enclosed in A do
   * not affect the concept of immediate enclosure. This definition includes
   * scopes that result from the &lt;invoke&gt; shorthand notation for fault
   * handlers and compensation handlers.<br/>
   * <br/>
   * <p align="right"><b><i>wsbpel-v2.0-OS, page 123, paragraph 2</i></b></p>
   */
  private List<Scope> getImmediatelyEnclosingScopes(final NodeImpl node) {
    final List<Scope> result = new ArrayList<Scope>();
    if (node.getNodes() != null) {
      for (final Node subNode : node.getNodes()) {
        final NodeImpl subNodeImpl = (NodeImpl) subNode;
        final AbstractActivity subNodeActivity = (AbstractActivity) subNodeImpl.getBehaviour();
        final ActivityType subNodeActivityType = subNodeActivity.getType();
        if (subNodeActivityType == ActivityType.SCOPE) {
          result.add((Scope) subNodeActivity);
        } else if (subNodeActivityType != ActivityType.CATCH_HANDLER && subNodeActivityType != ActivityType.CATCHALL_HANDLER
          && subNodeActivityType != ActivityType.TERMINATION_HANDLER && subNodeActivityType != ActivityType.COMPENSATION_HANDLER) {
          result.addAll(this.getImmediatelyEnclosingScopes(subNodeImpl));
        }
      }
    }
    return result;
  }


  /**
   * Parses all partnerLink elements of the fiven scope.
   *
   * @param partnerLinkElements
   *          list of all partnerLink elements of this scope
   * @param wsdlInfos
   *          wsdlInfos
   * @param scope
   *          current scope. if <code>null</code>, only parse partnerLinks, without adding it to the scope.
   *          (used in OnEvent parsing)
   */
  private Map<String, PartnerLink> parsePartnerLinks(final List<Element> partnerLinkElements,
      final WsdlsInfos wsdlInfos, final Scope scope, final Parse parse) {
    final Map<String, PartnerLink> partnerLinks = new HashMap<String, PartnerLink>();

    final List<PartnerLinkDefinition> partnerLinkDefinitionList = new ArrayList<PartnerLinkDefinition>();

    if (partnerLinkElements != null) {
      for (final Element partnerLinkElement : partnerLinkElements) {
        // variables use to construct the PartnerLinkDefinition
        String myRoleNameDefinition = null;
        String partnerRoleNameDefinition = null;
        QName partnerLinkTypeDefinition;

        final PartnerLink partnerLink = new PartnerLink();
        final String name = XmlUtil.attribute(partnerLinkElement, "name");
        if (partnerLinks.containsKey(name)) {
          // Static Analysis SA00018
          final String message = StaticAnalysisFault.getMessage("SA00018", "  PartnerLink " + name + " defined twice in scope "
              + scope != null ? scope.getName() : "");
          parse.addProblem(message, new StaticAnalysisException("SA00018"));
        } else {
          partnerLink.setName(name);

          final QName partnerLinkTypeName = XmlUtil.attributeQName(partnerLinkElement, "partnerLinkType");
          final PartnerLinkType partnerLinkType = wsdlInfos.getPartnerLinkType(partnerLinkTypeName);

          partnerLinkTypeDefinition = partnerLinkTypeName;
          boolean roleDef = false;
          if (partnerLinkType != null) {
            final String myRoleName = XmlUtil.attribute(partnerLinkElement, "myRole");
            if (myRoleName != null) {
              final QName myRolePortTypeQName = partnerLinkType.getPortTypeQName(myRoleName);
              partnerLink.setMyRolePortTypeQName(myRolePortTypeQName);
              roleDef = true;

              myRoleNameDefinition = myRoleName;
            }
            final String partnerRoleName = XmlUtil.attribute(partnerLinkElement, "partnerRole");
            if (partnerRoleName != null) {
              final QName partnerRolePortTypeQName = partnerLinkType.getPortTypeQName(partnerRoleName);
              partnerLink.setPartnerRolePortTypeQName(partnerRolePortTypeQName);
              roleDef = true;

              partnerRoleNameDefinition = partnerRoleName;
            }
            final String initializePartnerRole = XmlUtil.attribute(partnerLinkElement, "initializePartnerRole");
            if (initializePartnerRole != null) {
              if (partnerRoleName != null) {
                partnerLink.setInitializePartnerRole(initializePartnerRole.equals("true"));
              } else {
                // SA00017
                final String message = StaticAnalysisFault.getMessage("SA00017", "  PartnerLink " + name
                  + " has initializePartnerRole attribute used and does not have a partner role.");
                parse.addProblem(message, new StaticAnalysisException("SA00017"));
              }
            }
          }
          if (!roleDef) {
            // SA00016
            final String message = StaticAnalysisFault.getMessage("SA00016", "  PartnerLink " + name
              + " should specify myRole or partnerRole or both");
            parse.addProblem(message, new StaticAnalysisException("SA00016"));
          } else {
            partnerLinks.put(name, partnerLink);

            // Add variable definition to the scope definition list
            final PartnerLinkDefinitionImpl partnerLinkDefinition = new PartnerLinkDefinitionImpl(myRoleNameDefinition,
              partnerLink.getName(), partnerLinkTypeDefinition, partnerRoleNameDefinition, partnerLink.getInitializePartnerRole());

            partnerLinkDefinitionList.add(partnerLinkDefinition);
          }
        }
      }
    }
    if (scope != null) {
      scope.setPartnerLinks(partnerLinks);
      final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
      final ActivityFullDefinition activityFullDefinition = definitionKeeper.getActivityFullDefinition();
      ((ScopeActivityFullDefinition) activityFullDefinition).setPartnerLinkDefinition(partnerLinkDefinitionList);
    }
    return partnerLinks;
  }

  /**
   * Parses all messageExchanges of the given scope.
   *
   * @param messageExchangeElements
   *          list of all messageExchange elements of the given scope
   * @param scope
   *          scope
   */
  private Set<String> parseMessageExchanges(final List<Element> messageExchangeElements, final Scope scope, final Parse parse) {
    final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
    final ActivityFullDefinition activityFullDefinition = definitionKeeper.getActivityFullDefinition();

    final Set<String> messageExchanges = new HashSet<String>();
    if (messageExchangeElements != null) {
      for (final Element messageExchangeElement : messageExchangeElements) {
        final String name = XmlUtil.attribute(messageExchangeElement, "name");
        messageExchanges.add(name);
      }
    }
    if (scope != null) {
      scope.setMessageExchanges(messageExchanges);
      ((ScopeActivityFullDefinition) activityFullDefinition).setMessageExchange(messageExchanges);
    }
    return messageExchanges;
  }

  /**
   * Parses all variables of the given scope.
   *
   * @param variableElements
   *          list of all variable elements of the given scope
   * @param scope
   *          scope
   * @param parse
   *          parse
   */
  private void parseVariables(final List<Element> variableElements, final Scope scope, final Parse parse) {
    final Map<String, Variable> variables = new HashMap<String, Variable>();

    final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
    final ActivityFullDefinition activityFullDefinition = definitionKeeper.getActivityFullDefinition();
    final List<VariableDefinition> variableDefinitionList = new ArrayList<VariableDefinition>();

    if (variableElements != null) {
      for (final Element variableElement : variableElements) {
        final Variable variable = new Variable();
        final String name = XmlUtil.attribute(variableElement, "name");
        if (variables.containsKey(name)) {
          // Static Analysis SA00023
          final String message = StaticAnalysisFault.getMessage("SA00023", "  Variable " + name + " defined twice in scope "
            + scope.getName());
          parse.addProblem(message, new StaticAnalysisException("SA00023"));
        } else {
          variable.setName(name);
          variable.setMessageType(XmlUtil.attributeQName(variableElement, "messageType"));
          variable.setXmlType(XmlUtil.attributeQName(variableElement, "type"));
          variable.setElement(XmlUtil.attributeQName(variableElement, "element"));

          if (!(((variable.getElement() != null) && (variable.getMessageType() == null) && (variable.getXmlType() == null))
            || ((variable.getElement() == null) && (variable.getMessageType() != null) && (variable.getXmlType() == null)) || ((variable
            .getElement() == null)
            && (variable.getMessageType() == null) && (variable.getXmlType() != null)))) {
            // Static Analysis SA00025
            final String message = StaticAnalysisFault.getMessage("SA00025", "  Type of variable " + name
              + " not defined properly.");
            parse.addProblem(message, new StaticAnalysisException("SA00025"));
          } else {

            final Element fromElement = XmlUtil.element(variableElement, "from");
            if (fromElement != null) {
              final CopyFullDefinition copyFullDefinition = new CopyFullDefinitionImpl();
              final Copy virtualCopy = new Copy();
              virtualCopy.setKeepSrcElementName(true);
              AssignBinding.parseFrom(fromElement, virtualCopy, parse, scope, copyFullDefinition);
              variable.setVirtualCopy(virtualCopy);
              final To to = new To();
              to.setVariable(name);
              to.setVariant(Variant.VAR);
              virtualCopy.setTo(to);
              scope.addVariableToBeInitialized(name);
            }
            variables.put(name, variable);

            // Add variable definition to the scope definition list
            final VariableDefinitionImpl variableDefinition = new VariableDefinitionImpl(variable.getElement(), variable
              .getMessageType(), variable.getName(), variable.getXmlType());
            variableDefinitionList.add(variableDefinition);
          }
        }
      }
    }
    scope.setVariables(variables);
    ((ScopeActivityFullDefinition) activityFullDefinition).setVariableDefinition(variableDefinitionList);
  }

  /**
   * Parses all correlationSet of the given scope.
   *
   * @param correlationSetElements
   *          list of all correlationSet elements of the given scope
   * @param scope
   *          scope
   * @param wsdlInfos
   *          wsdlInfos
   */
  private Map<String, CorrelationSet> parseCorrelationSets(final List<Element> correlationSetElements,
      final Scope scope, final WsdlsInfos wsdlInfos, final Parse parse) {
    final Map<String, CorrelationSet> correlationSets = new HashMap<String, CorrelationSet>();

    final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
    final ActivityFullDefinition activityFullDefinition = definitionKeeper.getActivityFullDefinition();
    final List<CorrelationSetDefinition> correlationSetDefinitionList = new ArrayList<CorrelationSetDefinition>();

    if (correlationSetElements != null) {
      for (final Element correlationSetElement : correlationSetElements) {
        final CorrelationSet cs = new CorrelationSet();
        final String name = XmlUtil.attribute(correlationSetElement, "name");
        if (correlationSets.containsKey(name) && scope != null) {
          // Static Analysis SA00044
          final String message = StaticAnalysisFault.getMessage("SA00044", "  CorrelationSet " + name
            + " defined twice in scope " + scope.getName());
          parse.addProblem(message, new StaticAnalysisException("SA00044"));
        } else {
          cs.setName(name);
          final String propertiesSt = XmlUtil.attribute(correlationSetElement, "properties");
          final Map<QName, Property> properties = new HashMap<QName, Property>();
          final String[] propertyQNames = propertiesSt.split(" ");

          final List<QName> qNameList = new ArrayList<QName>();
          for (String prop : propertyQNames) {
            prop = prop.trim();
            if (!prop.equals("")) {
              final QName propertyQN = XmlUtil.getQNameFromString(correlationSetElement, prop);
              final Property property = wsdlInfos.getProperty(propertyQN);
              properties.put(propertyQN, property);
              qNameList.add(propertyQN);
            }
          }
          cs.setProperties(properties);
          correlationSets.put(name, cs);

          final CorrelationSetDefinitionImpl correlationSetDefinition = new CorrelationSetDefinitionImpl(name, qNameList);
          correlationSetDefinitionList.add(correlationSetDefinition);
        }
      }
    }
    if (scope != null) {
      scope.setCorrelationSets(correlationSets);
      ((ScopeActivityFullDefinition) activityFullDefinition).setCorrelationSetDefinition(correlationSetDefinitionList);
    }

    return correlationSets;
  }

  private int parseEventHandlers(final Element eventHandlersElement, final Scope scope, final Parse parse, final Parser parser,
    final EventHandlerActivity eventHandlerActivity) {

    final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
    final ScopeActivityFullDefinition activityFullDefinition =
      (ScopeActivityFullDefinition) definitionKeeper.getActivityFullDefinition();

    final List<OnEvent> onEvents = new ArrayList<OnEvent>();
    final List<OnEventFullDefinition> onEventFullDefinitions = new ArrayList<OnEventFullDefinition>();
    final List<Element> onEventElements = XmlUtil.elements(eventHandlersElement, "onEvent");

    if (onEventElements != null) {
      for (final Element onEventElement : onEventElements) {
        final OnEventFullDefinition onEventFullDefinition = new OnEventFullDefinitionImpl();

        final OnEvent onEvent = this.parseOnEvent(onEventElement, scope, parse, parser, onEventFullDefinition);
        onEvents.add(onEvent);

        onEventFullDefinitions.add(onEventFullDefinition);
      }
    }

    final List<Element> onAlarmElements = XmlUtil.elements(eventHandlersElement, "onAlarm");
    final List<OnAlarm> onAlarms = new ArrayList<OnAlarm>();
    final List<OnAlarmFullDefinition> onAlarmFullDefinitions = new ArrayList<OnAlarmFullDefinition>();
    if (onAlarmElements != null) {
      for (final Element onAlarmElement : onAlarmElements) {
        final OnAlarmFullDefinition onAlarmFullDefinition = new OnAlarmFullDefinitionImpl();

        final OnAlarm onAlarm = this.parseOnAlarm(onAlarmElement, scope, parse, parser, onAlarmFullDefinition);
        onAlarms.add(onAlarm);

        onAlarmFullDefinitions.add(onAlarmFullDefinition);
      }
    }

    activityFullDefinition.setOnEventEventHandlers(onEventFullDefinitions);
    activityFullDefinition.setOnAlarmEventHandlers(onAlarmFullDefinitions);

    eventHandlerActivity.setOnAlarms(onAlarms);
    eventHandlerActivity.setOnEvents(onEvents);

    return onAlarms.size() + onEvents.size();
  }

  private OnEvent parseOnEvent(final Element element, final Scope scope, final Parse parse, final Parser parser,
    final OnEventFullDefinition onEventFullDefinition) {
    if (ScopeBinding.log.isLoggable(Level.FINE)) {
      ScopeBinding.log.fine("parsing element = " + element);
    }
    // create onEvent element
    final OnEvent onEvent = new OnEvent();

    // Fetch enclosed scope ...
    final Element scopeElement = XmlUtil.element(element, XmlConstants.XMLNS_BPEL_2_0_EXECUTABLE, "scope");
    if (scopeElement == null) {
      parse.addProblem("No scope was specified under this onEvent element.");
      return null;
    }

    // Set partnerLink
    final String partnerLinkName = XmlUtil.attribute(element, "partnerLink");
    // get partnerLinkDef and get plnkType attribute and myRole attribute
    final BpelProcess bpelProcess = parse.findObject(BpelProcess.class);
    // partnerLinks
    // search partner link: first in associated scope, then in ancestors
    PartnerLink partnerLink = null;
    final Element partnerLinksElement = XmlUtil.element(scopeElement, "partnerLinks");
    if (partnerLinksElement != null) {
      final List<Element> partnerLinkElements = XmlUtil.elements(partnerLinksElement, "partnerLink");
      final Map<String, PartnerLink> associatedPartnerLinks =
        this.parsePartnerLinks(partnerLinkElements, bpelProcess.getWsdlInfos(), null, parse);
      if (associatedPartnerLinks != null) {
        partnerLink = associatedPartnerLinks.get(partnerLinkName);
      }
    }
    if (partnerLink == null) {
      partnerLink = this.findPartnerLinkDefinition(parse, partnerLinkName);
    }
    if (partnerLink == null) {
      // SA00084
      final String message = StaticAnalysisFault.getMessage("SA00084", "  onEvent " + onEvent
          + " defines an unknown partnerLink " + partnerLinkName + ".");
      parse.addProblem(message, new StaticAnalysisException("SA00084"));
      return null;
    }
    onEvent.setPartnerLinkName(partnerLinkName);
    onEventFullDefinition.setPartnerLink(partnerLinkName);

    // check portType
    final QName plportType = partnerLink.getMyRolePortTypeQName();
    if (plportType == null) {
      parse.addProblem("No myRole defined for partnerLink: " + partnerLinkName);
      return null;
    }
    final QName portType = XmlUtil.attributeQName(element, "portType");
    if (portType != null) {
      if (!plportType.equals(portType)) {
        // SA00005
        final String message = StaticAnalysisFault.getMessage("SA00005", "  onEvent " + onEvent
          + " has specified a PortType that is different from the one specified by the partnerLink");
        parse.addProblem(message, new StaticAnalysisException("SA00005"));
      }
    }
    onEventFullDefinition.setPortType(plportType);

    // Parse operation name
    // TODO: check operation exists
    final String operation = XmlUtil.attribute(element, "operation");
    onEventFullDefinition.setOperation(operation);
    onEvent.setOperationKey(new OperationKey(bpelProcess.getQName(), plportType, operation));

    final Operation wsdlOperation = bpelProcess.getWsdlInfos().getPortType(plportType).getOperation(operation, null, null);
    final Message inputMessage = wsdlOperation.getInput().getMessage();

    // Parse variable
    final Set<Variable> implicitVariables = new HashSet<Variable>();

    // Need to declare a variable that fits the needs ...
    final QName messageType = XmlUtil.attributeQName(element, "messageType");
    final QName elementAttribute = XmlUtil.attributeQName(element, "element");
    final String variable = XmlUtil.attribute(element, "variable");
    final Element fromPartsElement = XmlUtil.element(element, "fromParts");

    // check attributes
    if (variable != null) {
      // either both are set or none
      if (messageType != null && elementAttribute != null) {
        // SA00090
        final String message = StaticAnalysisFault.getMessage("SA00090", "  onEvent " + onEvent
          + " has specified both an element attribute and a messageType one.");
        parse.addProblem(message, new StaticAnalysisException("SA00090"));
      }
      if (messageType == null && elementAttribute == null) {
        // SA00090
        final String message = StaticAnalysisFault.getMessage("SA00090", "  onEvent " + onEvent
          + " has specified neither an element attribute nor a messageType one.");
        parse.addProblem(message, new StaticAnalysisException("SA00090"));
      }
      // Add input variable to onEvent enclosed scope
      // This is done by giving additional variables as input parameter during scope parsing.
      // declare the input variable
      final Variable variableInput = new Variable();
      variableInput.setName(variable);

      variableInput.setMessageType(messageType);
      variableInput.setElement(elementAttribute);

      implicitVariables.add(variableInput);

      if (elementAttribute != null) {
        // for element: create a virtual assign that copies the only message part to the element.
        // SA00087 check element type
        final Collection<Part> messageParts = inputMessage.getParts().values();
        if (messageParts.size() != 1) {
          final String message = StaticAnalysisFault.getMessage("SA00087",
              " onEvent " + onEvent
              + " has specified a variable element type but the incoming message has not a single part.");
          parse.addProblem(message, new StaticAnalysisException("SA00087"));
        } else {
          // get unique part
          final Part part = messageParts.iterator().next();
          // SA 00087
          if (part.getElementName() == null) {
            final String message = StaticAnalysisFault.getMessage("SA00087",
                " onEvent " + onEvent + " has specified a variable element type "
                + "but the message single part is not defined with an element type.");
            parse.addProblem(message, new StaticAnalysisException("SA00087"));
          } else if (!part.getElementName().equals(elementAttribute)) {
            final String message = StaticAnalysisFault.getMessage("SA00087",
                " onEvent " + onEvent + " has specified a variable element type " + elementAttribute + "."
                + " According to the wsdl definition, it should be " + part.getElementName());
            parse.addProblem(message, new StaticAnalysisException("SA00087"));
          } else {
            // ok, copy to element variable
            // create anonymous variable to store the incoming message
            onEvent.setAnonymousInboundMessageVariable(ReceiveBinding.getAnonymousVariable(onEvent, parse));
            // copy part to element variables
            final From from = new From();
            from.setVariable(Variable.INBOUND_ANONYMOUS_WSDL_MESSAGE_NAME);
            from.setPart(part.getName());
            from.setVariant(Variant.VAR);
            final To to = new To();
            to.setVariable(variable);
            to.setVariant(Variant.VAR);
            final Copy copy = new Copy();
            copy.setFrom(from);
            copy.setTo(to);
            onEvent.setInCopies(Collections.singletonList(copy));
          }
        }
      } else {
        // SA00087 check message type
        if (!inputMessage.getQName().equals(messageType)) {
          final String message = StaticAnalysisFault.getMessage("SA00087",
              " onEvent " + onEvent + " has specified a variable message type " + messageType + "."
              + " According to the wsdl definition, it should be " + inputMessage.getQName());
          parse.addProblem(message, new StaticAnalysisException("SA00087"));
        }
        onEvent.setVariable(variable);
      }
      onEventFullDefinition.setVariable(variable);

      if (fromPartsElement != null) {
        // SA00085
        final String message = StaticAnalysisFault.getMessage("SA00085",
            " onEvent " + onEvent + " has specified both a Variable and a fromPart");
        parse.addProblem(message, new StaticAnalysisException("SA00085"));
      }
    } else {
      if (messageType != null || elementAttribute != null) {
        // SA00085
        final String message = StaticAnalysisFault.getMessage("SA00085", "  onEvent " + onEvent
          + " has specified either an element attribute or a messageType one, but no variable attribute is set.");
        parse.addProblem(message, new StaticAnalysisException("SA00085"));
      }
      // fromParts
      if (fromPartsElement != null) {
        final List<Element> fromPartElements = XmlUtil.elements(fromPartsElement, "fromPart");
        // create anonymous variable to store the incoming message
        onEvent.setAnonymousInboundMessageVariable(ReceiveBinding.getAnonymousVariable(onEvent, parse));
        // copy parts to bpel variables
        final List<Copy> inCopies =
          ReceiveBinding.getInCopies(fromPartElements, scope, parse, onEventFullDefinition);
        if (!inCopies.isEmpty()) {
          onEvent.setInCopies(inCopies);
        }
        // create implicit scope variables
        for (final Element fromPartElement : fromPartElements) {
          final String part = XmlUtil.attribute(fromPartElement, "part");
          final String toVariable = XmlUtil.attribute(fromPartElement, "toVariable");

          final Part wsdlPart = inputMessage.getPart(part);
          if (wsdlPart == null) {
            parse.addProblem("No part " + part + " defined for message: " + inputMessage.getQName());
            return null;
          }
          // Add input variable to onEvent enclosed scope
          // This is done by giving additional variables as input parameter during scope parsing.
          // declare the input variable
          final Variable variableInput = new Variable();
          variableInput.setName(toVariable);
          variableInput.setXmlType(wsdlPart.getTypeName());
          variableInput.setElement(wsdlPart.getElementName());
          implicitVariables.add(variableInput);
        }
      }
    }

    // Parse enclosed scope
    // Get binding of enclosed scope
    final EventHandlerScopeBinding binding = new EventHandlerScopeBinding();

    // and parse the scope with method which allow to define some additional variables

    NodeImpl associatedScopeNode = null;
    final DefinitionKeeper defKeeper = parse.findObject(DefinitionKeeper.class);
    defKeeper.pushParent(onEventFullDefinition);
    try {
      associatedScopeNode = (NodeImpl) binding.parse(scopeElement, implicitVariables, parse, parser);
    } catch (final DuplicateVariableException e) {
      // SA00086
      final String message = StaticAnalysisFault.getMessage("SA00086",
        "enclosed scope declare a variable with input variable name of onEvent activity");
      parse.addProblem(message, new StaticAnalysisException("SA00086"));
      return null;
    } finally {
      defKeeper.popParent();
    }

    onEvent.setActivity(associatedScopeNode);

    final Scope associatedScope = (Scope) associatedScopeNode.getBehaviour();

    // Set message exchange
    final String messageExchange = XmlUtil.attribute(element, "messageExchange");
    if (messageExchange != null) {
      onEvent.setMessageExchange(messageExchange);
      onEventFullDefinition.setMessageExchange(messageExchange);

      // SA00089
      // search message exchange: first in associated scope, then in ancestors
      boolean found = false;
      if (!found) {
        Scope enclosingScope = associatedScope;
        while (!found && enclosingScope != null) {
          found = enclosingScope.getMessageExchanges() != null
            && enclosingScope.getMessageExchanges().contains(messageExchange);
          enclosingScope = enclosingScope.getEnclosingScope();
        }
      }
      if (!found) {
        // SA00089
        final String message = StaticAnalysisFault.getMessage("SA00089", "  onEvent " + onEvent
            + " defines an unknown messageExchange " + messageExchange + ".");
        parse.addProblem(message, new StaticAnalysisException("SA00089"));
      }
    }

    // Parse correlations
    // SA00088: first in associated scope, then in ancestors.
    final Element correlationsElement = XmlUtil.element(element, "correlations");
    if (correlationsElement != null) {
      final List<Element> correlationElements = XmlUtil.elements(correlationsElement, "correlation");
      try {
        onEvent.setCorrelations(
            ReceiveBinding.parseCorrelations(correlationElements, associatedScope, parse, onEventFullDefinition));
      } catch (final CorrelationSetNotFoundException e) {
        // SA00088
        final String message = StaticAnalysisFault.getMessage("SA00088", "  onEvent " + onEvent
            + " refers to an unknown correlation set.");
        parse.addProblem(message, new StaticAnalysisException("SA00088"));
      }
    }

    return onEvent;
  }

  private OnAlarm parseOnAlarm(final Element element, final Scope scope, final Parse parse,
      final Parser parser, final OnAlarmFullDefinition onAlarmFullDefinition) {
    if (ScopeBinding.log.isLoggable(Level.FINE)) {
      ScopeBinding.log.fine("parsing element = " + element);
    }
    final OnAlarm onAlarm = new OnAlarm();

    // Fetch enclosed scope ...
    final Element scopeElement = XmlUtil.element(element, XmlConstants.XMLNS_BPEL_2_0_EXECUTABLE, "scope");
    if (scopeElement == null) {
      parse.addProblem("No scope was specified under this onEvent element.");
      return null;
    }
    final DefinitionKeeper defKeeper = parse.findObject(DefinitionKeeper.class);
    defKeeper.pushParent(onAlarmFullDefinition);
    // Parse enclosed scope
    // Get binding of enclosed scope
    final EventHandlerScopeBinding binding = new EventHandlerScopeBinding();
    final NodeImpl associatedScopeNode = (NodeImpl) binding.parse(scopeElement, parse, parser);
    onAlarm.setActivity(associatedScopeNode);

    defKeeper.popParent();

    final BpelProcess bpelProcess = parse.findObject(BpelProcess.class);
    final Element deadlineElement = XmlUtil.element(element, "until");
    final Element durationElement = XmlUtil.element(element, "for");
    final Element repeatEveryElement = XmlUtil.element(element, "repeatEvery");
    if (durationElement != null && deadlineElement != null) {
      parse.addProblem("Error in element: " + element + " only one element 'for' or 'until' should be used.");
    }
    if (durationElement != null) {
      final Expression expression = PickBinding.parseExpression(durationElement, bpelProcess);
      onAlarm.setDurationExpression(expression);

      final ForDefinition forDefinition = new ForDefinitionImpl(expression.getLanguage(), expression.getText());

      onAlarmFullDefinition.setForDefinition(forDefinition);
    } else if (deadlineElement != null) {
      final Expression expression = PickBinding.parseExpression(deadlineElement, bpelProcess);
      onAlarm.setDeadlineExpression(expression);

      final UntilDefinition untilDefinition = new UntilDefinitionImpl(expression.getLanguage(), expression.getText());
      onAlarmFullDefinition.setUntilDefinition(untilDefinition);
    } else if (repeatEveryElement == null) {
      parse.addProblem("Error in element: " + element + " one element 'for' or 'until' should be used.");
    }

    if (repeatEveryElement != null) {
      final Expression expression = PickBinding.parseExpression(repeatEveryElement, bpelProcess);
      onAlarm.setRepeatEveryExpression(expression);

      final RepeatEveryDefinition repeatEveryDefinition =
        new RepeatEveryDefinitionImpl(expression.getLanguage(), expression.getText());

      onAlarmFullDefinition.setRepeatEveryDefinition(repeatEveryDefinition);
    }
    return onAlarm;
  }

  @Override
  public BpelActivityFullDefinition createActivityFullDefinition(final ProcessDefinitionUUID processDefinitionUUID,
    final ActivityDefinitionUUID activityDefinitionUUID, final String name, final Parse parse) {
    return new ScopeActivityFullDefinitionImpl(processDefinitionUUID, activityDefinitionUUID, name);
  }

}
