/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.persistence.db.hibernate;

import java.io.ByteArrayOutputStream;
import java.io.Reader;
import java.io.Serializable;
import java.io.StringReader;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;

import org.hibernate.usertype.UserType;
import org.ow2.orchestra.util.XmlConstants;
import org.ow2.orchestra.util.XmlUtil;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * Maps a DOM {@linkplain Node node} to a {@linkplain Types#CLOB} column.
 * @author Juan Cantu
 * @author Guillaume Porcher
 */
public class NodeType implements UserType {

  static final int[] SQL_TYPES = { Types.CLOB };

  public boolean equals(final Object x, final Object y) {
    if (x == null) {
      return y == null;
    }
    return x.equals(y);
  }

  public int hashCode(final Object x) {
    return x.hashCode();
  }

  public Object deepCopy(final Object value) {
    if (value != null) {
      return ((Node) value).cloneNode(true);
    }
    return null;
  }

  public boolean isMutable() {
    return true;
  }

  public Serializable disassemble(final Object value) {
    return (Serializable) this.deepCopy(value);
  }

  public Object assemble(final Serializable cached, final Object owner) {
    return this.deepCopy(cached);
  }

  public Object replace(final Object original, final Object target, final Object owner) {
    return this.deepCopy(original);
  }

  public int[] sqlTypes() {
    return NodeType.SQL_TYPES;
  }

  public Object nullSafeGet(final ResultSet rs, final String[] names, final Object owner) throws SQLException {
    // retrieve stream from database
    final String columnName = names[0];
    Reader reader = null;

    try {
      reader = rs.getCharacterStream(columnName);
      // if SQL value is NULL, element is null as well
      if (reader == null) {
        return null;
      }

      // parse XML text
      final Node node = XmlUtil.getDocumentFromReader(reader).getDocumentElement();


      return node.getFirstChild();
    } finally {
      if (reader != null) {
        try {
          reader.close();
        } catch (final Exception e) {
          // nothing
        }
      }
    }
  }

  public void nullSafeSet(final PreparedStatement st, final Object value, final int index) throws SQLException {
    // easy way out: null value
    if (value == null) {
      st.setNull(index, NodeType.SQL_TYPES[0]);
    } else {
      final Node node = (Node) value;
      final Element el = node.getOwnerDocument().createElementNS(
          XmlConstants.ELT_DEFAULT_NS,
          XmlConstants.ELT_DEFAULT_PREFIX + ":" + XmlConstants.ELT_DEFAULT_ELEMENT);
      el.appendChild(node);
      final ByteArrayOutputStream byteStream = XmlUtil.getNodeAsBAOS(el);
      final Reader sr = new StringReader(byteStream.toString());
      st.setCharacterStream(index, sr, byteStream.size());
    }
  }

  public Class returnedClass() {
    return Node.class;
  }
}


