/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.persistence.db.hibernate;

import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.sql.Blob;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;

import org.hibernate.Hibernate;
import org.hibernate.engine.Mapping;
import org.hibernate.engine.SessionImplementor;
import org.hibernate.type.BlobType;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;

/**
 * Maps a {@link Serializable} object to a {@linkplain Types#BLOB} column.
 * @author Guillaume Porcher
 */
public class SerializableToBlobType extends BlobType {

  static final int[] SQL_TYPES = { Types.BLOB };

  /* (non-Javadoc)
   * @see org.hibernate.type.BlobType#sqlTypes(org.hibernate.engine.Mapping)
   */
  @Override
  public int[] sqlTypes(final Mapping mapping) {
    // TODO Auto-generated method stub
    return SerializableToBlobType.SQL_TYPES;
  }

  /* (non-Javadoc)
   * @see org.hibernate.type.BlobType#get(java.sql.ResultSet, java.lang.String)
   */
  @Override
  public Object get(final ResultSet rs, final String name) throws SQLException {
    // retrieve stream from database
    InputStream inStream = null;
    ObjectInputStream ois = null;
    try {
      final Blob blob = rs.getBlob(name);
      if (blob == null) {
        return null;
      }
      inStream = blob.getBinaryStream();
      // if SQL value is NULL, element is null as well
      if (inStream == null) {
        return null;
      }

      ois = new ObjectInputStream(inStream);

      final Object wsdlInfos = ois.readObject();

      ois.close();

      return wsdlInfos;
    } catch (final Exception e) {
      throw new OrchestraRuntimeException(e);
    } finally {
      if (ois != null) {
        try {
          ois.close();
        } catch (final Exception e) {
          // nothing
        }
      }
      if (inStream != null) {
        try {
          inStream.close();
        } catch (final Exception e) {
          // nothing
        }
      }
    }
  }

  @Override
  public void set(final PreparedStatement st,
      final Object value, final int index, final SessionImplementor session) throws SQLException {
    // easy way out: null value

    if (value == null) {
      st.setNull(index, SerializableToBlobType.SQL_TYPES[0]);
    } else {
      final ByteArrayOutputStream byteStream = new ByteArrayOutputStream();
      try {
        final ObjectOutputStream sr = new ObjectOutputStream(byteStream);
        try {
          sr.writeObject(value);
        } finally {
          sr.close();
        }


        if (session.getFactory().getDialect().useInputStreamToInsertBlob()) {
          st.setBytes(index, byteStream.toByteArray());
        } else {
          final Blob blob = Hibernate.createBlob(byteStream.toByteArray());
          st.setBlob(index, blob);
        }

      } catch (final Exception e) {
        throw new OrchestraRuntimeException(e);
      }
    }
  }

}


