/**
 * Copyright (C) 2007-2009  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.persistence.log;

import java.util.logging.Level;
import java.util.logging.Logger;

import javax.transaction.Synchronization;

import org.ow2.orchestra.facade.data.runtime.ActivityInstanceData;
import org.ow2.orchestra.facade.data.runtime.CorrelationSetInitializedData;
import org.ow2.orchestra.facade.data.runtime.ForeachInitializedData;
import org.ow2.orchestra.facade.data.runtime.PartnerLinkUpdateData;
import org.ow2.orchestra.facade.data.runtime.ProcessInstanceData;
import org.ow2.orchestra.facade.data.runtime.ScopeStateUpdateData;
import org.ow2.orchestra.facade.data.runtime.TimerData;
import org.ow2.orchestra.facade.data.runtime.TransitionConditionUpdateData;
import org.ow2.orchestra.facade.data.runtime.VariableUpdateData;
import org.ow2.orchestra.facade.def.full.ProcessFullDefinition;
import org.ow2.orchestra.facade.exception.BpelFaultException;
import org.ow2.orchestra.facade.runtime.full.ProcessFullInstance;
import org.ow2.orchestra.facade.uuid.ActivityInstanceUUID;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.pvm.env.Environment;
import org.ow2.orchestra.pvm.env.Transaction;
import org.ow2.orchestra.services.Archivable;
import org.ow2.orchestra.services.Recorder;
import org.ow2.orchestra.util.XmlUtil;
import org.w3c.dom.Node;

/**
 * @author Gael Lalire, S. Ali Tokmen, Guillaume Renault
 */
public class LoggerRecorder implements Recorder {

  private static final Logger LOG = Logger.getLogger(LoggerRecorder.class.getName());

  private final Level level;

  public LoggerRecorder() {
    this(Level.FINE);
  }

  public LoggerRecorder(final String levelName) {
    this(Level.parse(levelName));
  }

  public LoggerRecorder(final Level level) {
    this.level = level;
    final Transaction tx = Environment.getFromCurrent(Transaction.class);
    tx.registerSynchronization(new Synchronization() {

      public void beforeCompletion() {
      }

      public void afterCompletion(final int status) {
        if (status == javax.transaction.Status.STATUS_COMMITTED) {
          LoggerRecorder.LOG.log(level, "Transaction committed");
        } else {
          LoggerRecorder.LOG.log(level, "Transaction rolledback");
        }
      }
    });
  }

  public void recordInstanceStarted(final ProcessInstanceData processInstanceData) {
    if (LoggerRecorder.LOG.isLoggable(this.level)) {
      LoggerRecorder.LOG.log(this.level, "Start [uuid:" + processInstanceData.getUUID() + "] an instance of process named '{"
        + processInstanceData.getProcessDefinitionData().getNameSpace() + "}"
        + processInstanceData.getProcessDefinitionData().getName() + "'");
    }
  }

  public void recordInstanceEnded(final ProcessInstanceData processInstanceData) {
    if (LoggerRecorder.LOG.isLoggable(this.level)) {
      LoggerRecorder.LOG.log(this.level, "Instance [uuid:" + processInstanceData.getUUID() + "] ended");
    }
  }

  /**
   * Record an instance state update.
   */
  public void recordInstanceUpdated(final ProcessInstanceData processInstanceData) {
    if (LoggerRecorder.LOG.isLoggable(this.level)) {
      LoggerRecorder.LOG.log(this.level, "Instance [uuid:" + processInstanceData.getUUID() + "] " + processInstanceData.getState().toString().toLowerCase());
    }
  }

  public void recordActivityStarted(final ActivityInstanceData activityInstanceData) {
    if (LoggerRecorder.LOG.isLoggable(this.level)) {
      LoggerRecorder.LOG.log(this.level, "Start [uuid:" + activityInstanceData.getUUID() + "] an activity of an instance of process named {"
        + activityInstanceData.getActivityDefinitionData().getProcessDefinitionData().getNameSpace() + "}"
        + activityInstanceData.getActivityDefinitionData().getProcessDefinitionData().getName() + ", this "
        + activityInstanceData.getActivityDefinitionData().getType() + " activity name is '"
        + activityInstanceData.getActivityDefinitionData().getName() + "'");
    }
  }

  public void recordEventStarted(final ActivityInstanceData activityInstanceData) {
    if (LoggerRecorder.LOG.isLoggable(this.level)) {
      LoggerRecorder.LOG.log(this.level, "Start [uuid:" + activityInstanceData.getUUID() + "] an event of an instance of process named {"
        + activityInstanceData.getActivityDefinitionData().getProcessDefinitionData().getNameSpace() + "}"
        + activityInstanceData.getActivityDefinitionData().getProcessDefinitionData().getName() + ", this "
        + activityInstanceData.getActivityDefinitionData().getType() + " activity name is '"
        + activityInstanceData.getActivityDefinitionData().getName() + "'");
    }
  }

  public void recordExceptionOccured(final ActivityInstanceUUID activityInstanceUUID, final BpelFaultException exception) {
    if (LoggerRecorder.LOG.isLoggable(this.level)) {
      LoggerRecorder.LOG.log(this.level, "Activity [uuid:" + activityInstanceUUID + "] has thrown an exception: " + exception);
    }
  }

  public void recordActivityEnded(final ActivityInstanceData activityInstanceData) {
    if (LoggerRecorder.LOG.isLoggable(this.level)) {
      LoggerRecorder.LOG.log(this.level, "Activity [uuid:" + activityInstanceData.getUUID() + "] ended");
    }
  }

  public void recordActivityExited(final ActivityInstanceData activityInstanceData) {
    if (LoggerRecorder.LOG.isLoggable(this.level)) {
      LoggerRecorder.LOG.log(this.level, "Activity [uuid:" + activityInstanceData.getUUID() + "] exited");
    }
  }

  public void recordActivityTerminated(final ActivityInstanceData activityInstanceData) {
    if (LoggerRecorder.LOG.isLoggable(this.level)) {
      LoggerRecorder.LOG.log(this.level, "Activity [uuid:" + activityInstanceData.getUUID() + "] terminated");
    }
  }

  public void remove(final Archivable archivable) {
    if (archivable instanceof ProcessFullInstance) {
      this.remove((ProcessFullInstance) archivable);
    } else if (archivable instanceof ProcessFullDefinition) {
      this.remove((ProcessFullDefinition) archivable);
    }
  }

  public void remove(final ProcessFullInstance processFullInstance) {
    if (LoggerRecorder.LOG.isLoggable(this.level)) {
      LoggerRecorder.LOG.log(this.level, "ProcessInstance " + processFullInstance.getUUID() + " removed.");
    }
  }

  public void remove(final ProcessFullDefinition processFullDefinition) {
    if (LoggerRecorder.LOG.isLoggable(this.level)) {
      LoggerRecorder.LOG.log(this.level, "Process " + processFullDefinition.getUUID() + " removed.");
    }
  }

  public void recordActivityPending(final ActivityInstanceData activityInstanceData) {

    if (LoggerRecorder.LOG.isLoggable(this.level)) {
      LoggerRecorder.LOG.log(this.level, "Activity [uuid:" + activityInstanceData.getUUID() + "] pending");
    }
  }

  public void recordTimerStarted(final TimerData timerData) {
    if (LoggerRecorder.LOG.isLoggable(this.level)) {
      LoggerRecorder.LOG.log(this.level, "Timer [expectedStartedDate: " + timerData.getExpectedStartedDate() + "] started for activity "
        + timerData.getActivityInstanceUUID());
    }
  }

  public void recordForeachInitialized(final ForeachInitializedData foreachInitializedData) {
    if (LoggerRecorder.LOG.isLoggable(this.level)) {
      LoggerRecorder.LOG.log(this.level, "Foreach [startValue: " + foreachInitializedData.getStartValue() + "]" + "[stopValue: "
        + foreachInitializedData.getStopValue() + "]");
    }
  }

  public void recordVariableUpdated(final VariableUpdateData variableUpdateData) {
    if (LoggerRecorder.LOG.isLoggable(this.level)) {
      String variableValueAsString = null;
      final Object variableValue = variableUpdateData.getValue();
      if (variableValue instanceof Node) {
        variableValueAsString = XmlUtil.toString((Node) variableValue);
      } else {
        variableValueAsString = variableValue.toString();
      }
      LoggerRecorder.LOG.log(this.level, "Variable [name: " + variableUpdateData.getName() + ", scopeUUID: "
        + variableUpdateData.getScopeActivityUUID() + "] is set to [" + variableValueAsString + "] by activity "
        + variableUpdateData.getActivityUUID());
    }
  }

  public void recordCorrelationSetInitialized(final CorrelationSetInitializedData correlationSetInitializedData) {
    if (LoggerRecorder.LOG.isLoggable(this.level)) {
      LoggerRecorder.LOG.log(this.level, "Correlation set [name: " + correlationSetInitializedData.getName() + ", scopeUUID: "
        + correlationSetInitializedData.getScopeActivityUUID() + "] is initialized by activity "
        + correlationSetInitializedData.getActivityUUID());
    }

  }

  public void recordPartnerLinkUpdated(final PartnerLinkUpdateData partnerLinkUpdateData) {
    if (LoggerRecorder.LOG.isLoggable(this.level)) {
      LoggerRecorder.LOG
      .log(this.level, "Partner Link [name: " + partnerLinkUpdateData.getName() + ", scopeUUID: "
        + partnerLinkUpdateData.getScopeActivityUUID() + "] is updated by activity "
        + partnerLinkUpdateData.getActivityUUID());
    }
  }

  public void recordTransitionConditionUpdated(final TransitionConditionUpdateData transitionConditionUpdateData) {
    if (LoggerRecorder.LOG.isLoggable(this.level)) {
      LoggerRecorder.LOG.log(this.level, "Transition Condition [name: " + transitionConditionUpdateData.getName() + ", status: "
        + transitionConditionUpdateData.getStatus() + ", flow activity UUID "
        + transitionConditionUpdateData.getFlowActivityUUID() + "]");
    }
  }

  public void recordScopeStateChanged(final ScopeStateUpdateData scopeStateUpdateData) {
    if (LoggerRecorder.LOG.isLoggable(this.level)) {
      LoggerRecorder.LOG.log(this.level, "Scope State Update [state: " + scopeStateUpdateData.getState() + ", scope activity UUID "
        + scopeStateUpdateData.getScopeActivityUUID() + "]");
    }
  }

  public void recordProcessDeployed(final ProcessFullDefinition processFullDefinition) {
    if (LoggerRecorder.LOG.isLoggable(this.level)) {
      LoggerRecorder.LOG.log(this.level, "Deploy process [name: " + processFullDefinition.getName() + ", processUUID "
        + processFullDefinition.getUUID() + "]");
    }
  }

  public void recordProcessUndeployed(final ProcessDefinitionUUID processDefinitionUUID) {
    if (LoggerRecorder.LOG.isLoggable(this.level)) {
      LoggerRecorder.LOG.log(this.level, "Undeploy process [processUUID " + processDefinitionUUID + "]");
    }

  }
}
