/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.runtime;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

import org.ow2.orchestra.definition.element.Correlation;
import org.ow2.orchestra.definition.element.CorrelationSet;
import org.ow2.orchestra.definition.element.Correlation.Initiate;
import org.ow2.orchestra.facade.data.runtime.CorrelationSetInitializedData;
import org.ow2.orchestra.facade.exception.BpelFaultException;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.facade.exception.BpelFaultException.BpelFault;
import org.ow2.orchestra.facade.uuid.ActivityInstanceUUID;
import org.ow2.orchestra.util.BpelUtil;
import org.ow2.orchestra.util.EnvTool;
import org.ow2.orchestra.util.XmlUtil;
import org.ow2.orchestra.var.MessageVariable;
import org.ow2.orchestra.wsdl.PropertyAlias;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * @author Charles Souillard, Guillaume Porcher
 */
public class CorrelationSetRuntime {

  protected static final Logger LOG = Logger.getLogger(CorrelationSetRuntime.class.getName());

  /** cs name. */
  protected String name;

  /** cs definition (the same for all instances of the process). */
  protected CorrelationSet correlationSetDefinition;

  /** map of runtime values of properties used by this cs. */
  protected Map<QName, Node> propertyValues;

  /** is cs initialized */
  protected boolean initialized = false;

  protected ActivityInstanceUUID scopeActivityInstanceUUID;

  protected CorrelationSetRuntime() {
  }

  /**
   * Default constructor.
   *
   * @param correlationSetDefinition correlationSetDefinition
   */
  public CorrelationSetRuntime(final CorrelationSet correlationSetDefinition,
    final ActivityInstanceUUID scopeActivityInstanceUUID) {
    this.scopeActivityInstanceUUID = scopeActivityInstanceUUID;
    this.name = correlationSetDefinition.getName();
    this.correlationSetDefinition = correlationSetDefinition;
    this.propertyValues = new HashMap<QName, Node>();
  }

  /**
   * Return the runtime value of the property identified by the given
   * propertyName.
   *
   * @param propertyName
   *            propertyName
   * @return the runtime value of the property identified by the given
   *         propertyName.
   */
  public Node getPropertyValue(final QName propertyName) {
    return this.propertyValues.get(propertyName);
  }

  /**
   * Set the runtime value of the property identified by the given propertyName.
   *
   * @param propertyName
   *            propertyName
   * @param value
   *            value to set
   */
  public void setPropertyValue(final QName propertyName, final Node value) {
    this.propertyValues.put(propertyName, value.cloneNode(true));
  }

  protected boolean validatePropertyValue(final PropertyAlias propertyAlias, final MessageVariable message,
    final BpelExecution execution) {
    try {
      final Node expectedValue = this.getPropertyValue(propertyAlias.getPropertyName());

      final Element partElement = message.getPartValue(propertyAlias.getPart());
      if (partElement == null) {
        return expectedValue == null;
      }

      Node value = partElement;
      if (propertyAlias.getQuery() != null) {
        value = propertyAlias.getQuery().getEvaluator().evaluate(partElement, execution);
      }

      if (value == null) {
        return expectedValue == null;
      } else if (expectedValue == null) {
        return false;
      }
      final boolean valid = BpelUtil.areNodeEquals(expectedValue, value);
      if (CorrelationSetRuntime.LOG.isLoggable(Level.FINEST)) {
        if (!valid) {
          CorrelationSetRuntime.LOG.finest("Trying to validate CS value: expected " + XmlUtil.toString(expectedValue) + " got "
            + XmlUtil.toString(value));
        } else {
          CorrelationSetRuntime.LOG.finest("CS value valid: " + XmlUtil.toString(expectedValue));
        }
      }
      return valid;
    } catch (final Exception ex) {
      throw new OrchestraRuntimeException("In Correlation: Function call exception:" + ex, ex);
    }
  }

  public boolean validateCSValues(final BpelExecution execution, final MessageVariable message,
      final QName messageType) {
    final List<PropertyAlias> propertyAliases = this.correlationSetDefinition.getPropertyAliasRefs(execution, messageType);

    //foreach propertyAlias, we extract the value from the message.
    for (final PropertyAlias propertyAlias : propertyAliases) {
      if (!this.validatePropertyValue(propertyAlias, message, execution)) {
        return false;
      }
    }
    return true;
  }

  public void checkCSValues(final BpelExecution execution, final MessageVariable message,
      final QName messageType, final Correlation.Initiate initiate) {
    // initialize CS
    if (Initiate.yes.equals(initiate) || !this.isInitialized()) {
      this.initiate(execution, message, messageType, initiate);
    } else {
      // check CS values
      if (!this.validateCSValues(execution, message, messageType)) {
        throw BpelFaultException.generateBpelFaultException(
            BpelFault.correlationViolation,
            "Correlation consistency constraint violated",
            null);
      }
    }
  }

  public void initiate(final BpelExecution execution, final MessageVariable message,
      final QName messageType, final Correlation.Initiate initiate) {

    if (this.isInitialized()) {
      if (Initiate.yes.equals(initiate)) {
        throw BpelFaultException.generateBpelFaultException(
            BpelFault.correlationViolation,
            "CorrelationSet already initialized and initiate='yes'",
            null);
      }
      return;
    }

    if (Initiate.no.equals(initiate)) {
      throw BpelFaultException.generateBpelFaultException(
          BpelFault.correlationViolation,
          "CorrelationSet not initialized and initiate='no'",
          null);
    }

    final List<PropertyAlias> propertyAliases = execution.getPropertyAliases(this.correlationSetDefinition, messageType);
    //TODO : see if we can have an Element type message

    //foreach propertyAlias, we extract the value from the message.
    for (final PropertyAlias propertyAlias : propertyAliases) {
      try {
        final Element partElement = message.getPartValue(propertyAlias.getPart());

        Node value = partElement;
        if (propertyAlias.getQuery() != null) {
          value = propertyAlias.getQuery().getEvaluator().evaluate(partElement, execution);
        }
        this.setPropertyValue(propertyAlias.getPropertyName(), value);
      } catch (final Exception ex) {
        throw new OrchestraRuntimeException("In Correlation: Function call exception:" + ex, ex);
      }
    }
    if (CorrelationSetRuntime.LOG.isLoggable(Level.FINEST)) {
      CorrelationSetRuntime.LOG.finest("Initialized correlation set with values: " + this.propertyValues);
    }
    this.setInitialized(true, execution.getActivityInstanceUUID());
  }

  @Override
  public boolean equals(final Object obj) {
    if (obj == null) {
      return false;
    }
    if (!(obj instanceof CorrelationSetRuntime)) {
      return false;
    }
    final CorrelationSetRuntime corr = (CorrelationSetRuntime) obj;
    return this.name.equals(corr.getName())
      && this.propertyValues.equals(corr.getPropertyValues());
  }

  @Override
  public int hashCode() {
    final StringBuffer result = new StringBuffer();
    result.append("name=\"" + this.name + "\" properties=\"");
    for (final QName property : this.propertyValues.keySet()) {
      result.append(property);
    }
    result.append("\"");

    return result.toString().hashCode();
  }

  public Map<QName, Node> getPropertyValues() {
    return this.propertyValues;
  }

  public void setPropertyValues(final Map<QName, Node> propertyValues) {
    this.propertyValues = propertyValues;
  }

  public String getName() {
    return this.name;
  }

  public void setName(final String name) {
    this.name = name;
  }

  public CorrelationSet getCorrelationSetDefinition() {
    return this.correlationSetDefinition;
  }

  public void setCorrelationSetDefinition(
      final CorrelationSet correlationSetDefinition) {
    this.correlationSetDefinition = correlationSetDefinition;
  }
  public boolean isInitialized() {
    return this.initialized;
  }
  public void setInitialized(final boolean initialized, final ActivityInstanceUUID activityInstanceUUID) {
    this.initialized = initialized;
    EnvTool.getRecorder().recordCorrelationSetInitialized(
      new CorrelationSetInitializedData(this.scopeActivityInstanceUUID, activityInstanceUUID, this.name, this.propertyValues));
  }
}
