/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.runtime;

import java.util.Hashtable;
import java.util.Map;

import org.ow2.orchestra.definition.element.Link;
import org.ow2.orchestra.facade.uuid.ActivityInstanceUUID;

/**
 * @author Charles Souillard
 */
public class FlowRuntime {

  // for persistence
  protected long dbid;

  /** all links of this flow with their runtime status. */
  protected Map<String, LinkRuntime> linkRuntimes;

  protected FlowRuntime() {
  }

  // TODO : add synchronisation on all these maps and linkStatus...

  /**
   * Default constructor.
   *
   * @param links
   *            links of this flow
   * @param flowActivityInstanceUUID
   * @param childrenNumber
   *            number of children activities of tjis flow
   */
  public FlowRuntime(final Map<String, Link> links, final ActivityInstanceUUID flowActivityInstanceUUID) {
    this.linkRuntimes = new Hashtable<String, LinkRuntime>();
    if (links != null) {
      for (final String linkName : links.keySet()) {
        this.linkRuntimes.put(linkName, new LinkRuntime(linkName, flowActivityInstanceUUID));
      }
    }
  }

  /**
   * Returns the link having the given linkName.
   *
   * @param linkName
   *            linkName
   * @return the link having the given linkName.
   */
  public LinkRuntime getLink(final String linkName) {
    return this.linkRuntimes.get(linkName);
  }

  /**
   * Returns the linkStatus of the link having the given linkName.
   *
   * @param linkName
   *            linkName
   * @return the linkStatus of the link having the given linkName.
   */
  public Boolean getLinkStatus(final String linkName) {
    return this.linkRuntimes.get(linkName).getStatus();
  }

  /**
   * Set the given linkStatus to the link having the given linkName.
   *
   * @param linkName
   *            linkName
   * @param status
   *            linkStatus
   */
  public void setLinkStatus(final String linkName, final Boolean status) {
    this.linkRuntimes.get(linkName).setStatus(status);
  }

  /**
   * returns true if this flowRuntime has a link with the given linkName.
   *
   * @param linkName
   *            linkName
   * @return true if this flowRuntime has a link with the given linkName.
   */
  public boolean hasLink(final String linkName) {
    return this.linkRuntimes.containsKey(linkName);
  }

  /**
   * return the bpelExecution waiting for an incoming link having the given
   * linkName. Null if no one was found
   *
   * @param linkName
   *            linkName
   * @return the bpelExecution waiting for an incoming link having the given
   *         linkName. Null if no one was found
   */
  public BpelExecution getWaitingExecution(final String linkName) {
    final LinkRuntime lr = this.linkRuntimes.get(linkName);
    if (lr != null) {
      final BpelExecution waitingExecution = lr.getWaitingExecution();
      if (waitingExecution != null) {
        lr.setWaitingExecution(null);
      }
      return waitingExecution;
    }
    return null;
  }

  /**
   * Called when an activity is waiting for an incoming link with the given
   * linkName (when its target are evaluated).
   *
   * @param linkName
   *            linkName
   * @param waitingExecution
   *            execution waiting for the link (execution which was trying to
   *            execute the activity that should wait...)
   */
  public void addWaitingExecution(final String linkName,
      final BpelExecution waitingExecution) {
    this.linkRuntimes.get(linkName).setWaitingExecution(waitingExecution);
  }
}
