/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.runtime;

import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.orchestra.definition.element.Copy;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.facade.data.runtime.VariableUpdateData;
import org.ow2.orchestra.facade.exception.BpelFaultException;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.facade.exception.BpelFaultException.BpelFault;
import org.ow2.orchestra.facade.uuid.ActivityInstanceUUID;
import org.ow2.orchestra.util.BpelXmlUtil;
import org.ow2.orchestra.util.EnvTool;
import org.ow2.orchestra.util.XmlUtil;
import org.ow2.orchestra.var.MessageVariable;
import org.w3c.dom.Document;

/**
 * @author Charles Souillard
 */
public class VariableRuntime {

  protected long dbid;

  /** log. */
  private static Logger log = Logger.getLogger(VariableRuntime.class.getName());

  /** variable name. */
  protected String name;

  /** variable value. */
  protected MessageVariable messageValue;

  protected Document xmlValue;

  /** variable definition. */
  protected Variable variableDefinition;

  protected ActivityInstanceUUID scopeActivityInstanceUUID;

  protected VariableRuntime() {
  }

  /**
   * Default constructor.
   *
   * @param variableDefinition
   *            variableDefinition
   * @param execution
   *            execution
   */
  public VariableRuntime(final Variable variableDefinition, final BpelExecution execution,
    final ActivityInstanceUUID scopeActivityInstanceUUID) {
    this.scopeActivityInstanceUUID = scopeActivityInstanceUUID;
    this.name = variableDefinition.getName();
    this.variableDefinition = variableDefinition;
    if (this.variableDefinition.getMessageType() != null) {
      final javax.wsdl.Message message =
        execution.getWsdlInfos().getMessage(this.variableDefinition.getMessageType());
      if (message != null) {
        this.messageValue = new MessageVariable(message);
      }
    } else if (this.variableDefinition.getXmlType() != null) {
      this.xmlValue = BpelXmlUtil.createDefaultElement().getOwnerDocument();
    } else if (this.variableDefinition.getElement() != null) {
      this.xmlValue = XmlUtil.getDocumentWithOneElement(this.variableDefinition
          .getElement()).getOwnerDocument();
    } else {
      if (VariableRuntime.log.isLoggable(Level.FINE)) {
        VariableRuntime.log
            .fine("initializeXmlType : variable is not a type and is not an element... : this case is forbiden");
      }
    }
  }

  public void initializeXmlType(final BpelExecution execution) {
    final Copy initCopy = this.variableDefinition.getVirtualCopy();
    if (initCopy != null) {
      initCopy.execute(execution);
    }
    this.updated(execution.getActivityInstanceUUID());
  }

  /**
   * This method must be called only if the variable is not a message.
   *
   * @return true is the variable is already initialized
   */
  public boolean isXmlInitialized() {
    if (this.variableDefinition.getMessageType() != null) {
      throw new OrchestraRuntimeException("Variable " + this.name
          + " is a message, we cannot call " + this.getClass().getName()
          + ".isXmlInitialized method."
          + "please call isPartInitialized method");
    }
    return this.xmlValue != null;
  }

  /**
   * This method must be called only if the variable is a message.
   *
   * @param partName
   *            name of the part we want to know if it is initialized or not
   * @return true is the given part is already initialized
   */
  public boolean isPartInitialized(final String partName) {
    if (this.variableDefinition.getMessageType() == null) {
      throw new OrchestraRuntimeException("Variable " + this.name
          + " is not a message, we cannot call " + this.getClass().getName()
          + ".isPartInitialized method."
          + "please call isXmlInitialized method");
    }
    if (this.messageValue == null) {
      return false;
    } else {
      return this.messageValue.isPartInitialized(partName);
    }
  }

  public Variable getVariableDefinition() {
    return this.variableDefinition;
  }

  public void setVariableDefinition(final Variable variableDefinition) {
    this.variableDefinition = variableDefinition;
  }

  public String getName() {
    return this.name;
  }

  public void setName(final String name) {
    this.name = name;
  }

  public Object getValue() {
    if (this.variableDefinition.getMessageType() != null) {
      return this.messageValue;
    }
    if (this.xmlValue == null) {
      return null;
    }
    return this.xmlValue.getDocumentElement();
  }

  public void setValue(final MessageVariable message) {
    if (this.variableDefinition.getMessageType() == null) {
      throw BpelFaultException.generateBpelFaultException(
          BpelFault.mismatchedAssignmentFailure,
          "trying to set a messageValue and the variableDefinition is not a messageType : " + this.toString(),
          null);
    } else {
      this.messageValue = message;
    }
  }

  public void setValue(final Object value, final boolean keepSrcElementName) {
    if (this.variableDefinition.getMessageType() != null) {
      throw BpelFaultException.generateBpelFaultException(
          BpelFault.mismatchedAssignmentFailure,
          "trying to set a xmlValue and the variableDefinition is a messageType : " + this.toString(),
          null);
    }
    if (this.xmlValue == null) {
      if (this.variableDefinition.getXmlType() != null) {
        this.xmlValue = BpelXmlUtil.createDefaultElement().getOwnerDocument();
      } else if (this.variableDefinition.getElement() != null) {
        this.xmlValue = XmlUtil.getDocumentWithOneElement(this.variableDefinition.getElement()).getOwnerDocument();
      } else {
        throw new OrchestraRuntimeException("trying to set a xmlValue and the variableDefinition type is unknown.");
      }
    }
    BpelXmlUtil.setObjectValue(this.xmlValue.getDocumentElement(), value, keepSrcElementName);
  }

  public void updated(final ActivityInstanceUUID activityInstanceUUID) {
    if (Variable.INBOUND_ANONYMOUS_WSDL_MESSAGE_NAME.equals(this.name)
        || Variable.OUTBOUND_ANONYMOUS_WSDL_MESSAGE_NAME.equals(this.name)) {
      //do not log anonymous variables
      return;
    }
    if (this.variableDefinition.getMessageType() == null) {
      EnvTool.getRecorder().recordVariableUpdated(
          new VariableUpdateData(
              this.scopeActivityInstanceUUID,
              activityInstanceUUID,
              this.name,
              this.xmlValue.getDocumentElement()));
    } else {
      EnvTool.getRecorder().recordVariableUpdated(
          new VariableUpdateData(
              this.scopeActivityInstanceUUID,
              activityInstanceUUID,
              this.name,
              this.messageValue));
    }
  }

  public long getDbid() {
    return this.dbid;
  }

//  public String toString() {
//    return VariableRuntime.class.getName()
//        + "[dbid: " + dbid
//        + ", name: " + name
//        + ", messageValue:" + messageValue
//        + ", xmlValue: " + xmlValue
//        + ", variableDefinition: " + variableDefinition
//        + "]";
//  }
}
