/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.services;

import java.util.Iterator;
import java.util.Set;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.logging.Logger;

import javax.transaction.Synchronization;
import javax.wsdl.Definition;

import org.ow2.orchestra.pvm.env.Environment;
import org.ow2.orchestra.pvm.env.Transaction;
import org.ow2.orchestra.pvm.internal.cmd.Command;
import org.ow2.orchestra.pvm.internal.cmd.CommandService;
import org.ow2.orchestra.pvm.job.Job;
import org.ow2.orchestra.pvm.session.MessageSession;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.services.itf.Invoker;
import org.ow2.orchestra.services.job.SignalInvokeJob;
import org.ow2.orchestra.util.EnvTool;
import org.ow2.orchestra.var.MessageVariable;
import org.w3c.dom.Element;


/**
 * @author Guillaume Porcher
 *
 */
public class InvokeExecutor {

  private InvokeExecutor() {}

  private static ExecutorService executorService = Executors.newCachedThreadPool();

  private static Logger LOG = Logger.getLogger(InvokeExecutor.class.getName());

  private static class InvokeCall implements Runnable {

    private final Invoker invoker;
    private final OperationKey operationKey;
    private final Element partnerEndPointReference;
    private final MessageVariable requestMessage;
    private final BpelExecution execution;
    private final CommandService commandService;
    private final Set<Definition> wsdlDefinitions;

    public InvokeCall(final BpelExecution execution,
        final Invoker invoker,
        final OperationKey operationKey,
        final Element partnerEndPointReference,
        final MessageVariable requestMessage,
        final Set<Definition> wsdlDefinitions,
        final CommandService commandService) {
      this.execution = execution;
      this.invoker = invoker;
      this.operationKey = operationKey;
      this.partnerEndPointReference = partnerEndPointReference;
      this.requestMessage = requestMessage;
      this.wsdlDefinitions = wsdlDefinitions;
      this.commandService = commandService;
    }

    public void run() {
      try {
        final MessageVariable responseMsg = this.invoker.invoke(this.operationKey, this.partnerEndPointReference, this.requestMessage, this.wsdlDefinitions);
        this.commandService.execute(new FinishInvokeCommand(this.execution, responseMsg));
      } catch (final Exception e) {
        this.commandService.execute(new FinishInvokeCommand(this.execution, e));
      }
    }

  }

  public static void executeInvoke(
      final Environment environment,
      final BpelExecution execution,
      final OperationKey operationKey,
      final Element partnerEndPointReference,
      final MessageVariable requestMessage,
      final Set<Definition> wsdlDefinitions) {
    final Invoker invoker = EnvTool.get(Invoker.class, Invoker.DEFAULT_KEY);
    final CommandService commandService = environment.get(CommandService.class);
    final Transaction transaction = environment.get(Transaction.class);

    if (transaction == null) {
      // we are outside a transaction, directly schedule task
      InvokeExecutor.executorService.execute(
        new InvokeCall(
            execution,
            invoker,
            operationKey,
            partnerEndPointReference,
            requestMessage,
            wsdlDefinitions,
            commandService));
    } else {
      // we are in a transaction, schedule task after the end of the transaction
      transaction.registerSynchronization(
          new Synchronization() {
            public void afterCompletion(final int status) {
              if (status == javax.transaction.Status.STATUS_COMMITTED) {
                InvokeExecutor.executorService.execute(
                    new InvokeCall(
                        execution,
                        invoker,
                        operationKey,
                        partnerEndPointReference,
                        requestMessage,
                        wsdlDefinitions,
                        commandService));
              }
            }

            public void beforeCompletion() {
              // do nothing
            }
          }
      );
    }
  }


  private static class FinishInvokeCommand implements Command<Void> {

    private final BpelExecution execution;
    private final MessageVariable responseMessage;
    private final Exception exceptionMessage;

    public FinishInvokeCommand(final BpelExecution execution, final MessageVariable responseMessage) {
      this.execution = execution;
      this.responseMessage = responseMessage;
      this.exceptionMessage = null;
    }

    public FinishInvokeCommand(final BpelExecution execution, final Exception exceptionMessage) {
      this.execution = execution;
      this.responseMessage = null;
      this.exceptionMessage = exceptionMessage;
    }

    public Void execute(final Environment environment) throws Exception {
      // synchronize execution with session
      final BpelExecution bpelExecution = EnvTool.getRepository().getUpdatedExecution(this.execution);

      if (bpelExecution == null) {
        InvokeExecutor.LOG.info("Invoke response not delivered to execution " + this.execution.getUUID() + ": execution has been removed.");
        return null;
      }
      if (bpelExecution.isFinished()) {
        InvokeExecutor.LOG.info("Invoke response not delivered to execution " + bpelExecution + ": execution is finished.");
        return null;
      }

      InvokeExecutor.LOG.info("Deliver invoke response to execution " + bpelExecution);
      boolean isSuspended = false;
      // look for ExecuteInvokeJob previously defined for this execution
      // remove it and save isSuspended value
      final Iterator<Job> jobIterator = bpelExecution.getJobs().iterator();
      while (jobIterator.hasNext()) {
        final Job job = jobIterator.next();
        isSuspended = isSuspended || job.isSuspended();
        jobIterator.remove();
      }

      // Create SignalInvokeJob
      // set same suspended mode
      final SignalInvokeJob signalInvokeJob = new SignalInvokeJob(this.responseMessage, this.exceptionMessage);
      signalInvokeJob.setExecution(bpelExecution);
      signalInvokeJob.setSuspended(isSuspended);

      // schedule job execution
      environment.get(MessageSession.class).send(signalInvokeJob);
      return null;
    }

  }
}
