/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.services;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.transaction.Synchronization;
import javax.xml.namespace.QName;

import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.pvm.env.Environment;
import org.ow2.orchestra.pvm.env.Transaction;
import org.ow2.orchestra.var.MessageVariable;

public final class MessageCarrierRepository {

  private static class TransactionnalMap<K,V> implements Map<K, V> {

    private final Map<K,V> mapToCopy;

    private Map<K,V> map = new HashMap<K, V>();
    private final List<K> addedEntries = new ArrayList<K>();
    private final List<K> removedEntries = new ArrayList<K>();

    public TransactionnalMap(final Transaction tx, final Map<K, V> mapToCopy) {
      synchronized (mapToCopy) {
        this.map = new HashMap<K, V>(mapToCopy);
      }
      this.mapToCopy = mapToCopy;
    }

    public void clear() {
      this.map.clear();
    }

    public boolean containsKey(final Object key) {
      return this.map.containsKey(key);
    }

    public boolean containsValue(final Object value) {
      return this.map.containsValue(value);
    }

    public Set<java.util.Map.Entry<K, V>> entrySet() {
      return Collections.unmodifiableSet(this.map.entrySet());
    }

    public V get(final Object key) {
      return this.map.get(key);
    }

    public boolean isEmpty() {
      return this.map.isEmpty();
    }

    public Set<K> keySet() {
      return Collections.unmodifiableSet(this.map.keySet());
    }

    public V put(final K key, final V value) {
      this.addedEntries.add(key);
      return this.map.put(key, value);
    }

    public void putAll(final Map< ? extends K, ? extends V> t) {
      for (final Entry< ? extends K, ? extends V> e : t.entrySet()) {
        this.put(e.getKey(), e.getValue());
      }
    }

    public V remove(final Object key) {
      if (this.map.containsKey(key)) {
        this.removedEntries.add((K) key);
      }
      return this.map.remove(key);
    }

    public int size() {
      return this.map.size();
    }

    public Collection<V> values() {
      return Collections.unmodifiableCollection(this.map.values());
    }

    void synchronize() {
      synchronized (this.mapToCopy) {
        for (final K key : this.addedEntries) {
          this.mapToCopy.put(key, this.get(key));
        }
        for (final K key : this.removedEntries) {
          this.mapToCopy.remove(key);
        }
      }
    }
  }

  private static Map<ReplierKey, MessageCarrier> messageCarriers =
    new HashMap<ReplierKey, MessageCarrier>();

  private static Map<String, MessageCarrier> pendingMessageCarriers =
    new HashMap<String, MessageCarrier>();

  private static Map<Transaction, TransactionnalMap<ReplierKey, MessageCarrier>> msgCarriersCopies =
    new HashMap<Transaction, TransactionnalMap<ReplierKey,MessageCarrier>>();
  private static Map<Transaction, TransactionnalMap<String, MessageCarrier>> pendingMsgCarriersCopies =
    new HashMap<Transaction, TransactionnalMap<String,MessageCarrier>>();

  private MessageCarrierRepository() {
  }

  private static synchronized Map<String, MessageCarrier> getPendingMessageCarriers() {
    final Transaction transaction = Environment.getCurrent().get(Transaction.class);
    if (transaction != null) {
      TransactionnalMap<String, MessageCarrier> pendingMsgCarriers =
        MessageCarrierRepository.pendingMsgCarriersCopies.get(transaction);
      if (pendingMsgCarriers == null) {
        pendingMsgCarriers =
          new TransactionnalMap<String, MessageCarrier>(transaction, MessageCarrierRepository.pendingMessageCarriers);
        MessageCarrierRepository.pendingMsgCarriersCopies.put(transaction, pendingMsgCarriers);

        final Synchronization synchro = new Synchronization() {
          public void afterCompletion(final int status) {
            final TransactionnalMap<String, MessageCarrier> tm =
              MessageCarrierRepository.pendingMsgCarriersCopies.remove(transaction);
            if (status == javax.transaction.Status.STATUS_COMMITTED) {
              tm.synchronize();
            }
          }
          public void beforeCompletion() {
          }
        };
        transaction.registerSynchronization(synchro);
      }
      return pendingMsgCarriers;
    } else {
      return MessageCarrierRepository.pendingMessageCarriers;
    }
  }

  static synchronized Map<ReplierKey, MessageCarrier> getMessageCarriers() {
    final Transaction transaction = Environment.getCurrent().get(Transaction.class);
    if (transaction != null) {
      TransactionnalMap<ReplierKey, MessageCarrier> msgCarriers = MessageCarrierRepository.msgCarriersCopies.get(transaction);
      if (msgCarriers == null) {
        msgCarriers = new TransactionnalMap<ReplierKey, MessageCarrier>(transaction, MessageCarrierRepository.messageCarriers);
        MessageCarrierRepository.msgCarriersCopies.put(transaction, msgCarriers);

        final Synchronization synchro = new Synchronization() {
          public void beforeCompletion() {
          }
          public void afterCompletion(final int status) {
            final TransactionnalMap<ReplierKey, MessageCarrier> tm =
              MessageCarrierRepository.msgCarriersCopies.remove(transaction);
            if (status == javax.transaction.Status.STATUS_COMMITTED) {
              tm.synchronize();
            }
          }
        };
        transaction.registerSynchronization(synchro);

      }
      return msgCarriers;
    } else {
      return MessageCarrierRepository.messageCarriers;
    }
  }

  public static synchronized void addMessageCarrier(final ReplierKey replierKey,
      final MessageCarrier messageCarrier) {
    MessageCarrierRepository.getMessageCarriers().put(replierKey, messageCarrier);
  }

  public static synchronized void associateMessage(final ReplierKey replierKey,
      final MessageVariable message) {
    final MessageCarrier messageCarrier = MessageCarrierRepository.getMessageCarriers().get(replierKey);
    if (messageCarrier == null) {
      throw new OrchestraRuntimeException("No MessageCarrier corresponding to key = "
          + replierKey);
    }
    MessageCarrierRepository.removeMessageCarrier(replierKey);
    messageCarrier.setMessage(message);
  }

  /**
   * @param replierKey
   * @param faultQName
   * @param message
   */
  public static void associateFaultMessage(final ReplierKey replierKey, final QName faultQName, final MessageVariable message) {
    final MessageCarrier messageCarrier = MessageCarrierRepository.getMessageCarriers().get(replierKey);
    if (messageCarrier == null) {
      throw new OrchestraRuntimeException("No MessageCarrier corresponding to key = " + replierKey);
    }
    MessageCarrierRepository.removeMessageCarrier(replierKey);
    messageCarrier.setFaultMessage(faultQName, message);

  }

  public static synchronized MessageCarrier removeMessageCarrier(final ReplierKey replierKey) {
    return MessageCarrierRepository.getMessageCarriers().remove(replierKey);
  }

  public static synchronized void addPendingMessageCarrier(
      final MessageCarrier messageCarrier) {
    MessageCarrierRepository.getPendingMessageCarriers().put(messageCarrier.getUuid(), messageCarrier);
  }

  public static synchronized MessageCarrier removePendingMessageCarrier(
      final String messageCarrierId) {
    return MessageCarrierRepository.getPendingMessageCarriers().remove(messageCarrierId);
  }

}
