/**
 * Copyright (C) 2007-2009  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.services;

import java.util.List;
import java.util.Set;

import javax.xml.namespace.QName;

import org.ow2.orchestra.facade.def.ProcessState;
import org.ow2.orchestra.facade.def.full.ActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.ProcessFullDefinition;
import org.ow2.orchestra.facade.exception.InstanceNotFoundException;
import org.ow2.orchestra.facade.runtime.ActivityState;
import org.ow2.orchestra.facade.runtime.full.ActivityFullInstance;
import org.ow2.orchestra.facade.runtime.full.ProcessFullInstance;
import org.ow2.orchestra.facade.uuid.ActivityDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessInstanceUUID;

/**
 * A Querier is responsible for getting <strong>runtime</strong> instances
 * informations such as currently running instances and activities.
 *
 * @author Alfonso Fuca
 *
 */

public interface Querier {

  String DEFAULT_KEY = "queryList";

  /**
   * @param processInstanceUUID a process instance uuid
   * @return the process instance which matches uuid
   * @throws InstanceNotFoundException if this instance cannot be found
   */
  ProcessFullInstance getProcessInstance(final ProcessInstanceUUID processInstanceUUID);

  /**
   * Get informations about activity instances
   *
   * @param processUUID the uuid of the process instance
   * @return a set of activity instances corresponding to processUUID
   */
  Set<ActivityFullInstance> findActivityInstances(ProcessInstanceUUID processUUID);

  /**
   * Get informations about activity instances
   *
   * @param processUUID the uuid of the process instance
   * @param activityState the state of activities
   * @return a set of activity instances corresponding to processUUID and state
   *         'active'
   */
  Set<ActivityFullInstance> findActivityInstances(ProcessInstanceUUID processUUID, ActivityState activityState);

  /**
   * Get informations about process instances
   *
   * @return a set of all process instances
   */
  Set<ProcessFullInstance> findProcessInstances();

  /**
   * Get informations about process instances
   *
   * @param processDefinitionUUID the uuid of the process definition
   * @return a set of process instances corresponding to processDefinitionUUID
   */
  Set<ProcessFullInstance> findProcessInstances(ProcessDefinitionUUID processDefinitionUUID);

  /**
   * Get informations about process instances
   *
   * @param processDefinitionUUID the uuid of the process definition
   * @param processInstanceState the state of process instances
   * @return a set of process instances corresponding to processDefinitionUUID
   *         and state 'active'
   */
  Set<ProcessFullInstance> findProcessInstances(
    ProcessDefinitionUUID processDefinitionUUID,
    ActivityState processInstanceState);

  /**
   * Get informations about process instances
   *
   * @param processInstanceState state of process instances
   * @return a set of process instances corresponding to state 'active'
   */
  Set<ProcessFullInstance> findProcessInstances(ActivityState processInstanceState);

  /**
   * @param name the process qualified name
   * @param state TODO
   * @return a list of process UUID, the latest process is the first in the list
   */
  List<ProcessFullDefinition> findProcessDefinitions(QName name, ProcessState state);

  /**
   * @param name the process qualified name
   * @return a list of process UUID, the latest process is the first in the list
   */
  List<ProcessFullDefinition> findProcessDefinitions(QName name);

  /**
   * @param state TODO
   * @return a list of process UUID, the latest process is the first in the list
   */
  List<ProcessFullDefinition> findProcessDefinitions(ProcessState state);

  /**
   * @return a list of process, the latest process is the first in the list
   */
  List<ProcessFullDefinition> findProcessDefinitions();

  /**
   * Get informations about activity definition
   *
   * @param activityDefinitionUUID an uuid of activity
   * @return the activity definition which matches the uuid
   */
  ActivityFullDefinition getActivityDefinition(ActivityDefinitionUUID activityDefinitionUUID);

  /**
   * @param processDefinitionUUID an uuid
   * @return the process definition which matches the uuid
   */
  ProcessFullDefinition getProcessDefinition(ProcessDefinitionUUID processDefinitionUUID);

}
