package org.ow2.orchestra.services;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.activity.AbstractBpelActivity;
import org.ow2.orchestra.definition.activity.Process;
import org.ow2.orchestra.definition.element.Copy;
import org.ow2.orchestra.definition.element.Correlation;
import org.ow2.orchestra.definition.element.CorrelationSet;
import org.ow2.orchestra.definition.element.InboundMessageElement;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.pvm.activity.Activity;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.services.impl.ReplierImpl;
import org.ow2.orchestra.services.itf.Replier;
import org.ow2.orchestra.util.ReceivingElementUtil;
import org.ow2.orchestra.var.MessageVariable;
/**
 * base implementation for the services to be published for a deployed bpel
 * process.
 *
 * @author Tom Baeyens, Charles Souillard, Guillaume Porcher
 */
public class Receiver {

  /** log. */
  private static Logger log = Logger.getLogger(Receiver.class.getName());

  protected long dbid;

  protected int dbversion;

  /**
   * maps ReceiveElementKeys(processQName, portTypeQName, operationName) to a
   * list of executions that are waiting for an incoming message.
   */
  private final Set<BpelExecution> waitingExecutions = new HashSet<BpelExecution>();

  /**
   * messages stored when there is no execution yet that matches this incoming
   * message.
   */
  private final List<PendingMessage> storedMessages = new ArrayList<PendingMessage>();

  private OperationKey operationKey;

  protected Receiver() {
    super();
  }

  public Receiver(final OperationKey operationKey) {
    this.operationKey = operationKey;
  }

  /**
   * Find if there is a stored message for one of the executions. If such an
   * execution is found, associate the stored message to this execution, and
   * return this execution.
   *
   * @param parent
   *          execution to check (recursive search in the child executions)
   * @return first execution found with a matching message.
   */
  static ExecElementToSignal findExecutionToSignal(final BpelProcess bpelProcess, final BpelExecution instance) {
    if (instance == null || instance.isEnded()) {
      return null;
    }
    ExecElementToSignal execToSignal = null;
    final Set<BpelExecution> waitingExec = new HashSet<BpelExecution>(instance.getWaitingExecutions());
    if (waitingExec != null) {
      for (final BpelExecution child : waitingExec) {
        final Activity activity = child.getNode().getBehaviour();
        if (activity instanceof InboundMessageElement) {
          for (final ReceivingElement receivingElement : ((InboundMessageElement) activity).getReceivingElements()) {
            final OperationKey opKey = receivingElement.getOperationKey();
            final Receiver receiver = bpelProcess.getReceiver(opKey);
            receiver.addWaitingExecution(child);
            if (execToSignal == null) {
              final PendingMessage pendingMessage = receiver.getMessageFromStoredMessage(child, receivingElement);
              if (pendingMessage != null) {
                receiver.removePendingMessage(pendingMessage);
                execToSignal = new ExecElementToSignal(child, receivingElement, pendingMessage);
                if (!receiver.removeWaitingExecution(child)) {
                  throw new OrchestraRuntimeException("No such waiting execution : " + child);
                }
              }
            }
          }
        }
      }
    }
    return execToSignal;
  }

  ExecElementToSignal associatePendingMessage(final PendingMessage pendingMessage, final BpelProcess bpelProcess) {
    ExecElementToSignal execElementToSignal = null;

    // find a waiting execution using the correlations
    execElementToSignal = this.getWaitingExecution(pendingMessage);
    // if an execution was waiting for such an incoming message
    if (execElementToSignal != null) {
      if (Receiver.log.isLoggable(Level.FINE)) {
        Receiver.log.fine("Matching execution found for incoming msg: " + execElementToSignal.getBpelExecution());
      }

      return execElementToSignal;
    } else {
      final InboundMessageElement startActivity = bpelProcess.getStartElement(this.operationKey);
      // if the incoming message matched a start activity
      if (startActivity != null) {
        ReceivingElement startElement = null;
        for (final ReceivingElement receivingElement : startActivity.getReceivingElements()) {
          if (receivingElement.getOperationKey().equals(this.operationKey)) {
            startElement = receivingElement;
          }
        }
        if (startElement == null) {
          throw new IllegalStateException("No start element found");
        }
        if (Receiver.log.isLoggable(Level.FINE)) {
          Receiver.log.fine("found start node " + startElement + " for incoming msg");
        }
        // create new instance
        final BpelExecution bpelExecution = this.createNewInstance(pendingMessage, bpelProcess, startActivity, startElement);

        final ExecElementToSignal receiveExecution = Receiver.getWaitingExecution(bpelExecution, startElement, pendingMessage);
        if (receiveExecution == null) {
          throw new OrchestraRuntimeException("No waiting execution found !, process = "
            + bpelExecution.getProcessDefinition().getQName());
        }
        return receiveExecution;
      } else { // no start node was found
        if (Receiver.log.isLoggable(Level.FINE)) {
          Receiver.log.fine("no matching execution found for pending msg.");
        }
        return null;
      }
    }
  }

  private BpelExecution createNewInstance(
      final PendingMessage pendingMessage,
      final BpelProcess bpelProcess,
      final InboundMessageElement startActivity,
      final ReceivingElement startElement) {
    // create instance
    final BpelExecution bpelExecution = (BpelExecution) bpelProcess.createProcessInstance();
    // initialise receive variables
    final Map<String, Object> additionnalVariables = new HashMap<String, Object>();
    final Process processActivity = (Process) bpelProcess.getInitial().getBehaviour();
    if (startElement.getVariable() != null) {
      final Variable var =
        ((AbstractBpelActivity) startActivity).getEnclosingScope().findVariable(startElement.getVariable());
      final Variable var1 = processActivity.getVariable(startElement.getVariable());
      if (var1.equals(var)) {
        additionnalVariables.put(startElement.getVariable(), pendingMessage.getMessage().duplicate());
      }
    } else {
      for (final Copy copy : startElement.getInCopies()) {
        final Variable var =
          ((AbstractBpelActivity) startActivity).getEnclosingScope().findVariable(copy.getTo().getVariable());
        final Variable var1 = processActivity.getVariable(copy.getTo().getVariable());
        if (var1.equals(var)) {
          additionnalVariables.put(copy.getTo().getVariable(),
              pendingMessage.getMessage().getPartValue(copy.getFrom().getPart()));
        }
      }
    }
    if (!additionnalVariables.isEmpty()) {
      bpelExecution.createVariable(BpelExecution.ADDITIONNAL_SCOPE_VARIABLES, additionnalVariables);
    }
    // start execution
    bpelExecution.begin();
    return bpelExecution;
  }

  /**
   * Associate the given replier to the corresponding runtimeScope (top level
   * scope if messageExchange is not specified, scopeRuntime declaring the used
   * messageExchange if there is one.
   */
  public static void addReplier(final BpelExecution bpelExecution,
      final ReceivingElement receivingElement,
      final MessageCarrier messageCarrier) {
    if (messageCarrier != null) {
      final String messageExchange = receivingElement.getMessageExchange();
      final ReplierKey replierKey =
        new ReplierKey(receivingElement.getOperationKey(), messageExchange, bpelExecution.getCurrentScopeRuntime());
      final Replier replier = Receiver.createReplier(replierKey);
      MessageCarrierRepository.addMessageCarrier(replierKey, messageCarrier);
      bpelExecution.addReplier(replier);
    }
  }

  /**
   * Search for an execution of the given process waiting for a message on the
   * given portType/operation. I one was found and the cs matches (if any),
   * associate this execution to the given message and return this execution.
   *
   * @param incomingMessage
   *          message received we want to give to an execution
   * @param operationKey
   *          operationKey
   * @param environment
   *          environment
   * @return the found execution or null if no one was found
   */
  private ExecElementToSignal getWaitingExecution(final PendingMessage pendingMessage) {
    final MessageVariable incomingMessage = pendingMessage.getMessage();
    if (this.getWaitingExecutions().isEmpty()) {
      if (Receiver.log.isLoggable(Level.FINE)) {
        Receiver.log.fine("No waiting executions for Receiver: " + this.getOperationKey());
      }
      return null;
    }
    if (Receiver.log.isLoggable(Level.FINE)) {
      Receiver.log.fine("Waiting executions for Receiver " + this.getOperationKey() + ": " + this.getWaitingExecutions());
    }
    for (final BpelExecution bpelExecution : this.getWaitingExecutions()) {

      if (!bpelExecution.isInstanceSuspended()) {
        final InboundMessageElement inboundMessageElement = (InboundMessageElement) bpelExecution.getNode().getBehaviour();
        for (final ReceivingElement receivingElement : inboundMessageElement.getReceivingElements()) {
          if (Receiver.log.isLoggable(Level.FINE)) {
            Receiver.log.fine("Receiving Element: "
              + receivingElement.getOperationKey() + " / OperationKey="
              + this.operationKey);
          }
          if (receivingElement.getOperationKey().equals(this.operationKey)) {
            if (Receiver.csMatch(incomingMessage, receivingElement, bpelExecution)) {
              final ExecElementToSignal res = new ExecElementToSignal(bpelExecution, receivingElement, pendingMessage);
              if (!this.removeWaitingExecution(bpelExecution)) {
                throw new OrchestraRuntimeException("No such waiting execution : " + bpelExecution);
              }
              return res;
            }
          }
        }
      }

    }
    return null;
  }

  /**
   * Search for an execution of the given process instance waiting for a message
   * on the given node. This method is used to associate an incoming message
   * which creates a new instance to the newly created instance.
   *
   * @return the found execution or null if no one was found
   */
  private static ExecElementToSignal getWaitingExecution(final BpelExecution parent, final ReceivingElement startElement,
    final PendingMessage pendingMessage) {
    // TODO: change this method
    if (parent == null) {
      return null;
    }
    final BpelExecution instance = parent.getProcessInstance();
    final BpelProcess bpelProcess = instance.getProcessDefinition();

    final Set<BpelExecution> waitingExec = new HashSet<BpelExecution>(instance.getWaitingExecutions());
    if (Receiver.log.isLoggable(Level.FINE)) {
      Receiver.log.fine(Receiver.class + ".addWaitingExecutions, execution = " + parent + " waiting=" + waitingExec);
    }
    ExecElementToSignal res = null;
    if (waitingExec != null) {
      for (final BpelExecution child : waitingExec) {
        final Activity activity = child.getNode().getBehaviour();
        if (activity instanceof InboundMessageElement) {
          final InboundMessageElement inboundMessageElement = (InboundMessageElement) activity;
          for (final ReceivingElement receivingElement : inboundMessageElement.getReceivingElements()) {
            final OperationKey opKey = receivingElement.getOperationKey();
            final Receiver receiver = bpelProcess.getReceiver(opKey);
            receiver.addWaitingExecution(child);
            if (res == null && receivingElement.equals(startElement)) {
              res = new ExecElementToSignal(child, startElement, pendingMessage, true);
              if (!receiver.removeWaitingExecution(child)) {
                throw new OrchestraRuntimeException("No such waiting execution : " + child);
              }
              if (bpelProcess.getCommonCorrelationSets() != null) {
                for (final CorrelationSet cs : bpelProcess.getCommonCorrelationSets()) {
                  ReceivingElementUtil.initiateCs(startElement, child, cs, pendingMessage.getMessage());
                }
              }
            }
          }
        }
      }
    }
    return res;
  }

  /**
   * Store the given message as a message that could be consumed by an activity
   * waiting for a message on the given portType/operation.
   *
   * @param message
   *          message to store
   * @param operationName
   *          operation
   * @param portType
   *          portType
   * @param environment
   *          environment
   *
   */
  PendingMessage storeIncomingMessage(final MessageVariable message, final MessageCarrier messageCarrier) {
    if (Receiver.log.isLoggable(Level.FINEST)) {
      Receiver.log.finest("Incoming msg: " + message);
    }
    PendingMessage pendingMessage = null;
    if (messageCarrier != null) {
      MessageCarrierRepository.addPendingMessageCarrier(messageCarrier);
      pendingMessage = new PendingMessage(message, messageCarrier.getUuid(), this);
    } else {
      pendingMessage = new PendingMessage(message, null, this);
    }
//    this.storePendingMessage(pendingMessage);
    return pendingMessage;
  }

  protected void storePendingMessage(final PendingMessage pendingMessage) {
    this.storedMessages.add(pendingMessage);
  }
  // ///////////////////////////////////////////////////////////////////////////////////////

  /**
   * Return true if the given message match the given startActivity for the
   * given execution. It takes in charge cs verification.
   *
   * @param messageToAnalyse
   *          messageToAnalyse
   * @param receivingElement
   *          startActivity
   * @param bpelExecution
   *          bpelExecution
   * @return true if the given message match the given startActivity for the
   *         given execution.
   */
  static boolean csMatch(final MessageVariable messageToAnalyse, final ReceivingElement receivingElement,
    final BpelExecution bpelExecution) {
    if (receivingElement.getCorrelations() == null) {
      return true;
    }
    // we calcul the values for each correlationSet
    for (final Correlation correlation : receivingElement.getCorrelations()) {
      if (!correlation.correlationOK(messageToAnalyse, receivingElement, bpelExecution)) {
        return false;
      }
    }
    return true;

  }

  /**
   * Search for a stored message matching the given parameters.
   *
   * @param bpelExecution
   *          bpelExecution
   * @param receivingElement
   *          startActivity
   * @return a stored message matching the given parameters. or null.
   */
  private PendingMessage getMessageFromStoredMessage(final BpelExecution bpelExecution, final ReceivingElement receivingElement) {
    if (this.storedMessages == null) {
      return null;
    }
    for (final PendingMessage pendingMessage : this.storedMessages) {
      final MessageVariable message = pendingMessage.getMessage();
      if (Receiver.csMatch(message, receivingElement, bpelExecution)) {
        return pendingMessage;
      }
    }
    return null;
  }

  /**
   * Add a waiting execution in the the list.
   *
   * @param bpelExecution
   *          bpelExecution
   * @param receivingElement
   *          startActivity
   */
  protected void addWaitingExecution(final BpelExecution bpelExecution) {
    this.getWaitingExecutions().add(bpelExecution);
    if (Receiver.log.isLoggable(Level.FINE)) {
      Receiver.log.fine("addWaitingExecution, operationKey=" + this.operationKey + ", execution = " + bpelExecution);
    }
  }

  public boolean removeWaitingExecution(final BpelExecution bpelExecution) {
    final boolean b1 = bpelExecution.getProcessInstance().getWaitingExecutions().remove(bpelExecution);
    final boolean b2 = this.getWaitingExecutions().remove(bpelExecution);
    return b1 && b2;
  }

  public boolean removePendingMessage(final PendingMessage pendingMessage) {
    return this.storedMessages.remove(pendingMessage);
  }

  public OperationKey getOperationKey() {
    return this.operationKey;
  }

  private static Replier createReplier(final ReplierKey replierKey) {
    return new ReplierImpl(replierKey);
  }

  protected Set<BpelExecution> getWaitingExecutions() {
    return this.waitingExecutions;
  }

}
