/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.services.impl;

import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.element.InboundMessageElement;
import org.ow2.orchestra.pvm.activity.Activity;
import org.ow2.orchestra.pvm.env.Environment;
import org.ow2.orchestra.pvm.internal.model.NodeImpl;
import org.ow2.orchestra.pvm.model.Node;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.services.Receiver;
import org.ow2.orchestra.services.ReceivingElement;
import org.ow2.orchestra.services.itf.Publisher;
import org.ow2.orchestra.services.itf.Repository;
import org.ow2.orchestra.util.Misc;

public class PublisherImpl implements Publisher {

  private Repository repository = null;

  private synchronized Repository getRepository() {
    if (this.repository == null) {
      this.repository = Environment.getCurrent().get(Repository.class);
    }
    return this.repository;
  }


  public void publishServices(final BpelProcess bpelProcess, final Environment environment) {
    Misc.log(Level.INFO, "Publishing services for Bpel process %s.", bpelProcess.getQName());
    // set of unique operation keys defined in the process
    final Set<OperationKey> operationKeys = this.getServiceOperationKeys(bpelProcess, bpelProcess.getInitial(),
        environment);
    // for each operation key, deploy a receiver
    for (final OperationKey operationKey : operationKeys) {
      final Receiver receiver = this.createReceiverInstance(operationKey);
      bpelProcess.addReceiver(receiver);
    }
  }

  protected Set<OperationKey> getServiceOperationKeys(final BpelProcess bpelProcess, final NodeImpl node,
      final Environment environment) {

    final Set<OperationKey> operationKeys = new HashSet<OperationKey>();
    // get list of operation keys for the current node
    final Activity behaviour = node.getBehaviour();
    if (behaviour instanceof InboundMessageElement) {
      final InboundMessageElement inboundMessageElement = (InboundMessageElement) behaviour;
      for (final ReceivingElement receivingElement : inboundMessageElement.getReceivingElements()) {
        final OperationKey operationKey = receivingElement.getOperationKey();
        operationKeys.add(operationKey);
      }
    }
    // Add list of operation keys defined in children nodes
    final List<Node> children = node.getNodes();
    if (children != null) {
      for (final Node child : children) {
        operationKeys.addAll(this.getServiceOperationKeys(bpelProcess, (NodeImpl) child, environment));
      }
    }

    return operationKeys;
  }

  public void unpublishServices(final BpelProcess bpelProcess, final Environment environment) {
    Misc.log(Level.INFO, "Unpublishing services for Bpel process %s.", bpelProcess.getQName());
    this.unpublishServices(bpelProcess, bpelProcess.getInitial(),
        environment);
  }

  public void unpublishServices(final BpelProcess bpelProcess, final NodeImpl node,
      final Environment environment) {
    final Activity behaviour = node.getBehaviour();
    if (behaviour instanceof InboundMessageElement) {
      final InboundMessageElement inboundMessageElement = (InboundMessageElement) behaviour;
      for (final ReceivingElement receivingElement : inboundMessageElement.getReceivingElements()) {
        final OperationKey operationKey = receivingElement.getOperationKey();
        bpelProcess.removeReceiver(operationKey);
      }
    }
    final List<Node> children = node.getNodes();
    if (children != null) {
      for (final Node child : children) {
        this.unpublishServices(bpelProcess, (NodeImpl) child, environment);
      }
    }
  }

  public void republishServices(final BpelProcess bpelProcess, final Environment environment) {
    Misc.log(Level.INFO, "Republishing services for Bpel process %s.", bpelProcess.getQName());
  }

  protected Receiver createReceiverInstance(final OperationKey operationKey) {
    return new Receiver(operationKey);
  }

}
