/**
 * Copyright (C) 2007-2009  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.services.job;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.HashMap;
import java.util.Map;

import javax.xml.namespace.QName;

import org.ow2.orchestra.definition.activity.Invoke;
import org.ow2.orchestra.exception.FaultWithMessageVariable;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.pvm.env.Environment;
import org.ow2.orchestra.pvm.internal.job.MessageImpl;
import org.ow2.orchestra.pvm.internal.jobexecutor.JobDbSession;
import org.ow2.orchestra.pvm.session.MessageSession;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.services.AsyncFindExecToSignal;
import org.ow2.orchestra.var.MessageVariable;

/**
 * @author Guillaume Porcher, S. Ali Tokmen
 */
public class SignalInvokeJob extends MessageImpl<Void> {

  private static final long serialVersionUID = 1796627705334254482L;

  MessageVariable responseMessage;
  String exceptionMessage;

  QName faultName;

  // for persistence
  protected SignalInvokeJob() {
  }

  public SignalInvokeJob(final MessageVariable responseMessage, final Exception exception) {
    this.responseMessage = responseMessage != null ? responseMessage.duplicate() : null;
    this.exceptionMessage = null;
    if (exception != null) {
      if (exception instanceof FaultWithMessageVariable) {
        final FaultWithMessageVariable fault = (FaultWithMessageVariable) exception;
        this.faultName = fault.getFaultName();
        this.responseMessage = fault.getFaultMessage();
      } else {
        // IMPORTANT: We assume FaultWithMessage can only be created in internal
        // by Orchestra, and will never be returned by an invoked WS
        //
        // serialize the stack trace
        final StringWriter sw = new StringWriter();
        exception.printStackTrace(new PrintWriter(sw));
        this.exceptionMessage = sw.toString();
      }
    }
  }

  /**
   * Signals execution waiting on an invoke activity. Provides received message
   * as parameter.
   */
  public Void execute(final Environment environment) throws Exception {
    final BpelExecution execution = (BpelExecution) this.getExecution();

    final Map<String, Object> signalParams = new HashMap<String, Object>();

    if (this.faultName != null) {
      signalParams.put(Invoke.RECEIVED_EXCEPTION, new FaultWithMessageVariable(this.faultName, this.responseMessage));
    } else if (this.exceptionMessage != null && !this.exceptionMessage.equals("")) {
      signalParams.put(Invoke.RECEIVED_EXCEPTION, new OrchestraRuntimeException(this.exceptionMessage));
    } else {
      signalParams.put(Invoke.RECEIVED_MESSAGE, this.responseMessage);
    }

    final BpelExecution bpelInstance = execution.getProcessInstance();
    execution.signal(signalParams);
    // signal executions waiting on a receive.
    if (!bpelInstance.isFinished()) {
      environment.get(MessageSession.class).send(
        new AsyncFindExecToSignal(bpelInstance, bpelInstance.getProcessDefinition().getQName()));
    }
    final JobDbSession jobDbSession = environment.get(JobDbSession.class);
    jobDbSession.delete(this);

    return null;
  }

}
