/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.util;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import javax.wsdl.Part;
import javax.wsdl.extensions.soap.SOAPAddress;
import javax.xml.XMLConstants;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;

import org.w3c.dom.Document;
import org.w3c.dom.DocumentType;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * @author Charles Souillard
 */
public final class BpelUtil {

  /**
   * Default private constructor to prevent this class to be instanciated.
   */
  private BpelUtil() { }

  /**
   * Return true if the given element has an attribute named createInstance and if
   * its value is 'yes'.
   * @param element dom element to analyse
   * @return true if the given element has an attribute named createInstance and if
   * its value is 'yes'.
   */
  public static boolean isCreateInstance(final Element element) {
    final String createInstance = XmlUtil.attribute(element, "createInstance");
    if (createInstance != null && createInstance.equals("yes")) {
      return true;
    }
    return false;
  }
  /**
   * Get all namespaces defined in the given node and all of its parents.
   * @param n node from where namespaces are taken
   * @return all namespaces defined in the given node and all of its parents.
   */
  public static Map<String, String> getAllNameSpaces(final org.w3c.dom.Node n) {
    final HashMap<String, String> result = new HashMap<String, String>();
    if (n == null) {
      return result;
    }
    //we look for the NameSpace in the parent Node
    if (n.getParentNode() != null) {
      final Map<String, String> resultParent = BpelUtil.getAllNameSpaces(n.getParentNode());
      for (final Entry<String, String> ns : resultParent.entrySet()) {
        if (!result.containsKey(ns.getKey())) {
          result.put(ns.getKey(), ns.getValue());
        }
      }

    }
    final NamedNodeMap attributes = n.getAttributes();
    if (attributes != null) {
      for (int i = 0; i < attributes.getLength(); i++) {
        if (attributes.item(i).getNodeName().contains("xmlns:")) {
          final String nodeName = attributes.item(i).getNodeName();
          final String prefix = nodeName.split(":")[1];
          result.put(prefix, attributes.item(i).getTextContent());
        }
      }
    }
    result.put(XmlConstants.BPEL_FUNC_PREFIX, XmlConstants.XMLNS_BPEL_2_0_EXECUTABLE);
    return result;
  }


  /**
   * Look for an instance of a class in a list of ExtensibilityElement.
   * @param listOfExtensibilityElement the list of ExtensibilityElement.
   * @param classToFind : the class we llok for
   * @return If found, return the element found, else, return null.
   */
  @SuppressWarnings("unchecked")
  public static Object findExtensibilityElement(final List listOfExtensibilityElement, final Class classToFind) {
    for (final Object extElem : listOfExtensibilityElement) {
      if (classToFind.isInstance(extElem)) {
        return extElem;
      }
    }
    return null;
  }
  /**
   * return true if the given ns is a xmlSchema.
   * @param ns ns to analyse
   * @return true if the given ns is a xmlSchema.
   */
  public static boolean isXmlSchemaNs(final String ns) {
    return ns != null && (ns.equals(XmlConstants.XMLNS_XMLSCHEMA) || ns.equals(XmlConstants.XMLNS_XMLSCHEMAXSD));
  }
  /**
   * Return a dom document conformant with ServiceRef xsd based on the given address and serviceQName.
   * @param address address to use to build the serviceRef document
   * @param serviceQN serviceQN to use to build the serviceRef document
   * @return a dom document conformant with ServiceRef xsd based on the given address and serviceQName.
   */
  public static Document getServiceRefDocument(final SOAPAddress address, final QName serviceQN) {
    final DocumentBuilder documentBuilder = XmlUtil.getDocumentBuilder();
    final Document document = documentBuilder.newDocument();
    XmlUtil.returnDocumentBuilder(documentBuilder);

    final Element serviceRef =
      document.createElementNS(XmlConstants.XMLNS_BPEL_2_0_SERVICEREF, XmlConstants.SERVICE_REF_TAG_NAME);
    document.appendChild(serviceRef);

    final Element epr =
      document.createElementNS(XmlConstants.ADDRESSING_2004_REF_NAMESPACE, XmlConstants.EPR_TAG_NAME);
    serviceRef.appendChild(epr);

    final Element addressElement =
      document.createElementNS(XmlConstants.ADDRESSING_2004_REF_NAMESPACE,XmlConstants.ADDRESS_TAG_NAME);
    addressElement.setTextContent(address.getLocationURI());
    epr.appendChild(addressElement);

    final Element serviceNameElement =
      document.createElementNS(XmlConstants.ADDRESSING_2004_REF_NAMESPACE,XmlConstants.SERVICE_NAME_TAG_NAME);
    serviceNameElement.setAttribute("xmlns:eprServicePrefix", serviceQN.getNamespaceURI());
    serviceNameElement.setTextContent("eprServicePrefix:" + serviceQN.getLocalPart());
    epr.appendChild(serviceNameElement);

    return document;
  }


  public static Element getPartElement(final Element father, final Part part) {
    final String partName = part.getName();
    final List<Element> childrenElements = XmlUtil.elements(father);
    for (final Element child : childrenElements) {
      if (child.getLocalName().equals(partName)) {
        return child;
      }
    }
    return null;
  }

  public static boolean booleanValue(final String text, final boolean defaultValue) {
    final Boolean result = BpelUtil.booleanValue(text);
    if (result == null) {
      return defaultValue;
    }
    return result.booleanValue();
  }

  public static Boolean booleanValue(final String text) {
    if (text == null) {
      return null;
    }

    // if we have to check for value true
    if (("true".equalsIgnoreCase(text))
        || ("enabled".equalsIgnoreCase(text))
        || ("on".equalsIgnoreCase(text))
        || ("yes".equalsIgnoreCase(text))
    ) {
      return Boolean.TRUE;
    }
    if (("false".equalsIgnoreCase(text))
        || ("disabled".equalsIgnoreCase(text))
        || ("off".equalsIgnoreCase(text))
        || ("no".equalsIgnoreCase(text))
    ) {
      return Boolean.FALSE;
    }

    return null;
  }

  /**
   * Test the equality of two nodes following same conditions that org.w3c.dom.Node.isEqualNode(Node n) except for prefix.
   * The two nodes prefix doesn't need to be equals.
   * @param n1 First node
   * @param n2 Second node
   * @return true if the two nodes are equals, false otherwise
   */
  public static boolean areNodeEquals(final Node n1, final Node n2) {
    Misc.checkArgsNotNull(n1, n2);

    Node n1Temp, n2Temp;
    int i = 0;
    NamedNodeMap n1NamedNodeMap;
    NamedNodeMap n2NamedNodeMap;

    if (n1.getNodeType() != n2.getNodeType()) {
      return false;
    }
    if (n1.getNodeType() == Node.DOCUMENT_TYPE_NODE && !BpelUtil.areDocumentTypeEquals((DocumentType) n1, (DocumentType) n2)) {
      return false;
    }

    if (!BpelUtil.areStringEquals(n1.getLocalName(), n2.getLocalName())) {
      return false;
    }
    if (!BpelUtil.areStringEquals(n1.getNamespaceURI(), n2.getNamespaceURI())) {
      return false;
    }
    if (!BpelUtil.areStringEquals(n1.getNodeValue(), n2.getNodeValue())) {
      return false;
    }
    n1NamedNodeMap = n1.getAttributes();
    n2NamedNodeMap = n2.getAttributes();
    if (!BpelUtil.areNamedNodeMapNullOrDiffLength(n1NamedNodeMap, n2NamedNodeMap)) {
      return false;
    }

    String n1Name, n2Name;
    boolean found = false;
    int j = 0;
    //Iterate the first NamedNodeMap and verify that the elements exist in the second NamedNodeMap
    if (n1NamedNodeMap != null && n1NamedNodeMap.getLength() > 0) {
      for (i = 0;i < n1NamedNodeMap.getLength();i++) {
        n1Temp = n1NamedNodeMap.item(i);
        if (XMLConstants.XMLNS_ATTRIBUTE_NS_URI.equals(n1Temp.getNamespaceURI())) {
          break;
        }
        if (n1Temp.getNodeName().contains(":")) {
          n1Name = n1Temp.getLocalName();
        } else {
          n1Name = n1Temp.getNodeName();
        }
        n1Temp.normalize();
        for (j = 0;!found && j < n2NamedNodeMap.getLength();j++) {
          n2Temp = n2NamedNodeMap.item(j);
          if (n2Temp.getNodeName().contains(":")) {
            n2Name = n2Temp.getLocalName();
          } else {
            n2Name = n2Temp.getNodeName();
          }
          n2Temp.normalize();
          found = n1Temp.getNodeValue().equals(n2Temp.getNodeValue()) && BpelUtil.areStringEquals(n1Name, n2Name)
            && BpelUtil.areStringEquals(n1Temp.getNamespaceURI(), n2Temp.getNamespaceURI());
        }
        if (!found) {
          return false;
        }
      }
    }

    final NodeList n1ChildNodes = n1.getChildNodes();
    final NodeList n2ChildNodes = n2.getChildNodes();
    if ((n1ChildNodes != null && n2ChildNodes == null) || (n1ChildNodes == null && n2ChildNodes != null)) {
      return false;
    }
    if (n1ChildNodes != null && n2ChildNodes != null) {
      if (n1ChildNodes.getLength() != n2ChildNodes.getLength()) {
        return false;
      }
      for (i = 0;i < n1ChildNodes.getLength();i++) {
        n1Temp = n1ChildNodes.item(i);
        n2Temp = n2ChildNodes.item(i);
        if (n2Temp == null) {
          return false;
        }
        n1Temp.normalize();
        n2Temp.normalize();
        if (!BpelUtil.areNodeEquals(n1Temp, n2Temp)) {
          return false;
        }
      }
    }
    return true;
  }

  //Supplementary checks in case of DocumentType Nodes
  private static boolean areDocumentTypeEquals(final DocumentType documentType1, final DocumentType documentType2) {
    NamedNodeMap n1NamedNodeMap;
    NamedNodeMap n2NamedNodeMap;
    if (!BpelUtil.areStringEquals(documentType1.getPublicId(), documentType2.getPublicId())) {
      return false;
    }
    if (!BpelUtil.areStringEquals(documentType1.getSystemId(), documentType2.getSystemId())) {
      return false;
    }
    if (!BpelUtil.areStringEquals(documentType1.getInternalSubset(), documentType2.getInternalSubset())) {
      return false;
    }
    n1NamedNodeMap = documentType1.getEntities();
    n2NamedNodeMap = documentType2.getEntities();
    if (!BpelUtil.areNamedNodeMapNullOrDiffLength(n1NamedNodeMap, n2NamedNodeMap)) {
      return false;
    }
    if (!BpelUtil.areNamedNodeMapEquals(n1NamedNodeMap, n2NamedNodeMap)) {
      return false;
    }

    n1NamedNodeMap = documentType1.getNotations();
    n2NamedNodeMap = documentType2.getNotations();
    if (!BpelUtil.areNamedNodeMapNullOrDiffLength(n1NamedNodeMap, n2NamedNodeMap)) {
      return false;
    }
    if (!BpelUtil.areNamedNodeMapEquals(n1NamedNodeMap, n2NamedNodeMap)) {
      return false;
    }
    return true;
  }

  //Return true if the two strings are null or equals, false otherwise
  private static boolean areStringEquals(final String str1, final String str2) {
    if (str1 == null && str2 == null) {
      return true;
    }
    if (str1 == null) {
      return false;
    }
    if (str2 == null) {
      return false;
    }
    return str1.equals(str2);
  }

  //Return true if the two NamedNodeMap are null or have the same size, false otherwise
  private static boolean areNamedNodeMapNullOrDiffLength(final NamedNodeMap n1NamedNodeMap, final NamedNodeMap n2NamedNodeMap) {
    if ((n1NamedNodeMap != null && n2NamedNodeMap == null)
        || (n1NamedNodeMap == null && n2NamedNodeMap != null)) {
      return false;
    }
    if (n1NamedNodeMap == null) {
      return true;
    }
    int n1size = 0;
    for (int i = 0; i < n1NamedNodeMap.getLength(); i++) {
      if (!XMLConstants.XMLNS_ATTRIBUTE_NS_URI.equals(n1NamedNodeMap.item(i).getNamespaceURI())) {
        n1size++;
      }
    }
    int n2size = 0;
    for (int i = 0; i < n2NamedNodeMap.getLength(); i++) {
      if (!XMLConstants.XMLNS_ATTRIBUTE_NS_URI.equals(n2NamedNodeMap.item(i).getNamespaceURI())) {
        n2size++;
      }
    }
    return n1size == n2size;
  }

  //Return true if the two NamedNodeMap are equals : equals element (using isEqualNode) at the same index, false otherwise
  private static boolean areNamedNodeMapEquals(final NamedNodeMap n1NamedNodeMap, final NamedNodeMap n2NamedNodeMap) {
    if (n1NamedNodeMap != null && n1NamedNodeMap.getLength() > 0) {
      Node n1Temp, n2Temp;
      int i = 0;
      for (i = 0;i < n1NamedNodeMap.getLength();i++) {
        n1Temp = n1NamedNodeMap.item(i);
        n2Temp = n2NamedNodeMap.item(i);
        if (n2Temp == null) {
          return false;
        }
        n1Temp.normalize();
        n2Temp.normalize();
        if (!n1Temp.isEqualNode(n2Temp)) {
          return false;
        }
      }
    }
    return true;
  }
}


