/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.util;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.List;

import org.hibernate.cfg.Configuration;
import org.hibernate.dialect.Dialect;
import org.ow2.orchestra.env.DefaultEnvXMLGenerator;
import org.ow2.orchestra.env.DefaultEnvXMLGenerator.InstallationType;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.pvm.internal.util.ReflectUtil;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;


/**
 * @author Guillaume Porcher
 * Helper class to initialize DB tables from the environment.
 */
public final class DbTool {

  private DbTool() { }

  public static String generateScript(final String create, final String envXml, final String confName, final Dialect dialect)
    throws MalformedURLException, ClassNotFoundException {
    final Document doc = XmlUtil.getDocumentFromString(envXml);
    final NodeList nl = doc.getElementsByTagName("hibernate-configuration");
    final Configuration configuration = new Configuration();
    for (int i = 0; i < nl.getLength(); i++) {
      final Element confElt = (Element) nl.item(i);
      if (confName.equals(XmlUtil.attribute(confElt, "name"))) {
        final List<Element> mappingsElts = XmlUtil.elements(confElt, "mappings");
        if (mappingsElts != null) {
          for (final Element element : mappingsElts) {
            final String resourceName = XmlUtil.attribute(element, "resource");
            final Document doc2 = XmlUtil.getDocumentFromInputStream(ReflectUtil.getResourceAsStream(null, resourceName));
            final List<Element> mappingElts = XmlUtil.elements(doc2.getDocumentElement(), "mapping");
            if (mappingElts != null) {
              for (final Element elt : mappingElts) {
                DbTool.addMapping(configuration, elt);
              }
            }
          }
        }
        final List<Element> mappingElts = XmlUtil.elements(confElt, "mapping");
        if (mappingElts != null) {
          for (final Element element : mappingElts) {
            DbTool.addMapping(configuration, element);
          }
        }
      }
    }
    String[] string = null;
    if (create.equals("create")) {
      string = configuration.generateSchemaCreationScript(dialect);
    } else if (create.equals("drop")) {
      string = configuration.generateDropSchemaScript(dialect);
    } else {
      throw new IllegalArgumentException("Invalid type: " + create + ", only 'create' or 'drop' allowed");
    }
    final StringBuffer stringBuffer = new StringBuffer();
    for (final String s : string) {
      stringBuffer.append(s).append(';').append(Misc.LINE_SEPARATOR);
    }
    return stringBuffer.toString();
  }

  private static void addMapping(final Configuration configuration, final Element element)
    throws ClassNotFoundException, MalformedURLException {
    if (element.hasAttribute("resource")) {
      final String resource = element.getAttribute("resource");
      configuration.addResource(resource);
    } else if (element.hasAttribute("file")) {
      final String fileName = element.getAttribute("file");
      configuration.addFile(fileName);
    } else if (element.hasAttribute("class")) {
      final String className = element.getAttribute("class");
      configuration.addClass(Class.forName(className));
    } else if (element.hasAttribute("url")) {
      final String urlText = element.getAttribute("url");
      configuration.addURL(new URL(urlText));
    } else {
      Misc.unreachableStatement();
    }
  }



  public static void main(final String[] args) throws ClassNotFoundException, IOException {
    switch (args.length) {
      case 2:
      {
        final Dialect dialect = (Dialect) ReflectUtil.instantiate(null, args[1]);
        System.out.println(DbTool.generateScript(args[0],
            new DefaultEnvXMLGenerator(null, InstallationType.INTERNAL_JOURNAL_WITH_HISTORY).createEnvironmentXml(),
            DefaultEnvXMLGenerator.CORE_HB_CONF_NAME,
            dialect));
        break;
      }
      case 3:
      {
        final Dialect dialect = (Dialect) ReflectUtil.instantiate(null, args[1]);
        final String s = DbTool.generateScript(args[0],
            new DefaultEnvXMLGenerator(null, InstallationType.INTERNAL_JOURNAL_WITH_HISTORY).createEnvironmentXml(),
            DefaultEnvXMLGenerator.CORE_HB_CONF_NAME,
            dialect);
        Misc.write(s, new File(args[2]));
        break;
      }
      case 4:
      {
        final Dialect dialect = (Dialect) ReflectUtil.instantiate(null, args[1]);
        final String s = DbTool.generateScript(args[0],
            new DefaultEnvXMLGenerator(null, InstallationType.INTERNAL_JOURNAL_WITH_HISTORY).createEnvironmentXml(),
            args[3],
            dialect);
        Misc.write(s, new File(args[2]));
        break;
      }
      case 5:
      {
        final Dialect dialect = (Dialect) ReflectUtil.instantiate(null, args[1]);
        final String s = DbTool.generateScript(args[0],
            new String(Misc.getAllContentFrom(new File(args[4]))),
            args[3],
            dialect);
        Misc.write(s, new File(args[2]));
        break;
      }
      default:
        throw new OrchestraRuntimeException(
            "Invalid syntax: " + DbTool.class.getName()
            + " dialectClass [outputFile [hibernateConfigName [environmentFileName]]]");
    }
  }
}
