/**
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2007 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag. 
 *
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.util;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import org.ow2.orchestra.pvm.internal.xml.Parse;

/**
 * maintains the mappings between prefix names and namespace URI's throughout
 * the whole bpel process deployment. The deployer will make a NamespaceMappings
 * object available in the object stack of the {@link Parse}.
 *
 * <pre>
 * Namespaces namespaces = parse.findObject(Namespaces.class);
 * </pre>
 *
 * @author Tom Baeyens, Charles Souillard
 */
public class Namespaces {

  /** parse object in which problems must be stored. */
  private Parse parse;

  /**
   * Default constructor.
   *
   * @param parse
   *            parse
   */
  public Namespaces(Parse parse) {
    this.parse = parse;
  }

  /** maps namespace prefixes to namespace uris. */
  private Map<String, String> uris = null;

  /**
   * add a namespace prefix mapping.
   *
   * @param prefix
   *            prefix
   * @param uri
   *            uri
   */
  protected void addNamespace(String prefix, String uri) {
    if (uris == null) {
      uris = new HashMap<String, String>();
    }
    if (uris.containsKey(prefix)) {
      String oldUri = uris.get(prefix);
      if (!oldUri.equals(uri)) {
        parse.addProblem("duplicate definition of namespace prefix: " + prefix
            + ":" + oldUri + " and " + prefix + ":" + uri);
      }
    }
    uris.put(prefix, uri);
  }

  /**
   * remove a namespace prefix mapping.
   *
   * @param prefix
   *            prefix
   */
  protected void removeNamespace(String prefix) {
    if (uris != null) {
      uris.remove(prefix);
    }
  }

  /**
   * adds all given namespaces and returns the prefixes that were added.
   *
   * @param namespaces
   *            maps prefixes to namespace uris.
   * @return the set of prefixes that were added or null in case no prefixes
   *         were added.
   */
  public Set<String> addNamespaces(Map<String, String> namespaces) {
    if (namespaces == null) {
      return null;
    }
    Set<String> prefixes = new HashSet<String>();
    for (Entry<String, String> ns : namespaces.entrySet()) {
      addNamespace(ns.getKey(), ns.getValue());
      prefixes.add(ns.getKey());
    }
    return prefixes;
  }

  /**
   * removes the mappings for the given prefixes.
   *
   * @param prefixes
   *            prefixes
   */
  public void removePrefixes(Set<String> prefixes) {
    if (prefixes != null && uris != null) {
      for (String prefix : prefixes) {
        uris.remove(prefix);
      }
    }
  }

  /**
   * looks up a namespace URI for the given prefix.
   *
   * @param prefix
   *            prefix
   * @return the ns corresponding to the given prefix
   */
  public String getUri(String prefix) {
    if (uris != null) {
      return uris.get(prefix);
    }
    return null;
  }

  /**
   * extracts a QName from a piece of text in the format prefix:localPart and
   * resolves the prefix against the known prefixes in the currently known
   * namespaces in this object. public QName getQName(String text) { QName qname =
   * null; if (text!=null) { int colonIndex = text.indexOf(':'); if
   * (colonIndex==-1) { qname = new QName(text); } else { String prefix =
   * text.substring(0, colonIndex); String localPart = null; if
   * (text.length()>colonIndex+1) { localPart = text.substring(colonIndex+1); }
   * String uri = getUri(prefix);
   *
   * if (uri==null) { parse.addProblem("unknown prefix in qname "+text); } else
   * if (localPart==null) { parse.addProblem("no local part in qname "+text); }
   * else { qname = new QName(uri, localPart); } } } return qname; }
   */
}
