/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.util;

import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

import org.ow2.orchestra.definition.element.Copy;
import org.ow2.orchestra.definition.element.Correlation;
import org.ow2.orchestra.definition.element.CorrelationSet;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.runtime.CorrelationSetRuntime;
import org.ow2.orchestra.runtime.ScopeRuntime;
import org.ow2.orchestra.runtime.VariableRuntime;
import org.ow2.orchestra.services.MessageCarrier;
import org.ow2.orchestra.services.Receiver;
import org.ow2.orchestra.services.ReceivingElement;
import org.ow2.orchestra.var.MessageVariable;


/**
 * @author Guillaume Porcher
 *
 */
public final class ReceivingElementUtil {

  private ReceivingElementUtil() { }

  private static final Logger LOG = Logger.getLogger(ReceivingElementUtil.class.getName());

  /**
   * This message is to be called from the receive implementation (above) in
   * case the message arrived in the engine before the execution arrived in the
   * receive. in case the message arrives later, the receive will be invoked by
   * the correlator.
   */
  public static void receive(
      final ReceivingElement receivingElement,
      final MessageVariable message,
      final MessageCarrier messageCarrier,
      final BpelExecution execution) {
    if (ReceivingElementUtil.LOG.isLoggable(Level.FINE)) {
      ReceivingElementUtil.LOG.fine("Receiving message : " + message);
    }

    // initiate correlationSets
    ReceivingElementUtil.initiateCs(receivingElement, execution, message);
    if (receivingElement.getVariable() != null) {
      // we store the variable
      final VariableRuntime variableRuntime = execution.getVariableByName(receivingElement.getVariable());
      variableRuntime.setValue(message);
      // record variable updated
      variableRuntime.updated(execution.getActivityInstanceUUID());
    } else if (receivingElement.getInCopies() != null && !receivingElement.getInCopies().isEmpty()) {
      final ScopeRuntime scopeRuntime = execution.getCurrentScopeRuntime();
      final VariableRuntime anonymousInboundMessageVariableRuntime = scopeRuntime
          .createVariable(execution, receivingElement.getAnonymousInboundMessageVariable());

      anonymousInboundMessageVariableRuntime.setValue(message);
      for (final Copy copy : receivingElement.getInCopies()) {
        copy.execute(execution);
      }
      scopeRuntime.removeVariable(execution,
          receivingElement.getAnonymousInboundMessageVariable());
    }

    // Add replier if needed
    Receiver.addReplier(execution, receivingElement, messageCarrier);
  }

  private static void initiateCs(
      final ReceivingElement receivingElement,
      final BpelExecution bpelExecution,
      final MessageVariable incomingMessage) {
    if (receivingElement.getCorrelations() != null) {
      // we calcul the values for each correlationSet
      for (final Correlation correlation : receivingElement.getCorrelations()) {
        final CorrelationSetRuntime correlationSetRuntime =
          bpelExecution.getCorrelationSetByName(correlation.getCorrelationSet().getName());
        final QName messageType = bpelExecution.getPortType(
            receivingElement.getPortTypeQName()).getOperation(receivingElement.getOperation(), null, null)
            .getInput().getMessage().getQName();
        correlationSetRuntime.checkCSValues(bpelExecution, incomingMessage, messageType,
            correlation.getInitiate());
      }
    }
  }

  public static void initiateCs(final ReceivingElement receivingElement,
      final BpelExecution bpelExecution,
      final CorrelationSet cs,
      final MessageVariable incomingMessage) {
    if (receivingElement.getCorrelations() != null) {
      // we calcul the values for each correlationSet
      for (final Correlation correlation : receivingElement.getCorrelations()) {
        if (correlation.getCorrelationSet().equals(cs)) {
          final CorrelationSetRuntime correlationSetRuntime =
            bpelExecution.getCorrelationSetByName(correlation.getCorrelationSet().getName());
          final QName messageType = bpelExecution.getPortType(
              receivingElement.getPortTypeQName()).getOperation(receivingElement.getOperation(), null, null)
              .getInput().getMessage().getQName();
          correlationSetRuntime.checkCSValues(bpelExecution, incomingMessage, messageType,
              correlation.getInitiate());
        }
      }
    }
  }

}
