/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.var;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import javax.wsdl.Part;
import javax.xml.namespace.QName;

import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.facade.runtime.Message;
import org.ow2.orchestra.facade.runtime.impl.MessageImpl;
import org.ow2.orchestra.util.BpelXmlUtil;
import org.ow2.orchestra.util.XmlUtil;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * Models a WSDL message type variable.
 *
 * @author Goulven Le Jeune, Charles Souillard, Candy Chlad, Stanislas Giraudet
 *         De Boudemange
 */
public class MessageVariable implements Message {

  protected Map<String, Document> parts;


  /**
   * Empty message constructor. Parts can be added with
   * {@link #addPart(MessagePart)}.
   */
  public MessageVariable() {
    this.parts = new HashMap<String, Document>();
  }

  /**
   * Create a multi-part message.
   *
   * @param parts
   *            HashMap containing the message parts
   */
  public MessageVariable(final Map<String, Document> parts) {
    this.parts = parts;
  }

  /**
   * Constructor based on the wsdl Message.
   *
   * @param wsdlMessage
   *            wsdlMessage that specify this message
   */
  @SuppressWarnings("unchecked")
  public MessageVariable(final javax.wsdl.Message wsdlMessage) {
    this();
    final Map<String, Part> wsdlMessageParts = wsdlMessage.getParts();
    for (final Entry<String, Part> entry : wsdlMessageParts.entrySet()) {
      final Part part = (entry).getValue();
      final String partName = part.getName();
      // this.parts.put(partName, null);
      final QName elementQName = part.getElementName();
      final QName typeQName = part.getTypeName();
      if (typeQName != null) {
        this.parts.put(partName, BpelXmlUtil.createDefaultElement().getOwnerDocument());
      } else {
        this.parts.put(partName, BpelXmlUtil.createElement(elementQName).getOwnerDocument());
      }
    }
  }

  /* (non-Javadoc)
   * @see org.ow2.orchestra.facade.runtime.impl.Message#getPartValue(java.lang.String)
   */
  public Element getPartValue(final String partName) {
    if (partName != null) {
      if (this.parts.get(partName) != null) {
        return this.parts.get(partName).getDocumentElement();
      }
      return null;
    }
    return this.parts.values().iterator().next().getDocumentElement();
  }

  /**
   * Return true if the part having the given name is initialized.
   *
   * @param partName
   *            partName
   * @return true if the part having the given name is initialized.
   */
  public boolean isPartInitialized(final String partName) {
    return this.parts.get(partName) != null;
  }

  /* (non-Javadoc)
   * @see org.ow2.orchestra.facade.runtime.impl.Message#hasPart(java.lang.String)
   */
  public boolean hasPart(final String partName) {
    if (partName != null) {
      return this.parts.containsKey(partName);
    }
    return false;
  }

  /**
   * Return a string representation of the message.
   *
   * @return a String representation of the message
   */
  @Override
  public String toString() {
    final StringBuffer buf = new StringBuffer();
    buf.append("<message>");
    for (final String part : this.parts.keySet()) {
      if (this.parts.get(part) != null) {
        buf.append("<" + part + ">");
        if (this.parts.get(part) == null) {
          buf.append("no_value_for_this_part");
        } else {
          buf.append(XmlUtil.toString(this.getPartValue(part)));
        }
        buf.append("</" + part + ">");
      }
    }
    buf.append("</message>");
    return buf.toString();
  }


  /**
   * Export to log message.
   *
   * @return The duplicate of Xml object
   */
  public MessageImpl copy() {
    if (this.parts == null) {
      return null;
    }
    final Map<String, Document> duplicateMessageParts = new HashMap<String, Document>();
    final Set<String> partNameSet = this.parts.keySet();

    for (final String partName : partNameSet) {
      Document duplicatePart = null;
      final Element oldElement = this.getPartValue(partName);
      if (oldElement != null) {
        duplicatePart = XmlUtil.getNewDocument();
        final Node n = duplicatePart.importNode(oldElement, true);
        duplicatePart.appendChild(n);
      }
      duplicateMessageParts.put(partName, duplicatePart);
    }
    return new MessageImpl(duplicateMessageParts);
  }


  /**
   * Make a deep copy of the Xml object.
   *
   * @return The duplicate of Xml object
   */
  public MessageVariable duplicate() {
    if (this.parts == null) {
      return null;
    }
    final MessageVariable duplicateMessage = new MessageVariable();
    final Set<String> partNameSet = this.parts.keySet();

    for (final String partName : partNameSet) {
      Document duplicatePart = null;
      final Element oldElement = this.getPartValue(partName);
      if (oldElement != null) {
        duplicatePart = XmlUtil.getNewDocument();
        final Node n = duplicatePart.importNode(oldElement, true);
        duplicatePart.appendChild(n);
      }
      duplicateMessage.parts.put(partName, duplicatePart);
    }
    return duplicateMessage;
  }

  /**
   * Compare the Message to another one.
   *
   * @param anotherMessage :
   *            the other Message to compare to
   * @return true if all parts are equals, else false
   */
  @Override
  public boolean equals(final Object anotherMessage) {
    if (anotherMessage == null) {
      return false;
    }
    if (!(anotherMessage instanceof MessageVariable)) {
      return false;
    }
    final MessageVariable message = (MessageVariable) anotherMessage;
    final Map<String, Element> otherParts = message.getParts();

    return otherParts.equals(this.getParts());
  }

  /**
   * Returns a hash code value for the Message.
   *
   * @return a hash code value for this Message.
   */
  @Override
  public int hashCode() {
    int hashCode = 0;
    for (final Document messagePart : this.parts.values()) {
      hashCode += messagePart.hashCode();
    }
    return hashCode;
  }

  public Map<String, Element> getParts() {
    final Map<String, Element> res = new HashMap<String, Element>();
    if (this.parts != null) {
      for (final Map.Entry<String, Document> entry : this.parts.entrySet()) {
        if (entry.getValue() != null) {
          res.put(entry.getKey(), entry.getValue().getDocumentElement());
        } else {
          res.put(entry.getKey(), null);
        }
      }
    }
    return res;
  }

  public void setPart(final String partName, final Object value, final boolean keepSrcElementName) {
    if (!this.hasPart(partName)) {
      throw new OrchestraRuntimeException("Invalid part name: " + partName);
    }
    final Element currentValue = this.getPartValue(partName);
    BpelXmlUtil.setObjectValue(currentValue, value, keepSrcElementName);
  }
}
