/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 *
 * WSDeployer.java
 *
 * @author Goulven Le Jeune, Charles Souillard, Candy Chlad, Stanislas Giraudet De Boudemange
 *
 * Created on : Mar 1, 2007
 *
 * $Id: WSDeployer.java 2955 2009-09-10 07:26:24Z porcherg $
 **/
package org.ow2.orchestra.ws;

import com.ibm.wsdl.extensions.soap.SOAPConstants;
import com.ibm.wsdl.extensions.soap12.SOAP12Constants;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Properties;
import java.util.logging.Logger;

import javax.wsdl.Binding;
import javax.wsdl.BindingFault;
import javax.wsdl.BindingInput;
import javax.wsdl.BindingOperation;
import javax.wsdl.BindingOutput;
import javax.wsdl.Fault;
import javax.wsdl.Input;
import javax.wsdl.Operation;
import javax.wsdl.OperationType;
import javax.wsdl.Output;
import javax.wsdl.Port;
import javax.wsdl.Service;
import javax.wsdl.extensions.ElementExtensible;
import javax.wsdl.extensions.ExtensibilityElement;
import javax.wsdl.extensions.soap.SOAPAddress;
import javax.wsdl.extensions.soap.SOAPBinding;
import javax.wsdl.extensions.soap.SOAPBody;
import javax.wsdl.extensions.soap.SOAPFault;
import javax.wsdl.extensions.soap.SOAPOperation;
import javax.wsdl.extensions.soap12.SOAP12Address;
import javax.wsdl.extensions.soap12.SOAP12Binding;
import javax.wsdl.extensions.soap12.SOAP12Body;
import javax.wsdl.extensions.soap12.SOAP12Fault;
import javax.wsdl.extensions.soap12.SOAP12Operation;
import javax.xml.namespace.QName;

import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.pvm.env.Environment;
import org.ow2.orchestra.pvm.internal.xml.Problem;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.util.OrchestraConstants;




/**
 *
 * This class will deploy a WebService on the choosen ws engine.
 */
public abstract class WSDeployer {

  private static Logger log = Logger.getLogger(WSDeployer.class.getName());

  /**
   * Ns of wsdl soap 1.1.
   */
  public static final String URI_WSDL11_SOAP11 = SOAPConstants.NS_URI_SOAP;

  /**
   * Ns of wsdl soap 1.2.
   */
  public static final String URI_WSDL11_SOAP12 = SOAP12Constants.NS_URI_SOAP12;

  /**
   * Default value for soap binding style.
   */
  public static final String DEFAULT_SOAP_BINDING_STYLE = "document";

  /**
   * soap over http uri used in soap binding.
   */
  public static final String SOAP_HTTP_TRANSPORT_URI = "http://schemas.xmlsoap.org/soap/http";

  /**
   * List of problems found in the wsdl file.
   */
  private final List<Problem> problems;

  /**
   * A string which identify this ws : it could be the processName.
   */
  protected QName processQName;

  protected final Properties orchestraProperties;

  protected final String expectedAddressPrefix;

  /**
   * Default constructor.
   *
   * @param wsId -
   *            wsId
   * @param orchestraDirectory
   *            Absolute path to the orchestra directory.
   */
  protected WSDeployer(final QName processQName) {
    this.processQName = processQName;
    this.problems = new ArrayList<Problem>();
    this.orchestraProperties = (Properties) Environment.getFromCurrent(OrchestraConstants.ORCHESTRA_PROPERTIES);
    this.expectedAddressPrefix = "http://" + this.orchestraProperties.getProperty(OrchestraConstants.SERVLET_HOST)
      + ":" + this.orchestraProperties.getProperty(OrchestraConstants.SERVLET_PORT)
      + "/" + this.orchestraProperties.getProperty(OrchestraConstants.SERVLET_PATH)
      + "/";
  }

  /**
   * Deploys the Web Services classes of the wsdl representing the interface of
   * a BPEL process.
   *
   * @param wsdlDef
   *            the wsdl definition of the process
   *
   * @return true, if deployed
   */
  public void deploy(final List<Service> services) {
    this.checkSupport(services);

    if (this.problems.size() > 0) {
      final String problems = this.printProblems();
      WSDeployer.log.severe(problems);
      throw new OrchestraRuntimeException(problems);
    } else if (services.size() == 0) {
      /*
       * throw new Exception( "No Service having at least one port defining a
       * soap adress was found in the top level wsdl file. " + "To be
       * compliant with WSI BAsic Profile, used services must be declared in
       * the top level wsdl file.");
       */
      throw new OrchestraRuntimeException("No service to deploy.");
    } else {

      this.checkWSEngineIsAvailable();
      // deploys all services
      this.deployServices(services);

    }
  }

  /**
   * This method will undeploy the ws from the choosen ws engine.
   *
   * @param wsdlDef -
   *            Definition object that represents the ws to undeploy.
   * @return true, if undeploy
   */
  public boolean undeploy(final List<Service> services) {
    this.checkSupport(services);
    if (this.problems.size() == 0) {
      this.checkWSEngineIsAvailable();
      // undeploys all services
      this.undeployServices(services);
    } else {
      WSDeployer.log.severe(this.printProblems());
      return false;
    }
    return true;
  }

  /**
   * Print problems list.
   *
   */
  protected String printProblems() {
    final StringBuffer buf = new StringBuffer();
    buf.append("Problems while deploying web services:").append(Misc.LINE_SEPARATOR);
    for (final Problem problem : this.problems) {
      buf.append("- ").append(problem).append(Misc.LINE_SEPARATOR);
    }
    return buf.toString();
  }

  /**
   * Utility method to add a problem.
   */
  protected void addProblem(final Binding binding, final String mess,
      final String severity) {
    if (mess != null) {
      this.addProblem("The following problem was found on binding : "
          + binding.getQName() + "\n" + mess, severity);
    }
  }

  /**
   * Utility method to add a problem.
   */
  protected void addProblem(final Binding binding,
      final BindingOperation bindingOperation, final String mess,
      final String severity) {
    if (mess != null) {
      final StringBuffer stringBuffer = new StringBuffer();
      stringBuffer.append("The following problem was found on binding : "
          + binding.getQName());
      stringBuffer.append("\nThe problem was found in operation : "
          + bindingOperation.getName());
      stringBuffer.append("\n");
      stringBuffer.append(mess);
      this.addProblem(stringBuffer.toString(), severity);
    }
  }

  /**
   * Utility method to add a problem.
   */
  protected void addProblem(final Binding binding, final Operation operation,
      final String mess, final String severity) {
    if (mess != null) {
      final StringBuffer stringBuffer = new StringBuffer();
      stringBuffer.append("The following problem was found on binding : "
          + binding.getQName());
      stringBuffer.append("\nThe problem was found in operation : "
          + operation.getName());
      stringBuffer.append("\n");
      stringBuffer.append(mess);
      this.addProblem(stringBuffer.toString(), severity);
    }
  }

  /**
   * Utility method to add a problem.
   */
  protected void addProblem(final String mess, final String severity) {
    this.problems.add(new Problem(mess, null, severity));
  }

  /**
   * This method will deploy the specified WS Service on the choosen ws engine.
   * This method must be overriden by each ws engine Deployer.
   *
   * @param def -
   *            Definition object that represents a WS Service to deploy.
   */
  protected abstract void deployServices(final List<Service> services);

  /**
   * This method will undeploy the specified WS Service from the choosen ws
   * engine. This method must be overriden by each ws engine Deployer.
   *
   * @param def -
   *            Definition object that represents a WS Service to undeploy.
   */
  protected abstract void undeployServices(final List<Service> services);

  /**
   * This method will add a problem if Ws Engine is not Available. This method
   * must be overriden by each ws engine Deployer.
   */
  protected abstract void checkWSEngineIsAvailable();

  /**
   * This method will add a problem if Ws Engine does not support the specified
   * operationType. This method must be overriden by each ws engine Deployer.
   *
   * @param operationType -
   *            operationType to check
   */
  protected abstract String checkOperationType(final OperationType operationType);

  /**
   * This method will add a problem if Ws Engine does not support this soap
   * binding version : soap 1.1 or soap 1.2 for example. This method must be
   * overriden by each ws engine Deployer.
   *
   * @param soapBinding -
   *            soapBinding to check
   */
  protected abstract String checkSoapVersion(
      final ExtensibilityElement soapBinding);

  /**
   * This method will add a problem if Ws Engine does not support this soap
   * binding transport. This method must be overriden by each ws engine
   * Deployer.
   *
   * @param soapBinding -
   *            soapBinding to check
   */
  protected abstract String checkTransport(
      final ExtensibilityElement soapBinding);

  /**
   * This method will add a problem if Ws Engine does not support this style
   * (rpc/document). This method must be overriden by each ws engine Deployer.
   *
   * @param soapOperation -
   *            soapOperation to check
   */
  protected abstract String checkOperationStyle(final String operationStyle);

  /**
   * This method will add a problem if Ws Engine does not support this use
   * (encoded/literal). This method must be overriden by each ws engine
   * Deployer.
   *
   * @param soapBody -
   *            soapBody to check
   */
  protected abstract String checkUse(final ExtensibilityElement soapBody);

  /**
   * This method will add a problem if Ws Engine does not support an attribute
   * of the soapBody. For examples, parts attribute. This method must be
   * overriden by each ws engine Deployer.
   *
   * @param soapBody -
   *            soapBody to check
   */
  protected abstract String checkSoapBody(final ExtensibilityElement soapBody);

  protected abstract String checkSoapFault(final ExtensibilityElement soapFault);

  @SuppressWarnings("unchecked")
  public List<Service> checkSupport(final List<Service> services) {
    final List<Service> soapServices = new ArrayList<Service>();
    for (final Service service : services) {
      // get all ports of this service
      final Collection<Port> ports = service.getPorts().values();
      for (final Port port : ports) {
        // if port has a soap address, then it could be a correct one
        final String address = this.getAddressFromPort(port);
        if (address != null) {
          if (!address.startsWith(this.expectedAddressPrefix)) {
            this.addProblem("Port " + port.getName()
                + " of service " + service.getQName()
                + " specifies an unexpected SOAP address " + address + "." + Misc.LINE_SEPARATOR
                + " SOAP address must start with " + this.expectedAddressPrefix + "."
                + " Please check the wsdl file and orchestra settings.", Problem.SEVERITY_ERROR);
          }
          soapServices.add(service);
        }
        final Binding binding = port.getBinding();
        this.checkBinding(binding);
      }
    }
    return soapServices;
  }

  @SuppressWarnings("unchecked")
  protected void checkBinding(final Binding binding) {
    // check that this binding is a soap binding
    // if yes, check soapversion used and transport
    final ExtensibilityElement soapBinding = WSDeployer.getSoapBindingFromBinding(binding);
    if (soapBinding == null) {
      this.addProblem(binding, "this binding is not a soap binding.",
          Problem.SEVERITY_ERROR);
    } else {
      String mess = this.checkSoapVersion(soapBinding);
      this.addProblem(binding, mess, Problem.SEVERITY_ERROR);
      mess = this.checkTransport(soapBinding);
      this.addProblem(binding, mess, Problem.SEVERITY_ERROR);

      String soapBindingStyle = WSDeployer.getSOAPStyle(soapBinding);
      if (soapBindingStyle == null) {
        soapBindingStyle = WSDeployer.DEFAULT_SOAP_BINDING_STYLE;
      }
      // check operations
      final List<BindingOperation> operations = binding
          .getBindingOperations();
      for (final BindingOperation bindingOperation : operations) {
        this.checkOperation(binding, bindingOperation, soapBindingStyle);
      }
    }
  }

  @SuppressWarnings("unchecked")
  protected void checkOperation(final Binding binding,
      final BindingOperation bindingOperation, final String soapBindingStyle) {
    final Operation operation = bindingOperation.getOperation();
    // check : operationType, style, use
    final OperationType operationType = operation.getStyle();
    String mess = this.checkOperationType(operationType);
    this.addProblem(binding, operation, mess, Problem.SEVERITY_ERROR);
    final ExtensibilityElement soapOperation = WSDeployer.getSoapOperationFromBindingOperation(bindingOperation);
    String operationStyle = soapBindingStyle;
    if (soapOperation != null) {
      final String operationStyle2 = WSDeployer.getSOAPStyle(soapOperation);
      if (operationStyle2 != null) {
        operationStyle = operationStyle2;
      }
    }
    if (operationStyle.equals("document")) {
      this.verifyThereIsOnlyOnePart(binding, operation);
      /*
       * Part part = (Part)
       * operation.getInput().getMessage().getParts().values().iterator().next();
       * String partName = part.getName(); QName elementQName =
       * part.getElementName(); QName typeQName = part.getTypeName(); if
       * (typeQName != null) { addProblem(binding, operation, "this operation is
       * of style document and its part is using " + "xmlType attribute :
       * forbidden", Problem.SEVERITY_FATALERROR); } if (elementQName == null) {
       * addProblem(binding, operation, "this operation is of style document and
       * its part is not using " + "element attribute : forbidden",
       * Problem.SEVERITY_FATALERROR); }
       */
    }
    mess = this.checkOperationStyle(operationStyle);
    this.addProblem(binding, bindingOperation, mess, Problem.SEVERITY_ERROR);
    final BindingInput input = bindingOperation.getBindingInput();
    if (input != null) {
      final ExtensibilityElement soapBody = this.getSoapBodyFromElement(input);
      this.verifySoapBody(binding, bindingOperation, soapBody);
    }
    final BindingOutput output = bindingOperation.getBindingOutput();
    if (output != null) {
      final ExtensibilityElement soapBody = this.getSoapBodyFromElement(output);
      this.verifySoapBody(binding, bindingOperation, soapBody);
    }
    final Collection<BindingFault> faults = bindingOperation
        .getBindingFaults().values();
    for (final BindingFault fault : faults) {
      final ExtensibilityElement soapBody = this.getSoapFaultFromElement(fault);
      this.verifySoapFault(binding, bindingOperation, soapBody);
    }
  }

  /**
   * For this soapBody, if it exists, check the use and verify that all defined
   * attributes are supported.
   *
   * @param binding -
   *            binding in which the soapBody is declared
   * @param bindingOperation -
   *            bindingOperation in which the soapBody is declared
   * @param soapBody -
   *            soapBody to analyse
   */
  protected void verifySoapBody(final Binding binding,
      final BindingOperation bindingOperation,
      final ExtensibilityElement soapBody) {
    String mess;
    if (soapBody != null) {
      mess = this.checkUse(soapBody);
      this.addProblem(binding, bindingOperation, mess, Problem.SEVERITY_ERROR);
      mess = this.checkSoapBody(soapBody);
      this.addProblem(binding, bindingOperation, mess, Problem.SEVERITY_ERROR);
    } else {
      mess = "There is a missing soapBody";
      this.addProblem(binding, bindingOperation, mess, Problem.SEVERITY_ERROR);
    }
  }

  protected void verifySoapFault(final Binding binding,
      final BindingOperation bindingOperation,
      final ExtensibilityElement soapFault) {
    String mess;
    if (soapFault != null) {
      mess = this.checkUse(soapFault);
      this.addProblem(binding, bindingOperation, mess, Problem.SEVERITY_ERROR);
      mess = this.checkSoapFault(soapFault);
      this.addProblem(binding, bindingOperation, mess, Problem.SEVERITY_ERROR);
    } else {
      mess = "There is a missing soapFault";
      this.addProblem(binding, bindingOperation, mess, Problem.SEVERITY_ERROR);
    }
  }

  /**
   * This method must be called only if the given operation is defined in
   * document style. In this case, only one part can be defined. Else it is not
   * supported.
   *
   * @param binding -
   *            binding in which the operation is declared
   * @param operation -
   *            operation in which verifications will be performed
   */
  @SuppressWarnings("unchecked")
  protected void verifyThereIsOnlyOnePart(final Binding binding,
      final Operation operation) {
    int nbOfParts;
    final Input input = operation.getInput();
    if (input != null) {
      nbOfParts = input.getMessage().getParts().size();
      if (nbOfParts > 1) {
        this.addProblem(binding, operation,
            "This operation is used in document style and is using an input message "
                + "containing more than one part", Problem.SEVERITY_ERROR);
      }
    }

    final Output output = operation.getOutput();
    if (output != null) {
      nbOfParts = output.getMessage().getParts().size();
      if (nbOfParts > 1) {
        this.addProblem(
            binding,
            operation,
            "This operation is used in document style and is using an output message containing more "
                + "than one part", Problem.SEVERITY_ERROR);
      }
    }

    final Collection<Fault> faults = operation.getFaults().values();
    for (final Fault fault : faults) {
      nbOfParts = fault.getMessage().getParts().size();
      if (nbOfParts > 1) {
        this.addProblem(
            binding,
            operation,
            "This operation is used in document style and is using a fault message containing more "
                + "than one part", Problem.SEVERITY_ERROR);
      }
    }
  }

  @SuppressWarnings("unchecked")
  public static <T> T getFromList(final List< ? > list,
      final Class< ? extends T>... types) {
    for (final Object o : list) {
      for (final Class< ? extends T> clazz : types) {
        if (clazz.isAssignableFrom(o.getClass())) {
          return (T) o;
        }
      }
    }
    return null;
  }

  @SuppressWarnings("unchecked")
  protected String getAddressFromPort(final Port p) {
    // Get the endpoint for a port
    final List<ExtensibilityElement> extensibilityList = p.getExtensibilityElements();
    for (final ExtensibilityElement obj : extensibilityList) {
      if (obj instanceof SOAPAddress) {
        return ((SOAPAddress) obj).getLocationURI();
      } else if (obj instanceof SOAP12Address) {
        return ((SOAP12Address) obj).getLocationURI();
      }
    }
    // didn't find it
    return null;
  }

  @SuppressWarnings("unchecked")
  /**
   * @return the SOAPBinding or SOAP12Binding element for the given Binding.
   */
  public static ExtensibilityElement getSoapBindingFromBinding(
      final Binding binding) {
    return WSDeployer.getFromList(binding.getExtensibilityElements(),
        SOAPBinding.class, SOAP12Binding.class);
  }

  @SuppressWarnings("unchecked")
  public static ExtensibilityElement getSoapOperationFromBindingOperation(
      final BindingOperation bindingOperation) {
    return WSDeployer
        .getFromList(bindingOperation.getExtensibilityElements(),
            SOAPOperation.class, SOAP12Operation.class);
  }

  public static String getSOAPStyle(final ExtensibilityElement element) {
    if (element instanceof SOAPBinding) {
      return ((SOAPBinding) element).getStyle();
    } else if (element instanceof SOAP12Binding) {
      return ((SOAP12Binding) element).getStyle();
    } else if (element instanceof SOAPOperation) {
      return ((SOAPOperation) element).getStyle();
    } else if (element instanceof SOAP12Operation) {
      return ((SOAP12Operation) element).getStyle();
    }
    return null;
  }

  @SuppressWarnings("unchecked")
  protected ExtensibilityElement getSoapBodyFromElement(
      final ElementExtensible element) {
    return WSDeployer.getFromList(element.getExtensibilityElements(),
        SOAPBody.class, SOAP12Body.class);
  }

  @SuppressWarnings("unchecked")
  protected ExtensibilityElement getSoapFaultFromElement(
      final ElementExtensible element) {
    return WSDeployer.getFromList(element.getExtensibilityElements(),
        SOAPFault.class, SOAP12Fault.class);
  }

  /**
   * @return the processQName
   */
  public QName getProcessQName() {
    return this.processQName;
  }

  /**
   * @param processQName
   *            the processQName to set
   */
  public void setProcessQName(final QName processQName) {
    this.processQName = processQName;
  }

}
