package org.ow2.orchestra.wsdl;

import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.wsdl.Definition;
import javax.wsdl.Import;
import javax.wsdl.Message;
import javax.wsdl.extensions.ExtensibilityElement;
import javax.wsdl.extensions.UnknownExtensibilityElement;
import javax.xml.namespace.QName;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.lang.PropertyQuery;
import org.ow2.orchestra.pvm.env.Environment;
import org.ow2.orchestra.pvm.internal.xml.Parse;
import org.ow2.orchestra.util.BpelUtil;
import org.ow2.orchestra.util.Namespaces;
import org.ow2.orchestra.util.XmlConstants;
import org.ow2.orchestra.util.XmlUtil;
import org.ow2.orchestra.util.wsdl.WsdlUtil;
import org.w3c.dom.Element;

/**
 * parses the wsdl4j structures and translates them into the BpelProcess model.
 *
 * @author Tom Baeyens
 */
public class BpelWsdlParser {

  /** log. */
  private static Logger log = Logger.getLogger(BpelWsdlParser.class.getName());

  /**
   *
   * @param documentBaseUri
   *            documentBaseUri
   * @param wsdlInputSource
   *            wsdl inputSource to analyse
   * @param parse
   *            parse object : use it to add problems
   * @param processQueryLanguage
   *            queryLanguage attribute value of the process element
   * @param styleSheetRepository
   *            styleSheetRepository
   * @return a WsdlInfos object that contains all necessary wsdl details
   */
  @SuppressWarnings("unchecked")
  public Definition parse(final URL wsdlURL, final Parse parse,
      final String processQueryLanguage, final BpelProcess bpelProcess, final WsdlsInfos wsdlInfos) {
    final Definition wsdlDefinition = WsdlUtil.readWsdl(wsdlURL);
    return this.parse(wsdlDefinition, parse, processQueryLanguage, bpelProcess,
        wsdlInfos);
  }

  public Definition parse(final Definition wsdlDefinition, final Parse parse,
      final String processQueryLanguage, final BpelProcess bpelProcess, final WsdlsInfos wsdlInfos) {
    try {
      final Namespaces namespaces = parse.findObject(Namespaces.class);
      final Set<String> addedPrefixes = namespaces.addNamespaces(wsdlDefinition
          .getNamespaces());

      final List<Definition> allDefinitions = new ArrayList<Definition>();
      this.retriveAllDefinitions(wsdlDefinition, allDefinitions);

      try {

        // add the wsdl to the process definition
        wsdlInfos.addWsdlDefinition(wsdlDefinition);

        // extract the partner link types from the extensibility elements.
        final List<ExtensibilityElement> extensibilityElements = this.getAllExtensibilityElements(allDefinitions);
        if (extensibilityElements != null) {
          for (final ExtensibilityElement extensibilityElement : extensibilityElements) {
            if (extensibilityElement instanceof UnknownExtensibilityElement) {
              final UnknownExtensibilityElement unknownExtensibilityElement = (UnknownExtensibilityElement) extensibilityElement;
              final QName elementQName = extensibilityElement.getElementType();

              if (this.isPartnerLinkType(elementQName)) {
                final Element partnerLintTypeElement = unknownExtensibilityElement
                    .getElement();
                final PartnerLinkType partnerLinkType = this.createPartnerLinkType(
                    partnerLintTypeElement,
                    wsdlDefinition.getTargetNamespace(), parse);
                wsdlInfos.addPartnerLinkType(partnerLinkType);
              } else if (this.isVarProperty(elementQName)) {
                this.addProperty(wsdlDefinition.getTargetNamespace(),
                    unknownExtensibilityElement,
                    wsdlInfos);
              } else if (this.isVarPropertyAlias(elementQName)) {
                this.addPropertyAlias(wsdlDefinition.getTargetNamespace(),
                    unknownExtensibilityElement,
                    processQueryLanguage, wsdlInfos, bpelProcess, parse
                        .findObject(Environment.class));
              } else if (this.isServiceRef(elementQName)) {
                // TODO : manage service-ref
                throw new OrchestraRuntimeException("serviceRef not yet implemented !");
              }
            } else {
              if (BpelWsdlParser.log.isLoggable(Level.FINE)) {
                BpelWsdlParser.log.fine("unknown extensibility element: "
                    + extensibilityElement);
              }
            }
          }
        }
      } finally {
        namespaces.removePrefixes(addedPrefixes);
      }
    } catch (final Exception e) {
      parse.addProblem("couldn't parse wsdl "
          + wsdlDefinition.getDocumentBaseURI(), e);
    }
    return wsdlDefinition;
  }

  /**
   * Add a property defined by the given prop parameter to the given wsdlInfos.
   *
   * @param targetNamespace
   *            tns of the property
   * @param prop
   *            prop element to analyse
   * @param wsdlInfos
   *            wsdlInfos on which the property will be added
   */
  private void addProperty(final String targetNamespace,
      final UnknownExtensibilityElement prop, final WsdlsInfos wsdlInfos) {
    final Element propertyElement = prop.getElement();
    final String name = propertyElement
        .getAttribute(XmlConstants.VARPROP_PROPERTY_NAME);
    final Property property = new Property(new QName(targetNamespace, name));
    property.setType(XmlUtil.attributeQName(propertyElement,
        XmlConstants.VARPROP_PROPERTY_TYPE));
    property.setElement(XmlUtil.attributeQName(propertyElement,
        XmlConstants.VARPROP_PROPERTY_ELEMENT));
    wsdlInfos.addProperty(property);
  }

  /**
   * Add a propertyAlias defined by the given propAlias parameter to the given
   * wsdlInfos.
   *
   * @param targetNamespace
   *            tns of the property
   * @param propAlias
   *            propAlias element to analyse
   * @param processQueryLanguage
   *            queryLanguage attribute set on the process element
   * @param wsdlInfos
   *            wsdlInfos on which the propertyAlias will be added
   * @param styleSheetRepository
   *            styleSheetRepository
   * @param environment
   *            environment
   */
  private void addPropertyAlias(final String targetNamespace,
      final UnknownExtensibilityElement propAlias, final String processQueryLanguage,
      final WsdlsInfos wsdlInfos, final BpelProcess bpelProcess, final Environment environment) {
    final Element propertyAliasElement = propAlias.getElement();
    final PropertyAlias propertyAlias = new PropertyAlias(XmlUtil.attributeQName(
        propertyAliasElement, XmlConstants.VARPROP_PROPERTYALIAS_PROPNAME));
    propertyAlias.setMessageType(XmlUtil.attributeQName(propertyAliasElement,
        XmlConstants.VARPROP_PROPERTYALIAS_MESSAGETYPE));
    final String partAttributeValue = propertyAliasElement
        .getAttribute(XmlConstants.VARPROP_PROPERTYALIAS_PART);
    if (!"".equals(partAttributeValue)) {
      propertyAlias.setPart(partAttributeValue);
    }
    propertyAlias.setType(XmlUtil.attributeQName(propertyAliasElement,
        XmlConstants.VARPROP_PROPERTYALIAS_TYPE));
    propertyAlias.setElement(XmlUtil.attributeQName(propertyAliasElement,
        XmlConstants.VARPROP_PROPERTY_ELEMENT));

    final List<Element> queries = XmlUtil.elements(propertyAliasElement,
        XmlConstants.XMLNS_BPEL_2_0_VARPROP,
        XmlConstants.VARPROP_PROPERTYALIAS_QUERY);
    if (queries.size() != 0) {
      final Element queryElement = queries.get(0);

      String queryLanguage = XmlUtil.attribute(queryElement,
          XmlConstants.VARPROP_PROPERTYALIAS_QUERY_LANGUAGE);
      if (queryLanguage == null) {
        queryLanguage = processQueryLanguage;
      }
      /*
       * Query query = new Query(bpelProcess,
       * BpelUtil.getAllNameSpaces(queryElement), queryElement.getTextContent(),
       * null, null, queryLanguage); propertyAlias.setQuery(query);
       */
      final PropertyQuery propertyQuery = new PropertyQuery();
      propertyQuery.setLanguage(queryLanguage);
      propertyQuery.setNamespaces(BpelUtil.getAllNameSpaces(queryElement));
      propertyQuery.setText(queryElement.getTextContent());
      propertyAlias.setQuery(propertyQuery);
    }
    wsdlInfos.addPropertyAlias(propertyAlias);
  }

  /**
   * return true if the given qn is conformant with a bpel plnk.
   *
   * @param qn
   *            qname to analyse
   * @return true if the given qn is conformant with a bpel plnk.
   */
  protected boolean isPartnerLinkType(final QName qn) {
    return XmlConstants.XMLNS_BPEL_2_0_PLNKTYPE.equals(qn.getNamespaceURI())
        && qn.getLocalPart().equals("partnerLinkType");
  }

  /**
   * return true if the given qn is conformant with a bpel serviceRef.
   *
   * @param qn
   *            qname to analyse
   * @return true if the given qn is conformant with a bpel serviceRef.
   */
  protected boolean isServiceRef(final QName qn) {
    return XmlConstants.XMLNS_BPEL_2_0_SERVICEREF.equals(qn.getNamespaceURI())
        && qn.getLocalPart().equals("service-ref");
  }

  /**
   * return true if the given qn is conformant with a bpel property.
   *
   * @param qn
   *            qname to analyse
   * @return true if the given qn is conformant with a bpel property.
   */
  protected boolean isVarProperty(final QName qn) {
    return XmlConstants.XMLNS_BPEL_2_0_VARPROP.equals(qn.getNamespaceURI())
        && qn.getLocalPart().equals("property");
  }

  /**
   * return true if the given qn is conformant with a bpel propertyAlias.
   *
   * @param qn
   *            qname to analyse
   * @return true if the given qn is conformant with a bpel propertyAlias.
   */
  protected boolean isVarPropertyAlias(final QName qn) {
    return XmlConstants.XMLNS_BPEL_2_0_VARPROP.equals(qn.getNamespaceURI())
        && qn.getLocalPart().equals("propertyAlias");
  }

  /**
   * Create a new PartnerLinkType base on the given partnerLinkTypeElement.
   *
   * @param partnerLinkTypeElement
   *            dom element to analyse
   * @param targetNamespace
   *            targetNamespace in which the plnk is defined
   * @param parse
   *            parse object to use to add problems
   * @return the created PartnerLinkType
   */
  private PartnerLinkType createPartnerLinkType(final Element partnerLinkTypeElement,
      final String targetNamespace, final Parse parse) {
    final String name = partnerLinkTypeElement.getAttribute("name");
    final QName partnerLinkTypeName = new QName(targetNamespace, name);
    final PartnerLinkType partnerLinkType = new PartnerLinkType(partnerLinkTypeName);
    final List<Element> roleElements = XmlUtil.elements(partnerLinkTypeElement,
        "role");
    if (roleElements != null) {
      for (final Element roleElement : roleElements) {
        final String roleName = roleElement.getAttribute("name");
        try {
          // NOTE it seems that the values in the WSDL have to be resolved
          // programmatically with the
          // javax.wsdl.Definition.getNamespaces().
          //
          // Regrettably, this is a different technique then the one in parsing
          // the bpel document, where
          // it is necessary to first set the namespace awareness to true and
          // then use the
          // org.w3c.dom.Element.lookupNamespaceURI(prefix);

          final XmlUtil.NamespaceValue portTypeValue = XmlUtil.attributeNamespaceValue(roleElement, "portType");
          final String uri = roleElement.lookupNamespaceURI(portTypeValue.getPrefix());
          final QName portTypeQName = new QName(uri, portTypeValue.getLocalPart());
          partnerLinkType.addRole(roleName, portTypeQName);
        } catch (final OrchestraRuntimeException e) {
          parse.addProblem(e.getMessage(), e);
        }
      }
    }
    return partnerLinkType;
  }

  private void retriveAllDefinitions(final Definition def,
      final List<Definition> allDefinitions) {
    allDefinitions.add(def);
    final Map<String, List<Import>> importsMap = def.getImports();
    if (importsMap != null) {
      final Collection<List<Import>> importLists = importsMap.values();
      if (importLists != null) {
        for (final List<Import> importList : importLists) {
          if (importList != null) {
            for (final Import i : importList) {
              allDefinitions.add(i.getDefinition());
              this.retriveAllDefinitions(i.getDefinition(), allDefinitions);
            }
          }
        }
      }
    }
  }

  private Map<QName, Message> getAllMessages(final List<Definition> definitions) {
    final Map<QName, Message> messages = new HashMap<QName, Message>();
    for (final Definition definition : definitions) {
      messages.putAll(definition.getMessages());
    }
    return messages;
  }

  private List<ExtensibilityElement> getAllExtensibilityElements(
      final List<Definition> definitions) {
    final List<ExtensibilityElement> extensibilityElements = new ArrayList<ExtensibilityElement>();
    for (final Definition definition : definitions) {
      extensibilityElements.addAll(definition.getExtensibilityElements());
    }
    return extensibilityElements;
  }
}
