/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.wsdl;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.wsdl.Binding;
import javax.wsdl.Definition;
import javax.wsdl.Fault;
import javax.wsdl.Input;
import javax.wsdl.Message;
import javax.wsdl.Operation;
import javax.wsdl.Output;
import javax.wsdl.Port;
import javax.wsdl.PortType;
import javax.wsdl.Service;
import javax.xml.namespace.QName;

import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.util.wsdl.WsdlUtil;

/**
 *
 * @author Charles Souillard
 */
public class WsdlsInfos implements Serializable {

  /**   */
  private static final long serialVersionUID = 2688615599128088745L;

  protected long dbid;

  protected Set<Definition> wsdlDefinitions = new HashSet<Definition>();

  /** partnerLinkTypes of the corresponding wsdl. */
  protected Map<QName, PartnerLinkType> partnerLinkTypes;

  /** properties of the corresponding wsdl. */
  protected Map<QName, Property> properties;

  /** propertyAliases of the corresponding wsdl. */
  protected List<PropertyAlias> propertyAliases;

  /**
   * Default constructor.
   */
  public WsdlsInfos() {
  }

  /**
   * Add a PartnerLinkType.
   *
   * @param partnerLinkType
   *            partnerLinkType to add
   */
  public void addPartnerLinkType(final PartnerLinkType partnerLinkType) {
    if (this.partnerLinkTypes == null) {
      this.partnerLinkTypes = new HashMap<QName, PartnerLinkType>();
    }
    this.partnerLinkTypes.put(partnerLinkType.getName(), partnerLinkType);
  }

  /**
   * Return the partnerLinkType having the given QName.
   *
   * @param partnerLinkTypeName
   *            partnerLinkTypeName
   * @return the partnerLinkType having the given QName.
   */
  public PartnerLinkType getPartnerLinkType(final QName partnerLinkTypeName) {
    if (this.partnerLinkTypes != null) {
      return this.partnerLinkTypes.get(partnerLinkTypeName);
    }
    return null;
  }

  public void addWsdlDefinition(final Definition wsdlDefinitionToAdd) {
    this.wsdlDefinitions.add(wsdlDefinitionToAdd);
  }

  /**
   * Return the portType having the given QName.
   *
   * @param portTypeQName
   *            portTypeName
   * @return the portType having the given QName.
   */
  public PortType getPortType(final QName portTypeQName) {
    PortType pt = null;
    for (final Definition wsdlDefinition : this.wsdlDefinitions) {
      pt = wsdlDefinition.getPortType(portTypeQName);
      if (pt != null) {
        break;
      }
    }
    return pt;
  }

  /**
   * Return the message having the given QName.
   *
   * @param messageName
   *            messageName
   * @return the message having the given QName.
   */
  public Message getMessage(final QName messageName) {
    Message msg = null;
    for (final Definition wsdlDefinition : this.wsdlDefinitions) {
      msg = wsdlDefinition.getMessage(messageName);
      if (msg != null) {
        break;
      }
    }
    return msg;
  }

  /**
   * Return the service having the given QName.
   *
   * @param serviceName
   *            serviceName
   * @return the service having the given QName.
   */
  public Service getService(final QName serviceName) {
    Service service = null;
    for (final Definition wsdlDefinition : this.wsdlDefinitions) {
      service = wsdlDefinition.getService(serviceName);
      if (service != null) {
        break;
      }
    }
    return service;
  }

  /**
   * Return the service that contains the given port.
   *
   * @param port
   *            port
   * @return the service that contains the given port.
   */
  public Service getServiceOfPort(final Port port) {
    final String portName = port.getName();
    for (final Definition wsdlDefinition : this.wsdlDefinitions) {
      for (final Service service : (Collection<Service>) wsdlDefinition.getServices().values()) {
        if (service.getPort(portName) != null) {
          return service;
        }
      }
    }
    return null;
  }

  /**
   * Return the service that contains the given port.
   *
   * @param port
   *            port
   * @return the service that contains the given port.
   */
  public List<Service> getServicesOfPortType(final QName portTypeQName) {
    final List<Service> matchinServices = WsdlUtil.getServicesOfPortType(portTypeQName, this.wsdlDefinitions);
    if (matchinServices.size() > 0) {
      return matchinServices;
    }
    return null;
  }

  /**
   * Return all ports implementing the given PortType.
   *
   * @param portTypeQName
   *            portType
   * @return all ports implementing the given PortType
   */
  @SuppressWarnings("unchecked")
  public List<Port> getPorts(final QName portTypeQName) {
    final List<Port> ports = new ArrayList<Port>();
    for (final Definition wsdlDefinition : this.wsdlDefinitions) {
      for (final Service service : (Collection<Service>) wsdlDefinition.getServices().values()) {
        final Map<QName, Port> servicePorts = service.getPorts();
        for (final Port port : servicePorts.values()) {
          if (port.getBinding().getPortType().getQName().equals(portTypeQName)) {
            ports.add(port);
          }
        }
      }
    }
    return ports;
  }

  /**
   * Add the given property.
   *
   * @param property
   *            property to add
   */
  public void addProperty(final Property property) {
    if (this.properties == null) {
      this.properties = new HashMap<QName, Property>();
    }
    this.properties.put(property.getName(), property);
  }

  /**
   * Return the property having the given QName.
   *
   * @param propertyQName
   *            propertyQName
   * @return the property having the given QName.
   */
  public Property getProperty(final QName propertyQName) {
    if (this.properties == null) {
      return null;
    }
    return this.properties.get(propertyQName);
  }

  /**
   * Add the given PrpertyAlias.
   *
   * @param propertyAlias
   *            propertyAlias to add
   */
  public void addPropertyAlias(final PropertyAlias propertyAlias) {
    if (this.propertyAliases == null) {
      this.propertyAliases = new ArrayList<PropertyAlias>();
    }
    this.propertyAliases.add(propertyAlias);
  }

  /**
   * Get the propertyAlias associated with the propertyName and the messageType
   * given in parameter. Or null if there is no propertyAlias associated with
   * this propertyName.
   *
   * @param propertyName :
   *            name of the property
   * @param messageType :
   *            messageType of the propertyAlias
   * @return the propertyAlias
   */
  public PropertyAlias getPropertyAliasWithMessageType(final QName propertyName,
      final QName messageType) {
    for (final PropertyAlias propertyAlias : this.propertyAliases) {
      if (propertyAlias.getMessageType() != null
          && propertyAlias.getMessageType().equals(messageType)
          && propertyAlias.getPropertyName().equals(propertyName)) {
        return propertyAlias;
      }
    }
    return null;
  }

  /**
   * Return all propertyAlias using the given messageType.
   *
   * @param messageType
   *            messageType
   * @return all propertyAlias using the given messageType.
   */
  public List<PropertyAlias> getPropertyAliasesWithMessageType(final QName messageType) {
    final List<PropertyAlias> matchingPropertyAliases = new ArrayList<PropertyAlias>();
    for (final PropertyAlias propertyAlias : this.propertyAliases) {
      if (propertyAlias.getMessageType() != null
          && propertyAlias.getMessageType().equals(messageType)) {
        matchingPropertyAliases.add(propertyAlias);
      }
    }
    return matchingPropertyAliases;
  }

  /**
   * Return all propertyAliases using the given messageType and properties.
   *
   * @param propertiesToScan
   *            properties we need to scan.
   * @param messageType
   *            messageType that returned propertyAlias should conform
   * @return all propertyAliases using the given messageType and properties.
   */
  public List<PropertyAlias> getPropertyAliases(
      final Collection<Property> propertiesToScan, final QName messageType) {
    final List<PropertyAlias> result = new ArrayList<PropertyAlias>();
    for (final Property property : propertiesToScan) {
      final PropertyAlias propertyAlias = this.getPropertyAliasWithMessageType(property
          .getName(), messageType);
      if (propertyAlias == null) {
        throw new OrchestraRuntimeException("propertyAlias is null, property = "
            + property);
      }
      result.add(propertyAlias);
    }
    return result;
  }

  /**
   * Get the propertyAlias associated with the propertyName and the element
   * given in parameter. Or null if there is no propertyAlias associated with
   * this propertyName.
   *
   * @param propertyName :
   *            name of the property
   * @param element :
   *            element of the propertyAlias
   * @return the propertyAlias
   */
  public PropertyAlias getPropertyAliasWithElement(final QName propertyName,
      final QName element) {
    for (final PropertyAlias propertyAlias : this.propertyAliases) {
      if (propertyAlias.getElement() != null
          && propertyAlias.getElement().equals(element)
          && propertyAlias.getPropertyName().equals(propertyName)) {
        return propertyAlias;
      }
    }
    return null;
  }

  /**
   * Get the propertyAlias associated with the propertyName and the xmlType given
   * in parameter. Or null if there is no propertyAlias associated with this
   * propertyName.
   *
   * @param propertyName :
   *            name of the property
   * @param xmlType :
   *            xmlType of the propertyAlias
   * @return the propertyAlias
   */
  public PropertyAlias getPropertyAliasWithType(final QName propertyName, final QName type) {
    for (final PropertyAlias propertyAlias : this.propertyAliases) {
      if (propertyAlias.getType() != null
          && propertyAlias.getType().equals(type)
          && propertyAlias.getPropertyName().equals(propertyName)) {
        return propertyAlias;
      }
    }
    return null;
  }


  public Set<Definition> getWsdlDefinitions() {
    return this.wsdlDefinitions;
  }

  public Definition extractServiceWsdlDefinition(final Service service) {
    for (final Definition definition : this.wsdlDefinitions) {
      if (definition.getService(service.getQName()) != null) {
        return definition;
      }
    }
    return null;
  }

  public Set<Definition> extractServiceWsdlDefinitions(final Service service) {
    final ServiceWsdlGenerator serviceWsdlGenerator = new ServiceWsdlGenerator(this.wsdlDefinitions);
    return serviceWsdlGenerator.generateWsdlsForService(service);
  }

  private static class ServiceWsdlGenerator {

    private final Set<Definition> wsdlDefinitions;

    public ServiceWsdlGenerator(final Set<Definition> wsdlDefinitions) {
      this.wsdlDefinitions = wsdlDefinitions;
    }

    public Set<Definition> generateWsdlsForService(final Service service) {
      final Set<Definition> defs = new HashSet<Definition>();
      // add service
      this.addServiceToDefinitions(defs, service);
      return defs;
    }

    private void addServiceToDefinitions(final Set<Definition> definitions, final Service service) {
      // add definition for service
      for (final Definition definition : this.wsdlDefinitions) {
        if (definition.getService(service.getQName()) != null) {
          definitions.add(definition);
        }
      }
      // for each service port
      final Map<String, Port> ports = service.getPorts();
      for (final Port port : ports.values()) {
        // add binding
        final Binding binding = port.getBinding();
        this.addBindingToDefinition(definitions, binding);
      }
    }

    private void addBindingToDefinition(final Set<Definition> definitions, final Binding binding) {
      // add definition for binding
      for (final Definition definition : this.wsdlDefinitions) {
        if (definition.getBinding(binding.getQName()) != null) {
          definitions.add(definition);
        }
      }
      // add portType
      final PortType portType = binding.getPortType();
      this.addPortTypeToDefinition(definitions, portType);
    }

    private void addPortTypeToDefinition(final Set<Definition> definitions, final PortType portType) {
      // add definition for portType
      for (final Definition definition : this.wsdlDefinitions) {
        if (definition.getPortType(portType.getQName()) != null) {
          definitions.add(definition);
        }
      }
      // add messages
      final List<Operation> operations = portType.getOperations();
      for (final Operation operation : operations) {
        final Input input = operation.getInput();
        if (input != null) {
          final Message inputMessage = input.getMessage();
          if (inputMessage != null) {
            this.addMessageToDefinition(definitions, inputMessage);
          }
        }
        final Output output = operation.getOutput();
        if (output != null) {
          final Message outputMessage = output.getMessage();
          if (outputMessage != null) {
            this.addMessageToDefinition(definitions, outputMessage);
          }
        }
        final Map<String, Fault> faults = operation.getFaults();
        if (faults != null) {
          for (final Fault fault : faults.values()) {
            // add fault message
            final Message faultMessage = fault.getMessage();
            this.addMessageToDefinition(definitions, faultMessage);
          }
        }
      }
    }

    private void addMessageToDefinition(final Set<Definition> definitions, final Message message) {
      // add definition for message
      for (final Definition definition : this.wsdlDefinitions) {
        if (definition.getMessage(message.getQName()) != null) {
          definitions.add(definition);
        }
      }
    }

  }
}
